/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     element.c                                                      */
/*                                                                          */
/*                                                                          */
/* description:  routines on elements that depend on the dimension in 1d    */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include "alberta.h"

/****************************************************************************/
/*  world_to_coord():  return -1 if inside, otherwise index of lambda<0     */
/****************************************************************************/

int world_to_coord(const EL_INFO *el_info, const REAL *x, REAL lambda[DIM+1])
{
  FUNCNAME("world_to_coord");
  REAL lmin;
  REAL a = el_info->coord[0][0];
  REAL length = (el_info->coord[1][0] - a);
  REAL det = ABS(length);
  int i, j, k;

#if (DIM != DIM_OF_WORLD)
  ERROR_EXIT("not yet for DIM != DIM_OF_WORLD\n");
#endif

  TEST_FLAG(FILL_COORDS, el_info);
  TEST_EXIT(!el_info->mesh->parametric)("not for parametric meshes\n");

  if (det < 1.E-20) {
    ERROR_EXIT("length = %le; abort\n", length);
    return(-2);
  }

  lambda[1] = (x[0]-a)/length;
  lambda[0] = 1.0 - lambda[1];

  k = -1;
  lmin = 0.0;
  j = 0;
  for (i = 0; i <= DIM; i++) {
    if ((lambda[i]*det) < -1.E-15) {
      if (lambda[i] < lmin) {
        k = i;
        lmin = lambda[i];
      }
      j++;
    }
  }

  return(k);
}

/****************************************************************************/
/*  transform local coordintes l to world coordinates; if w is non nil      */
/*  store them at w otherwise return a pointer to some local static         */
/*  area containing world coordintes                                        */
/****************************************************************************/

const REAL *coord_to_world(const EL_INFO *el_info, const REAL *l, REAL_D w)
{
  FUNCNAME("coord_to_world");
  static REAL world[DIM_OF_WORLD];
  REAL        c, *ret;
  const REAL  *v;
  int         i, j;

  TEST_FLAG(FILL_COORDS, el_info);
  TEST_EXIT(!el_info->mesh->parametric)("not for parametric meshes\n");

  ret = w ? w : world;

  v = el_info->coord[0];
  c = l[0];
  for (j = 0; j < DIM_OF_WORLD; j++)
    ret[j] = c*v[j];
  
  for (i = 1; i < N_VERTICES; i++)
  {
    v = el_info->coord[i];
    c = l[i];
    for (j = 0; j < DIM_OF_WORLD; j++)
      ret[j] += c*v[j];
  }
  return((const REAL *) ret);
}

/*--------------------------------------------------------------------------*/
/*  compute volume of an element                                            */
/*--------------------------------------------------------------------------*/

REAL el_det(const EL_INFO *el_info)
{
  FUNCNAME("el_det");

#if DIM != DIM_OF_WORLD
  ERROR_EXIT("not for DIM != DIM_OF_WORLD\n");
#endif

  TEST_FLAG(FILL_COORDS, el_info);
  TEST_EXIT(!el_info->mesh->parametric)("not for parametric meshes\n");

  return(el_info->coord[1][0] - el_info->coord[0][0]);
}

REAL el_volume(const EL_INFO *el_info)
{
  return(el_det(el_info));
}

/*--------------------------------------------------------------------------*/
/*  compute gradients of basis functions on element; return the absulute    */
/*  value of the determinante from the transformation to the reference      */
/*  element                                                                 */
/*--------------------------------------------------------------------------*/

REAL el_grd_lambda(const EL_INFO *el_info, REAL grd_lam[DIM+1][DIM_OF_WORLD])
{
  FUNCNAME("el_grd_lambda");
  REAL   det;

#if DIM != DIM_OF_WORLD
  ERROR_EXIT("not for DIM != DIM_OF_WORLD\n");
#endif

  TEST_FLAG(FILL_COORDS, el_info);
  TEST_EXIT(!el_info->mesh->parametric)("not for parametric meshes\n");

  det = el_info->coord[1][0] - el_info->coord[0][0];
  grd_lam[0][0] = -1.0/det;
  grd_lam[1][0] =  1.0/det;

  return(det);
}

#if DIM_OF_WORLD == 3
/****************************************************************************/
/*  vector product in R^3                                                   */
/****************************************************************************/

void vector_product(const REAL_D x, const REAL_D y, REAL_D z)
{
  FUNCNAME("vector_product");

  z[0] = x[1]*y[2] - x[2]*y[1];
  z[1] = x[2]*y[0] - x[0]*y[2];
  z[2] = x[0]*y[1] - x[1]*y[0];

  return;
}
#endif

/*---8<---------------------------------------------------------------------*/
/*---  face normal in 1d is really rather easy....                       ---*/
/*--------------------------------------------------------------------->8---*/

REAL get_face_normal(const EL_INFO *el_info, int i0, REAL *normal)
{
#if DIM_OF_WORLD == 1
  normal[0] = i0 ? -1.0 : 1.0;
#else
  int          n;
  REAL         det;
  const REAL_D *coord = el_info->coord;

  TEST_FLAG(FILL_COORDS, el_info);

  for (n = 0; n < DIM_OF_WORLD; n++)
    normal[n] = i0? coord[0][n] - coord[1][n] : coord[1][n] - coord[0][n];

  det = NORM_DOW(normal);
  for (n = 0; n < DIM_OF_WORLD; n++)
    normal[n] /= det;
#endif
  return(1.0);
}

/*--------------------------------------------------------------------------*/
/* orient the vertices of edges/faces                                       */
/* used by estimator for the jumps => same quadrature nodes from both sides!*/
/*--------------------------------------------------------------------------*/

const int *sort_face_indices(const EL *el, int face, int *vec)
{
  static int out[1];
  int        *ret = vec ? vec : out;

  ret[0] = 1-face;

  return(ret);
}
