/*--------------------------------------------------------------------------*/
/*--- ALBERTA:  an Adaptive multi Level finite element toolbox using     ---*/
/*---           Bisectioning refinement and Error control by Residual    ---*/
/*---           Techniques for scientific Applications                   ---*/
/*---                                                                    ---*/
/*--- file: lagrange_1.c                                                 ---*/
/*---                                                                    ---*/
/*--- description: implementation of the basis functions                 ---*/
/*---              lagrange4 in 1d                                       ---*/
/*---                                                                    ---*/
/*--- created by: kgs on host enigma                                     ---*/
/*---           at 17:19 on 28 of March 2003                             ---*/
/*--------------------------------------------------------------------------*/
/*---                                                                    ---*/
/*--- authors:   Alfred Schmidt                                          ---*/
/*---            Zentrum fuer Technomathematik                           ---*/
/*---            Fachbereich 3 Mathematik/Informatik                     ---*/
/*---            Universitaet Bremen                                     ---*/
/*---            Bibliothekstr. 2                                        ---*/
/*---            D-28359 Bremen, Germany                                 ---*/
/*---                                                                    ---*/
/*---            Kunibert G. Siebert                                     ---*/
/*---            Institut fuer Mathematik                                ---*/
/*---            Universitaet Augsburg                                   ---*/
/*---            Universitaetsstr. 14                                    ---*/
/*---            D-86159 Augsburg, Germany                               ---*/
/*---                                                                    ---*/
/*---    http://www.alberta-fem.de/                                      ---*/
/*---                                                                    ---*/
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include <alberta.h>

/*--------------------------------------------------------------------------*/
/*---  basisfunction 0 located at vertex 0                               ---*/
/*--------------------------------------------------------------------------*/

static REAL phi1_0(const REAL lambda[DIM+1])
{
  return(lambda[0]);
}

static const REAL *grd_phi1_0(const REAL lambda[DIM+1])
{
  static const REAL  grd[DIM+1] = {1.0, 0.0};

  return(grd);
}

static const REAL (*D2_phi1_0(const REAL *lambda))[DIM+1]
{
  static const REAL D2[DIM+1][DIM+1] = {{0.0}};

  return(D2);
}

/*--------------------------------------------------------------------------*/
/*---  basisfunction 1 located at vertex 1                               ---*/
/*--------------------------------------------------------------------------*/

static REAL phi1_1(const REAL lambda[DIM+1])
{
  return(lambda[1]);
}

static const REAL *grd_phi1_1(const REAL lambda[DIM+1])
{
  static const REAL  grd[DIM+1] = {0.0, 1.0};

  return(grd);
}

static const REAL (*D2_phi1_1(const REAL *lambda))[DIM+1]
{
  static const REAL D2[DIM+1][DIM+1] = {{0.0}};

  return(D2);
}

/*--------------------------------------------------------------------*/
/*--- function for accessing local DOFs on an element              ---*/
/*--------------------------------------------------------------------*/

static const DOF *get_dof_indices1(const EL *el, const DOF_ADMIN *admin,
                                  DOF *idof)
{
  static DOF  dof_vec[2];
  DOF         *rvec = idof ? idof : dof_vec;
  int         ibas = 0, i, n0, node;
  DOF         **dof = el->dof;

/*--------------------------------------------------------------------*/
/*--- DOFs at vertices                                             ---*/
/*--------------------------------------------------------------------*/

  node = admin->mesh->node[VERTEX];
  n0   = admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[ibas++] = dof[node+i][n0];


  return(rvec);
}

/*--------------------------------------------------------------------*/
/*--- function for accessing boundary type of DOFs                 ---*/
/*--------------------------------------------------------------------*/

static const S_CHAR *get_bound1(const EL_INFO *el_info, S_CHAR *vec)
{
  FUNCNAME("get_bound1");
  static S_CHAR  my_vec[2];
  S_CHAR         *rvec = vec ? vec : my_vec;
  int            ibas = 0, i;

  TEST_FLAG(FILL_BOUND, el_info);

/*--------------------------------------------------------------------*/
/*--- basis functions at vertices                                  ---*/
/*--------------------------------------------------------------------*/

  for (i = 0; i < N_VERTICES; i++)
    rvec[ibas++] = el_info->bound[i];

  return(rvec);
}

/*--------------------------------------------------------------------*/
/*--- vector of barycentric coordinates                            ---*/
/*--------------------------------------------------------------------*/

static const REAL coords1[2][DIM+1] = 
{
  {1.000000000000000, 0.000000000000000},
  {0.000000000000000, 1.000000000000000}
};

/*--------------------------------------------------------------------*/
/*--- function for local interpolaton of scalar functions          ---*/
/*--------------------------------------------------------------------*/

static const REAL *interpol1(const EL_INFO *el_info,
                             int no, const int *b_no,
                             REAL (*f)(const REAL_D),
                             REAL *vec)
{
  FUNCNAME("interpol1");
  static REAL   my_vec[2];
  REAL          *rvec = vec ? vec : my_vec;
  int              i, j;
  const PARAMETRIC *parametric = el_info->mesh->parametric;

  if (b_no)
  {
     TEST_EXIT(no > 0 && no <= 2)("not for %d points\n", no);
  }

  if (parametric  &&  parametric->init_element(el_info, parametric))
  {
    REAL_D     world[2];
    REAL       coord_no[2][DIM+1];
    const REAL (*coord)[DIM+1];

    if (b_no)
    {
      for (i = 0; i < no; i++)
      {
        for (j = 0; j <= DIM; j++)
          coord_no[i][j] = coords1[b_no[i]][j];
      }
      coord = (const REAL (*)[DIM+1]) coord_no;
    }
    else
    {
       no = 2;
       coord = coords1;
    }

    (*parametric->coord_to_world)(el_info, nil, no, (void *) coord, world);

    for (i = 0; i < no; i++)
      rvec[i] = (*f)(world[i]);
  }
  else
  {
    TEST_FLAG(FILL_COORDS, el_info);

    if (b_no)
    {
      for (i = 0; i < no; i++)
      {
        const REAL *world = coord_to_world(el_info, coords1[b_no[i]], nil);
        rvec[i] = (*f)(world);
      }
    }
    else
    {
      for (i = 0; i < 2; i++)
      {
        const REAL *world = coord_to_world(el_info, coords1[i], nil);
        rvec[i] = (*f)(world);
      }
    }
  }

  return(rvec);
}
/*--------------------------------------------------------------------*/
/*--- function for local interpolaton of vector functions          ---*/
/*--------------------------------------------------------------------*/

static const REAL_D *interpol_d1(const EL_INFO *el_info,
                                 int no, const int *b_no,
                                 const REAL *(*f)(const REAL_D, REAL_D),
                                 REAL_D *vec)
{
  FUNCNAME("interpol_d1");
  static REAL_D my_vec[2];
  REAL_D        *rvec = vec ? vec : my_vec;
  int              i, j;
  const PARAMETRIC *parametric = el_info->mesh->parametric;

  if (b_no)
  {
     TEST_EXIT(no > 0 && no <= 2)("not for %d points\n", no);
  }

  if (parametric  &&  parametric->init_element(el_info, parametric))
  {
    REAL_D     world[2];
    REAL       coord_no[2][DIM+1];
    const REAL (*coord)[DIM+1];

    if (b_no)
    {
      for (i = 0; i < no; i++)
      {
        for (j = 0; j <= DIM; j++)
          coord_no[i][j] = coords1[b_no[i]][j];
      }
      coord = (const REAL (*)[DIM+1]) coord_no;
    }
    else
    {
       no = 2;
       coord = coords1;
    }

    (*parametric->coord_to_world)(el_info, nil, no, (void *) coord, world);

    for (i = 0; i < no; i++)
      (*f)(world[i], rvec[i]);
  }
  else
  {
    TEST_FLAG(FILL_COORDS, el_info);

    if (b_no)
    {
      for (i = 0; i < no; i++)
      {
        const REAL *world = coord_to_world(el_info, coords1[b_no[i]], nil);
        (*f)(world, rvec[i]);
      }
    }
    else
    {
      for (i = 0; i < 2; i++)
      {
        const REAL *world = coord_to_world(el_info, coords1[i], nil);
        (*f)(world, rvec[i]);
      }
    }
  }

  return((const REAL_D *) rvec);
}
/*--------------------------------------------------------------------*/
/*--- function for accessing a local DOF_INT_VEC                   ---*/
/*--------------------------------------------------------------------*/

static const int *get_int_vec1(const EL *el, const DOF_INT_VEC *dv,
                               int *vec)
{
  FUNCNAME("get_int_vec1");
  static int     my_vec[2];
  int            *rvec = vec ? vec : my_vec;
  int            *v = dv ? dv->vec : nil;
  int            ibas = 0, i, n0, node;
  DOF            **dof = el->dof;

  TEST_EXIT(v)("no DOF_INT_VEC dv, or no dv->vec\n");

/*--------------------------------------------------------------------*/
/*--- DOFs at vertices                                             ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[VERTEX];
  n0   = dv->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[ibas++] = v[dof[node+i][n0]];

  return(rvec);
}

/*--------------------------------------------------------------------*/
/*--- function for accessing a local DOF_REAL_VEC                  ---*/
/*--------------------------------------------------------------------*/

static const REAL *get_real_vec1(const EL *el, const DOF_REAL_VEC *dv,
                                 REAL *vec)
{
  FUNCNAME("get_real_vec1");
  static REAL    my_vec[2];
  REAL           *rvec = vec ? vec : my_vec;
  REAL           *v = dv ? dv->vec : nil;
  int            ibas = 0, i, n0, node;
  DOF            **dof = el->dof;

  TEST_EXIT(v)("no DOF_REAL_VEC dv, or no dv->vec\n");

/*--------------------------------------------------------------------*/
/*--- DOFs at vertices                                             ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[VERTEX];
  n0   = dv->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[ibas++] = v[dof[node+i][n0]];

  return(rvec);
}

/*--------------------------------------------------------------------*/
/*--- function for accessing a local DOF_REAL_D_VEC                ---*/
/*--------------------------------------------------------------------*/

static const REAL_D *get_real_d_vec1(const EL *el, const DOF_REAL_D_VEC *dv,
                                     REAL_D *vec)
{
  FUNCNAME("get_real_d_vec1");
  static REAL_D  my_vec[2];
  REAL_D         *rvec = vec ? vec : my_vec;
  REAL_D         *v = dv ? dv->vec : nil;
  int            ibas = 0, i, n, n0, node;
  DOF            **dof = el->dof;

  TEST_EXIT(v)("no DOF_REAL_D_VEC dv, or no dv->vec\n");

/*--------------------------------------------------------------------*/
/*--- DOFs at vertices                                             ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[VERTEX];
  n0   = dv->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
  {
    for (n = 0; n < DIM_OF_WORLD; n++)
      rvec[ibas][n] = v[dof[node+i][n0]][n];
    ibas++;
  }

  return((const REAL_D *) rvec);
}

/*--------------------------------------------------------------------*/
/*--- function for accessing a local DOF_SCHAR_VEC                 ---*/
/*--------------------------------------------------------------------*/

static const S_CHAR *get_schar_vec1(const EL *el, const DOF_SCHAR_VEC *dv,
                                    S_CHAR *vec)
{
  FUNCNAME("get_schar_vec1");
  static S_CHAR  my_vec[2];
  S_CHAR         *rvec = vec ? vec : my_vec;
  S_CHAR         *v = dv ? dv->vec : nil;
  int            ibas = 0, i, n0, node;
  DOF            **dof = el->dof;

  TEST_EXIT(v)("no DOF_SCHAR_VEC dv, or no dv->vec\n");

/*--------------------------------------------------------------------*/
/*--- DOFs at vertices                                             ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[VERTEX];
  n0   = dv->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[ibas++] = v[dof[node+i][n0]];

  return(rvec);
}

/*--------------------------------------------------------------------*/
/*--- function for accessing a local DOF_UCHAR_VEC                 ---*/
/*--------------------------------------------------------------------*/

static const U_CHAR *get_uchar_vec1(const EL *el, const DOF_UCHAR_VEC *dv,
                                    U_CHAR *vec)
{
  FUNCNAME("get_uchar_vec1");
  static U_CHAR  my_vec[2];
  U_CHAR         *rvec = vec ? vec : my_vec;
  U_CHAR         *v = dv ? dv->vec : nil;
  int            ibas = 0, i, n0, node;
  DOF            **dof = el->dof;

  TEST_EXIT(v)("no DOF_UCHAR_VEC dv, or no dv->vec\n");

/*--------------------------------------------------------------------*/
/*--- DOFs at vertices                                             ---*/
/*--------------------------------------------------------------------*/

  node = dv->fe_space->mesh->node[VERTEX];
  n0   = dv->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[ibas++] = v[dof[node+i][n0]];

  return(rvec);
}

/*--------------------------------------------------------------------*/
/*--- function for interpolaton DOF_REAL_VECs during refinement    ---*/
/*--------------------------------------------------------------------*/

static void refine_inter1(DOF_REAL_VEC *dv, RC_LIST_EL *list, int n_el)
{
  EL              *el;
  const DOF       *cdof;
  const REAL      *pvec;
  const BAS_FCTS  *bas_fcts = dv->fe_space->bas_fcts;
  const DOF_ADMIN *admin = dv->fe_space->admin;
  REAL            *v = dv->vec;

  el = list->el;
  pvec = (*bas_fcts->get_real_vec)(el, dv, nil);

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, first child                           ---*/
/*--------------------------------------------------------------------*/

  cdof = (*bas_fcts->get_dof_indices)(el->child[0], admin, nil);

  v[cdof[1]] = (0.5*pvec[0] + 0.5*pvec[1]);

  return;
}

/*--------------------------------------------------------------------*/
/*--- function for interpolaton DOF_REAL_D_VECs during refinement  ---*/
/*--------------------------------------------------------------------*/

static void refine_inter_d1(DOF_REAL_D_VEC *dv, RC_LIST_EL *list, int n_el)
{
  EL              *el;
  const DOF       *cdof;
  const REAL_D    *pvec;
  int             n;
  const BAS_FCTS  *bas_fcts = dv->fe_space->bas_fcts;
  const DOF_ADMIN *admin = dv->fe_space->admin;
  REAL_D          *v = dv->vec;

  el = list->el;
  pvec = (*bas_fcts->get_real_d_vec)(el, dv, nil);

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, first child                           ---*/
/*--------------------------------------------------------------------*/

  cdof = (*bas_fcts->get_dof_indices)(el->child[0], admin, nil);

  for (n = 0; n < DIM_OF_WORLD; n++)
  {
    v[cdof[1]][n] = (0.5*pvec[0][n] + 0.5*pvec[1][n]);
  }

  return;
}

/*--------------------------------------------------------------------*/
/*--- function for interpolaton DOF_REAL_VECs during coarsening    ---*/
/*--------------------------------------------------------------------*/

static void coarse_restr1(DOF_REAL_VEC *dv, RC_LIST_EL *list, int n_el)
{
  EL              *el;
  const DOF       *pdof;
  const REAL      *cvec;
  const BAS_FCTS  *bas_fcts = dv->fe_space->bas_fcts;
  const DOF_ADMIN *admin = dv->fe_space->admin;
  REAL            *v = dv->vec;

  el = list->el;
  pdof = (*bas_fcts->get_dof_indices)(el, admin, nil);

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, from first child                      ---*/
/*--------------------------------------------------------------------*/

  cvec = (*bas_fcts->get_real_vec)(el->child[0], dv, nil);

  v[pdof[0]] += (0.5*cvec[1]);
  v[pdof[1]] += (0.5*cvec[1]);

  return;
}

/*--------------------------------------------------------------------*/
/*--- function for interpolaton DOF_REAL_D_VECs during coarsening  ---*/
/*--------------------------------------------------------------------*/

static void coarse_restr_d1(DOF_REAL_D_VEC *dv, RC_LIST_EL *list, int n_el)
{
  EL              *el;
  const DOF       *pdof;
  const REAL_D    *cvec;
  int             n;
  const BAS_FCTS  *bas_fcts = dv->fe_space->bas_fcts;
  const DOF_ADMIN *admin = dv->fe_space->admin;
  REAL_D          *v = dv->vec;

  el = list->el;
  pdof = (*bas_fcts->get_dof_indices)(el, admin, nil);

/*--------------------------------------------------------------------*/
/*--- DOFs on first element, from first child                      ---*/
/*--------------------------------------------------------------------*/

  cvec = (*bas_fcts->get_real_d_vec)(el->child[0], dv, nil);

  for (n = 0; n < DIM_OF_WORLD; n++)
  {
    v[pdof[0]][n] += (0.5*cvec[1][n]);
    v[pdof[1]][n] += (0.5*cvec[1][n]);
  }

  return;
}


/*--------------------------------------------------------------------*/
/*--- Collect all information about basis functions                ---*/
/*--------------------------------------------------------------------*/

static BAS_FCT *phi1[2] =
{
  phi1_0, phi1_1
};

static GRD_BAS_FCT *grd_phi1[2] =
{
  grd_phi1_0, grd_phi1_1
};
static D2_BAS_FCT *D2_phi1[2] =
{
  D2_phi1_0, D2_phi1_1
};

static BAS_FCTS lagrange1 =
{
  "lagrange1", 2, 1, {1, 0},
  nil, phi1, grd_phi1, D2_phi1,
  get_dof_indices1,
  get_bound1,
  interpol1,
  interpol_d1,
  get_int_vec1,
  get_real_vec1,
  get_real_d_vec1,
  get_uchar_vec1,
  get_schar_vec1,
  refine_inter1,
  nil,
  coarse_restr1,
  refine_inter_d1,
  nil,
  coarse_restr_d1,
  (void *) coords1
};
