/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     coarsen.c                                                      */
/*                                                                          */
/*    !!!  includes ../Common/coarsen_common.c  !!!                         */
/*                                                                          */
/* description:  coarsening of 2 dim. hierarchical meshes;                  */
/*               at the end of file Common/coarsen_common.c is included;    */
/*               file contains all routines depending on DIM == 2;          */
/*               routines for both dimensions in ../Common/coarsen_common.c */
/*               are included at the end of the file                        */
/*                                                                          */
/* global variables that have to be defined for Common/coarsen_common.c:    */
/*  static MESH           *coarse_mesh;                                     */
/*  static int            do_more;                                          */
/*  static int            call_coarse_restrict;                             */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include "alberta.h"

/****************************************************************************/
/*  global variables used for coarsening                                    */
/****************************************************************************/
static MESH           *coarse_mesh;
static int            do_more = true;
static int            call_coarse_restrict;

#if !NEIGH_IN_EL
static TRAVERSE_STACK *stack;
#endif

/****************************************************************************/
/*  coarse_element:  coarses a single element of the coarsening patch; dofs */
/*  in the interior of the element are removed; dofs for higher order       */
/*  at the boundary or the coarsening patch still belong to                 */
/*  the parent. Do not remove them form the mesh!!!                         */
/*                                                                          */
/*  if NEIGH_IN_EL neighbour information to elements outside the patch is   */
/*  set on the parent; neighbour information inside the patch is still      */
/*  valid!                                                                  */
/****************************************************************************/

static void coarse_element(EL *el)
{
  FUNCNAME("coarse_element");
  EL       *child[2];

  child[0] = el->child[0];   child[1] = el->child[1]; 

  TEST_EXIT(child[0]->mark < 0  &&  child[1]->mark < 0)
    ("element %d with children[%d,%d] must not be coarsend!\n",
     INDEX(el), INDEX(child[0]), INDEX(child[1]));
  
#if NEIGH_IN_EL
/****************************************************************************/
/*  adjust neighbour connectivity  at the boundary of the coarsening patch  */
/****************************************************************************/
  {
    EL     *neigh;
    int    opp_v, i_child;

    for (i_child = 0; i_child < 2; i_child++)
    {
      if (neigh = (el->neigh[1-i_child] = child[i_child]->neigh[2]))
      {
	opp_v = el->opp_vertex[1-i_child] = child[i_child]->opp_vertex[2];
	neigh->neigh[opp_v] = el;
	neigh->opp_vertex[opp_v] = 1-i_child;
      }
    }
  }
#endif

  if (coarse_mesh->n_dof[EDGE])
  {
/****************************************************************************/
/*  remove dof from common edge of child[0] and child[1]                    */
/****************************************************************************/
    free_dof(child[0]->dof[4], coarse_mesh, EDGE);
  }

  if (coarse_mesh->n_dof[CENTER])
  {
/****************************************************************************/
/*  remove dof from the barycenters of child[0] and child[1]                */
/****************************************************************************/
    int  node = coarse_mesh->node[CENTER];

    free_dof(child[0]->dof[node], coarse_mesh, CENTER);
    free_dof(child[1]->dof[node], coarse_mesh, CENTER);
  }

  el->child[0] = nil;
  el->child[1] = (EL *) get_leaf_data(coarse_mesh);

  if (el->child[1]  &&  coarse_mesh->leaf_data_info->coarsen_leaf_data)
    coarse_mesh->leaf_data_info->coarsen_leaf_data(el, child);

  free_element(child[0], coarse_mesh);
  free_element(child[1], coarse_mesh);

  el->mark++;

  coarse_mesh->n_elements--;
  coarse_mesh->n_hier_elements -= 2;
  coarse_mesh->n_edges--;

  return;
}

/****************************************************************************/
/*  if DOFs for higher order are been removed on parents during refinement  */
/*  they are now added again by add_dof_parents                             */
/****************************************************************************/

static void add_dof_parents(RC_LIST_EL c_list[], int n_neigh)
{
  int    i, j, node;


  if (coarse_mesh->n_dof[EDGE])
  {
    node = coarse_mesh->node[EDGE];

/****************************************************************************/
/*  get dofs on the boundary of the coarsening patch from the children      */
/****************************************************************************/
    for (i = 0; i < n_neigh; i++)
    {
      for (j = 0; j < 2; j++)
	c_list[i].el->dof[node+j] = c_list[i].el->child[1-j]->dof[node+2];
    }
  }

  if (coarse_mesh->n_dof[CENTER])
  {
    int   node = coarse_mesh->node[CENTER];

/****************************************************************************/
/*  get new dof on parents at the barycenter                                */
/****************************************************************************/
    for (i = 0; i < n_neigh; i++)
    {
      if (!c_list[i].el->dof[node])
	c_list[i].el->dof[node] = get_dof(coarse_mesh, CENTER);
    }
  }

  return;
}

/****************************************************************************/
/*  coarse_patch: first rebuild the dofs on the parents then do restriction */
/*  of data (if possible) and finally coarsen the patch elements            */
/****************************************************************************/

static void  coarse_restrict(RC_LIST_EL *list, int n_el);

static void coarse_patch(RC_LIST_EL coarse_list[], int n_neigh, int bound)
{
  EL       *el = coarse_list[0].el, *neigh = coarse_list[1].el;
  REAL     *new_coord = el->new_coord;
  DOF      *dof[3];

  dof[0] = el->child[0]->dof[2];
  if (coarse_mesh->n_dof[EDGE])
  {
    dof[1] = el->child[0]->dof[3];
    dof[2] = el->child[1]->dof[4];
  }

  if(!coarse_mesh->preserve_coarse_dofs) {
    if (coarse_mesh->n_dof[EDGE])
      {
	int   node = coarse_mesh->node[EDGE];
/****************************************************************************/
/*  get new dof on el at the midpoint of the coarsening edge                */
/****************************************************************************/
	if (!el->dof[node+2])
	  {
	    el->dof[node+2] = get_dof(coarse_mesh, EDGE);
	    if (neigh) neigh->dof[node+2] = el->dof[node+2];
	  }
      }
    if (coarse_mesh->n_dof[EDGE]  ||  coarse_mesh->n_dof[CENTER])
      add_dof_parents(coarse_list, n_neigh);
  }

  if (call_coarse_restrict)
  {
/****************************************************************************/
/*  restrict dof vectors to the parents on the patch                        */
/****************************************************************************/
    coarse_restrict(coarse_list, n_neigh);
  }

/*--------------------------------------------------------------------------*/
/*--- and now coarse single elements                                     ---*/
/*--------------------------------------------------------------------------*/

  coarse_element(el);
  el->new_coord = nil;

  if (neigh) 
  {
    coarse_element(neigh);
    neigh->new_coord = nil;
  }

/*--------------------------------------------------------------------------*/
/*---  if there is a projected boundary node, free memory                ---*/
/*--------------------------------------------------------------------------*/

  if (new_coord) free_real_d(coarse_mesh, new_coord);

/****************************************************************************/
/*  now, remove those dofs in the coarcening edge                           */
/****************************************************************************/
  free_dof(dof[0], coarse_mesh, VERTEX);
  if (coarse_mesh->n_dof[EDGE])
  {
    free_dof(dof[1], coarse_mesh, EDGE);
    free_dof(dof[2], coarse_mesh, EDGE);
  }

  coarse_mesh->n_vertices--;
  coarse_mesh->n_edges--;

  return;
}

/****************************************************************************/
/* include common traversal routines for 2d and 3d coarsening               */
/****************************************************************************/

#include "coarsen_common.c"
