/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     lagrange_0.c                                                   */
/*                                                                          */
/* description:  piecewise constant Lagrange elements in 2d                 */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

static const REAL   bary0[1][DIM+1] = {{1.0/3.0, 1.0/3.0, 1.0/3.0}};
#if 0
static const int    fct_at_node0[1] = {6};
static const int    n_dof0[DIM+1] = {0, 0, 1};
#endif

static REAL phi0c0(const REAL lambda[DIM+1])
{
  return(1.0);
}

static const REAL *grd_phi0c0(const REAL lambda[DIM+1])
{
  static const REAL  grd[DIM+1] = {0.0, 0.0, 0.0};

  return(grd);
}

static const REAL (*D2_phi0c0(const REAL *lambda))[DIM+1]
{
  static  const REAL D2[DIM+1][DIM+1] = {{0, 0, 0},{0, 0, 0}, {0, 0, 0}};

  return(D2);
}

/*--------------------------------------------------------------------------*/
/*  functions for combining basisfunctions with coefficients                */
/*--------------------------------------------------------------------------*/

static const DOF *get_dof_indices0(const EL *el, const DOF_ADMIN *admin,
				   DOF *idof)
{
  static DOF  index_vec[1];
  DOF         *rvec = idof ? idof : index_vec;
  int         i = admin->mesh->node[CENTER], n0 = admin->n0_dof[CENTER];
  DOF         **dof = el->dof;

  rvec[0] = dof[i][n0];

  return((const DOF *) rvec);
}

static const S_CHAR *get_bound0(const EL_INFO *el_info, S_CHAR *bound)
{
  static S_CHAR  bound_vec[1];
  S_CHAR         *rvec = bound ? bound : bound_vec;

  rvec[0] = INTERIOR;

  return((const S_CHAR *) rvec);
}

static const int *get_int_vec0(const EL *el, const DOF_INT_VEC *vec,
			       int *ivec)
{
  FUNCNAME("get_int_vec0");
  static int  local_vec[1];
  int         i, n0, *v = nil;
  int         *rvec = ivec ? ivec : local_vec;
  DOF         **dof = el->dof;

  GET_DOF_VEC(v, vec);
  i = vec->fe_space->admin->mesh->node[CENTER];
  n0 = vec->fe_space->admin->n0_dof[CENTER];
  
  rvec[0] = v[dof[i][n0]];

  return((const int *) rvec);
}

static const REAL *get_real_vec0(const EL *el, const DOF_REAL_VEC *vec,
				 REAL *Rvec)
{
  FUNCNAME("get_real_v0");
  int          i, n0;
  static REAL  local_vec[1];
  REAL         *v = nil, *rvec = Rvec ? Rvec : local_vec;
  DOF          **dof = el->dof;

  GET_DOF_VEC(v, vec);
  i = vec->fe_space->admin->mesh->node[CENTER];
  n0 = vec->fe_space->admin->n0_dof[CENTER];
  
  rvec[0] = v[dof[i][n0]];

  return((const REAL *) rvec);
}

static const REAL_D *get_real_d_vec0(const EL *el, const DOF_REAL_D_VEC *vec,
				     REAL_D *Rvec)
{
  FUNCNAME("get_real_d_vec0");
  int           i, k, n0;
  static REAL_D local_vec[1];
  REAL_D        *v = nil, *rvec = Rvec ? Rvec : local_vec;
  DOF           **dof = el->dof;

  GET_DOF_VEC(v, vec);
  i = vec->fe_space->admin->mesh->node[CENTER];
  n0 = vec->fe_space->admin->n0_dof[CENTER];
  
  for (k = 0; k < DIM_OF_WORLD; k++)
    rvec[0][k] = v[dof[i][n0]][k];

  return((const REAL_D *) rvec);
}

static const U_CHAR *get_uchar_vec0(const EL *el, const DOF_UCHAR_VEC *vec,
				     U_CHAR *uvec)
{
  FUNCNAME("get_uchar_vec0");
  int           i, n0;
  static U_CHAR local_vec[1];
  U_CHAR        *v = nil, *rvec = uvec ? uvec : local_vec;
  DOF           **dof = el->dof;

  GET_DOF_VEC(v, vec);
  i = vec->fe_space->admin->mesh->node[CENTER];
  n0 = vec->fe_space->admin->n0_dof[CENTER];
  
  rvec[0] = v[dof[i][n0]];

  return((const U_CHAR *) rvec);
}

static const S_CHAR *get_schar_vec0(const EL *el, const DOF_SCHAR_VEC *vec,
				     S_CHAR *svec)
{
  FUNCNAME("get_schar_vec0");
  int            i, n0;
  static S_CHAR  local_vec[1];
  S_CHAR         *v = nil, *rvec = svec ? svec : local_vec;
  DOF            **dof = el->dof;

  GET_DOF_VEC(v, vec);
  i = vec->fe_space->admin->mesh->node[CENTER];
  n0 = vec->fe_space->admin->n0_dof[CENTER];
  
  rvec[0] = v[dof[i][n0]];

  return((const S_CHAR *) rvec);
}

static const REAL *interpol0(const EL_INFO *el_info, int no, 
			     const int *b_no, REAL (*f)(const REAL_D),
			     REAL *vec)
{
  FUNCNAME("interpol0");
  static REAL  inter[1];
  REAL         *rvec = vec ? vec : inter;
  REAL_D       x[1];
  const PARAMETRIC *parametric = el_info->mesh->parametric;

  if (b_no)
  {
    if (no != 1)
    {
      ERROR("something is wrong, doing nothing\n");
      rvec[0] = 0.0;
      return((const REAL *) rvec);
    }
  }

  if (parametric  &&  parametric->init_element(el_info, parametric))
  {
    static const REAL lambda[1][DIM+1] = {{1.0/3.0,1.0/3.0,1.0/3.0}};

    parametric->coord_to_world(el_info, nil, 1, lambda, x);
  }
  else
  {
    int  i, j;

    TEST_FLAG(FILL_COORDS, el_info);

    for (j = 0; j < DIM_OF_WORLD; j++)
    {
      x[0][j] = 0.0;
      for (i = 0; i < N_VERTICES; i++)
	x[0][j] += el_info->coord[i][j];
      x[0][j] /= (REAL) N_VERTICES;
    }
  }

  rvec[0] = (*f)(x[0]);
  
  return((const REAL *) rvec);
}

static const REAL_D *interpol_d0(const EL_INFO *el_info, int no, 
				 const int *b_no,
				 const REAL *(*f)(const REAL_D, REAL_D),
				 REAL_D *vec)
{
  FUNCNAME("interpol_d0");
  static REAL_D  inter[1];
  REAL_D         *rvec = vec ? vec : inter;
  REAL_D         x[1];
  const PARAMETRIC *parametric = el_info->mesh->parametric;

  if (b_no)
  {
    if (no != 1)
    {
      int i;
      ERROR("something is wrong, doing nothing\n");
      for (i = 0; i < DIM_OF_WORLD; i++)
	rvec[0][i] = 0.0;
      return((const REAL_D *) rvec);
    }
  }

  if (parametric  &&  parametric->init_element(el_info, parametric))
  {
    static const REAL  lambda[1][3] = {{1.0/3.0,1.0/3.0,1.0/3.0}};

    parametric->coord_to_world(el_info, nil, 1, lambda, x);
  }
  else
  {
    int  i, j;

    TEST_FLAG(FILL_COORDS, el_info);

    for (j = 0; j < DIM_OF_WORLD; j++)
    {
      x[0][j] = 0.0;
      for (i = 0; i < N_VERTICES; i++)
	x[0][j] += el_info->coord[i][j];
      x[0][j] /= (REAL) N_VERTICES;
    }
  }

  (*f)(x[0], rvec[0]);
  
  return((const REAL_D *) rvec);
}

/*--------------------------------------------------------------------------*/
/*  functions for interpolation/ restriction during refinement/coarsening   */
/*--------------------------------------------------------------------------*/

static void real_refine_inter0(DOF_REAL_VEC *drv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_refine_inter0");
  EL      *el;
  REAL    *vec = nil;
  int     cdof, pdof, node0, n0, i;

  if (n < 1) return;
  GET_DOF_VEC(vec, drv);
  node0 = drv->fe_space->mesh->node[CENTER];        
  n0 = drv->fe_space->admin->n0_dof[CENTER];

  for (i = 0; i < n; i++)
  {
    el = list[i].el;

    pdof = el->dof[node0][n0];
    cdof = el->child[0]->dof[node0][n0];
    vec[cdof] = vec[pdof];
    cdof = el->child[1]->dof[node0][n0];
    vec[cdof] = vec[pdof];
  }
  return;
}

static void real_coarse_inter0(DOF_REAL_VEC *drv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_coarse_inter0");
  EL      *el;
  REAL    *vec = nil;
  int     cdof0, cdof1, pdof, node0, n0, i;

  if (n < 1) return;
  GET_DOF_VEC(vec, drv);
  node0 = drv->fe_space->mesh->node[CENTER];        
  n0 = drv->fe_space->admin->n0_dof[CENTER];
  el = list->el;

  for (i = 0; i < n; i++)
  {
    el = list[i].el;

    pdof = el->dof[node0][n0];
    cdof0 = el->child[0]->dof[node0][n0];
    cdof1 = el->child[1]->dof[node0][n0];

    vec[pdof] = 0.5*(vec[cdof0] + vec[cdof1]);
  }
  return;
}

static void real_coarse_restr0(DOF_REAL_VEC *drv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_coarse_restr0");
  EL      *el;
  REAL    *vec = nil;
  int     cdof0, cdof1, pdof, node0, n0, i;

  if (n < 1) return;
  GET_DOF_VEC(vec, drv);
  node0 = drv->fe_space->mesh->node[CENTER];        
  n0 = drv->fe_space->admin->n0_dof[CENTER];
  el = list->el;

  for (i = 0; i < n; i++)
  {
    el = list[i].el;

    pdof = el->dof[node0][n0];
    cdof0 = el->child[0]->dof[node0][n0];
    cdof1 = el->child[1]->dof[node0][n0];

    vec[pdof] = vec[cdof0] + vec[cdof1];
  }
  return;
}

static void real_d_refine_inter0(DOF_REAL_D_VEC *drdv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_d_refine_inter0");
  EL      *el;
  REAL_D  *vec = nil;
  int     cdof, pdof, node0, n0, i, k;

  if (n < 1) return;
  GET_DOF_VEC(vec, drdv);
  node0 = drdv->fe_space->mesh->node[CENTER];        
  n0 = drdv->fe_space->admin->n0_dof[CENTER];

  for (i = 0; i < n; i++)
  {
    el = list[i].el;

    pdof = el->dof[node0][n0];
    cdof = el->child[0]->dof[node0][n0];
    for (k = 0; k < DIM_OF_WORLD; k++)
      vec[cdof][k] = vec[pdof][k];
    cdof = el->child[1]->dof[node0][n0];
    for (k = 0; k < DIM_OF_WORLD; k++)
      vec[cdof][k] = vec[pdof][k];
  }
  return;
}

static void real_d_coarse_inter0(DOF_REAL_D_VEC *drdv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_coarse_inter0");
  EL      *el;
  REAL_D  *vec = nil;
  int     cdof0, cdof1, pdof, node0, n0, i, k;

  if (n < 1) return;
  GET_DOF_VEC(vec, drdv);
  node0 = drdv->fe_space->mesh->node[CENTER];        
  n0 = drdv->fe_space->admin->n0_dof[CENTER];
  el = list->el;

  for (i = 0; i < n; i++)
  {
    el = list[i].el;

    pdof = el->dof[node0][n0];
    cdof0 = el->child[0]->dof[node0][n0];
    cdof1 = el->child[1]->dof[node0][n0];

    for (k = 0; k < DIM_OF_WORLD; k++)
      vec[pdof][k] = 0.5*(vec[cdof0][k] + vec[cdof1][k]);
  }
  return;
}

static void real_d_coarse_restr0(DOF_REAL_D_VEC *drdv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_d_coarse_restr0");
  EL      *el;
  REAL_D  *vec = nil;
  int     cdof0, cdof1, pdof, node0, n0, i, k;

  if (n < 1) return;
  GET_DOF_VEC(vec, drdv);
  node0 = drdv->fe_space->mesh->node[CENTER];        
  n0 = drdv->fe_space->admin->n0_dof[CENTER];
  el = list->el;

  for (i = 0; i < n; i++)
  {
    el = list[i].el;

    pdof = el->dof[node0][n0];
    cdof0 = el->child[0]->dof[node0][n0];
    cdof1 = el->child[1]->dof[node0][n0];

    for (k = 0; k < DIM_OF_WORLD; k++)
      vec[pdof] [k]= vec[cdof0][k] + vec[cdof1][k];
  }
  return;
}

static BAS_FCT      *phi0[1]     = {phi0c0};
static GRD_BAS_FCT  *grd_phi0[1] = {grd_phi0c0};
static D2_BAS_FCT   *D2_phi0[1]  = {D2_phi0c0};

static BAS_FCTS  lagrange0 = {"lagrange0", 1, 0, {0, 0, 1},
			      nil, phi0, grd_phi0, D2_phi0,
			      get_dof_indices0, 
			      get_bound0,
			      interpol0,
			      interpol_d0,
			      get_int_vec0,
			      get_real_vec0,
			      get_real_d_vec0,
			      get_uchar_vec0,
			      get_schar_vec0,
			      real_refine_inter0,
			      real_coarse_inter0,
			      real_coarse_restr0,
			      real_d_refine_inter0,
			      real_d_coarse_inter0,
			      real_d_coarse_restr0,
			      (void *) bary0};

