/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     lagrange_3.c                                                   */
/*                                                                          */
/* description:  piecewise cubic Lagrange elements in 2d                    */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#define N_BAS3  10

static const REAL   bary3[N_BAS3][DIM+1] = {{1.0, 0.0, 0.0},
					    {0.0, 1.0, 0.0},
					    {0.0, 0.0, 1.0},
					    {0.0, 2.0/3.0, 1.0/3.0},
					    {0.0, 1.0/3.0, 2.0/3.0},
					    {1.0/3.0, 0.0, 2.0/3.0},
					    {2.0/3.0, 0.0, 1.0/3.0},
					    {2.0/3.0, 1.0/3.0, 0.0},
					    {1.0/3.0, 2.0/3.0, 0.0},
					    {1.0/3.0, 1.0/3.0, 1.0/3.0}};

#if 0
static const int    fct_at_node3[N_BAS3] = {0,1,2,3,3,4,4,5,5,6};
static const int    n_dof3[DIM+1] = {1, 2, 1};
#endif

/*--------------------------------------------------------------------------*/
/*  basisfunction at vertex 0                                               */
/*--------------------------------------------------------------------------*/

static REAL phi3v0(const REAL lambda[DIM+1])
{
  return((4.5*(lambda[0] - 1.0)*lambda[0] + 1.0)*lambda[0]);
}

static const REAL *grd_phi3v0(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0};
  grd[0] = (13.5*lambda[0] - 9.0)*lambda[0] + 1;
  return((const REAL *) grd);
}

static const REAL (*D2_phi3v0(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0, 0, 0},{0, 0, 0}, {0, 0, 0}};
  D2[0][0] = 27.0*lambda[0] - 9.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at vertex 1                                               */
/*--------------------------------------------------------------------------*/

static REAL phi3v1(const REAL lambda[DIM+1])
{
  return((4.5*(lambda[1] - 1.0)*lambda[1] + 1.0)*lambda[1]);
}

static const REAL *grd_phi3v1(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0};
  grd[1] = (13.5*lambda[1] - 9.0)*lambda[1] + 1;
  return((const REAL *) grd);
}

static const REAL (*D2_phi3v1(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0, 0, 0},{0, 0, 0}, {0, 0, 0}};
  D2[1][1] = 27.0*lambda[1] - 9.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at vertex 2                                               */
/*--------------------------------------------------------------------------*/

static REAL phi3v2(const REAL lambda[DIM+1])
{
  return((4.5*(lambda[2] - 1.0)*lambda[2] + 1.0)*lambda[2]);
}

static const REAL *grd_phi3v2(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0};
  grd[2] = (13.5*lambda[2] - 9.0)*lambda[2] + 1;
  return((const REAL *) grd);
}

static const REAL (*D2_phi3v2(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0, 0, 0},{0, 0, 0}, {0, 0, 0}};
  D2[2][2] = 27.0*lambda[2] - 9.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 0 at edge 0                                               */
/*--------------------------------------------------------------------------*/

static REAL phi3e0(const REAL lambda[DIM+1])
{
  return((13.5*lambda[1] - 4.5)*lambda[1]*lambda[2]);
}

static const REAL *grd_phi3e0(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0};
  grd[1] = (27.0*lambda[1] - 4.5)*lambda[2];
  grd[2] = (13.5*lambda[1] - 4.5)*lambda[1];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3e0(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0, 0, 0},{0, 0, 0}, {0, 0, 0}};
  D2[1][1] = 27.0*lambda[2];
  D2[1][2] = D2[2][1] = 27.0*lambda[1] - 4.5;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 1 at edge 0                                               */
/*--------------------------------------------------------------------------*/

static REAL phi3e1(const REAL lambda[DIM+1])
{
  return((13.5*lambda[2] - 4.5)*lambda[2]*lambda[1]);
}

static const REAL *grd_phi3e1(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0};
  grd[1] = (13.5*lambda[2] - 4.5)*lambda[2];
  grd[2] = (27.0*lambda[2] - 4.5)*lambda[1];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3e1(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0, 0, 0},{0, 0, 0}, {0, 0, 0}};
  D2[1][2] = D2[2][1] = 27.0*lambda[2] - 4.5;
  D2[2][2] = 27.0*lambda[1];
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 0 at edge 1                                               */
/*--------------------------------------------------------------------------*/

static REAL phi3e2(const REAL lambda[DIM+1])
{
  return((13.5*lambda[2] - 4.5)*lambda[2]*lambda[0]);
}

static const REAL *grd_phi3e2(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0};
  grd[0] = (13.5*lambda[2] - 4.5)*lambda[2];
  grd[2] = (27.0*lambda[2] - 4.5)*lambda[0];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3e2(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0, 0, 0},{0, 0, 0}, {0, 0, 0}};
  D2[0][2] = D2[2][0] = 27.0*lambda[2] - 4.5;
  D2[2][2] = 27.0*lambda[0];
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 1 at edge 1                                               */
/*--------------------------------------------------------------------------*/

static REAL phi3e3(const REAL lambda[DIM+1])
{
  return((13.5*lambda[0] - 4.5)*lambda[0]*lambda[2]);
}

static const REAL *grd_phi3e3(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0};
  grd[0] = (27.0*lambda[0] - 4.5)*lambda[2];
  grd[2] = (13.5*lambda[0] - 4.5)*lambda[0];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3e3(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0, 0, 0},{0, 0, 0}, {0, 0, 0}};
  D2[0][0] = 27.0*lambda[2];
  D2[0][2] = D2[2][0] = 27.0*lambda[0] - 4.5;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 0 at edge 2                                               */
/*--------------------------------------------------------------------------*/

static REAL phi3e4(const REAL lambda[DIM+1])
{
  return((13.5*lambda[0] - 4.5)*lambda[0]*lambda[1]);
}

static const REAL *grd_phi3e4(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0};
  grd[0] = (27.0*lambda[0] - 4.5)*lambda[1];
  grd[1] = (13.5*lambda[0] - 4.5)*lambda[0];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3e4(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0, 0, 0},{0, 0, 0}, {0, 0, 0}};
  D2[0][0] = 27.0*lambda[1];
  D2[0][1] = D2[1][0] = 27.0*lambda[0] - 4.5;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 1 at edge 2                                               */
/*--------------------------------------------------------------------------*/

static REAL phi3e5(const REAL lambda[DIM+1])
{
  return((13.5*lambda[1] - 4.5)*lambda[1]*lambda[0]);
}

static const REAL *grd_phi3e5(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0};
  grd[0] = (13.5*lambda[1] - 4.5)*lambda[1];
  grd[1] = (27.0*lambda[1] - 4.5)*lambda[0];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3e5(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0, 0, 0},{0, 0, 0}, {0, 0, 0}};
  D2[0][1] = D2[1][0] = 27.0*lambda[1] - 4.5;
  D2[1][1] = 27.0*lambda[0];
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at barycenter                                             */
/*--------------------------------------------------------------------------*/

static REAL phi3c0(const REAL lambda[DIM+1])
{
  return(27.0*lambda[0]*lambda[1]*lambda[2]);
}

static const REAL *grd_phi3c0(const REAL lambda[DIM+1])
{
  static REAL grd[DIM+1] = {0.0, 0.0, 0.0};
  grd[0] = 27.0*lambda[1]*lambda[2];
  grd[1] = 27.0*lambda[0]*lambda[2];
  grd[2] = 27.0*lambda[0]*lambda[1];
  return((const REAL *) grd);
}

static const REAL (*D2_phi3c0(const REAL *lambda))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0, 0, 0},{0, 0, 0}, {0, 0, 0}};
  D2[0][1] = D2[1][0] = 27.0*lambda[2];
  D2[0][2] = D2[2][0] = 27.0*lambda[1];
  D2[1][2] = D2[2][1] = 27.0*lambda[0];
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  functions for combining basisfunctions with coefficients                */
/*--------------------------------------------------------------------------*/


static const DOF *get_dof_indices3(const EL *el, const DOF_ADMIN *admin,
				   DOF *idof)
{
  static DOF  index_vec[N_BAS3];
  DOF         *rvec = idof ? idof : index_vec;
  int         i, j, n0;
  DOF         **dof = el->dof;

  n0 = admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = dof[i][n0];
  n0 = admin->n0_dof[EDGE];
  for (i = 0, j = N_VERTICES; i < N_EDGES; i++)
  {
    if (dof[vertex_of_edge[i][0]][0] < dof[vertex_of_edge[i][1]][0])
    {
      rvec[j++] = dof[N_VERTICES+i][n0];
      rvec[j++] = dof[N_VERTICES+i][n0+1];
    }
    else
    {
      rvec[j++] = dof[N_VERTICES+i][n0+1];
      rvec[j++] = dof[N_VERTICES+i][n0];
    }
  }
  n0 = admin->n0_dof[CENTER];
  rvec[9] = dof[6][n0];

  return((const DOF *) rvec);
}


static const S_CHAR *get_bound3(const EL_INFO *el_info, S_CHAR *bound)
{
  FUNCNAME("get_bound3");
  static S_CHAR  bound_vec[N_BAS3];
  S_CHAR         *rvec = bound ? bound : bound_vec;
  int            i, j;

  TEST_FLAG(FILL_BOUND, el_info);

  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = el_info->bound[i];
  for (j = N_VERTICES, i = 0; i < N_EDGES; j += 2, i++)
    if (el_info->boundary[i])
      rvec[j] = rvec[j+1] = el_info->boundary[i]->bound;
    else
      rvec[j] = rvec[j+1] = INTERIOR;
  rvec[9] = INTERIOR;

  return((const S_CHAR *) rvec);
}

static const int *get_int_vec3(const EL *el, const DOF_INT_VEC *vec,
			       int *ivec)
{
  FUNCNAME("get_int_vec3");
  int            i, j, n0;
  static int     local_vec[N_BAS3];
  int            *v = nil, *rvec = ivec ? ivec : local_vec;
  DOF            **dof = el->dof;

  GET_DOF_VEC(v, vec);

  n0 = vec->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = v[dof[i][n0]];
  n0 = vec->fe_space->admin->n0_dof[EDGE];
  for (i = 0, j = N_VERTICES; i < N_EDGES; i++)
  {
    if (dof[vertex_of_edge[i][0]][0] < dof[vertex_of_edge[i][1]][0])
    {
      rvec[j++] = v[dof[N_VERTICES+i][n0]];
      rvec[j++] = v[dof[N_VERTICES+i][n0+1]];
    }
    else
    {
      rvec[j++] = v[dof[N_VERTICES+i][n0+1]];
      rvec[j++] = v[dof[N_VERTICES+i][n0]];
    }
  }
  n0 = vec->fe_space->admin->n0_dof[CENTER];
  rvec[9] = v[dof[6][n0]];

  return((const int *) rvec);
}

static const REAL *get_real_vec3(const EL *el, const DOF_REAL_VEC *vec,
				 REAL *Rvec)
{
  FUNCNAME("get_real_vec3");
  int            i, j, n0;
  static REAL    local_vec[N_BAS3];
  REAL           *v = nil, *rvec = Rvec ? Rvec : local_vec;
  DOF            **dof = el->dof;

  GET_DOF_VEC(v, vec);

  n0 = vec->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = v[dof[i][n0]];
  n0 = vec->fe_space->admin->n0_dof[EDGE];
  for (i = 0, j = N_VERTICES; i < N_EDGES; i++)
  {
    if (dof[vertex_of_edge[i][0]][0] < dof[vertex_of_edge[i][1]][0])
    {
      rvec[j++] = v[dof[N_VERTICES+i][n0]];
      rvec[j++] = v[dof[N_VERTICES+i][n0+1]];
    }
    else
    {
      rvec[j++] = v[dof[N_VERTICES+i][n0+1]];
      rvec[j++] = v[dof[N_VERTICES+i][n0]];
    }
  }
  n0 = vec->fe_space->admin->n0_dof[CENTER];
  rvec[9] = v[dof[6][n0]];

  return((const REAL *) rvec);
}

static const REAL_D *get_real_d_vec3(const EL *el, const DOF_REAL_D_VEC *vec,
				     REAL_D *Rvec)
{
  FUNCNAME("get_real_d_vec3");
  int            i, j, k, n0;
  static REAL_D  local_vec[N_BAS3];
  REAL_D         *v = nil, *rvec = Rvec ? Rvec : local_vec;
  DOF            **dof = el->dof;

  GET_DOF_VEC(v, vec);

  n0 = vec->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    for (k = 0; k < DIM_OF_WORLD; k++)
      rvec[i][k] = v[dof[i][n0]][k];

  n0 = vec->fe_space->admin->n0_dof[EDGE];
  for (i = 0, j = N_VERTICES; i < N_EDGES; i++)
  {
    if (dof[vertex_of_edge[i][0]][0] < dof[vertex_of_edge[i][1]][0])
    {
      for (k = 0; k < DIM_OF_WORLD; k++)
      {
	rvec[j][k] = v[dof[N_VERTICES+i][n0]][k];
	rvec[j+1][k] = v[dof[N_VERTICES+i][n0+1]][k];
      }
      j += 2;
    }
    else
    {
      for (k = 0; k < DIM_OF_WORLD; k++)
      {
	rvec[j][k] = v[dof[N_VERTICES+i][n0+1]][k];
	rvec[j+1][k] = v[dof[N_VERTICES+i][n0]][k];
      }
      j += 2;
    }
  }
  n0 = vec->fe_space->admin->n0_dof[CENTER];
  for (k = 0; k < DIM_OF_WORLD; k++)
    rvec[9][k] = v[dof[6][n0]][k];

  return((const REAL_D *) rvec);
}

static const U_CHAR *get_uchar_vec3(const EL *el, const DOF_UCHAR_VEC *vec,
				    U_CHAR *uvec)
{
  FUNCNAME("get_uchar_vec3");
  int            i, j, n0;
  static U_CHAR  local_vec[N_BAS3];
  U_CHAR         *v = nil, *rvec = uvec ? uvec : local_vec;
  DOF            **dof = el->dof;

  GET_DOF_VEC(v, vec);

  n0 = vec->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = v[dof[i][n0]];
  n0 = vec->fe_space->admin->n0_dof[EDGE];
  for (i = 0, j = N_VERTICES; i < N_EDGES; i++)
  {
    if (dof[vertex_of_edge[i][0]][0] < dof[vertex_of_edge[i][1]][0])
    {
      rvec[j++] = v[dof[N_VERTICES+i][n0]];
      rvec[j++] = v[dof[N_VERTICES+i][n0+1]];
    }
    else
    {
      rvec[j++] = v[dof[N_VERTICES+i][n0+1]];
      rvec[j++] = v[dof[N_VERTICES+i][n0]];
    }
  }
  n0 = vec->fe_space->admin->n0_dof[CENTER];
  rvec[9] = v[dof[6][n0]];

  return((const U_CHAR *) rvec);
}

static const S_CHAR *get_schar_vec3(const EL *el, const DOF_SCHAR_VEC *vec,
				    S_CHAR *svec)
{
  FUNCNAME("get_schar_vec3");
  int            i, j, n0;
  static S_CHAR  local_vec[N_BAS3];
  S_CHAR         *v = nil, *rvec = svec ? svec : local_vec;
  DOF            **dof = el->dof;

  GET_DOF_VEC(v, vec);

  n0 = vec->fe_space->admin->n0_dof[VERTEX];
  for (i = 0; i < N_VERTICES; i++)
    rvec[i] = v[dof[i][n0]];
  n0 = vec->fe_space->admin->n0_dof[EDGE];
  for (i = 0, j = N_VERTICES; i < N_EDGES; i++)
  {
    if (dof[vertex_of_edge[i][0]][0] < dof[vertex_of_edge[i][1]][0])
    {
      rvec[j++] = v[dof[N_VERTICES+i][n0]];
      rvec[j++] = v[dof[N_VERTICES+i][n0+1]];
    }
    else
    {
      rvec[j++] = v[dof[N_VERTICES+i][n0+1]];
      rvec[j++] = v[dof[N_VERTICES+i][n0]];
    }
  }
  n0 = vec->fe_space->admin->n0_dof[CENTER];
  rvec[9] = v[dof[6][n0]];

  return((const S_CHAR *) rvec);
}

static const REAL *interpol3(const EL_INFO *el_info, int no, 
			     const int *b_no, REAL (*f)(const REAL_D),
			     REAL *vec)
{
  FUNCNAME("interpol3");
  static REAL  inter[N_BAS3];
  REAL         *rvec = vec ? vec : inter;
  REAL_D       x;
  const REAL_D *coord;
  int          i;
  const PARAMETRIC *parametric = el_info->mesh->parametric;

  if (parametric) 
    ERROR_EXIT("Sorry, parametric support for lagrange3 not implemented yet!\n");

  TEST_FLAG(FILL_COORDS, el_info);
  
  coord = el_info->coord;
  if (b_no)
  {
    if (no <= 0  ||  no > N_BAS3)
    {
      ERROR("something is wrong, doing nothing\n");
      rvec[0] = 0.0;
      return((const REAL *) rvec);
    }

    for (i = 0; i < no; i++)
    {
      if (b_no[i] < N_VERTICES)
      {
	rvec[i] = f(coord[b_no[i]]);
      }
      else 
      {
	coord_to_world(el_info, bary3[b_no[i]], x);
	rvec[i] = f(x);
      }
    }
  }
  else
  {
    for (i = 0; i < N_VERTICES; i++)
      rvec[i] = f(coord[i]);
    for (i = N_VERTICES; i < N_BAS3; i++)
    {
      coord_to_world(el_info, bary3[i], x);
      rvec[i] = f(x);
    }
  }  
    
  return((const REAL *) rvec);
}

static const REAL_D *interpol_d3(const EL_INFO *el_info, int no, 
				 const int *b_no,
				 const REAL *(*f)(const REAL_D, REAL_D),
				 REAL_D *vec)
{
  FUNCNAME("interpol3");
  static REAL_D  inter[N_BAS3];
  REAL_D         *rvec = vec ? vec : inter;
  REAL_D         x;
  const REAL_D   *coord;
  int            i;
  const PARAMETRIC *parametric = el_info->mesh->parametric;

  if (parametric) 
    ERROR_EXIT("Sorry, parametric support for lagrange3 not implemented yet!\n");

  TEST_FLAG(FILL_COORDS, el_info);
  
  coord = el_info->coord;
  if (b_no)
  {
    if (no <= 0  ||  no > N_BAS3)
    {
      ERROR("something is wrong, doing nothing\n");
      for (i = 0; i < DIM_OF_WORLD; i++)
	rvec[0][i] = 0.0;
      return((const REAL_D *) rvec);
    }

    for (i = 0; i < no; i++)
    {
      if (b_no[i] < N_VERTICES)
      {
	(*f)(coord[b_no[i]], rvec[i]);
      }
      else 
      {
	coord_to_world(el_info, bary3[b_no[i]], x);
	(*f)(x, rvec[i]);
      }
    }
  }
  else
  {
    for (i = 0; i < N_VERTICES; i++)
      (*f)(coord[i], rvec[i]);
    for (i = N_VERTICES; i < N_BAS3; i++)
    {
      coord_to_world(el_info, bary3[i], x);
      (*f)(x, rvec[i]);
    }
  }  
    
  return((const REAL_D *) rvec);
}

/*--------------------------------------------------------------------------*/
/*  functions for interpolation/ restriction during refinement/coarsening   */
/*--------------------------------------------------------------------------*/

static void real_refine_inter3(DOF_REAL_VEC *drv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_refine_inter3");
  EL        *el;
  REAL      *v = nil;
  int       dof9, node, n0;
  const DOF *cdof;
  DOF       pdof[N_BAS3];
  const DOF       *(*get_dof_indices)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF_ADMIN *admin;

  if (n < 1) return;
  el = list->el;

  GET_DOF_VEC(v, drv);
  if (!drv->fe_space)
  {
    ERROR("no fe_space in dof_real_vec %s\n", NAME(drv));
    return;
  }
  else if (!drv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drv->fe_space));
    return;
  }
  get_dof_indices = drv->fe_space->bas_fcts->get_dof_indices;
  GET_STRUCT(admin,drv->fe_space);

  get_dof_indices(el, admin, pdof);
  
/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/
  cdof = get_dof_indices(el->child[0], admin, nil);

  v[cdof[2]] = (-0.0625*(v[pdof[0]] + v[pdof[1]]) 
		+ 0.5625*(v[pdof[7]] + v[pdof[8]]));
  v[cdof[3]] = (0.3125*(v[pdof[0]] - v[pdof[8]]) + 0.0625*v[pdof[1]]
		+ 0.9375*v[pdof[7]]);
  v[cdof[4]] =  v[pdof[7]];
  v[cdof[5]] =  v[pdof[9]];
  v[cdof[6]] =  (0.0625*(v[pdof[0]] + v[pdof[1]])
		 - 0.25*(v[pdof[3]] + v[pdof[6]])
		 + 0.5*(v[pdof[4]] + v[pdof[5]] + v[pdof[9]])
		 - 0.0625*(v[pdof[7]] + v[pdof[8]]));
  v[cdof[9]] = (0.0625*(-v[pdof[0]] + v[pdof[1]]) - 0.125*v[pdof[3]]
		+ 0.375*v[pdof[6]] + 0.1875*(v[pdof[7]] - v[pdof[8]])
		+ 0.75*v[pdof[9]]);

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

  cdof = get_dof_indices(el->child[1], admin, nil);

  v[cdof[5]] =  v[pdof[8]];
  v[cdof[6]] =  (0.0625*v[pdof[0]] + 0.9375*v[pdof[8]]
		 + 0.3125*(v[pdof[1]] - v[pdof[7]]));
  v[cdof[9]] =  (0.0625*(v[pdof[0]] - v[pdof[1]]) + 0.375*v[pdof[3]]
		 - 0.125*v[pdof[6]] + 0.1875*(-v[pdof[7]] + v[pdof[8]])
		 + 0.75*v[pdof[9]]);

  if (n <= 1)  return;

/*--------------------------------------------------------------------------*/
/*  adjust the value on the neihgbour                                       */
/*--------------------------------------------------------------------------*/

  el = list[1].el;
  get_dof_indices(el, admin, pdof);

/*--------------------------------------------------------------------------*/
/*  values on neigh's child[0]                                              */
/*--------------------------------------------------------------------------*/
  cdof = get_dof_indices(el->child[0], admin, nil);
  
  v[cdof[5]] =  v[pdof[9]];
  v[cdof[6]] =  (0.0625*(v[pdof[0]] + v[pdof[1]]) 
		 - 0.25*(v[pdof[3]] + v[pdof[6]])
		 + 0.5*(v[pdof[4]] + v[pdof[5]] + v[pdof[9]])
		 - 0.0625*(v[pdof[7]] + v[pdof[8]]));
  v[cdof[9]] = (0.0625*(-v[pdof[0]] + v[pdof[1]]) - 0.12500*v[pdof[3]]
		+ 0.375*v[pdof[6]] + 0.1875*(v[pdof[7]] - v[pdof[8]])
		+ 0.75*v[pdof[9]]);
/*--------------------------------------------------------------------------*/
/*  values on neigh's child[0]                                              */
/*--------------------------------------------------------------------------*/

  node = drv->fe_space->mesh->node[CENTER];  
  n0 = admin->n0_dof[CENTER];
  dof9 = el->child[1]->dof[node][n0];

  v[dof9] =  (0.0625*(v[pdof[0]] - v[pdof[1]]) +  0.375*v[pdof[3]]
	      - 0.125*v[pdof[6]] + 0.1875*(-v[pdof[7]] + v[pdof[8]])
	      + 0.75*v[pdof[9]]);
  return;
}

static void real_coarse_inter3(DOF_REAL_VEC *drv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_coarse_inter3");
  EL        *el, *child;
  REAL      *v = nil;
  int       cdof, pdof, node, n0;
  const DOF_ADMIN *admin;
  MESH      *mesh = nil;

  if (n < 1) return;
  el = list->el;

  GET_DOF_VEC(v, drv);
  if (!drv->fe_space)
  {
    ERROR("no fe_space in dof_real_vec %s\n", NAME(drv));
    return;
  }
  else if (!drv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drv->fe_space));
    return;
  }
  GET_STRUCT(admin,drv->fe_space);
  GET_STRUCT(mesh,drv->fe_space);

  node = mesh->node[EDGE];        
  n0 = admin->n0_dof[EDGE];

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

  child = el->child[0];

  if (el->dof[0][0] < el->dof[1][0])
    pdof = el->dof[node+2][n0];
  else
    pdof = el->dof[node+2][n0+1];

  if (child->dof[1][0] < child->dof[2][0])
    cdof = child->dof[node][n0+1];
  else
    cdof = child->dof[node][n0];
      
  v[pdof] = v[cdof];

  if (child->dof[2][0] < child->dof[0][0])
    cdof = child->dof[node+1][n0];
  else
    cdof = child->dof[node+1][n0+1];
      
  v[el->dof[mesh->node[CENTER]][admin->n0_dof[CENTER]]] = v[cdof];

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/
  
  child = el->child[1];

  if (el->dof[0][0] < el->dof[1][0])
    pdof = el->dof[node+2][n0+1];
  else
    pdof = el->dof[node+2][n0];

  if (child->dof[2][0] < child->dof[0][0])
    cdof = child->dof[node+1][n0];
  else
    cdof = child->dof[node+1][n0+1];

  v[pdof] = v[cdof];

  if (n <= 1)  return;

/*--------------------------------------------------------------------------*/
/*   adjust neighbour values                                                */
/*--------------------------------------------------------------------------*/

  el = list[1].el;
  child = el->child[0];

  if (child->dof[2][0] < child->dof[0][0])
    cdof = child->dof[node+1][n0];
  else
    cdof = child->dof[node+1][n0+1];
      
  v[el->dof[mesh->node[CENTER]][admin->n0_dof[CENTER]]] = v[cdof];

  return;
}

static void real_coarse_restr3(DOF_REAL_VEC *drv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_coarse_restr3");
  EL        *el;
  REAL      *v = nil;
  int       node, n0;
  const DOF *cdof;
  DOF       pdof[N_BAS3], dof9;
  const DOF       *(*get_dof_indices)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF_ADMIN *admin;

  if (n < 1) return;
  el = list->el;

  GET_DOF_VEC(v, drv);
  if (!drv->fe_space)
  {
    ERROR("no fe_space in dof_real_vec %s\n", NAME(drv));
    return;
  }
  else if (!drv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drv->fe_space));
    return;
  }
  get_dof_indices = drv->fe_space->bas_fcts->get_dof_indices;
  GET_STRUCT(admin,drv->fe_space);

  get_dof_indices(el, admin, pdof);
  
/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/
  cdof = get_dof_indices(el->child[0], admin, nil);

  v[pdof[0]] += (0.0625*(-v[cdof[2]] + v[cdof[6]] - v[cdof[9]])
		 + 0.3125*v[cdof[3]]);
  v[pdof[1]] += 0.0625*(-v[cdof[2]] +  v[cdof[3]] + v[cdof[6]] + v[cdof[9]]);
  v[pdof[3]] += -0.25*v[cdof[6]] - 0.125*v[cdof[9]];
  v[pdof[4]] +=  0.5*v[cdof[6]];
  v[pdof[5]] +=  0.5*v[cdof[6]];
  v[pdof[6]] += -0.25*v[cdof[6]] + 0.375*v[cdof[9]];
  v[pdof[7]] = (0.5625*v[cdof[2]] + 0.9375*v[cdof[3]] + v[cdof[4]] 
		- 0.0625*v[cdof[6]] + 0.1875*v[cdof[9]]);
  v[pdof[8]] = (0.5625*v[cdof[2]] - 0.3125*v[cdof[3]] 
		- 0.0625*v[cdof[6]] - 0.1875*v[cdof[9]]);
  v[pdof[9]] =  v[cdof[5]] + 0.5*v[cdof[6]] + 0.75*v[cdof[9]];

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

  cdof = get_dof_indices(el->child[1], admin, nil);

  v[pdof[0]] += 0.0625*v[cdof[6]] + 0.0625*v[cdof[9]];
  v[pdof[1]] += 0.3125*v[cdof[6]] - 0.0625*v[cdof[9]];
  v[pdof[3]] += 0.375*v[cdof[9]];
  v[pdof[6]] += -0.125*v[cdof[9]];
  v[pdof[7]] += -0.3125*v[cdof[6]] - 0.18750*v[cdof[9]];
  v[pdof[8]] += v[cdof[5]] + 0.9375*v[cdof[6]] + 0.1875*v[cdof[9]];
  v[pdof[9]] += 0.75*v[cdof[9]];

  if (n <= 1)  return;

/*--------------------------------------------------------------------------*/
/*  adjust the value on the neihgbour                                       */
/*--------------------------------------------------------------------------*/

  el = list[1].el;
  get_dof_indices(el, admin, pdof);

/*--------------------------------------------------------------------------*/
/*  values on neigh's child[0]                                              */
/*--------------------------------------------------------------------------*/
  cdof = get_dof_indices(el->child[0], admin, nil);
  
  v[pdof[0]] += 0.0625*(v[cdof[6]] - v[cdof[9]]);
  v[pdof[1]] += 0.0625*(v[cdof[6]] + v[cdof[9]]);
  v[pdof[3]] += -0.25*v[cdof[6]] - 0.12500*v[cdof[9]];
  v[pdof[4]] += 0.5*v[cdof[6]];
  v[pdof[5]] += 0.5*v[cdof[6]];
  v[pdof[6]] += -0.25*v[cdof[6]] + 0.375*v[cdof[9]];
  v[pdof[7]] += -0.0625*v[cdof[6]] + 0.1875*v[cdof[9]];
  v[pdof[8]] += -0.0625*v[cdof[6]] - 0.1875*v[cdof[9]];
  v[pdof[9]] =  v[cdof[5]] + 0.5*v[cdof[6]] + 0.75*v[cdof[9]];

/*--------------------------------------------------------------------------*/
/*  values on neigh's child[1]                                              */
/*--------------------------------------------------------------------------*/

  node = drv->fe_space->mesh->node[CENTER];  
  n0 = admin->n0_dof[CENTER];
  dof9 = el->child[1]->dof[node][n0];

  v[pdof[0]] += 0.0625*v[dof9];
  v[pdof[1]] -= 0.0625*v[dof9];
  v[pdof[3]] += 0.375*v[dof9];
  v[pdof[6]] -= 0.125*v[dof9];
  v[pdof[7]] -= 0.1875*v[dof9];
  v[pdof[8]] += 0.1875*v[dof9];
  v[pdof[9]] += 0.75*v[dof9];

  return;
}

static void real_d_refine_inter3(DOF_REAL_D_VEC *drdv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_d_refine_inter3");
  EL        *el;
  REAL_D    *v = nil;
  int       node, n0, j;
  const DOF *cdof;
  DOF       pdof[N_BAS3], dof9;
  const DOF       *(*get_dof_indices)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF_ADMIN *admin;

  if (n < 1) return;
  el = list->el;

  GET_DOF_VEC(v, drdv);
  if (!drdv->fe_space)
  {
    ERROR("no fe_space in dof_real_d_vec %s\n", NAME(drdv));
    return;
  }
  else if (!drdv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drdv->fe_space));
    return;
  }
  get_dof_indices = drdv->fe_space->bas_fcts->get_dof_indices;
  GET_STRUCT(admin,drdv->fe_space);

  get_dof_indices(el, admin, pdof);
  
/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/
  cdof = get_dof_indices(el->child[0], admin, nil);

  for (j = 0; j < DIM_OF_WORLD; j++)
  {
    v[cdof[2]][j] = (-0.0625*(v[pdof[0]][j] + v[pdof[1]][j]) 
		     + 0.5625*(v[pdof[7]][j] + v[pdof[8]][j]));
    v[cdof[3]][j] = (0.3125*(v[pdof[0]][j] - v[pdof[8]][j])
		     + 0.0625*v[pdof[1]][j] + 0.9375*v[pdof[7]][j]);
    v[cdof[4]][j] =  v[pdof[7]][j];
    v[cdof[5]][j] =  v[pdof[9]][j];
    v[cdof[6]][j] =  (0.0625*(v[pdof[0]][j] + v[pdof[1]][j])
		      - 0.25*(v[pdof[3]][j] + v[pdof[6]][j])
		      + 0.5*(v[pdof[4]][j] + v[pdof[5]][j] + v[pdof[9]][j])
		      - 0.0625*(v[pdof[7]][j] + v[pdof[8]][j]));
    v[cdof[9]][j] = (0.0625*(-v[pdof[0]][j] + v[pdof[1]][j])
		     - 0.125*v[pdof[3]][j] + 0.375*v[pdof[6]][j]
		     + 0.1875*(v[pdof[7]][j] - v[pdof[8]][j])
		     + 0.75*v[pdof[9]][j]);
  }

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

  cdof = get_dof_indices(el->child[1], admin, nil);

  for (j = 0; j < DIM_OF_WORLD; j++)
  {
    v[cdof[5]][j] =  v[pdof[8]][j];
    v[cdof[6]][j] =  (0.0625*v[pdof[0]][j] + 0.9375*v[pdof[8]][j]
		      + 0.3125*(v[pdof[1]][j] - v[pdof[7]][j]));
    v[cdof[9]][j] =  (0.0625*(v[pdof[0]][j] - v[pdof[1]][j]) 
		      + 0.375*v[pdof[3]][j] - 0.125*v[pdof[6]][j]
		      + 0.1875*(-v[pdof[7]][j] + v[pdof[8]][j])
		      + 0.75*v[pdof[9]][j]);
  }

  if (n <= 1)  return;

/*--------------------------------------------------------------------------*/
/*  adjust the value on the neihgbour                                       */
/*--------------------------------------------------------------------------*/

  el = list[1].el;
  get_dof_indices(el, admin, pdof);

/*--------------------------------------------------------------------------*/
/*  values on neigh's child[0]                                              */
/*--------------------------------------------------------------------------*/
  cdof = get_dof_indices(el->child[0], admin, nil);
  

  for (j = 0; j < DIM_OF_WORLD; j++)
  {
    v[cdof[5]][j] =  v[pdof[9]][j];
    v[cdof[6]][j] =  (0.0625*(v[pdof[0]][j] + v[pdof[1]][j]) 
		      - 0.25*(v[pdof[3]][j] + v[pdof[6]][j])
		      + 0.5*(v[pdof[4]][j] + v[pdof[5]][j] + v[pdof[9]][j])
		      - 0.0625*(v[pdof[7]][j] + v[pdof[8]][j]));
    v[cdof[9]][j] = (0.0625*(-v[pdof[0]][j] + v[pdof[1]][j])
		     - 0.12500*v[pdof[3]][j] + 0.375*v[pdof[6]][j]
		     + 0.1875*(v[pdof[7]][j] - v[pdof[8]][j])
		     + 0.75*v[pdof[9]][j]);
  }

/*--------------------------------------------------------------------------*/
/*  values on neigh's child[0]                                              */
/*--------------------------------------------------------------------------*/

  node = drdv->fe_space->mesh->node[CENTER];  
  n0 = admin->n0_dof[CENTER];

  for (j = 0; j < DIM_OF_WORLD; j++)
  {
    dof9 = el->child[1]->dof[node][n0];

    v[dof9][j] =  (0.0625*(v[pdof[0]][j] - v[pdof[1]][j])
		   +  0.375*v[pdof[3]][j] - 0.125*v[pdof[6]][j]
		   + 0.1875*(-v[pdof[7]][j] + v[pdof[8]][j])
		   + 0.75*v[pdof[9]][j]);
  }
  return;
}

static void real_d_coarse_inter3(DOF_REAL_D_VEC *drdv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_d_coarse_inter3");
  EL        *el, *child;
  REAL_D    *v = nil;
  int       cdof, pdof, node, n0, j;
  const DOF_ADMIN *admin;
  MESH      *mesh = nil;

  if (n < 1) return;
  el = list->el;

  GET_DOF_VEC(v, drdv);
  if (!drdv->fe_space)
  {
    ERROR("no fe_space in dof_real_d_vec %s\n", NAME(drdv));
    return;
  }
  else if (!drdv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drdv->fe_space));
    return;
  }
  GET_STRUCT(admin,drdv->fe_space);
  GET_STRUCT(mesh,drdv->fe_space);

  node = mesh->node[EDGE];        
  n0 = admin->n0_dof[EDGE];

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

  child = el->child[0];

  if (el->dof[0][0] < el->dof[1][0])
    pdof = el->dof[node+2][n0];
  else
    pdof = el->dof[node+2][n0+1];

  if (child->dof[1][0] < child->dof[2][0])
    cdof = child->dof[node][n0+1];
  else
    cdof = child->dof[node][n0];

  for (j = 0; j < DIM_OF_WORLD; j++)
    v[pdof][j] = v[cdof][j];

  if (child->dof[2][0] < child->dof[0][0])
    cdof = child->dof[node+1][n0];
  else
    cdof = child->dof[node+1][n0+1];
      
  for (j = 0; j < DIM_OF_WORLD; j++)
    v[el->dof[mesh->node[CENTER]][admin->n0_dof[CENTER]]][j] = v[cdof][j];

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/
  
  child = el->child[1];

  if (el->dof[0][0] < el->dof[1][0])
    pdof = el->dof[node+2][n0+1];
  else
    pdof = el->dof[node+2][n0];

  if (child->dof[2][0] < child->dof[0][0])
    cdof = child->dof[node+1][n0];
  else
    cdof = child->dof[node+1][n0+1];

  for (j = 0; j < DIM_OF_WORLD; j++)
    v[pdof][j] = v[cdof][j];

  if (n <= 1)  return;

/*--------------------------------------------------------------------------*/
/*   adjust neighbour values                                                */
/*--------------------------------------------------------------------------*/

  el = list[1].el;
  child = el->child[0];

  if (child->dof[2][0] < child->dof[0][0])
    cdof = child->dof[node+1][n0];
  else
    cdof = child->dof[node+1][n0+1];
      
  for (j = 0; j < DIM_OF_WORLD; j++)
    v[el->dof[mesh->node[CENTER]][admin->n0_dof[CENTER]]][j] = v[cdof][j];

  return;
}

static void real_d_coarse_restr3(DOF_REAL_D_VEC *drdv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_d_coarse_restr3");
  EL        *el;
  REAL_D    *v = nil;
  int       node, n0, j;
  const DOF *cdof;
  DOF       pdof[N_BAS3], dof9;
  const DOF       *(*get_dof_indices)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF_ADMIN *admin;

  if (n < 1) return;
  el = list->el;

  GET_DOF_VEC(v, drdv);
  if (!drdv->fe_space)
  {
    ERROR("no fe_space in dof_real_d_vec %s\n", NAME(drdv));
    return;
  }
  else if (!drdv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drdv->fe_space));
    return;
  }
  get_dof_indices = drdv->fe_space->bas_fcts->get_dof_indices;
  GET_STRUCT(admin,drdv->fe_space);

  get_dof_indices(el, admin, pdof);
  
/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/
  cdof = get_dof_indices(el->child[0], admin, nil);

  for (j = 0; j < DIM_OF_WORLD; j++)
  {
    v[pdof[0]][j] += (0.0625*(-v[cdof[2]][j] + v[cdof[6]][j] - v[cdof[9]][j])
		      + 0.3125*v[cdof[3]][j]);
    v[pdof[1]][j] += 0.0625*(-v[cdof[2]][j] +  v[cdof[3]][j] + v[cdof[6]][j]
			     + v[cdof[9]][j]);
    v[pdof[3]][j] += -0.25*v[cdof[6]][j] - 0.125*v[cdof[9]][j];
    v[pdof[4]][j] +=  0.5*v[cdof[6]][j];
    v[pdof[5]][j] +=  0.5*v[cdof[6]][j];
    v[pdof[6]][j] += -0.25*v[cdof[6]][j] + 0.375*v[cdof[9]][j];
    v[pdof[7]][j] = (0.5625*v[cdof[2]][j] + 0.9375*v[cdof[3]][j]
		     + v[cdof[4]][j] - 0.0625*v[cdof[6]][j]
		     + 0.1875*v[cdof[9]][j]);
    v[pdof[8]][j] = (0.5625*v[cdof[2]][j] - 0.3125*v[cdof[3]][j] 
		     - 0.0625*v[cdof[6]][j] - 0.1875*v[cdof[9]][j]);
    v[pdof[9]][j] =  v[cdof[5]][j] + 0.5*v[cdof[6]][j] + 0.75*v[cdof[9]][j];
  }

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

  cdof = get_dof_indices(el->child[1], admin, nil);

  for (j = 0; j < DIM_OF_WORLD; j++)
  {
    v[pdof[0]][j] += 0.0625*v[cdof[6]][j] + 0.0625*v[cdof[9]][j];
    v[pdof[1]][j] += 0.3125*v[cdof[6]][j] - 0.0625*v[cdof[9]][j];
    v[pdof[3]][j] += 0.375*v[cdof[9]][j];
    v[pdof[6]][j] += -0.125*v[cdof[9]][j];
    v[pdof[7]][j] += -0.3125*v[cdof[6]][j] - 0.18750*v[cdof[9]][j];
    v[pdof[8]][j] += (v[cdof[5]][j] + 0.9375*v[cdof[6]][j]
		      + 0.1875*v[cdof[9]][j]);
    v[pdof[9]][j] += 0.75*v[cdof[9]][j];
  }

  if (n <= 1)  return;

/*--------------------------------------------------------------------------*/
/*  adjust the value on the neihgbour                                       */
/*--------------------------------------------------------------------------*/

  el = list[1].el;
  get_dof_indices(el, admin, pdof);

/*--------------------------------------------------------------------------*/
/*  values on neigh's child[0]                                              */
/*--------------------------------------------------------------------------*/
  cdof = get_dof_indices(el->child[0], admin, nil);
  
  for (j = 0; j < DIM_OF_WORLD; j++)
  {
    v[pdof[0]][j] += 0.0625*(v[cdof[6]][j] - v[cdof[9]][j]);
    v[pdof[1]][j] += 0.0625*(v[cdof[6]][j] + v[cdof[9]][j]);
    v[pdof[3]][j] += -0.25*v[cdof[6]][j] - 0.12500*v[cdof[9]][j];
    v[pdof[4]][j] += 0.5*v[cdof[6]][j];
    v[pdof[5]][j] += 0.5*v[cdof[6]][j];
    v[pdof[6]][j] += -0.25*v[cdof[6]][j] + 0.375*v[cdof[9]][j];
    v[pdof[7]][j] += -0.0625*v[cdof[6]][j] + 0.1875*v[cdof[9]][j];
    v[pdof[8]][j] += -0.0625*v[cdof[6]][j] - 0.1875*v[cdof[9]][j];
    v[pdof[9]][j] =  v[cdof[5]][j] + 0.5*v[cdof[6]][j] + 0.75*v[cdof[9]][j];
  }

/*--------------------------------------------------------------------------*/
/*  values on neigh's child[1]                                              */
/*--------------------------------------------------------------------------*/

  node = drdv->fe_space->mesh->node[CENTER];  
  n0 = admin->n0_dof[CENTER];
  dof9 = el->child[1]->dof[node][n0];

  for (j = 0; j < DIM_OF_WORLD; j++)
  {
    v[pdof[0]][j] += 0.0625*v[dof9][j];
    v[pdof[1]][j] -= 0.0625*v[dof9][j];
    v[pdof[3]][j] += 0.375*v[dof9][j];
    v[pdof[6]][j] -= 0.125*v[dof9][j];
    v[pdof[7]][j] -= 0.1875*v[dof9][j];
    v[pdof[8]][j] += 0.1875*v[dof9][j];
    v[pdof[9]][j] += 0.75*v[dof9][j];
  }

  return;
}

static BAS_FCT      *phi3[N_BAS3]     = {phi3v0, phi3v1, phi3v2,
					 phi3e0, phi3e1, phi3e2,
					 phi3e3, phi3e4, phi3e5,
					 phi3c0};
static GRD_BAS_FCT  *grd_phi3[N_BAS3] = {grd_phi3v0, grd_phi3v1, 
					 grd_phi3v2, grd_phi3e0,
					 grd_phi3e1, grd_phi3e2,
					 grd_phi3e3, grd_phi3e4,
					 grd_phi3e5, grd_phi3c0};
static D2_BAS_FCT   *D2_phi3[N_BAS3]  = {D2_phi3v0, D2_phi3v1, 
					 D2_phi3v2, D2_phi3e0,
					 D2_phi3e1, D2_phi3e2,
					 D2_phi3e3, D2_phi3e4,
					 D2_phi3e5, D2_phi3c0};

static BAS_FCTS  lagrange3 = {"lagrange3", N_BAS3, 3, {1, 2, 1},
			      nil, phi3, grd_phi3, D2_phi3,
			      get_dof_indices3, 
			      get_bound3,
			      interpol3,
			      interpol_d3,
			      get_int_vec3,
			      get_real_vec3,
			      get_real_d_vec3,
			      get_uchar_vec3,
			      get_schar_vec3,
			      real_refine_inter3,
			      real_coarse_inter3,
			      real_coarse_restr3,
			      real_d_refine_inter3,
			      real_d_coarse_inter3,
			      real_d_coarse_restr3,
			      (void *) bary3};

