/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     refine.c                                                       */
/*                                                                          */
/*    !!!  includes ../Common/refine_common.c  !!!                          */
/*                                                                          */
/* description:  recursive refinement of 3 dim. hierarchical meshes;        */
/*               implementation of the newest vertex bisection              */
/*               file contains all routines depending on DIM == 3;          */
/*               routines for both dimensions in ../Common/refine_common.c  */
/*               are included at the end of the file                        */
/*                                                                          */
/* global variables that have to be defined for ../Common/refine_common.c   */
/*  static int            call_refine_interpol;                             */
/*  static int            do_more_refine;                                   */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include "alberta.h"

void free_dof(DOF *dof, MESH *mesh, int position);
void free_leaf_data(void *leaf_data, MESH *mesh);

static MESH          *refine_mesh;
static int           do_more_recursive_refine;
static int           new_coords;
static int           call_refine_interpol;

#if !NEIGH_IN_EL
static TRAVERSE_STACK *stack;
#endif

/*--------------------------------------------------------------------------*/
/*  sets neighbour connectivity inside the refinement/coarsening patch      */
/*--------------------------------------------------------------------------*/

void  get_neigh_on_patch(RC_LIST_EL ref_list[], int n_neigh, int bound)
{
  FUNCNAME("get_neigh_on_patch");
  int      i, j, k, dir;
  EL       *el, *neigh;

  for (i = 0; i < n_neigh; i++)
  {
    el = ref_list[i].el;
    ref_list[i].no = i;

    for (dir = 0; dir < 2; dir++)
    {
      for (j = 0; j < n_neigh; j++)
      {
	if ((neigh = ref_list[j].el) == el)  continue;
    
	for (k = 0; k < 2; k++)
	{
	  if (neigh->dof[2+k] == el->dof[3-dir])
	  {
	    ref_list[i].neigh[dir] = ref_list+j;
	    ref_list[i].opp_vertex[dir] = 3-k;
	    break;
	  }
	}

	if (k < 2) break;
      }

      if (j >= n_neigh)
      {
	TEST_EXIT(bound)
	  ("neighbour of element %d in list not found\n", INDEX(el));
	ref_list[i].neigh[dir] = nil;
	ref_list[i].opp_vertex[dir] = -1;
      }
    }
  }
  return;
}

/*--------------------------------------------------------------------------*/
/*  edge_of_dofs[i][j]: gives the local index of edge with vertices i and j */
/*--------------------------------------------------------------------------*/

U_CHAR edge_of_dofs[4][4] = {{255,0,1,2},{0,255,3,4},{1,3,255,5},{2,4,5,255}};

/*--------------------------------------------------------------------------*/
/*  n_child_edge[el_type][ichild][dir]                                      */
/*  gives local index of new edge on child[ichild] part of face [2+dir] on  */
/*  the parent                                                              */
/*--------------------------------------------------------------------------*/
static U_CHAR  n_child_edge[3][2][2] = {{{5,4},{4,5}},{{5,4},{5,4}},
					{{5,4},{5,4}}};

/*--------------------------------------------------------------------------*/
/*  n_child_face[el_type][ichild][dir]                                      */
/*  gives local index of sub-face on child[ichild] part of face [2+dir] on  */
/*  the parent                                                              */
/*--------------------------------------------------------------------------*/
static U_CHAR  n_child_face[3][2][2] = {{{1,2},{2,1}},{{1,2},{1,2}},
					{{1,2},{1,2}}};

/*--------------------------------------------------------------------------*/
/*  adjacent_child[position][ichild]                                        */
/*  gives number of the adjacent child on a neighbour element               */
/*     position = 0  same position of element and neigh at refinement edge  */
/*     position = 1  different ...                                          */
/*--------------------------------------------------------------------------*/
static U_CHAR  adjacent_child[2][2] = {{0,1}, {1,0}};

static void  fill_patch_connectivity(RC_LIST_EL *ref_list)
{
  FUNCNAME("fill_patch_connectivity");
  EL      *el = ref_list->el, *neigh;
  int     dir, el_type = ref_list->el_type, n_type = 0;
  int     adjc, i, j, i_neigh, j_neigh;
  int     node0, node1, opp_v = 0;

  for (dir = 0; dir < 2; dir++)
  {
    if (ref_list->neigh[dir])
    {
      neigh = ref_list->neigh[dir]->el;
      n_type = ref_list->neigh[dir]->el_type;
      opp_v = ref_list->opp_vertex[dir];
    }
    else
      neigh = nil;


    if (!neigh  ||  !neigh->child[0])
    {

#if NEIGH_IN_EL
/*--------------------------------------------------------------------------*/
/*  fill neighbour information (in the case that neighbour isn't refined    */
/*  set pointers to nil. Information is produced later on)                  */
/*--------------------------------------------------------------------------*/

      for (i = 0; i < 2; i++)
      {
	el->child[i]->neigh[n_child_face[el_type][i][dir]] = nil;
	el->child[i]->opp_vertex[n_child_face[el_type][i][dir]] = -1;
      }
#endif

/*--------------------------------------------------------------------------*/
/*  get new dof's in the midedge of the face of el and for the two midpoints*/
/*  of the sub-faces. If face is an interior face those pointers have to be */
/*  adjusted by the neighbour element also (see below)                      */
/*--------------------------------------------------------------------------*/

      if (refine_mesh->n_dof[EDGE])
      {
	node0 = node1 = refine_mesh->node[EDGE];
	node0 += n_child_edge[el_type][0][dir];
	node1 += n_child_edge[el_type][1][dir];
	el->child[0]->dof[node0] = el->child[1]->dof[node1] = 
	  get_dof(refine_mesh, EDGE);
      }
      if (refine_mesh->n_dof[FACE])
      {
	node0 = refine_mesh->node[FACE] + n_child_face[el_type][0][dir];
	el->child[0]->dof[node0] = get_dof(refine_mesh, FACE);
	node1 = refine_mesh->node[FACE] + n_child_face[el_type][1][dir];
	el->child[1]->dof[node1] = get_dof(refine_mesh, FACE);
      }
    }
    else     /*   if (!neigh  ||  !neigh->child[0])                         */
    {
/*--------------------------------------------------------------------------*/
/*  interior face and neighbour has been refined, look for position at the  */
/*  refinement edge                                                         */
/*--------------------------------------------------------------------------*/
      
      if (el->dof[0] == neigh->dof[0])
      {
/*--------------------------------------------------------------------------*/
/* same position at refinement edge                                         */
/*--------------------------------------------------------------------------*/
	adjc = 0;
      }
      else
      {
/*--------------------------------------------------------------------------*/
/* different position at refinement edge                                    */
/*--------------------------------------------------------------------------*/
	adjc = 1;
      }

      for (i = 0; i < 2; i++)
      {
	j = adjacent_child[adjc][i];

	i_neigh = n_child_face[el_type][i][dir];
	j_neigh = n_child_face[n_type][j][opp_v-2];

#if NEIGH_IN_EL
/*--------------------------------------------------------------------------*/
/*  fill neighbour information                                              */
/*--------------------------------------------------------------------------*/

	el->child[i]->neigh[i_neigh] = neigh->child[j];
	neigh->child[j]->neigh[j_neigh] = el->child[i];
	el->child[i]->opp_vertex[i_neigh] = j_neigh;
	neigh->child[j]->opp_vertex[j_neigh] = i_neigh;
#endif

/*--------------------------------------------------------------------------*/
/*  adjust dof pointer in the edge in the common face of el and neigh and   */
/*  the dof pointer in the sub-face child_i-child_j (allocated by neigh!)   */
/*--------------------------------------------------------------------------*/

	if (refine_mesh->n_dof[EDGE])
	{
	  node0 = refine_mesh->node[EDGE] + n_child_edge[el_type][i][dir];
	  node1 = refine_mesh->node[EDGE] + n_child_edge[n_type][j][opp_v-2];

	  TEST_EXIT(neigh->child[j]->dof[node1])
	    ("no dof on neighbour %d at node %d\n",
	     INDEX(neigh->child[j]), node1);

	  el->child[i]->dof[node0] = neigh->child[j]->dof[node1];
	}
	if (refine_mesh->n_dof[FACE])
	{
	  node0 = refine_mesh->node[FACE] + i_neigh;
	  node1 = refine_mesh->node[FACE] + j_neigh;

	  TEST_EXIT(neigh->child[j]->dof[node1])
	    ("no dof on neighbour %d at node %d\n",
	     INDEX(neigh->child[j]), node1);

	  el->child[i]->dof[node0] = neigh->child[j]->dof[node1];
	}

      }  /*   for (i = 0; i < 2; i++)                                       */
    }    /*   else of   if (!neigh  ||  !neigh->child[0])                   */
  }      /*   for (dir = 0; dir < 2; dir++)                                 */
  
  return;
}

static void  bisect_element(RC_LIST_EL *ref_list, DOF *dof[3], DOF *edge[2])
{
  EL        *el = ref_list->el, *child[2];
  int       i, node;
  int       el_type = ref_list->el_type;

  child[0] = get_element(refine_mesh);
  child[1] = get_element(refine_mesh);
  

  child[0]->mark = child[1]->mark = MAX(0, el->mark-1);
  el->mark = 0;
#if NEIGH_IN_EL
  child[0]->el_type = child[1]->el_type = ((int)(el->el_type + 1)) % 3;
#endif

/*--------------------------------------------------------------------------*/
/*  transfer hidden data from parent to children                            */
/*--------------------------------------------------------------------------*/

  if (el->child[1]  &&  refine_mesh->leaf_data_info->refine_leaf_data)
    refine_mesh->leaf_data_info->refine_leaf_data(el, child);
  free_leaf_data((void *) el->child[1], refine_mesh);

  el->child[0] = child[0];
  el->child[1] = child[1];

  if (child[0]->mark > 0)  do_more_recursive_refine = true;

  child[0]->dof[N_VERTICES-1] = child[1]->dof[N_VERTICES-1] = dof[0];
  for (i = 0; i < N_VERTICES-1; i++)
  {
    child[0]->dof[i] = el->dof[child_vertex[el_type][0][i]];
    child[1]->dof[i] = el->dof[child_vertex[el_type][1][i]];
  }
/*--------------------------------------------------------------------------*/
/*  there is one more leaf element and two more hierachical elements        */
/*--------------------------------------------------------------------------*/

  refine_mesh->n_elements++;
  refine_mesh->n_hier_elements +=2;

/*--------------------------------------------------------------------------*/
/* first set those dof pointers for higher order without neighbour          */
/* information                                                              */
/*--------------------------------------------------------------------------*/

  if (refine_mesh->n_dof[EDGE])
  {
    node = refine_mesh->node[EDGE];

/*--------------------------------------------------------------------------*/
/*  set pointers to those dof's that are handed on from the parant          */
/*--------------------------------------------------------------------------*/

    child[0]->dof[node] = el->dof[node+child_edge[el_type][0][0]];
    child[1]->dof[node] = el->dof[node+child_edge[el_type][1][0]];
    child[0]->dof[node+1] = el->dof[node+child_edge[el_type][0][1]];
    child[1]->dof[node+1] = el->dof[node+child_edge[el_type][1][1]];
    child[0]->dof[node+3] = el->dof[node+child_edge[el_type][0][3]];
    child[1]->dof[node+3] = el->dof[node+child_edge[el_type][1][3]];

/*--------------------------------------------------------------------------*/
/*  adjust pointers to the dof's in the refinement edge                     */
/*--------------------------------------------------------------------------*/

    if (el->dof[0] == edge[0])
    {
      child[0]->dof[node+2] = dof[1];
      child[1]->dof[node+2] = dof[2];
    }
    else
    {
      child[0]->dof[node+2] = dof[2];
      child[1]->dof[node+2] = dof[1];
    }
  }

  if (refine_mesh->n_dof[FACE])
  {
    node = refine_mesh->node[FACE];    

/*--------------------------------------------------------------------------*/
/*  set pointers to those dof's that are handed on from the parant          */
/*--------------------------------------------------------------------------*/

    child[0]->dof[node+3] = el->dof[node+1];
    child[1]->dof[node+3] = el->dof[node+0];

/*--------------------------------------------------------------------------*/
/*  get new dof for the common face of child0 and child1                    */
/*--------------------------------------------------------------------------*/

    child[0]->dof[node] = child[1]->dof[node] = get_dof(refine_mesh, FACE);
  }
  
  if (refine_mesh->n_dof[CENTER])
  {
    node = refine_mesh->node[CENTER];
    child[0]->dof[node] = get_dof(refine_mesh, CENTER);
    child[1]->dof[node] = get_dof(refine_mesh, CENTER);
  }

#if NEIGH_IN_EL
/*--------------------------------------------------------------------------*/
/*  fill information about neighbours at the conforming faces of children   */
/*--------------------------------------------------------------------------*/
  {
    EL  *neigh;
    int ov;

    child[0]->neigh[0] = child[1];
    child[0]->opp_vertex[0] = 0;
    child[0]->neigh[3] = el->neigh[1];
    if ((neigh = child[0]->neigh[3]))
    {
      ov = child[0]->opp_vertex[3] = el->opp_vertex[1];
      neigh->neigh[ov] = child[0];
      neigh->opp_vertex[ov] = 3;
    }

    child[1]->neigh[0] = child[0];
    child[1]->opp_vertex[0] = 0;
    child[1]->neigh[3] = el->neigh[0];
    if ((neigh = child[1]->neigh[3]))
    {
      ov = child[1]->opp_vertex[3] = el->opp_vertex[0];
      neigh->neigh[ov] = child[1];
      neigh->opp_vertex[ov] = 3;
    }

    child[0]->neigh[1] = nil;
    child[0]->neigh[2] = nil;
    child[1]->neigh[1] = nil;
    child[1]->neigh[2] = nil;
  }
#endif

  if (NEIGH_IN_EL  || refine_mesh->n_dof[EDGE]  || refine_mesh->n_dof[FACE])
    fill_patch_connectivity(ref_list);

  return;
}

/*--------------------------------------------------------------------------*/
/*  remove dof pointers from parent                                         */
/*--------------------------------------------------------------------------*/

static void remove_dof_parent(RC_LIST_EL *ref_list)
{
  EL         *el = ref_list->el;
  RC_LIST_EL *neigh;
  int        j, node;

  if (refine_mesh->n_dof[EDGE])
  {
    node = refine_mesh->node[EDGE];
    for (j = 0; j < N_EDGES; j++)
      el->dof[node+j] = nil;
  }

  if (refine_mesh->n_dof[FACE])
  {
    node = refine_mesh->node[FACE];
    
    neigh = ref_list->neigh[0];
    if ((!neigh) || (neigh > ref_list))
      free_dof(el->dof[node+2], refine_mesh, FACE);           /* face  2 */

    neigh = ref_list->neigh[1];
    if ((!neigh) || (neigh > ref_list))
      free_dof(el->dof[node+3], refine_mesh, FACE);           /* face  3 */

    for (j = 0; j < N_FACES; j++)
      el->dof[node+j] = nil;
  }
  
  if (refine_mesh->n_dof[CENTER])
  {
    node = refine_mesh->node[CENTER];
    free_dof(el->dof[node], refine_mesh, CENTER);
    el->dof[node] = nil;
  }
  return;
}

static void  refine_interpol(RC_LIST_EL *list, int n_el);

static void  bisect_patch(DOF *edge[2], RC_LIST_EL ref_list[], int n_neigh, 
			  int bound)
{
  int    i;
  EL     *el = ref_list->el;  /*    first element in the list               */
  DOF    *dof[3] = {nil, nil, nil};

/*--------------------------------------------------------------------------*/
/*  get new dof's in the refinement edge                                    */
/*--------------------------------------------------------------------------*/

  dof[0] = get_dof(refine_mesh, VERTEX);
  refine_mesh->n_vertices++;
  
  if (refine_mesh->n_dof[EDGE])
  {
    dof[1] = get_dof(refine_mesh, EDGE);
    dof[2] = get_dof(refine_mesh, EDGE);
  }

  for (i = 0; i < n_neigh; i++)
    bisect_element(ref_list+i, dof, edge);


/*--------------------------------------------------------------------------*/
/*  if there are functions to interpolate data to the finer grid, do so     */
/*--------------------------------------------------------------------------*/
  if (call_refine_interpol)
    refine_interpol(ref_list, n_neigh);

  if (!refine_mesh->preserve_coarse_dofs)
  {
/*--------------------------------------------------------------------------*/
/*  if there should be no dof information on interior leaf elements remove  */
/*  dofs from edges, faces and the centers of parents                       */
/*--------------------------------------------------------------------------*/
    int   *n_dof = refine_mesh->n_dof;
    if (n_dof[EDGE])
    {
/*--------------------------------------------------------------------------*/
/*  remove dof of the midpoint of the common refinement edge                */
/*--------------------------------------------------------------------------*/

      el = ref_list->el;
      free_dof(el->dof[refine_mesh->node[EDGE]], refine_mesh, EDGE);
    }

    if (n_dof[EDGE]  ||  n_dof[FACE]  ||  n_dof[CENTER])
    {
      for (i = 0; i < n_neigh; i++)
	remove_dof_parent(ref_list+i);
    }
  }


/*--------------------------------------------------------------------------*/
/*  update the number of edges and faces; depends whether refinement edge   */
/*  is a boundary edge or not                                               */
/*--------------------------------------------------------------------------*/

  if (bound)
  {
    refine_mesh->n_edges += n_neigh + 2;
    refine_mesh->n_faces += 2*n_neigh + 1;
    new_coords++;
  }
  else
  {
    refine_mesh->n_edges += n_neigh + 1;
    refine_mesh->n_faces += 2*n_neigh;
  }

  return;
}

static const EL_INFO  *refine_function(const EL_INFO *el_info); 

/*--------------------------------------------------------------------------*/
/*  get_refine_patch:                                                       */
/*  gets the elements around the refinement edge with vertices  node[0] and */
/*  node[1] ; refines those elements at this edge are not compatible        */
/*  devisible;  			                                    */
/*  dir determines the direction of the first neighbour			    */
/*  get_refine_patch returns  1  if the domain's boundary is reached while  */
/*  looping around the refinement edge, otherwise  0                        */
/*--------------------------------------------------------------------------*/

static int  get_refine_patch(EL_INFO **el_info, DOF *edge[2], int dir,
			     RC_LIST_EL ref_list[], int *n_neigh)
{
  FUNCNAME("get_refine_patch");
#if NEIGH_IN_EL
  const EL_INFO neigh_info[1];
#else
  const EL_INFO *neigh_info;
#endif
  EL        *el, *neigh;
  int       i, j, k, opp_v;
  int       edge_no, neigh_el_type;

  el = (*el_info)->el;
  
  if (NEIGH(el,(*el_info))[3-dir] == nil)  {return(1);}
  
#if NEIGH_IN_EL
  opp_v = el->opp_vertex[3-dir];
  neigh_info->el = el->neigh[3-dir];
  neigh_el_type = neigh_info->el->el_type;
#else
  opp_v = (*el_info)->opp_vertex[3-dir];
  neigh_info = traverse_neighbour(stack, *el_info, 3-dir);
  neigh_el_type = neigh_info->el_type;
#endif
  neigh = neigh_info->el;
  
  while (neigh != el)
  {
    for (j = 0; j < N_VERTICES; j++)
      if (neigh->dof[j] == edge[0])  break;
    for (k = 0; k < N_VERTICES; k++)
      if (neigh->dof[k] == edge[1])  break;

    TEST_EXIT(j < N_VERTICES  &&  k < N_VERTICES)
      ("dof %d or dof %d not found on element %d with nodes (%d %d %d %d)\n", 
       edge[0][0], edge[1][0], INDEX(neigh), neigh->dof[0][0],
       neigh->dof[1][0], neigh->dof[2][0], neigh->dof[3][0]);

    if ((edge_no = edge_of_dofs[j][k]))
    {
/*--------------------------------------------------------------------------*/
/*  neigh has not a compatible refinement edge                              */
/*--------------------------------------------------------------------------*/
      neigh->mark = MAX(neigh->mark, 1);
#if NEIGH_IN_EL
      refine_function(neigh_info);
#else
      neigh_info = refine_function(neigh_info);
#endif
/*--------------------------------------------------------------------------*/
/*  now, go to a child at the edge and determine the opposite vertex for    */
/*  this child; continue the looping around the edge with this element      */
/*--------------------------------------------------------------------------*/

#if NEIGH_IN_EL
      neigh_info->el = neigh->child[0];
      neigh_el_type = neigh_info->el->el_type;
#else
      neigh_info = traverse_next(stack, neigh_info);
      neigh_el_type = neigh_info->el_type;
#endif
      switch (edge_no)
      {
      case 1: 
	opp_v = opp_v == 1 ? 3 : 2;
	break;
      case 2: 
	opp_v = opp_v == 2 ? 1 : 3;
	break;
      case 3: 
#if NEIGH_IN_EL
	neigh_info->el = neigh->child[1];
	neigh_el_type = neigh_info->el->el_type;
#else
	neigh_info = traverse_next(stack, neigh_info);
	neigh_el_type = neigh_info->el_type;
#endif
	if (neigh_el_type != 1)
	  opp_v = opp_v == 0 ? 3 : 2;
	else
	  opp_v = opp_v == 0 ? 3 : 1;
	break;
      case 4:
#if NEIGH_IN_EL
	neigh_info->el = neigh->child[1];
	neigh_el_type = neigh_info->el->el_type;
#else
	neigh_info = traverse_next(stack, neigh_info);
	neigh_el_type = neigh_info->el_type;
#endif
	if (neigh_el_type != 1)
	  opp_v = opp_v == 0 ? 3 : 1;
	else
	  opp_v = opp_v == 0 ? 3 : 2;
	break;
      case 5:
	if (neigh_el_type != 1)
	{
#if NEIGH_IN_EL
	  neigh_info->el = neigh->child[1];
	  neigh_el_type = neigh_info->el->el_type;
#else
	  neigh_info = traverse_next(stack, neigh_info);
	  neigh_el_type = neigh_info->el_type;
#endif
	}
	opp_v = 3;
	break;
      }
      neigh = neigh_info->el;
    }
    else
    {
/*--------------------------------------------------------------------------*/
/*  neigh is compatible devisible; put neigh to the list of patch elements; */
/*  go to next neighbour                                                    */
/*--------------------------------------------------------------------------*/
      TEST_EXIT(*n_neigh < refine_mesh->max_edge_neigh)
	("too many neighbours %d in refine patch\n", *n_neigh);

      
      ref_list[*n_neigh].el = neigh;
      ref_list[*n_neigh].el_type = neigh_el_type;
#if !NEIGH_IN_EL
/*--------------------------------------------------------------------------*/
/*  we have to go back to the starting element via opp_v values             */
/*  correct information is produce by get_neigh_on_patch()                  */
/*--------------------------------------------------------------------------*/
      ref_list[*n_neigh].opp_vertex[0] = opp_v; 
#endif
      ++*n_neigh;

      if (opp_v != 3)
	i = 3;
      else
	i = 2;

      if (NEIGH(neigh, neigh_info)[i])
      {
	opp_v = OPP_VERTEX(neigh,neigh_info)[i];
#if NEIGH_IN_EL
	neigh_info->el = neigh->neigh[i];
	neigh_el_type = neigh_info->el->el_type;
#else
	neigh_info = traverse_neighbour(stack, neigh_info, i);
	neigh_el_type = neigh_info->el_type;
#endif
	neigh = neigh_info->el;
      }
      else
	break;
    }
  }

  if (neigh == el)
  {
#if !NEIGH_IN_EL
    *el_info = (EL_INFO *) neigh_info;
#endif
    return(0);
  }

#if ! NEIGH_IN_EL
/*--------------------------------------------------------------------------*/
/*  the domain's boundary is reached; loop back to the starting el          */
/*--------------------------------------------------------------------------*/

  i = *n_neigh-1;
  opp_v = ref_list[i].opp_vertex[0];
  do
  {
    TEST_EXIT(NEIGH(neigh,neigh_info)[opp_v]  &&  i > 0)
      ("while looping back domains boundary was reached or i == 0\n");
    opp_v = ref_list[i--].opp_vertex[0];
    neigh_info = traverse_neighbour(stack, neigh_info, opp_v);
  } while(neigh_info->el != el);
  *el_info = (EL_INFO *) neigh_info;
#endif

  return(1);
}

#if NEIGH_IN_EL == 100
/*--------------------------------------------------------------------------*/
/* set node->bound values for newly created nodes                           */
/*--------------------------------------------------------------------------*/

static void refine_set_node_fct(const EL_INFO *el_info)
{
  FUNCNAME("refine_set_node_fct");
  EL      *el = el_info->el;
  NODE    *node;

  if (el->child[0])
  {
    node = el->child[0]->node[3];
    if (node->bound == BOUND_UNKNOWN)
    {
      if (el_info->boundary[N_FACES+0])
	node->bound = el_info->boundary[N_FACES+0]->bound;
      else
	node->bound = INTERIOR;
    }
  }
}
#endif

/*--------------------------------------------------------------------------*/
/* include common traversal routines for 2d and 3d refinement               */
/*--------------------------------------------------------------------------*/

#include "refine_common.c"
