/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     traverse_r.c                                                   */
/*                                                                          */
/* description:                                                             */
/*           recursive mesh traversal - 3d routines:                        */
/*           fill_macro_info(), fill_elinfo()                               */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include "alberta.h"
#include "traverse_r_common.c"

/*--------------------------------------------------------------------------*/
/*   Refinement/Coarsening vertex numbering             		    */
/*   like Baensch / Kossaczky                                               */
/*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/*  vertex_of_edge[edge][i], i = 1,2 are the two vertices of edge           */
/*--------------------------------------------------------------------------*/
int vertex_of_edge[6][2]={{0,1},{0,2},{0,3},{1,2},{1,3},{2,3}};


/*--------------------------------------------------------------------------*/
/*   fill_macro_info:                                		            */
/*   ----------------                                   		    */
/*   fill EL_INFO structure for a macro element         		    */
/*--------------------------------------------------------------------------*/

void fill_macro_info(MESH *mesh, const MACRO_EL *mel, EL_INFO *elinfo)
{
  FUNCNAME("fill_macro_info");
  EL       *nb;
  MACRO_EL *mnb;
  int      i, j, k;
#if !NEIGH_IN_EL
  FLAGS    fill_opp_coords;
#endif

  elinfo->mesh     = mesh;
  elinfo->macro_el = mel;
  elinfo->el       = mel->el;
  elinfo->parent   = nil;
  elinfo->level    = 0;
#if !NEIGH_IN_EL
  elinfo->el_type  = mel->el_type;
#endif

  if (elinfo->fill_flag & FILL_COORDS) {

    for (i=0; i<N_VERTICES; i++) {
      if (mel->coord[i]) {
	for (j=0;j<DIM_OF_WORLD;j++) elinfo->coord[i][j] = mel->coord[i][j];
      }
      else {
	TEST_EXIT(0)("no mel->coord[%d]\n",i);
      }
    }
  }

#if NEIGH_IN_EL
  if (elinfo->fill_flag & FILL_OPP_COORDS) {

    for (i=0; i < N_NEIGH; i++) {
      if ((mnb = mel->neigh[i])) {
	nb = mel->neigh[i]->el;
	k  = mel->opp_vertex[i];

	if (nb->child[0] && (k < 2)) {   /*make nb nearest el.*/
	  if (k==1) {
	    nb = nb->child[0];
	  } else {
	    nb = nb->child[1];
	  }
	  k  = 3;
/* 	  if (mnb->coord[N_VERTICES]) { */
/* 	    for (j=0;j<DIM_OF_WORLD;j++) */
/* 	      elinfo->opp_coord[i][j] = mnb->coord[N_VERTICES][j]; */
/* 	  } */
/* 	  else */
	  { /* always edge between vertices 0 and 1 is bisected! */
	    if (mnb->el->new_coord)
	      for (j=0;j<DIM_OF_WORLD;j++)
		elinfo->opp_coord[i][j] = mnb->el->new_coord[j];
	    else
	      for (j=0;j<DIM_OF_WORLD;j++) elinfo->opp_coord[i][j] =
		(mnb->coord[0][j] + mnb->coord[1][j]) / 2;
	  }
	}
	else {
	  for (j=0;j<DIM_OF_WORLD;j++)
	    elinfo->opp_coord[i][j] = mnb->coord[k][j];
	}
      }
    }
  }

#else   /* ! NEIGH_IN_EL */

  if (elinfo->fill_flag & (FILL_OPP_COORDS | FILL_NEIGH)) {

    fill_opp_coords = (elinfo->fill_flag & FILL_OPP_COORDS);
    for (i=0; i < N_NEIGH; i++) {
      if ((mnb = mel->neigh[i])) {
	nb = elinfo->neigh[i]      = mel->neigh[i]->el;
	k  = elinfo->opp_vertex[i] = mel->opp_vertex[i];

	if (nb->child[0] && (k < 2)) {   /*make nb nearest el.*/
	  if (k==1) {
	    nb = elinfo->neigh[i]      = nb->child[0];
	  } else {
	    nb = elinfo->neigh[i]      = nb->child[1];
	  }
	  k  = elinfo->opp_vertex[i] = 3;
	  if  (fill_opp_coords)
	  {
	    /* always edge between vertices 0 and 1 is bisected! */
	    if (mnb->el->new_coord)
	      for (j=0;j<DIM_OF_WORLD;j++)
		elinfo->opp_coord[i][j] = mnb->el->new_coord[j];
	    else
	      for (j=0;j<DIM_OF_WORLD;j++) elinfo->opp_coord[i][j] =
			 (mnb->coord[0][j] + mnb->coord[1][j]) / 2;
	  }
	}
	else {
	  if  (fill_opp_coords) {
	    for (j=0;j<DIM_OF_WORLD;j++)
	      elinfo->opp_coord[i][j] = mnb->coord[k][j];
	  }
	}
      }
      else {
	elinfo->neigh[i] = nil;
      }
    }
  }
#endif

  if (elinfo->fill_flag & FILL_BOUND)
  {
    for (i = 0; i < N_VERTICES; i++)
    {
      elinfo->bound[i] = mel->bound[i];
    }
    for (i = 0; i < N_FACES+N_EDGES; i++) 
    {
      elinfo->boundary[i] = mel->boundary[i];
    }
  }

  if (elinfo->fill_flag & FILL_ORIENTATION) 
    elinfo->orientation = mel->orientation;

  return;
}


/*--------------------------------------------------------------------------*/
/*   child_vertex[el_type][child][i] =                                      */
/*       father's local vertex index of new vertex i                        */
/*       4 stands for the newly generated vertex                            */
/*   (declared external in tria.h)                                          */
/*--------------------------------------------------------------------------*/
int  child_vertex[3][2][4] = {{{0,2,3,4},{1,3,2,4}},
			      {{0,2,3,4},{1,2,3,4}},
			      {{0,2,3,4},{1,2,3,4}}};

/*--------------------------------------------------------------------------*/
/*   child_edge[el_type][child][i] =                                        */
/*       father's local edge index of new edge i                            */
/*       new edge 2 is half of old edge 0,                                  */
/*       new edges 4,5 are really new edges, and value is different:        */
/*         child_edge[][][4,5] = index of same edge in other child	    */
/*   (declared external in tria.h)                                          */
/*--------------------------------------------------------------------------*/
int child_edge[3][2][6] = {{{1,2,0,5,5,4},{4,3,0,5,5,4}},
			   {{1,2,0,5,4,5},{3,4,0,5,4,5}},
			   {{1,2,0,5,4,5},{3,4,0,5,4,5}}};

#if 0  /* Kuni 12.03.97: gcc warning: `refi_opp_edge' defined but not used  */

/*--------------------------------------------------------------------------*/
/*   refi_opp_edge[el_type][newedge][ichild][new_level][k] =                */
/*       beginning with father with el_type in {0,1,2},                     */
/*       looking for child elements around edge newedge (0=[V2V5],1=[V3V5]),*/
/*       beginning with child ichild,                                       */
/*       a path through the hierarchical tree is given                      */
/*       by refi_opp_edge[][][][0,1,2,...][k]  with values                  */
/*         refi_opp_edge[][][][][0] = index of child to use (=2 if both)    */
/*         refi_opp_edge[][][][][1] = index of opposite edge (=5 if final)  */
/*                                                                          */
/*--------------------------------------------------------------------------*/

static int refi_opp_edge[3][2][2][4][2] = {
 { { { {0,1},{1,4},{0,5},{-1,-1}},    /* typ0 edgeV2 child0 */
     { {1,0},{2,2},{1,5},{-1,-1}}},   /* typ0 edgeV2 child1 */
   { { {0,0},{2,2},{1,5},{-1,-1}},    /* typ0 edgeV3 child0 */
     { {1,1},{1,4},{0,5},{-1,-1}}}},  /* typ0 edgeV3 child1 */
 { { { {0,1},{1,4},{0,5},{-1,-1}},    /* typ1 edgeV2 child0 */
     { {1,1},{1,4},{0,5},{-1,-1}}},   /* typ1 edgeV2 child1 */
   { { {0,0},{2,2},{1,4},{0,5}},      /* typ1 edgeV3 child0 */
     { {1,0},{2,2},{1,4},{0,5}}}},    /* typ1 edgeV3 child1 */
 { { { {0,1},{1,5},{-1,-1},{-1,-1}},  /* typ2 edgeV2 child0 */
     { {1,1},{1,5},{-1,-1},{-1,-1}}}, /* typ2 edgeV2 child1 */
   { { {0,0},{2,2},{1,5},{-1,-1}},    /* typ2 edgeV3 child0 */
     { {1,0},{2,2},{1,5},{-1,-1}}}}}; /* typ2 edgeV3 child1 */
#endif

/*--------------------------------------------------------------------------*/
/*   child_orientation[el_type][child] =                                    */
/*      +1 if orientation is not changed during refinement                  */
/*      -1 if orientation is changed during refinement                      */
/*--------------------------------------------------------------------------*/
static S_CHAR child_orientation[3][2] = {{1,1}, {1,-1}, {1,-1}};


#if 0  /* Kuni 12.03.97: gcc warning: `test_edge' defined but not used      */

static int test_edge(EL *el, int oe, DOF *n1, DOF *n2)
{
  DOF *nn1=el->dof[vertex_of_edge[5-oe][0]];
  DOF *nn2=el->dof[vertex_of_edge[5-oe][1]];
  return ((nn1 == n1 || nn1 == n2) && (nn2 == n1 || nn2 == n2));
}
#define TEST_EDGE(el,oe,n1,n2) \
     TEST_EXIT(test_edge(el,oe,n1,n2))("invalid edge neighbour...\n")
#endif

/*--------------------------------------------------------------------------*/
/*   fill_elinfo:                                    		            */
/*   ------------                                       		    */
/*   fill EL_INFO structure for one child of an element   		    */
/*--------------------------------------------------------------------------*/

void fill_elinfo(int ichild, const EL_INFO *elinfo_old, EL_INFO *elinfo)
{
  FUNCNAME("fill_elinfo");
  int     i,j,k;
  int     el_type = -1;                            /* el_type in {0,1,2} */
  int     ochild = -1;                /* index of other child = 1-ichild */
  int     *cv = nil;               /* cv = child_vertex[el_type][ichild] */
  int     (*cvg)[4] = nil;                /* cvg = child_vertex[el_type] */
  int     *ce;                       /* ce = child_edge[el_type][ichild] */
  int     iedge;
  EL      *nb, *nbk, **neigh_old;
  EL      *el_old = elinfo_old->el;
  FLAGS   fill_flag = elinfo_old->fill_flag;
  DOF    *dof;
#if !NEIGH_IN_EL
  int     ov;
  EL      **neigh;
  FLAGS   fill_opp_coords;
  U_CHAR  *opp_vertex;
#endif

  TEST_EXIT(el_old->child[0])("missing child?\n");  /* Kuni 22.08.96 */

  elinfo->el        = el_old->child[ichild];
  elinfo->macro_el  = elinfo_old->macro_el;
  elinfo->fill_flag = fill_flag;
  elinfo->mesh      = elinfo_old->mesh;
  elinfo->parent    = el_old;
  elinfo->level     = elinfo_old->level + 1;
#if !NEIGH_IN_EL
  elinfo->el_type   = (elinfo_old->el_type + 1) % 3;
#endif

  TEST_EXIT(elinfo->el)("missing child %d?\n", ichild);

  if (fill_flag) {
    el_type = EL_TYPE(elinfo_old->el, elinfo_old);
    cvg = child_vertex[el_type];
    cv = cvg[ichild];
    ochild = 1-ichild;
  }

  if (fill_flag & FILL_COORDS)
  {
    for (i=0; i<3; i++) {
      for (j = 0; j < DIM_OF_WORLD; j++) {
	elinfo->coord[i][j] = elinfo_old->coord[cv[i]][j];
      }
    }
    if (el_old->new_coord)
      for (j = 0; j < DIM_OF_WORLD; j++) 
	elinfo->coord[3][j] = el_old->new_coord[j];
    else
      for (j = 0; j < DIM_OF_WORLD; j++)
	elinfo->coord[3][j] =
	  (elinfo_old->coord[0][j] + elinfo_old->coord[1][j]) / 2;
  }


#if NEIGH_IN_EL
  if (fill_flag & FILL_OPP_COORDS)
  {
    neigh_old = el_old->neigh;

    /*----- nb[0] is other child -------------------------------------------*/

/*    if (nb = el_old->child[ochild]) {        old version    */
    if (el_old->child[0]  &&  (nb = el_old->child[ochild])) /* Kuni 22.08.96*/
    {        
      if (nb->child[0]) {     /* go down one level for direct neighbour */
	k = cvg[ochild][1];
	if (nb->new_coord)
	  for (j = 0; j < DIM_OF_WORLD; j++)
	    elinfo->opp_coord[0][j] = nb->new_coord[j];
	else
	  for (j = 0; j < DIM_OF_WORLD; j++)
	    elinfo->opp_coord[0][j] =
	      (elinfo_old->coord[ochild][j] + elinfo_old->coord[k][j]) / 2;
      }
      else {
	for (j = 0; j < DIM_OF_WORLD; j++) {
	  elinfo->opp_coord[0][j] = elinfo_old->coord[ochild][j];
	}
      }
    }
    else {
      ERROR_EXIT("no other child");
    }


/*----- nb[1],nb[2] are childs of old neighbours nb_old[cv[i]] ----------*/

    for (i=1; i<3; i++) 
    { 
      if (nb = neigh_old[cv[i]]) 
      {
	TEST_EXIT(nb->child[0])("nonconforming triangulation\n");

	for (k=0; k<2; k++) {     /* look at both children of old neighbour */

	  nbk = nb->child[k];
	  if (nbk->dof[0] == el_old->dof[ichild]) {
	    dof = nb->dof[el_old->opp_vertex[cv[i]]];      /* opp. vertex */
	    if (dof == nbk->dof[1]) {
	      if (nbk->child[0]) {
		if (nbk->new_coord)
		  for (j = 0; j < DIM_OF_WORLD; j++)
		    elinfo->opp_coord[i][j] = nbk->new_coord[j];
		else
		  for (j = 0; j < DIM_OF_WORLD; j++)
		    elinfo->opp_coord[i][j] = (elinfo_old->opp_coord[cv[i]][j]
				       + elinfo_old->coord[ichild][j]) / 2;
		break;
	      }
	    }
	    else {
	      TEST_EXIT(dof == nbk->dof[2])("opp_vertex not found\n");
	    }

	    for (j = 0; j < DIM_OF_WORLD; j++) {
	      elinfo->opp_coord[i][j] = elinfo_old->opp_coord[cv[i]][j];
	    }
	    break;
	  }
	  
	} /* end for k */
	TEST_EXIT(k<2)("child not found with vertex\n");

      }
    }  /* end for i */


/*----- nb[3] is old neighbour neigh_old[ochild] ------------------------*/

    if (neigh_old[ochild]) {
      for (j = 0; j < DIM_OF_WORLD; j++) {
	elinfo->opp_coord[3][j] = elinfo_old->opp_coord[ochild][j];
      }
    }

  }

#else  /* ! NEIGH_IN_EL */

  if (fill_flag & (FILL_NEIGH | FILL_OPP_COORDS))
  {
    neigh      = elinfo->neigh;
    neigh_old  = (EL **) elinfo_old->neigh;
    opp_vertex = elinfo->opp_vertex;
    fill_opp_coords = (fill_flag & FILL_OPP_COORDS);

/*----- nb[0] is other child --------------------------------------------*/

/*    if (nb = el_old->child[ochild])   old version  */
    if (el_old->child[0]  &&  (nb = el_old->child[ochild])) /*Kuni 22.08.96*/
    {
      if (nb->child[0])
      {     /* go down one level for direct neighbour */
	if (fill_opp_coords)
	{
	  if (nb->new_coord)
	  {
	    for (j = 0; j < DIM_OF_WORLD; j++)
	      elinfo->opp_coord[0][j] = nb->new_coord[j];
	  }
	  else
	  {
	    k = cvg[ochild][1];
	    for (j = 0; j < DIM_OF_WORLD; j++)
	      elinfo->opp_coord[0][j] =
		(elinfo_old->coord[ochild][j] + elinfo_old->coord[k][j]) / 2;
	  }
	}
	neigh[0]      = nb->child[1];
	opp_vertex[0] = 3;
      }
      else {
	if (fill_opp_coords) {
	  for (j = 0; j < DIM_OF_WORLD; j++) {
	    elinfo->opp_coord[0][j] = elinfo_old->coord[ochild][j];
	  }
	}
	neigh[0]      = nb;
	opp_vertex[0] = 0;
      }
    }
    else {
      ERROR_EXIT("no other child");
      neigh[0] = nil;
    }


/*----- nb[1],nb[2] are childs of old neighbours nb_old[cv[i]] ----------*/

    for (i=1; i<3; i++)
    { 
      if ((nb = neigh_old[cv[i]])) 
      {
	TEST_EXIT(nb->child[0])("nonconforming triangulation\n");

	for (k=0; k<2; k++)  /* look at both childs of old neighbour */
	{       
	  nbk = nb->child[k];
	  if (nbk->dof[0] == el_old->dof[ichild]) {
	    dof = nb->dof[elinfo_old->opp_vertex[cv[i]]]; /* opp. vertex */
	    if (dof == nbk->dof[1]) 
	    {
	      ov = 1;
	      if (nbk->child[0])
	      {
		if (fill_opp_coords)
		{
		  if (nbk->new_coord)
		    for (j = 0; j < DIM_OF_WORLD; j++)
		      elinfo->opp_coord[i][j] = nbk->new_coord[j];
		  else
		    for (j = 0; j < DIM_OF_WORLD; j++)
		      elinfo->opp_coord[i][j] =
			(elinfo_old->opp_coord[cv[i]][j]
			 + elinfo_old->coord[ichild][j]) / 2;
		}
		neigh[i]      = nbk->child[0];
		opp_vertex[i] = 3;
		break;
	      }
	    }
	    else
	    {
	      TEST_EXIT(dof == nbk->dof[2])("opp_vertex not found\n");
	      ov = 2;
	    }

	    if (fill_opp_coords)
	    {
	      for (j = 0; j < DIM_OF_WORLD; j++)
	      {
		elinfo->opp_coord[i][j] = elinfo_old->opp_coord[cv[i]][j];
	      }
	    }
	    neigh[i]      = nbk;
	    opp_vertex[i] = ov;
	    break;
	  }
	  
	} /* end for k */
	TEST_EXIT(k<2)("child not found with vertex\n");
      }
      else 
      {
	neigh[i] = nil;
      }
    }  /* end for i */


/*----- nb[3] is old neighbour neigh_old[ochild] ------------------------*/

    if ((neigh[3] = neigh_old[ochild]))
    {
      opp_vertex[3] = elinfo_old->opp_vertex[ochild];
      if (fill_opp_coords) {
	for (j = 0; j < DIM_OF_WORLD; j++) {
	  elinfo->opp_coord[3][j] = elinfo_old->opp_coord[ochild][j];
	}
      }
    }
  }
#endif

  if (fill_flag & FILL_BOUND)
  {
    for (i = 0; i < 3; i++)
    {
      elinfo->bound[i] = elinfo_old->bound[cv[i]];
    }
    elinfo->bound[3] = GET_BOUND(elinfo_old->boundary[N_FACES+0]);

    elinfo->boundary[0] = nil;
    elinfo->boundary[1] = elinfo_old->boundary[cv[1]];
    elinfo->boundary[2] = elinfo_old->boundary[cv[2]];
    elinfo->boundary[3] = elinfo_old->boundary[ochild];

    ce = child_edge[el_type][ichild];
    for (iedge=0; iedge<4; iedge++) {
      elinfo->boundary[N_FACES+iedge] = 
	                 elinfo_old->boundary[N_FACES+ce[iedge]];
    }
    for (iedge=4; iedge<6; iedge++) {
      i = 5 - cv[iedge-3];                /* old vertex opposite new edge */
      elinfo->boundary[N_FACES+iedge] = elinfo_old->boundary[i];
    }
  }


  if (elinfo->fill_flag & FILL_ORIENTATION) {
    elinfo->orientation =
      elinfo_old->orientation * child_orientation[el_type][ichild];
  }

}

/*--------------------------------------------------------------------------*/
/*   update_elinfo:                                    		            */
/*   --------------                                       		    */
/*   update EL_INFO structure after refinement (of some neighbours)	    */
/*--------------------------------------------------------------------------*/

void update_elinfo(EL_INFO *elinfo)
{
  FUNCNAME("update_elinfo");
  EL      *el = elinfo->el;

  TEST_EXIT(el)("missing element?\n");

#if ! NEIGH_IN_EL
  if (elinfo->fill_flag & (FILL_NEIGH | FILL_OPP_COORDS))
  {
    EL    *nb;
    int   ineigh, ov, j, k;
    FLAGS fill_opp_coords = (elinfo->fill_flag & FILL_OPP_COORDS);
    
    for (ineigh=0; ineigh < N_NEIGH; ineigh++) {
      if ((nb = elinfo->neigh[ineigh])) {
	ov = elinfo->opp_vertex[ineigh];
	if (ov < 2 && nb->child[0]) {
	  if (fill_opp_coords) {
	    k = -1;
	    for (j=0; j<N_VERTICES; j++)
	      if (el->dof[j] == nb->dof[1-ov]) k = j;
	    TEST_EXIT(k >= 0)("neighbour dof not found\n");

	    if (nb->new_coord)
	      for (j = 0; j < DIM_OF_WORLD; j++)
		elinfo->opp_coord[ineigh][j] = nb->new_coord[j];
	    else
	      for (j = 0; j < DIM_OF_WORLD; j++)
		elinfo->opp_coord[ineigh][j] = 
		  (elinfo->opp_coord[ineigh][j] + elinfo->coord[k][j]) / 2;
	  }
	  elinfo->neigh[ineigh]      = nb->child[1-ov];
	  elinfo->opp_vertex[ineigh] = 3;
	}
      }
    }
  }
#endif

}
