/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file: alberta.h                                                          */
/*                                                                          */
/*                                                                          */
/* description: public header file of the ALBERTA package                   */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/*  Global configuration header                                             */
/*--------------------------------------------------------------------------*/

#ifdef HAVE_CONFIG_H        /* only defined during build of ALBERTA libs */
#include <config.h>
#endif

/*--------------------------------------------------------------------------*/
/*  Header--File for an hierarchical 2d and 3d simplicial triangulation     */
/*--------------------------------------------------------------------------*/
#ifndef _ALBERTA_H_
#define _ALBERTA_H_

#include "alberta_util.h"

#ifdef __cplusplus
extern "C" {
#endif

#define ALBERTA_VERSION  "ALBERTA: Version 1.2 "  /* must be 20 chars! */

/*--------------------------------------------------------------------------*/
/* NEIGH_IN_EL is not fully supported at the moment.                        */
/*--------------------------------------------------------------------------*/

#define NEIGH_IN_EL 0

/*--------------------------------------------------------------------------*/
/*  Definition of the space dimension and of parameters depending on the    */
/*  space dimension:							    */
/*									    */
/*  DIM:            simplex dimension					    */
/*  DIM_OF_WORLD:   space dimension					    */
/*--------------------------------------------------------------------------*/

#ifndef DIM
#error DIM UNDEFINED
#endif
#ifndef DIM_OF_WORLD
#error DIM_OF_WORLD UNDEFINED
#endif
#ifndef NEIGH_IN_EL
#error NEIGH_IN_EL UNDEFINED
#endif
#ifndef EL_INDEX
#error EL_INDEX UNDEFINED
#endif

#if DIM == 1
#define N_VERTICES_1D    2
#define N_NEIGH_1D       2

#define N_VERTICES N_VERTICES_1D
#define N_NEIGH    N_NEIGH_1D
#endif

#if DIM == 2
#define N_VERTICES_2D    3
#define N_EDGES_2D       3
#define N_NEIGH_2D       3

#define N_VERTICES N_VERTICES_2D
#define N_EDGES    N_EDGES_2D
#define N_NEIGH    N_NEIGH_2D
#endif

#if DIM == 3
#define N_VERTICES_3D    4
#define N_EDGES_3D       6
#define N_FACES_3D       4
#define N_NEIGH_3D       4

#define N_VERTICES N_VERTICES_3D
#define N_EDGES    N_EDGES_3D
#define N_FACES    N_FACES_3D
#define N_NEIGH    N_NEIGH_3D
#endif

typedef REAL          REAL_D[DIM_OF_WORLD];
typedef REAL          REAL_DD[DIM_OF_WORLD][DIM_OF_WORLD];

/*--------------------------------------------------------------------------*/
/*  some usefull macro depending on DIM_OF_WORLD                            */
/*--------------------------------------------------------------------------*/

#if DIM_OF_WORLD == 1
#define COPY_DOW(x,y) ((y)[0] = (x)[0])
#define SCP_DOW(x,y)  ((x)[0]*(y)[0])
#define SET_DOW(val,x)  ((x)[0] = (val))
#define SCAL_DOW(val,x)  ((x)[0] *= (val))
#define DIST_DOW(x,y)  ABS((x)[0]-(y)[0])
#endif

#if DIM_OF_WORLD == 2
#define COPY_DOW(x,y) ((y)[0] = (x)[0], (y)[1] = (x)[1])
#define SCP_DOW(x,y)  ((x)[0]*(y)[0] + (x)[1]*(y)[1])
#define SET_DOW(val,x)  ((x)[0] = (x)[1] = (val))
#define SCAL_DOW(val,x)  ((x)[0] *= (val), (x)[1] *= (val))
#define DIST_DOW(x,y)  sqrt(SQR((x)[0]-(y)[0])+SQR((x)[1]-(y)[1]))
#endif

#if DIM_OF_WORLD == 3
#define COPY_DOW(x,y) ((y)[0] = (x)[0], (y)[1] = (x)[1], (y)[2] = (x)[2])
#define SCP_DOW(x,y)  ((x)[0]*(y)[0] + (x)[1]*(y)[1] + (x)[2]*(y)[2])
#define SET_DOW(val,x)  ((x)[0] = (x)[1] = (x)[2] = (val))
#define SCAL_DOW(val,x)  ((x)[0] *= (val), (x)[1] *= (val), (x)[2] *= (val))
#define DIST_DOW(x,y)\
   sqrt(SQR((x)[0]-(y)[0])+SQR((x)[1]-(y)[1])+SQR((x)[2]-(y)[2]))
#endif

#if DIM_OF_WORLD > 3
#error not for DIM_OF_WORLD > 3 now
#endif

#define NORM_DOW(x)   (sqrt(SCP_DOW((x),(x))))

/*--------------------------------------------------------------------------*/
/*  access to neighbours via element or element_info structure		    */
/*--------------------------------------------------------------------------*/

#if NEIGH_IN_EL
#define NEIGH(el,el_info)      (el)->neigh
#define OPP_VERTEX(el,el_info) (el)->opp_vertex
#define EL_TYPE(el,el_info)    (el)->el_type
#else
#define NEIGH(el,el_info)      (el_info)->neigh
#define OPP_VERTEX(el,el_info) (el_info)->opp_vertex
#define EL_TYPE(el,el_info)    (el_info)->el_type
#endif

/*--------------------------------------------------------------------------*/
/*  access to element index via element or element_info structure           */
/*--------------------------------------------------------------------------*/

#if EL_INDEX
#define INDEX(el)   ((el) ? (el)->index : -1)
#else
#define INDEX(el)   -1
#endif

/*--------------------------------------------------------------------------*/
/*  access to leaf data (only for leaf elements)                            */
/*--------------------------------------------------------------------------*/

#define IS_LEAF_EL(el) (!(el)->child[0])
#define LEAF_DATA(el)  ((void *)(el)->child[1])

/*--------------------------------------------------------------------------*/
/*  boundary types							    */
/*--------------------------------------------------------------------------*/

#define INTERIOR      0
#define DIRICHLET     1
#define NEUMANN       -1

/*--------------------------------------------------------------------------*/
/*  node types (indices in n_dof[] vectors, e.g.)			    */
/*--------------------------------------------------------------------------*/

#if DIM == 1
#define VERTEX        0
#define CENTER        1
#endif

#if DIM == 2
#define VERTEX        0
#define EDGE          1
#define CENTER        2
#endif

#if DIM == 3
#define VERTEX        0
#define EDGE          1
#define FACE          2
#define CENTER        3
#endif

/*--------------------------------------------------------------------------*/
/*  basic types of the grid						    */
/*--------------------------------------------------------------------------*/

#if DIM == 1
typedef void                    BOUNDARY; /*--- not used in 1d           ---*/
#else
typedef struct boundary         BOUNDARY;
#endif

typedef signed int              DOF;
typedef struct el               EL;
typedef struct macro_el         MACRO_EL;
typedef struct mesh_mem_info    MESH_MEM_INFO;
typedef struct mesh             MESH;
typedef struct el_info          EL_INFO;
typedef struct leaf_data_info   LEAF_DATA_INFO;
typedef struct rc_list_el       RC_LIST_EL;
typedef struct parametric       PARAMETRIC;
typedef struct traverse_stack   TRAVERSE_STACK;

typedef struct adapt_stat       ADAPT_STAT;
typedef struct adapt_instat     ADAPT_INSTAT;

#ifndef DOF_ADMIN_DEF
typedef struct dof_admin        DOF_ADMIN;
typedef struct dof_int_vec      DOF_INT_VEC;
typedef struct dof_dof_vec      DOF_DOF_VEC;
typedef struct dof_uchar_vec    DOF_UCHAR_VEC;
typedef struct dof_schar_vec    DOF_SCHAR_VEC;
typedef struct dof_real_vec     DOF_REAL_VEC;
typedef struct dof_real_d_vec   DOF_REAL_D_VEC;
typedef struct matrix_row       MATRIX_ROW;
typedef struct dof_matrix       DOF_MATRIX;
#endif

typedef struct bas_fcts         BAS_FCTS;
typedef struct fe_space         FE_SPACE;

typedef struct quadrature       QUAD;
typedef struct quadrature       QUADRATURE;
typedef struct quad_fast        QUAD_FAST;

typedef struct macro_data MACRO_DATA;
#if DIM > 1
/*--------------------------------------------------------------------------*/
/*  boundary descriptor							    */
/*--------------------------------------------------------------------------*/
/*  param_bound: pointer to a function to calculate a new point of a        */
/*               curved boundary					    */
/*  type:        one of INTERIOR : DIRICHLET : NEUMANN			    */
/*--------------------------------------------------------------------------*/

struct boundary
{
  void      (*param_bound)(REAL_D  );
  S_CHAR    bound;
};

#define IS_INTERIOR(boundary)\
   ((boundary) ? (boundary)->bound == INTERIOR : true)
#define IS_DIRICHLET(boundary)\
   ((boundary) ? (boundary)->bound >= DIRICHLET : false)
#define IS_NEUMANN(boundary)\
   ((boundary) ? (boundary)->bound <= NEUMANN : false)
#define GET_BOUND(boundary) ((boundary) ? (boundary)->bound : INTERIOR)
#endif

/*--------------------------------------------------------------------------*/
/*  one single element (triangle) of the grid:			            */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  position of the nodes in 1d:                                            */
/*                                                                          */
/*  0 _____ 1   or  0 _____ 1                                               */
/*                      2                                                   */
/*                                                                          */
/*                          child[0]     child[1]                           */
/*  refinement:  0 _____ 1    0 ___ 1  0 ___ 1                              */
/*                               2        2                                 */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  position of the nodes in 2d                                             */
/*         2                 2                   2                 2        */
/*        /\       or       /\        or        /\       or       /\        */
/*       /  \             4/  \ 3              /  \             4/  \ 3     */
/*      /    \            /    \              /  3 \            /  6 \      */
/*    0/______\1        0/______\1          0/______\1        0/______\1    */
/*                           5                                     5        */
/*                                                                          */
/*  refinement:           2          child[0]  0    1   child[1]            */
/*                       /\                   /|    |\                      */
/*                     4/  \ 3  -->         5/ |4  3| \ 5                   */
/*                     /  6 \               /6 |    |6 \                    */
/*                   0/______\1           1/___|    |___\0                  */
/*                        5                  3  2  2  4                     */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  3d refinement: vertex numbering after (Baensch +) Kossaczky             */
/*                                                                          */
/*  edges:                                                                  */
/*  E0:  between V0, V1                                                     */
/*  E1:  between V0, V2                                                     */
/*  E2:  between V0, V3                                                     */
/*  E3:  between V1, V2                                                     */
/*  E4:  between V1, V3                                                     */
/*  E5:  between V2, V3                                                     */
/*                                                                          */
/*  Always edge 0 (between vertices 0 and 1) is bisected.                   */
/*                                                                          */
/*                           V1                                             */
/*                         -+                                               */
/*               E0    ----||                                               */
/*                 ----   | |E3                                             */
/*             ----       | |                                               */
/*         ----          |  |                                               */
/*     V0 +. . . . .E1 . | .+V2                                             */
/*         ---          |  /                                                */
/*            ---      E4 /                                                 */
/*               ---   | /E5                                                */
/*              E2  ---|/                                                   */
/*                     +                                                    */
/*                     V3                                                   */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  child:        pointers to the two children of the element		    */
/*                if (child[0]==child[1]==nil) element is a leaf of the     */
/*                        tree                                              */
/*  dof:          vector of pointers to dof vectors :-)                     */
/*  new_coord:    in case of curved boundary, coords of ref.edge midpoint   */
/*  index:        global element index (only for test purposes)		    */
/*  mark:         element is a leaf:   					    */
/*                   mark == 0         do not refine/coarsen		    */
/*                   mark > 0          refine (mark times)                  */
/*                   mark < 0          may be coarsened (mark times)        */
/*--------------------------------------------------------------------------*/

struct el
{
  EL         *child[2];
  DOF        **dof;
#if EL_INDEX
  int        index;
#endif
  S_CHAR     mark;
#if DIM > 1
  REAL       *new_coord;
#endif

#if NEIGH_IN_EL
  EL         *neigh[N_NEIGH];
  U_CHAR     opp_vertex[N_NEIGH];
#if DIM == 3
  U_CHAR     el_type;
#endif
#endif
};

#if DIM == 2
extern int vertex_of_edge[3][2];                  /* declared in refine.c   */
#endif

#if DIM == 3
/*--------------------------------------------------------------------------*/
/*   child_vertex[el_type][child][i] =                                      */
/*       parent's local vertex index of new vertex i                        */
/*       4 stands for the newly generated vertex                            */
/*--------------------------------------------------------------------------*/
/*   child_edge[el_type][child][i] =                                        */
/*       parent's local edge index of new edge i                            */
/*       new edge 2 is half of old edge 0,                                  */
/*       new edges 4,5 are really new edges, and value is different:        */
/*         child_edge[][][4,5] = index of same edge in other child	    */
/*--------------------------------------------------------------------------*/
/*  vertex_of_edge[edge][i], i = 1,2 are the two vertices of edge           */
/*--------------------------------------------------------------------------*/
extern int child_vertex[3][2][4];               /* declared in traverse.c   */
extern int child_edge[3][2][6];                 /* declared in traverse.c   */
extern int vertex_of_edge[6][2];                /* declared in traverse.c   */

/*--------------------------------------------------------------------------*/
/*  edge_of_dofs[i][j]: gives the local index of edge with vertices i and j */
/*--------------------------------------------------------------------------*/
extern U_CHAR  edge_of_dofs[4][4];               /* declared in refine.c  */
#endif

struct rc_list_el
{
  EL           *el;
  int          no;
  int          flag;
#if DIM == 3
  RC_LIST_EL   *neigh[2];
  int          opp_vertex[2];
  U_CHAR       el_type;
#endif
};


/*--------------------------------------------------------------------------*/
/* PARAMETRIC structure, entry in MESH structure                            */
/* description of parametric meshes and elements                            */
/*--------------------------------------------------------------------------*/

struct parametric 
{
  int (*init_element)(const EL_INFO *info, const PARAMETRIC *parametric);
  /* init_el = 0 : non-parametric element */
  /* init_el = 1 : parametric element     */

  void (*coord_to_world)(const EL_INFO *info, const QUAD *quad,
			 int n, const REAL lambda[][DIM+1], REAL_D *world);
  int (*world_to_coord)(const EL_INFO *info,
			int n, const REAL_D world[], REAL lambda[][DIM+1]);
  void (*det)(const EL_INFO *info, const QUAD *quad,
	      int n, const REAL lambda[][DIM+1], REAL dets[]);
  void (*grd_lambda)(const EL_INFO *info,  const QUAD *quad, 
		     int n, const REAL lambda[][DIM+1],
		     REAL_D grd_lam[][DIM+1], REAL dets[]);

  int not_all;      /* true:  some elements may be non-parametric */
  int fill_coords;  /* false: ignore FILL_COORDS in mesh_traverse */

  void *data;
};

/*--------------------------------------------------------------------------*/
/*  additional information to elements during hierarchy traversal	    */
/*--------------------------------------------------------------------------*/
/*  mesh:          pointer to the mesh structure                      	    */
/*  coord:         world coordinates of the nodes                     	    */
/*  boundary:      boundary description of an edge resp. face    	    */
/*  bound:         boundary type of the vertices                 	    */
/*  neigh:         pointer to the adjacent elements       		    */
/*                 nil-pointer for a part of the boundary		    */
/*  opp_coord:     world coordinates of opposite vertices 		    */
/*  opp_vertex:    local indices of opposite vertices       		    */
/*  edge_boundary: boundary description of the 6 edges (only 3d)	    */
/*  orientation:   orientation of the tetrahedron (+1|-1) (only 3d)	    */
/*--------------------------------------------------------------------------*/

struct el_info
{
  MESH             *mesh;
  REAL_D           coord[N_VERTICES];  /* parametric is done via dof_vector */
  const MACRO_EL   *macro_el;
  EL               *el, *parent;
  FLAGS            fill_flag;
  S_CHAR           bound[N_VERTICES];
#if DIM == 2
  const BOUNDARY   *boundary[N_EDGES];
#endif

#if DIM == 3
  const BOUNDARY   *boundary[N_FACES+N_EDGES];
#endif

  U_CHAR           level;

#if ! NEIGH_IN_EL
  EL               *neigh[N_NEIGH];
  U_CHAR           opp_vertex[N_NEIGH];
#if DIM == 3
  U_CHAR           el_type;
#endif
#endif
  REAL_D           opp_coord[N_NEIGH];

#if DIM == 3
  S_CHAR           orientation;
#endif
};

/*--------------------------------------------------------------------------*/
/*  flags, which information should be present in the EL_INFO structure     */
/*--------------------------------------------------------------------------*/

#define FILL_NOTHING               0X00L
#define FILL_COORDS                0X01L
#define FILL_BOUND                 0X02L
#define FILL_NEIGH                 0X04L
#define FILL_OPP_COORDS            0X08L
#define FILL_ORIENTATION           0X10L
#define FILL_EL_TYPE               0X20L

#if DIM == 2
#define FILL_ANY_2D         (0X01L|0X02L|0X04L|0X08L)
#define FILL_ANY FILL_ANY_2D
#else
#define FILL_ANY_3D         (0X01L|0X02L|0X04L|0X08L|0X10L|0X20L)
#define FILL_ANY FILL_ANY_3D
#endif

/*--------------------------------------------------------------------------*/
/*  flags for mesh traversal                                                */
/*--------------------------------------------------------------------------*/

#define CALL_EVERY_EL_PREORDER      0X0100L
#define CALL_EVERY_EL_INORDER       0X0200L
#define CALL_EVERY_EL_POSTORDER     0X0400L
#define CALL_LEAF_EL                0X0800L
#define CALL_LEAF_EL_LEVEL          0X1000L
#define CALL_EL_LEVEL               0X2000L
#define CALL_MG_LEVEL               0X4000L    /* used in multigrid methods */

#define TEST_FLAG(flags, el_info) \
        TEST_EXIT(!((((el_info)->fill_flag)^(flags)) & (flags)))\
                 ("flag %d not set\n", (flags))

/*--------------------------------------------------------------------------*/
/*  one single element of the macro triangulation:			    */
/*--------------------------------------------------------------------------*/
/*  next, last:    double linked list of macro elements	     		    */
/*  el:            pointer to the element data of the macro element         */
/*  coord:         world coordinates of the nodes on the macro element	    */
/*  bound:         boundary type of the vertices                 	    */
/*  boundary:      boundary description of an edge resp. face    	    */
/*  neigh:         pointer to the adjacent macro elements		    */
/*                 nil-pointer for a part of the boundary		    */
/*  opp_vertex:    local indices of opposite vertices       		    */
/*  edge_boundary: boundary description of the 6 edges (only 3d)	    */
/*--------------------------------------------------------------------------*/

struct macro_el
{
  MACRO_EL        *next, *last;
  EL              *el;
  REAL            *coord[N_VERTICES];
  S_CHAR          bound[N_VERTICES];
#if DIM == 2
  const BOUNDARY  *boundary[N_EDGES];
#endif
#if DIM == 3
  const BOUNDARY  *boundary[N_FACES+N_EDGES];
#endif
  MACRO_EL        *neigh[N_NEIGH];
  U_CHAR          opp_vertex[N_NEIGH];

  int             index;

#if DIM == 3
  U_CHAR          el_type;
  S_CHAR          orientation;
#endif
};

/*--------------------------------------------------------------------------*/
/*---  index based storage of macro triangulations                       ---*/
/*--------------------------------------------------------------------------*/

struct macro_data
{
  int n_total_vertices;
  int n_macro_elements;

  REAL_D *coords;                   /* Length will be n_total_vertices      */ 

  int (*mel_vertices)[N_VERTICES];  /* mel_vertices[i][j]: global index of  */
                                    /* jth vertex of element i              */ 

  int (*neigh)[N_NEIGH];            /* neigh[i][j]: global index of jth     */
                                    /* neighbor of element i                */
                                    /* or -1 at boundaries                  */

  S_CHAR (*boundary)[N_NEIGH];      /* boundary[i][j]: boundary type of jth */
                                    /* edge/face of element i               */ 

#if DIM == 3
  U_CHAR *el_type;                  /* el_type[i]: type of element i        */
                                    /* only used in 3d!                     */
#endif
};

#ifndef DOF_ADMIN_DEF
#define DOF_ADMIN_DEF

/*--------------------------------------------------------------------------*/
/*  dof handling                 				            */
/*--------------------------------------------------------------------------*/
#define DOF_FREE_SIZE 32
#define DOF_UNIT_ALL_FREE 0xFFFFFFFF
typedef unsigned int DOF_FREE_UNIT;
extern  DOF_FREE_UNIT dof_free_bit[DOF_FREE_SIZE];        /* in dof_admin.c */

#define FOR_ALL_DOFS(admin, todo) \
  if ((admin)->hole_count == 0) { int dof; \
  for (dof = 0; dof < (admin)->used_count; dof++) { todo; } } \
      else { DOF_FREE_UNIT _dfu, *_dof_free = (admin)->dof_free; \
        int _i, _ibit, dof=0; \
        int _n= ((admin)->size_used + DOF_FREE_SIZE-1) / DOF_FREE_SIZE; \
        for (_i = 0; _i < _n; _i++) { \
	  if ((_dfu = _dof_free[_i])) { \
	    if (_dfu == DOF_UNIT_ALL_FREE) dof += DOF_FREE_SIZE; \
	    else for (_ibit = 0; _ibit < DOF_FREE_SIZE; _ibit++) { \
               if (!(_dfu & dof_free_bit[_ibit])) { todo; } dof++; } } \
          else { \
            for (_ibit = 0; _ibit < DOF_FREE_SIZE; _ibit++){ todo; dof++;}}}}

#define FOR_ALL_FREE_DOFS(admin, todo) \
      if ((admin)->hole_count == 0) { int dof; \
        for (dof = (admin)->used_count; \
	     dof < (admin)->size; dof++)   { todo; } } \
      else { DOF_FREE_UNIT _dfu, *_dof_free = (admin)->dof_free; \
        int _i, _ibit, dof=0; \
        int _n= ((admin)->size + DOF_FREE_SIZE-1) / DOF_FREE_SIZE; \
        for (_i = 0; _i < _n; _i++) { \
	  if ((_dfu = _dof_free[_i])) { \
	    if (_dfu == DOF_UNIT_ALL_FREE) \
              for (_ibit = 0 ; _ibit < DOF_FREE_SIZE; _ibit++)\
                { todo; dof++; } \
	    else for (_ibit = 0; _ibit < DOF_FREE_SIZE; _ibit++) { \
                   if (_dfu & dof_free_bit[_ibit]) { todo; } dof++; } } \
	  else dof += DOF_FREE_SIZE; }}

struct dof_admin
{
  MESH         *mesh;
  const char   *name;

  DOF_FREE_UNIT *dof_free;    /* flag bit vector                           */
  unsigned int  dof_free_size;/* flag bit vector size                      */
  unsigned int  first_hole;   /* index of first non-zero dof_free entry    */

  DOF  size;                 /* allocated size of dof_list vector          */
  DOF  used_count;           /* number of used dof indices                 */
  DOF  hole_count;           /* number of FREED dof indices (NOT size-used)*/
  DOF  size_used;            /* > max. index of a used entry               */

  int  n_dof[DIM+1];         /* dofs from THIS dof_admin                   */
  int  n0_dof[DIM+1];        /* dofs from previous dof_admins              */

  DOF_INT_VEC     *dof_int_vec;           /* linked list of int vectors    */
  DOF_DOF_VEC     *dof_dof_vec;           /* linked list of dof vectors    */
  DOF_DOF_VEC     *int_dof_vec;           /* linked list of dof vectors    */
  DOF_UCHAR_VEC   *dof_uchar_vec;         /* linked list of u_char vectors */
  DOF_SCHAR_VEC   *dof_schar_vec;         /* linked list of s_char vectors */
  DOF_REAL_VEC    *dof_real_vec;          /* linked list of real vectors   */
  DOF_REAL_D_VEC  *dof_real_d_vec;        /* linked list of real_d vectors */
  DOF_MATRIX      *dof_matrix;            /* linked list of matrices       */

/*--------------------------------------------------------------------------*/
/*---  pointer for administration; don't touch!                          ---*/
/*--------------------------------------------------------------------------*/

  void            *mem_info;
};


/*--------------------------------------------------------------------------*/
/*  dof vector structures            					    */
/*--------------------------------------------------------------------------*/
/*  next:        pointer to next structure containing vector of same type   */
/*  fe_space:    pointer to fe_space  structure                             */
/*  refine_interpol: dof interpolation during refinement                    */
/*  coarse_restrict: restriction of linear functionals evaluated on a finer */
/*                   grid and stored in dof vector to the coarser grid      */
/*                   during coarsening                                      */
/*                or dof interpolation during coarsening                    */
/*  size:        allocated size of vector                         	    */
/*  vec[]:    vector entries (entry is used if dof index is used)	    */
/*--------------------------------------------------------------------------*/

struct dof_int_vec
{
  DOF_INT_VEC    *next;
  const FE_SPACE *fe_space;

  const char     *name;

  DOF            size;
  int            *vec;

  void  (*refine_interpol)(DOF_INT_VEC *, RC_LIST_EL *, int n);
  void  (*coarse_restrict)(DOF_INT_VEC *, RC_LIST_EL *, int n);
};

struct dof_dof_vec
{
  DOF_DOF_VEC    *next;
  const FE_SPACE *fe_space;

  const char     *name;

  DOF            size;
  DOF            *vec;

  void  (*refine_interpol)(DOF_DOF_VEC *, RC_LIST_EL *, int n);
  void  (*coarse_restrict)(DOF_DOF_VEC *, RC_LIST_EL *, int n);
};

struct dof_uchar_vec
{
  DOF_UCHAR_VEC  *next;
  const FE_SPACE *fe_space;

  const char     *name;
  DOF            size;
  U_CHAR         *vec;

  void  (*refine_interpol)(DOF_UCHAR_VEC *, RC_LIST_EL *, int n);
  void  (*coarse_restrict)(DOF_UCHAR_VEC *, RC_LIST_EL *, int n);
};

struct dof_schar_vec
{
  DOF_SCHAR_VEC  *next;
  const FE_SPACE *fe_space;

  const char     *name;

  DOF            size;
  S_CHAR         *vec;

  void  (*refine_interpol)(DOF_SCHAR_VEC *, RC_LIST_EL *, int n);
  void  (*coarse_restrict)(DOF_SCHAR_VEC *, RC_LIST_EL *, int n);
};

struct dof_real_vec
{
  DOF_REAL_VEC   *next;
  const FE_SPACE *fe_space;

  const char     *name;
  DOF            size;
  REAL           *vec;

  void  (*refine_interpol)(DOF_REAL_VEC *, RC_LIST_EL *, int n);
  void  (*coarse_restrict)(DOF_REAL_VEC *, RC_LIST_EL *, int n);
};

struct dof_real_d_vec
{
  DOF_REAL_D_VEC *next;
  const FE_SPACE *fe_space;

  const char     *name;
  DOF            size;
  REAL_D         *vec;

  void  (*refine_interpol)(DOF_REAL_D_VEC *, RC_LIST_EL *, int n);
  void  (*coarse_restrict)(DOF_REAL_D_VEC *, RC_LIST_EL *, int n);
};

/*--------------------------------------------------------------------------*/
/*  row structure for sparse matrix					    */
/*--------------------------------------------------------------------------*/
/*  next:        pointer to next structure containing entries of same row   */
/*  col[]:       column indices of entries (if >= 0; else unused)	    */
/*  entry[]:     matrix entries                        			    */
/*--------------------------------------------------------------------------*/

#define ROW_LENGTH 9

struct matrix_row
{
  MATRIX_ROW *next;
  DOF        col[ROW_LENGTH];    /* column indices */
  REAL       entry[ROW_LENGTH];  /* matrix entries */
};

#define ENTRY_USED(col)         ((col) >= 0)
#define ENTRY_NOT_USED(col)     ((col) < 0)
#define UNUSED_ENTRY    -1
#define NO_MORE_ENTRIES -2

#ifndef __CBLAS_H__
typedef enum { NoTranspose,
               Transpose,
               ConjugateTranspose } MatrixTranspose;
#endif

/*--------------------------------------------------------------------------*/
/*  sparse matrix with one row for each dof				    */
/*--------------------------------------------------------------------------*/
/*  next:        pointer to next matrix (linked list in MESH)               */
/*  matrix_row[]: pointers to row structures (or nil if row index is unused)*/
/*  size:         currently allocated size of matrix_row[]		    */
/*--------------------------------------------------------------------------*/

struct dof_matrix
{
  DOF_MATRIX     *next;
  const FE_SPACE *fe_space;
  const char     *name;

  MATRIX_ROW     **matrix_row;                 /* lists of matrix entries   */
  DOF            size;                         /* size of vector matrix_row */

  void       (*refine_interpol)(DOF_MATRIX *, RC_LIST_EL *, int n);
  void       (*coarse_restrict)(DOF_MATRIX *, RC_LIST_EL *, int n);
};

#endif  /* DOF_ADMIN_DEF */


/*--------------------------------------------------------------------------*/
/* information about size of leaf data, which will be hiden at child[1] and */
/* pointers to functions which will be used for transferring such           */
/* information during refinement/coarsening                                 */
/*--------------------------------------------------------------------------*/

struct leaf_data_info
{
  const char      *name;
  unsigned        leaf_data_size;
  void            (*refine_leaf_data)(EL *parent, EL *child[2]);
  void            (*coarsen_leaf_data)(EL *parent, EL *child[2]);
};

/*--------------------------------------------------------------------------*/
/* the structure stored in the entry mem_info of mesh.                      */
/* It stores information for memory management and macro elements.          */
/*--------------------------------------------------------------------------*/

struct mesh_mem_info
{
  void *dof_ptrs;
  void *dofs[DIM+1];
  void *element;

  void *leaf_data;
#if DIM == 3
  void *rc_list;
#endif
  void *real_d;
/*---8<---------------------------------------------------------------------*/
/*--- These last two entries are not really for memory management.       ---*/
/*--- They describe the array used to store macro element coordinate     ---*/
/*--- information, which is filled in macro_common.c.                    ---*/
/*--------------------------------------------------------------------->8---*/
  int   count;  
  REAL_D *coords;
};

/*--------------------------------------------------------------------------*/
/*  Here comes the MESH (giving access to the whole triangulation)	    */
/*--------------------------------------------------------------------------*/

struct mesh
{
  const char      *name;
  int             n_vertices;
  int             n_elements;
  int             n_hier_elements;

#if DIM == 2
  int             n_edges;
#endif

#if DIM == 3
  int             n_edges;
  int             n_faces;
  int             max_edge_neigh;
#endif

  int             n_macro_el;
  MACRO_EL        *first_macro_el;

  REAL            diam[DIM_OF_WORLD];
  PARAMETRIC      *parametric;

  LEAF_DATA_INFO  leaf_data_info[1];
  U_CHAR          preserve_coarse_dofs;

  DOF_ADMIN       **dof_admin;
  int             n_dof_admin;
  
  int             n_dof_el;    /* sum of all dofs from all admins           */
  int             n_dof[DIM+1];/* sum of vertex/edge/..dofs from all admins */
  int             n_node_el;   /* number of used nodes on each element      */
  int             node[DIM+1]; /* index of first vertex/edge/... node       */

/*--------------------------------------------------------------------------*/
/*---  pointer for administration; don't touch!                          ---*/
/*--------------------------------------------------------------------------*/

  void            *mem_info;
};

/*--------------------------------------------------------------------------*/
/*  stack data structure for non-recursive mesh traversal                   */
/*--------------------------------------------------------------------------*/

struct traverse_stack 
{
  MESH       *traverse_mesh;
  int        traverse_level;
  FLAGS      traverse_fill_flag;

  const MACRO_EL *traverse_mel;
  int        stack_size;
  int        stack_used;
  EL_INFO    *elinfo_stack;
  U_CHAR     *info_stack;
  
  const MACRO_EL *save_traverse_mel;
  EL_INFO    *save_elinfo_stack;
  U_CHAR     *save_info_stack;
  int        save_stack_used;

  int        el_count;

  TRAVERSE_STACK *next;
};

/*--------------------------------------------------------------------------*/
/*---  data structure for basis function representation                  ---*/
/*--------------------------------------------------------------------------*/

typedef REAL        BAS_FCT(const REAL[DIM+1]);
typedef const REAL  *GRD_BAS_FCT(const REAL[DIM+1]);
typedef const REAL  (*D2_BAS_FCT(const REAL[DIM+1]))[DIM+1];

struct bas_fcts
{
  char         *name;       /*  textual description                         */
  int          n_bas_fcts;  /*  number of basisfunctions on one el          */
  int          degree;      /*  maximal degree of the basis functions       */
  int          n_dof[DIM+1];/* dofs from these bas_fcts                     */

  void         (*init_element)(const EL_INFO *, const FE_SPACE *, U_CHAR);
  BAS_FCT      **phi;
  GRD_BAS_FCT  **grd_phi;
  D2_BAS_FCT   **D2_phi;

  const DOF    *(*get_dof_indices)(const EL *, const DOF_ADMIN *, DOF *);
  const S_CHAR *(*get_bound)(const EL_INFO *, S_CHAR *);

/*************** entries must be set for interpolation   ********************/

  const REAL   *(*interpol)(const EL_INFO *, int, const int *, 
			    REAL (*)(const REAL_D), REAL *);
  const REAL_D *(*interpol_d)(const EL_INFO *, int, const int *b_no,
			       const REAL *(*)(const REAL_D, REAL_D), 
			       REAL_D *);

/********************   optional entries  ***********************************/

  const int    *(*get_int_vec)(const EL *, const DOF_INT_VEC *, int *);
  const REAL   *(*get_real_vec)(const EL *, const DOF_REAL_VEC *, REAL *);
  const REAL_D *(*get_real_d_vec)(const EL *, const DOF_REAL_D_VEC *, REAL_D *);
  const U_CHAR *(*get_uchar_vec)(const EL *, const DOF_UCHAR_VEC *, U_CHAR *);
  const S_CHAR *(*get_schar_vec)(const EL *, const DOF_SCHAR_VEC *, S_CHAR *);

  void  (*real_refine_inter)(DOF_REAL_VEC *, RC_LIST_EL *, int);
  void  (*real_coarse_inter)(DOF_REAL_VEC *, RC_LIST_EL *, int);
  void  (*real_coarse_restr)(DOF_REAL_VEC *, RC_LIST_EL *, int);

  void  (*real_d_refine_inter)(DOF_REAL_D_VEC *, RC_LIST_EL *, int);
  void  (*real_d_coarse_inter)(DOF_REAL_D_VEC *, RC_LIST_EL *, int);
  void  (*real_d_coarse_restr)(DOF_REAL_D_VEC *, RC_LIST_EL *, int);


  void  *bas_fcts_data;
};


/*--------------------------------------------------------------------------*/
/*---  FE spaces are a triple of DOFs and BAS_FCTs on a MESH             ---*/
/*--------------------------------------------------------------------------*/

struct fe_space
{
  const char       *name;
  const DOF_ADMIN  *admin;
  const BAS_FCTS   *bas_fcts;
  MESH             *mesh;
};

/*--------------------------------------------------------------------------*/
/*---  data structures for numerical integration                         ---*/
/*--------------------------------------------------------------------------*/

struct quadrature
{
  char          *name;
  int           degree;

  int           dim;
  int           n_points;
  const double  **lambda;
  const double  *w;
};


#if DIM == 1
#define MAX_N_QUAD_POINTS  10
#endif
#if DIM == 2
#define MAX_N_QUAD_POINTS  61
#endif
#if DIM == 3
#define MAX_N_QUAD_POINTS  64
#endif

/*--------------------------------------------------------------------------*/
/*--- data structure with precomputed values of basis functions at       ---*/
/*--- quadrature nodes on the standard element                           ---*/
/*--------------------------------------------------------------------------*/

#define INIT_PHI       1
#define INIT_GRD_PHI   2
#define INIT_D2_PHI    4

struct quad_fast
{
  const QUAD        *quad;
  const BAS_FCTS    *bas_fcts;

  int               n_points;
  int               n_bas_fcts;
  const double      *w;

  U_CHAR            init_flag;

  REAL              **phi;
  REAL              (**grd_phi)[DIM+1];
  REAL              (**D2_phi)[DIM+1][DIM+1];
};

/*--------------------------------------------------------------------------*/
/*---  data structure for adaptive methods                               ---*/
/*--------------------------------------------------------------------------*/

struct adapt_stat
{
  const char  *name;
  REAL        tolerance;
  REAL        p;                         /* power in estimator norm        */
  int         max_iteration;
  int         info;

  REAL   (*estimate)(MESH *mesh, ADAPT_STAT *adapt);
  REAL   (*get_el_est)(EL *el);          /* local error estimate           */
  REAL   (*get_el_estc)(EL *el);         /* local coarsening error estimate*/
  U_CHAR (*marking)(MESH *mesh, ADAPT_STAT *adapt);

  void   *est_info;                      /* estimator parameters           */
  REAL   err_sum, err_max;               /* sum and max of el_est          */

  void   (*build_before_refine)(MESH *mesh, U_CHAR flag);
  void   (*build_before_coarsen)(MESH *mesh, U_CHAR flag);
  void   (*build_after_coarsen)(MESH *mesh, U_CHAR flag);
  void   (*solve)(MESH *mesh);

  int    refine_bisections;
  int    coarsen_allowed;                /* 0 : 1                          */
  int    coarse_bisections;

  int    strategy;                       /* 1=GR, 2=MS, 3=ES, 4=GERS       */
  REAL   MS_gamma, MS_gamma_c;           /* maximum strategy               */
  REAL   ES_theta, ES_theta_c;           /* equidistribution strategy      */
  REAL   GERS_theta_star, GERS_nu, GERS_theta_c;  /* willy's strategy      */
};


struct adapt_instat
{
  const char  *name;

  ADAPT_STAT adapt_initial[1];
  ADAPT_STAT adapt_space[1];

  REAL   time;
  REAL   start_time, end_time;
  REAL   timestep;

  void   (*init_timestep)(MESH *mesh, ADAPT_INSTAT *adapt);
  void   (*set_time)(MESH *mesh, ADAPT_INSTAT *adapt);
  void   (*one_timestep)(MESH *mesh, ADAPT_INSTAT *adapt);
  REAL   (*get_time_est)(MESH *mesh, ADAPT_INSTAT *adapt);
  void   (*close_timestep)(MESH *mesh, ADAPT_INSTAT *adapt);

  int    strategy; 
  int    max_iteration;

  REAL   tolerance;
  REAL   rel_initial_error;
  REAL   rel_space_error;
  REAL   rel_time_error;
  REAL   time_theta_1;
  REAL   time_theta_2;
  REAL   time_delta_1;
  REAL   time_delta_2;
  int    info;
};

#define MESH_REFINED   1
#define MESH_COARSENED 2

#define H1_NORM  1
#define L2_NORM  2

/*--------------------------------------------------------------------------*/
/*---  data structures for matrix and vector update                      ---*/
/*--------------------------------------------------------------------------*/

typedef struct el_matrix_info  EL_MATRIX_INFO;
struct el_matrix_info
{
  int              n_row;
  const DOF_ADMIN  *row_admin;
  const DOF        *(*get_row_dof)(const EL *,const DOF_ADMIN *, DOF *);

  int              n_col;
  const DOF_ADMIN  *col_admin;
  const DOF        *(*get_col_dof)(const EL *,const DOF_ADMIN *, DOF *);

  const S_CHAR     *(*get_bound)(const EL_INFO *, S_CHAR *);

  REAL             factor;

  const REAL       **(*el_matrix_fct)(const EL_INFO *, void *);
  void             *fill_info;

  FLAGS            fill_flag;
};

typedef struct el_vec_info  EL_VEC_INFO;
struct el_vec_info
{
  int              n_dof;
  const DOF_ADMIN  *admin;
  const DOF        *(*get_dof)(const EL *,const DOF_ADMIN *, DOF *);
  const S_CHAR     *(*get_bound)(const EL_INFO *, S_CHAR *);

  REAL             factor;

  const REAL       *(*el_vec_fct)(const EL_INFO *, void *);
  void             *fill_info;

  FLAGS            fill_flag;
};

typedef struct el_vec_d_info  EL_VEC_D_INFO;
struct el_vec_d_info
{
  int              n_dof;
  const DOF_ADMIN  *admin;
  const DOF        *(*get_dof)(const EL *,const DOF_ADMIN *, DOF *);
  const S_CHAR     *(*get_bound)(const EL_INFO *, S_CHAR *);

  REAL             factor;

  const REAL_D     *(*el_vec_fct)(const EL_INFO *, void *);
  void             *fill_info;

  FLAGS            fill_flag;
};

/*--------------------------------------------------------------------------*/
/*---  data structure about the differential operator for matrix assemblage */
/*--------------------------------------------------------------------------*/

typedef struct operator_info  OPERATOR_INFO;
struct operator_info
{
  const FE_SPACE *row_fe_space;
  const FE_SPACE *col_fe_space;

  const QUAD *quad[3];

  void       (*init_element)(const EL_INFO *, const QUAD *[3], void *);
  const REAL (*(*LALt)(const EL_INFO *, const QUAD *, int, void *))[DIM+1];
  int        LALt_pw_const;
  int        LALt_symmetric;
  const REAL *(*Lb0)(const EL_INFO *, const QUAD *, int, void *);
  int        Lb0_pw_const;
  const REAL *(*Lb1)(const EL_INFO *, const QUAD *, int, void *);
  int        Lb1_pw_const;
  int        Lb0_Lb1_anti_symmetric;
  REAL       (*c)(const EL_INFO *, const QUAD *, int, void *);
  int        c_pw_const;

  int        use_get_bound;
  void       *user_data;
  FLAGS      fill_flag;
};

/*--------------------------------------------------------------------------*/
/*  calculate element stiffness matrices by preevaluated integrals over the */
/*  the reference element.                                                  */
/*--------------------------------------------------------------------------*/

typedef struct q11_psi_phi   Q11_PSI_PHI;
typedef struct q01_psi_phi   Q01_PSI_PHI;
typedef struct q10_psi_phi   Q10_PSI_PHI;
typedef struct q00_psi_phi   Q00_PSI_PHI;

struct q11_psi_phi
{
  const BAS_FCTS     *psi;
  const BAS_FCTS     *phi;
  const QUAD         *quad;

  const int   **n_entries;
  const REAL  ***values;
  const int   ***k;
  const int   ***l;
};

struct q01_psi_phi
{
  const BAS_FCTS     *psi;
  const BAS_FCTS     *phi;
  const QUAD         *quad;

  const int   **n_entries;
  const REAL  ***values;
  const int   ***l;
};

struct q10_psi_phi
{
  const BAS_FCTS     *psi;
  const BAS_FCTS     *phi;
  const QUAD         *quad;

  const int   **n_entries;
  const REAL  ***values;
  const int   ***k;
};

struct q00_psi_phi
{
  const BAS_FCTS     *psi;
  const BAS_FCTS     *phi;
  const QUAD         *quad;

  const REAL  **values;
};


/*--------------------------------------------------------------------------*/
/*   data type for preconditioning                                          */
/*--------------------------------------------------------------------------*/

typedef struct precon PRECON;
struct precon
{
  void    *precon_data;

  int     (*init_precon)(void *precon_data);
  void    (*precon)(void *precon_data, int n, REAL *vec);
  void    (*exit_precon)(void *precon_data);
};

const PRECON *get_diag_precon_s(const DOF_MATRIX *, const DOF_SCHAR_VEC *);
const PRECON *get_diag_precon_d(const DOF_MATRIX *, const DOF_SCHAR_VEC *);

const PRECON *get_HB_precon_s(const FE_SPACE *, const DOF_SCHAR_VEC *, int, int);
const PRECON *get_HB_precon_d(const FE_SPACE *, const DOF_SCHAR_VEC *, int, int);

const PRECON *get_BPX_precon_s(const FE_SPACE *, const DOF_SCHAR_VEC *, int, int);
const PRECON *get_BPX_precon_d(const FE_SPACE *, const DOF_SCHAR_VEC *, int, int);

/*--------------------------------------------------------------------------*/
/*  solver identification for interface to OEM-lib                          */
/*--------------------------------------------------------------------------*/
typedef enum {NoSolver, BiCGStab, CG, GMRes, ODir, ORes} OEM_SOLVER;

/*--------------------------------------------------------------------------*/
/*--- abstract multigrid                                                 ---*/
/*--------------------------------------------------------------------------*/

typedef struct multi_grid_info MULTI_GRID_INFO;

struct multi_grid_info
{
  REAL             tolerance;                     /* tol. for resid         */
  REAL             exact_tolerance;               /* tol. for exact_solver  */

  int              cycle;                         /* 1=V-cycle, 2=W-cycle   */
  int              n_pre_smooth, n_in_smooth;     /* no of smoothing loops  */
  int              n_post_smooth;                 /* no of smoothing loops  */
  int              mg_levels;                     /* current no. of levels  */
  int              exact_level;                   /* level for exact_solver */
  int              max_iter;                      /* max. no of MG iter's   */
  int              info;

  int              (*init_multi_grid)(MULTI_GRID_INFO *mg_info);
  void             (*pre_smooth)(MULTI_GRID_INFO *mg_info, int level, int n);
  void             (*in_smooth)(MULTI_GRID_INFO *mg_info, int level, int n);
  void             (*post_smooth)(MULTI_GRID_INFO *mg_info, int level, int n);
  void             (*mg_restrict)(MULTI_GRID_INFO *mg_info, int level);
  void             (*mg_prolongate)(MULTI_GRID_INFO *mg_info, int level);
  void             (*exact_solver)(MULTI_GRID_INFO *mg_info, int level);
  REAL             (*mg_resid)(MULTI_GRID_INFO *mg_info, int level);
  void             (*exit_multi_grid)(MULTI_GRID_INFO *mg_info);

  void             *data;                         /* application dep. data */

};
int MG(MULTI_GRID_INFO *mg_info);

/*--------------------------------------------------------------------------*/
/*--- concrete multigrid                                                 ---*/
/*--------------------------------------------------------------------------*/

typedef struct mg_s_info MG_S_INFO;
struct mg_s_info
{
  MULTI_GRID_INFO  *mg_info;                      /* abstract MG info       */

  const FE_SPACE      *fe_space;
  DOF_MATRIX          *mat;
  const DOF_REAL_VEC  *f;
  DOF_REAL_VEC        *u;
  const DOF_SCHAR_VEC *bound;

  int              smoother,     exact_solver;
  REAL             smooth_omega, exact_omega;

  int              size;                          /* current size of vectors*/
  DOF_MATRIX       **matrix;                      /* one for each level     */
  REAL             **f_h;                         /* one for each level     */
  REAL             **u_h;                         /* one for each level     */
  REAL             **r_h;                         /* one for each level     */
  int              *dofs_per_level;               /* count dofs per level   */

  int              sort_size;                     /* size of sort vectors   */
  DOF              *sort_dof;                     /* dofs in order of levels*/
  DOF              *(dof_parent[2]);              /* (for linear elements)  */
  U_CHAR           *dof_level;
  S_CHAR           *sort_bound;                   /* sorted bound           */

  int              sort_invers_size;              /* size of inv. sort list */
  int              *sort_dof_invers;              /* inverse sort list      */
};
/*--------------------------------------------------------------------------*/
/*  sort_dof[ sorted dof ]          = unsorted dof                          */
/*  sort_dof_invers[ unsorted dof ] = sorted dof                            */
/*--------------------------------------------------------------------------*/

/* file MG_s1.c DOF_sort routines *******************************************/
void MG_s_setup_levels(MG_S_INFO *mg_s_info);
void MG_s_setup_mat_b(MG_S_INFO *mg_s_info,
		      DOF_MATRIX *mat, const DOF_SCHAR_VEC *bound);
void MG_s_dof_copy_to_sparse(MG_S_INFO *mg_s_info,
			     const DOF_REAL_VEC *x, REAL *y);
void MG_s_dof_copy_from_sparse(MG_S_INFO *mg_s_info, 
			       const REAL *x, DOF_REAL_VEC *y);
void MG_s_reset_mat(MG_S_INFO *mg_s_info);
void MG_s_sort_mat(MG_S_INFO *mg_s_info);
void MG_s_free_mem(MG_S_INFO *mg_s_info);

/* file MG_s2.c: DOF_sort independent routines ******************************/
void MG_s_restrict_mg_matrices(MG_S_INFO *mg_s_info);
void MG_s_restrict(MULTI_GRID_INFO *mg_info, int mg_level);
void MG_s_prolongate(MULTI_GRID_INFO *mg_info, int mg_level);
REAL MG_s_resid(MULTI_GRID_INFO *mg_info, int mg_level);
void MG_s_smoother(MULTI_GRID_INFO *mg_info, int mg_level, int n);
void MG_s_exact_solver(MULTI_GRID_INFO *mg_info, int mg_level);
void MG_s_gemv(MG_S_INFO *mg_s_info, int mg_level, MatrixTranspose transpose, 
	       REAL alpha, DOF_MATRIX *a, REAL *x, REAL beta, REAL *y);


/* file MG_s.c: *************************************************************/
int mg_s(DOF_MATRIX *matrix, DOF_REAL_VEC *u, const DOF_REAL_VEC *f,
	 const DOF_SCHAR_VEC *bound, 
	 REAL tol, int max_iter, int info, char *prefix);
MG_S_INFO *mg_s_init(DOF_MATRIX *matrix, const DOF_SCHAR_VEC *bound, 
		     int info, char *prefix);
int mg_s_solve(MG_S_INFO *mg_s_info,
	       DOF_REAL_VEC *u, const DOF_REAL_VEC *f, REAL tol, int max_iter);
void mg_s_exit(MG_S_INFO *mg_s_info);

/*--------------------------------------------------------------------------*/
/* Graphic output Definitions                                               */
/*--------------------------------------------------------------------------*/

typedef void * GRAPH_WINDOW;
typedef float  GRAPH_RGBCOLOR[3];

/** flags used by graph_mesh(): ****/
#define GRAPH_MESH_BOUNDARY        1
#define GRAPH_MESH_ELEMENT_MARK    2
#define GRAPH_MESH_VERTEX_DOF      4
#define GRAPH_MESH_ELEMENT_INDEX   8

/*--------------------------------------------------------------------------*/
/*  very usefull macro definitons                                           */
/*--------------------------------------------------------------------------*/

#define GET_MESH(na,f1,f2)\
 check_and_get_mesh(DIM,DIM_OF_WORLD,NEIGH_IN_EL,EL_INDEX,ALBERTA_VERSION,(na),(f1),(f2))

#define GET_DOF_VEC(ptr, dof_vec)\
   TEST_EXIT((dof_vec)  && ((ptr) = (dof_vec)->vec))\
      ("%s == nil\n", (dof_vec) ? NAME(dof_vec) : #dof_vec)


/*--------------------------------------------------------------------------*/
/*  defined in graphXO.c                                                    */
/*--------------------------------------------------------------------------*/
extern const GRAPH_RGBCOLOR rgb_black; 
extern const GRAPH_RGBCOLOR rgb_white;  
extern const GRAPH_RGBCOLOR rgb_red;    
extern const GRAPH_RGBCOLOR rgb_green;  
extern const GRAPH_RGBCOLOR rgb_blue;   
extern const GRAPH_RGBCOLOR rgb_yellow; 
extern const GRAPH_RGBCOLOR rgb_magenta;
extern const GRAPH_RGBCOLOR rgb_cyan;   
extern const GRAPH_RGBCOLOR rgb_grey50; 

extern const GRAPH_RGBCOLOR rgb_albert;
extern const GRAPH_RGBCOLOR rgb_alberta;

/*--------------------------------------------------------------------------*/
/*  functions supplied by ALBERTA                                           */
/*--------------------------------------------------------------------------*/

/***   file coarsen_common.c   **********************************************/
extern U_CHAR coarsen(MESH *mesh);
extern U_CHAR global_coarsen(MESH *mesh, int no);
extern int  get_max_level(MESH *mesh);
/***   file dof_admin.c   ***************************************************/
/*extern void add_bas_fcts_to_admin(DOF_ADMIN *admin, const BAS_FCTS *bas_fcts);*/
void add_element_matrix(DOF_MATRIX *, REAL, int, int, const DOF *, 
			const DOF *, const REAL **, const S_CHAR *);
void add_element_vec(DOF_REAL_VEC *, REAL, int, const DOF *,
		     const REAL *, const S_CHAR *);
void add_element_d_vec(DOF_REAL_D_VEC *, REAL, int, const DOF *,
		       const REAL_D *, const S_CHAR *);

extern void dof_compress(MESH *mesh);
extern void clear_dof_matrix(DOF_MATRIX *matrix);
/*  fill_dof_matrix is obsolete, only maintained for compatibility   */
extern void print_dof_matrix(const DOF_MATRIX *matrix);
extern void print_dof_real_vec(const DOF_REAL_VEC *drv);
extern void print_dof_real_d_vec(const DOF_REAL_D_VEC *drdv);
extern void print_dof_int_vec(const DOF_INT_VEC *div);
extern void print_dof_uchar_vec(const DOF_UCHAR_VEC *div);
extern void print_dof_schar_vec(const DOF_SCHAR_VEC *div);
/* BLAS 1 */
extern REAL dof_nrm2(const DOF_REAL_VEC *x);
extern REAL dof_asum(const DOF_REAL_VEC *x);
extern void dof_set(REAL alpha, DOF_REAL_VEC *x);
extern void dof_scal(REAL alpha, DOF_REAL_VEC *x);
extern REAL dof_dot(const DOF_REAL_VEC *x, const DOF_REAL_VEC *y);
extern void dof_copy(const DOF_REAL_VEC *x, DOF_REAL_VEC *y);
extern void dof_axpy(REAL alpha, const DOF_REAL_VEC *x, DOF_REAL_VEC *y);
/*  some non BLAS  */
extern void dof_xpay(REAL alpha, const DOF_REAL_VEC *x, DOF_REAL_VEC *y);
extern REAL dof_min(const DOF_REAL_VEC *x);
extern REAL dof_max(const DOF_REAL_VEC *x);
/* BLAS 2 */
extern void dof_gemv(MatrixTranspose transpose, REAL alpha,
		     const DOF_MATRIX *a, const DOF_REAL_VEC *x,
		     REAL beta, DOF_REAL_VEC *y);
extern void dof_mv(MatrixTranspose transpose, const DOF_MATRIX *a,
		   const DOF_REAL_VEC *x, DOF_REAL_VEC *y);

/* now the same for REAL_D */
extern void dof_axpy_d(REAL alpha, const DOF_REAL_D_VEC *x, DOF_REAL_D_VEC *y);
extern void dof_copy_d(const DOF_REAL_D_VEC *x, DOF_REAL_D_VEC *y);
extern REAL dof_dot_d(const DOF_REAL_D_VEC *x, const DOF_REAL_D_VEC *y);
extern REAL dof_nrm2_d(const DOF_REAL_D_VEC *x);
extern void dof_scal_d(REAL alpha, DOF_REAL_D_VEC *x);
extern void dof_set_d(REAL alpha, DOF_REAL_D_VEC *x);
extern void dof_xpay_d(REAL alpha, const DOF_REAL_D_VEC *x, DOF_REAL_D_VEC *y);
extern REAL dof_min_d(const DOF_REAL_D_VEC *x);
extern REAL dof_max_d(const DOF_REAL_D_VEC *x);
/*---    BLAS 2 for REAL_D                                               ---*/
extern void dof_mv_d(MatrixTranspose transpose, const DOF_MATRIX *a,
		     const DOF_REAL_D_VEC *x, DOF_REAL_D_VEC *y);
extern void dof_gemv_d(MatrixTranspose transpose, REAL alpha,
		       const DOF_MATRIX *a, const DOF_REAL_D_VEC *x,
		       REAL beta, DOF_REAL_D_VEC *y);

void update_matrix(DOF_MATRIX *dof_matrix, const EL_MATRIX_INFO *minfo);
void update_real_vec(DOF_REAL_VEC *drv, const EL_VEC_INFO *vec_info);
void update_real_d_vec(DOF_REAL_D_VEC *drdv, const EL_VEC_D_INFO *vecd_info);

/***   file macro.c   *******************************************************/
extern void macro_test(MACRO_DATA *data, const char *new_filename);
extern void read_macro(MESH *,const char *,const BOUNDARY *(*)(MESH *, int));
extern void read_macro_bin(MESH *, const char *, const BOUNDARY *(*)(MESH *, int));
extern void read_macro_xdr(MESH *, const char *, const BOUNDARY *(*)(MESH *, int));

extern int  write_macro(MESH *, const char *);
extern int  write_macro_bin(MESH *, const char *);
extern int  write_macro_xdr(MESH *, const char *);

extern int write_macro_data(MACRO_DATA *, const char *);
extern int write_macro_data_bin(MACRO_DATA *, const char *);
extern int write_macro_data_xdr(MACRO_DATA *, const char *);

extern MACRO_DATA *alloc_macro_data(int nv, int ne, FLAGS);
extern void free_macro_data(MACRO_DATA *);
extern void macro_data2mesh(MESH *, MACRO_DATA *, const BOUNDARY *(*)(MESH *, int));

extern MESH *check_and_get_mesh(int dim, int dow, int neigh, int index, 
				const char *version, const char *name,
				void (*init_dof_admins)(MESH *), 
				void (*init_leaf_data)(LEAF_DATA_INFO *));
const BOUNDARY *default_boundary(MESH *mesh, int bound);

/***   file memory.c   ******************************************************/
extern void free_dof_admin(DOF_ADMIN *admin, MESH *mesh);
extern void free_int_dof_vec(DOF_DOF_VEC *vec);
extern void free_dof_int_vec(DOF_INT_VEC *vec);
extern void free_dof_dof_vec(DOF_DOF_VEC *vec);
extern void free_dof_matrix(DOF_MATRIX *mat);
extern void free_dof_real_vec(DOF_REAL_VEC *vec);
extern void free_dof_real_d_vec(DOF_REAL_D_VEC *vec);
extern void free_dof_schar_vec(DOF_SCHAR_VEC *vec);
extern void free_dof_uchar_vec(DOF_UCHAR_VEC *vec);
extern void free_real_d(MESH *mesh, REAL *ptr);
extern void free_matrix_row(const FE_SPACE *, MATRIX_ROW *);
extern void free_element(EL *el, MESH *mesh);
#if DIM == 3
extern void free_rc_list(MESH *mesh, RC_LIST_EL *list);
#endif
extern void free_mesh(MESH *);
extern void free_dof(DOF *dof, MESH *mesh, int position);

/*
  extern DOF_ADMIN *get_dof_admin(MESH *mesh, const char *name,
                                  const int ndof[DIM+1], const BAS_FCTS *);
*/
extern DOF            *get_dof(MESH *mesh, int position);
extern const FE_SPACE *get_fe_space(MESH *mesh, const char *name,
				    const int ndof[DIM+1], const BAS_FCTS *);
extern DOF_INT_VEC    *get_dof_int_vec(const char *name, const FE_SPACE *);
extern DOF_DOF_VEC    *get_int_dof_vec(const char *name, const FE_SPACE *);
extern DOF_DOF_VEC    *get_dof_dof_vec(const char *name, const FE_SPACE *);
extern DOF_MATRIX     *get_dof_matrix(const char *name, const FE_SPACE *);
extern DOF_REAL_VEC   *get_dof_real_vec(const char *name, const FE_SPACE *);
extern DOF_REAL_D_VEC *get_dof_real_d_vec(const char *name, const FE_SPACE *);
extern DOF_SCHAR_VEC  *get_dof_schar_vec(const char *name, const FE_SPACE *);
extern DOF_UCHAR_VEC  *get_dof_uchar_vec(const char *name, const FE_SPACE *);
extern REAL           *get_real_d(MESH *mesh);
extern MATRIX_ROW     *get_matrix_row(const FE_SPACE *);
extern EL             *get_element(MESH *mesh);
extern void           *get_leaf_data(MESH *mesh);
#if DIM == 3
extern RC_LIST_EL     *get_rc_list(MESH *mesh);
#endif
extern MESH *get_mesh(const char *name, void (*init_dof_admins)(MESH *),
		      void (*init_leaf_data)(LEAF_DATA_INFO *));

/*--   file level.c   ------------------------------------------------------*/
int find_level(MESH *mesh, FLAGS fill_flag, const DOF_REAL_VEC *Level,
               REAL value, int (*init)(const EL_INFO *, REAL [], int, int,
                                       const REAL [][DIM+1]),
               void (*)(const EL_INFO *, REAL [], int, int,
                        const REAL [][DIM+1], const REAL_D [DIM]));

extern REAL level_element_det(const REAL_D [3]);
extern void level_coord_to_world(const REAL_D [DIM], const REAL [DIM],
                                 REAL_D);
extern void level_coord_to_el_coord(const REAL [][DIM+1], const REAL [DIM], 
                                    REAL [DIM+1]);
extern void set_element_mark(MESH *mesh, FLAGS fill_flag, S_CHAR mark);

#if DIM == 3
void movi_find_level(MESH *, const DOF_REAL_VEC *, REAL, int (*)(const EL *),
                     const char *);
#endif
#if DIM == 2  && DIM_OF_WORLD == 2
void write_ps_level(MESH *mesh, const char *filename, const char *mode,
                    const char *title, const REAL x[2], const REAL y[2], 
                    int keepaspect, const DOF_REAL_VEC *level, REAL value, 
                    int (*is_d)(const EL *), int draw_bound);
#endif

/***   file numint.c   ******************************************************/
const QUAD *get_quadrature(unsigned dim, unsigned degree);
const QUAD *get_lumping_quadrature(int dim);
void print_quadrature(const QUAD *quad);
REAL  integrate_std_simp(const QUAD *quad, REAL (*f)(const REAL *));
const REAL   *f_at_qp(const QUAD*, REAL (*f)(const REAL [DIM+1]), REAL*);
int   max_quad_points(void);
const REAL_D *f_d_at_qp(const QUAD *quad, const REAL *(*f)(const REAL[DIM+1]),
			REAL_D *vec);
const REAL_D *grd_f_at_qp(const QUAD *, const REAL *(*)(const REAL [DIM+1]), 
			  REAL_D *);
const REAL_DD *grd_f_d_at_qp(const QUAD *, 
		           const REAL_D *(*)(const REAL [DIM+1]), REAL_DD *);
const QUAD_FAST *get_quad_fast(const BAS_FCTS *, const QUAD *, 
				      U_CHAR init_flag);
/***   file refine_common.c   ***********************************************/
extern U_CHAR refine(MESH *mesh);
extern U_CHAR global_refine(MESH *mesh, int mark);

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/***   file adapt.c   *******************************************************/
extern void adapt_method_stat(MESH *mesh, ADAPT_STAT *adapt);
extern void adapt_method_instat(MESH *mesh, ADAPT_INSTAT *adapt);
extern int marking(MESH *mesh, ADAPT_STAT *adapt);
extern ADAPT_INSTAT *get_adapt_instat(const char *name, const char *prefix,
				      int info, ADAPT_INSTAT *adapt_instat);
extern ADAPT_STAT *get_adapt_stat(const char *name, const char *prefix,
				  int info, ADAPT_STAT *adapt_stat);

/***   file assemble.c   ****************************************************/
const EL_MATRIX_INFO *fill_matrix_info(const OPERATOR_INFO *, 
				       EL_MATRIX_INFO *);
const Q00_PSI_PHI *get_q00_psi_phi(const BAS_FCTS *psi, const BAS_FCTS *phi,
				   const QUAD *quad);
const Q01_PSI_PHI *get_q01_psi_phi(const BAS_FCTS *psi, const BAS_FCTS *phi,
				   const QUAD *quad);
const Q10_PSI_PHI *get_q10_psi_phi(const BAS_FCTS *psi, const BAS_FCTS *phi,
				   const QUAD *quad);
const Q11_PSI_PHI *get_q11_psi_phi(const BAS_FCTS *psi, const BAS_FCTS *phi,
				   const QUAD *quad);

/***   file bas_fct.c   *****************************************************/
extern const BAS_FCTS *get_bas_fcts(const char *name);
extern const BAS_FCTS *get_lagrange(int degree);
extern int new_bas_fcts(const BAS_FCTS * bas_fcts);

int cghs(const DOF_MATRIX *mat, DOF_REAL_VEC *u, const DOF_REAL_VEC *f,
	 const DOF_SCHAR_VEC *b, REAL tol, int icon, int max_iter, int info);
int cgn(const DOF_MATRIX *mat, DOF_REAL_VEC *u, const DOF_REAL_VEC *f,
	const DOF_SCHAR_VEC *b, REAL tol, int icon, int max_iter, int info);

/***   file check.c   *******************************************************/
extern void check_mesh(MESH *mesh);
extern void interpol(REAL (*fct)(const REAL_D), DOF_REAL_VEC *);

/***   file dof_admin.c   ***************************************************/
extern void free_dof_index(DOF_ADMIN *admin, int dof);
extern int  get_dof_index(DOF_ADMIN *admin);
extern void enlarge_dof_lists(DOF_ADMIN *admin, int minsize);
extern void update_dof_matrix(DOF_MATRIX *, REAL, int, const REAL **, 
			      const DOF *, const S_CHAR *);
extern void update_dof_real_vec(DOF_REAL_VEC *, REAL, int, const REAL *,
				const DOF *, const S_CHAR *);
extern void update_dof_real_d_vec(DOF_REAL_D_VEC *, REAL, int, const REAL_D *,
				  const DOF *, const S_CHAR *);
extern void test_dof_matrix(DOF_MATRIX *matrix);

/***   file element.c   ***************************************************/
extern REAL el_volume(const EL_INFO *);
extern REAL el_det(const EL_INFO *);
extern REAL el_grd_lambda(const EL_INFO *, REAL grd_lam[DIM+1][DIM_OF_WORLD]);
extern REAL get_face_normal(const EL_INFO *, int , REAL *);
extern const int *sort_face_indices(const EL *el, int face, int *vec);
#if DIM_OF_WORLD == 3
extern void vector_product(const REAL_D x, const REAL_D y, REAL_D z);
#endif
/***   file estimator.c   **************************************************/
#define INIT_UH 1
#define INIT_GRD_UH 2
REAL ellipt_est(const DOF_REAL_VEC *uh, ADAPT_STAT *adapt,
		REAL *(rw_est)(EL *), REAL *(rw_estc)(EL *),
                int degree, int norm, REAL C[3], const REAL_DD A,
                REAL (*f)(const EL_INFO *,const QUAD *,int,REAL,const REAL_D),
		FLAGS f_flag);
REAL heat_est(const DOF_REAL_VEC *uh, ADAPT_INSTAT *adapt,
	      REAL *(rw_est)(EL *), REAL *(rw_estc)(EL *),
	      int degree, REAL C[4], 
	      const DOF_REAL_VEC *uh_old, const REAL_DD A,
	      REAL (*f)(const EL_INFO *, const QUAD *, int iq, REAL t,
			REAL u, const REAL_D grd_u),
	      FLAGS f_flag);
/***   file error.c   *******************************************************/
REAL max_err_at_qp(REAL (*)(const REAL_D), const DOF_REAL_VEC *, const QUAD *);
REAL max_err_d_at_qp(const REAL *(*)(const REAL_D, REAL_D), 
		     const DOF_REAL_D_VEC *, const QUAD *);
REAL max_err_at_vert(REAL (*)(const REAL_D), const DOF_REAL_VEC *);
REAL L2_err(REAL (*)(const REAL_D), const DOF_REAL_VEC *, const QUAD *, int, 
	     REAL *(*)(EL *), REAL *);
REAL H1_err(const REAL *(*)(const REAL_D), const DOF_REAL_VEC *, 
	     const QUAD *, int, REAL *(*)(EL *), REAL *);
REAL H1_err_d(const REAL_D *(*grd_u)(const REAL_D, REAL_DD),
	       const DOF_REAL_D_VEC *uh, const QUAD *quad, int rel_err,
	       REAL *(*rw_err_el)(EL *), REAL *max_h1_err2);
REAL L2_err_d(const REAL *(*)(const REAL_D, REAL_D), const DOF_REAL_D_VEC *,
	       const QUAD *, int, REAL *(*)(EL *), REAL *);

/***   file eval.c   ********************************************************/
REAL eval_uh(const REAL [DIM+1], const REAL *, const BAS_FCTS *);
REAL eval_uh_fast(const REAL *, const REAL *, int);
const REAL *eval_grd_uh(const REAL [DIM+1], const REAL_D [DIM+1],
			const REAL *, const BAS_FCTS *, REAL_D);
const REAL *eval_grd_uh_fast(const REAL_D [DIM+1], const REAL *,
			     const REAL (*)[DIM+1], int , REAL_D);
const REAL_D *eval_D2_uh(const REAL [DIM+1], const REAL_D [DIM+1],
			 const REAL *, const BAS_FCTS *, REAL_DD);
const REAL_D *eval_D2_uh_fast(const REAL_D [DIM+1], const REAL *, 
			      const REAL (*)[DIM+1][DIM+1], int , REAL_DD);

const REAL *eval_uh_d(const REAL [DIM+1], const REAL_D *, const BAS_FCTS *,
		      REAL_D);
const REAL *eval_uh_d_fast(const REAL_D *, const REAL *, int, REAL_D);
const REAL_D *eval_grd_uh_d(const REAL [DIM+1], const REAL_D [DIM+1],
			     const REAL_D *, const BAS_FCTS *, REAL_DD);
const REAL_D *eval_grd_uh_d_fast(const REAL_D [DIM+1], const REAL_D *, 
				 const REAL (*)[DIM+1], int, REAL_DD);
const REAL_DD *eval_D2_uh_d(const REAL [DIM+1], const REAL_D [DIM+1],
			    const REAL_D *, const BAS_FCTS *, REAL_DD *);
const REAL_DD *eval_D2_uh_d_fast(const REAL_D [DIM+1], const REAL_D *, 
				 const REAL (*)[DIM+1][DIM+1], int,REAL_DD *);


const REAL *uh_at_qp(const QUAD_FAST *, const REAL *uh_loc, REAL *vec);
const REAL_D *grd_uh_at_qp(const QUAD_FAST *, const REAL_D grd_lambda[DIM+1],
			   const REAL *uh_loc, REAL_D *vec);
const REAL_DD *D2_uh_at_qp(const QUAD_FAST *, const REAL_D [DIM+1],
			   const REAL *, REAL_DD *);
const REAL_D *uh_d_at_qp(const QUAD_FAST *fast, const REAL_D *uh_loc,
			 REAL_D *vec);
const REAL_DD *grd_uh_d_at_qp(const QUAD_FAST *fast,
			      const REAL_D grd_lambda[DIM+1], 
			      const REAL_D *uh_loc, REAL_DD *vec);
const REAL_DD (*D2_uh_d_at_qp(const QUAD_FAST *, const REAL_D [DIM+1],
			      const REAL_D *, 
			      REAL_DD (*)[DIM_OF_WORLD]))[DIM_OF_WORLD];
REAL eval_div_uh_d(const REAL lambda[DIM+1], const REAL_D grd_lambda[DIM+1],
		   const REAL_D *uh_loc, const BAS_FCTS *b);
REAL eval_div_uh_d_fast(const REAL_D grd_lambda[DIM+1], const REAL_D *uh_loc, 
			const REAL (*grd_phi)[DIM+1], int n_bas_fcts);
const REAL *div_uh_d_at_qp(const QUAD_FAST *fast, const REAL_D Lambda[DIM+1], 
			   const REAL_D *uh_loc, REAL *vec);

const REAL *coord_to_world(const EL_INFO *, const REAL *, REAL_D);
REAL H1_norm_uh(const QUAD *quad, const DOF_REAL_VEC *u_h);
REAL L2_norm_uh(const QUAD *quad, const DOF_REAL_VEC *u_h);
REAL H1_norm_uh_d(const QUAD *quad, const DOF_REAL_D_VEC *u_h);
REAL L2_norm_uh_d(const QUAD *quad, const DOF_REAL_D_VEC *u_h);

/***   file graphXO.c   ******************************************************/
GRAPH_WINDOW graph_open_window(const char *title, const char *geometry,
			       REAL *world, MESH *mesh);
void graph_close_window(GRAPH_WINDOW win);
void graph_clear_window(GRAPH_WINDOW win, const GRAPH_RGBCOLOR c);
void graph_mesh(GRAPH_WINDOW win, MESH *mesh, const GRAPH_RGBCOLOR c,
		FLAGS flag);
void graph_drv(GRAPH_WINDOW win, const DOF_REAL_VEC *uh,
	       REAL min, REAL max, int refine);
void graph_drv_d(GRAPH_WINDOW win, const DOF_REAL_D_VEC *uh,
		 REAL min, REAL max, int refine);
void graph_el_est(GRAPH_WINDOW win, MESH *mesh, REAL (*get_el_est)(EL *el),
		  REAL min, REAL max);
void graph_fvalues(GRAPH_WINDOW win, MESH *mesh,
		   REAL(*fct)(const EL_INFO *el_info, const REAL *lambda),
		   FLAGS flags, REAL min, REAL max, int refine);
void graph_point(GRAPH_WINDOW, const REAL [2], const GRAPH_RGBCOLOR, float);
void graph_line(GRAPH_WINDOW, const REAL [2], const REAL [2], 
		const GRAPH_RGBCOLOR, float);
#if DIM == 1
void graph1d_drv(GRAPH_WINDOW win, const DOF_REAL_VEC *uh, 
		 REAL min, REAL max, int refine, const GRAPH_RGBCOLOR c);
void graph1d_drv_d(GRAPH_WINDOW win, const DOF_REAL_D_VEC *uh, 
		   REAL min, REAL max, int refine, const GRAPH_RGBCOLOR c);
void graph1d_el_est(GRAPH_WINDOW win, MESH *mesh, REAL (*get_el_est)(EL *el),
		    REAL min, REAL max, const GRAPH_RGBCOLOR c);
#endif
#if DIM == 2
void graph_level(GRAPH_WINDOW win, const DOF_REAL_VEC *v, REAL level,
		 const GRAPH_RGBCOLOR c, int refine);
void graph_levels(GRAPH_WINDOW win, const DOF_REAL_VEC *v,
		  int n, REAL const *levels, const GRAPH_RGBCOLOR *color,
		  int refine);
void graph_level_d(GRAPH_WINDOW, const DOF_REAL_D_VEC *,
		   REAL, const GRAPH_RGBCOLOR, int);
void graph_levels_d(GRAPH_WINDOW, const DOF_REAL_D_VEC *,
		    int, const REAL *, const GRAPH_RGBCOLOR *, int);
void graph_values(GRAPH_WINDOW, const DOF_REAL_VEC *, REAL, REAL, int);
void graph_values_d(GRAPH_WINDOW, const DOF_REAL_D_VEC *, REAL, REAL, int);
/* multigrid level display routines:  */
void graph_mesh_mg(GRAPH_WINDOW win, MESH *mesh, const GRAPH_RGBCOLOR c,
		    FLAGS flags, int mg_level);
void graph_values_mg(GRAPH_WINDOW win, const DOF_REAL_VEC *v, 
		     REAL min, REAL max, int refine,
		     int mg_level, const FE_SPACE *fe_space,
		     const int *sort_dof_invers);
#endif

/***   file l2scp.c   *******************************************************/
void dirichlet_bound(REAL (*)(const REAL_D), DOF_REAL_VEC *, DOF_REAL_VEC *,
		     DOF_SCHAR_VEC *);
void L2scp_fct_bas(REAL (*f)(const REAL_D), const QUAD *, DOF_REAL_VEC *fh);

void dirichlet_bound_d(const REAL *(*)(const REAL_D, REAL_D),
		       DOF_REAL_D_VEC *, DOF_REAL_D_VEC *, DOF_SCHAR_VEC *);
void interpol_d(const REAL *(*)(const REAL_D, REAL_D), DOF_REAL_D_VEC *);
void L2scp_fct_bas_d(const REAL *(*f)(const REAL_D, REAL_D), const QUAD *,
		     DOF_REAL_D_VEC *fhd);

/*---  file oem_solve_s.c  -------------------------------------------------*/
int oem_solve_s(const DOF_MATRIX *A, const DOF_REAL_VEC *f, 
		DOF_REAL_VEC *u,  OEM_SOLVER, REAL, int, int, int, int);
int mat_vec_s(void *ud, int dim, const REAL *x, REAL *y);
void *init_mat_vec_s(MatrixTranspose, const DOF_MATRIX *, const DOF_SCHAR_VEC *);
void exit_mat_vec_s(void *);

/*---  file oem_solve_d.c  -------------------------------------------------*/
int oem_solve_d(const DOF_MATRIX *A, const DOF_REAL_D_VEC *f, 
		DOF_REAL_D_VEC *u,  OEM_SOLVER, REAL, int, int, int, int);
int mat_vec_d(void *ud, int dim, const REAL *x, REAL *y);
void *init_mat_vec_d(MatrixTranspose, const DOF_MATRIX *, const DOF_SCHAR_VEC *);
void exit_mat_vec_d(void *);

/***  file sor.c  ***********************************************************/
int sor_d(DOF_MATRIX *a, const DOF_REAL_D_VEC *f, const DOF_SCHAR_VEC *b,
	  DOF_REAL_D_VEC *u, REAL omega, REAL tol, int max_iter, int info);
int sor_s(DOF_MATRIX *a, const DOF_REAL_VEC *f, const DOF_SCHAR_VEC *b,
	  DOF_REAL_VEC *u, REAL omega, REAL tol, int max_iter, int info);

/***  file ssor.c  **********************************************************/
int ssor_d(DOF_MATRIX *a, const DOF_REAL_D_VEC *f, const DOF_SCHAR_VEC *b,
	   DOF_REAL_D_VEC *u, REAL omega, REAL tol, int max_iter, int info);
int ssor_s(DOF_MATRIX *a, const DOF_REAL_VEC *f, const DOF_SCHAR_VEC *b,
	   DOF_REAL_VEC *u, REAL omega, REAL tol, int max_iter, int info);

/***   file traverse_r.c  ***************************************************/
void mesh_traverse(MESH *mesh, int level, FLAGS fill_flag,
		   void (*el_fct)(const EL_INFO *));
void fill_macro_info(MESH *mesh, const MACRO_EL *mel, EL_INFO *elinfo);
void fill_elinfo(int ichild, const EL_INFO *parent_info, EL_INFO *elinfo);
void update_elinfo(EL_INFO *elinfo);  /* only 3d */
void test_traverse(MESH *mesh, int level, FLAGS fill_flag);

/***   file traverse_nr.c ***************************************************/
TRAVERSE_STACK *get_traverse_stack(void);
void free_traverse_stack(TRAVERSE_STACK *stack);
const EL_INFO *traverse_first(TRAVERSE_STACK *stack,
			MESH *mesh, int level, FLAGS fill_flag);
const EL_INFO *traverse_next(TRAVERSE_STACK *stack, const EL_INFO *);
const EL_INFO *traverse_neighbour(TRAVERSE_STACK *stack, const EL_INFO *,
				  int neighbour);
void update_elinfo_stack(TRAVERSE_STACK *stack);  /* only 3d */
void test_traverse_nr(MESH *mesh, int level, FLAGS fill_flag);

/*---  file trav_xg.c  -----------------------------------------------------*/
int world_to_coord(const EL_INFO *el_info, const REAL *, REAL [DIM+1]);
extern int find_el_at_pt(MESH *mesh, const REAL_D xy,
			 EL_INFO **el_info_p, FLAGS flag, REAL bary[DIM+1],
			 const MACRO_EL *start_mel, const REAL_D xy0, REAL *sp);

/***   file read_mesh.c  *********************************************/
MESH *read_mesh(const char *, REAL *, void (*)(LEAF_DATA_INFO *),
		const BOUNDARY *(*)(MESH *, int));
DOF_REAL_VEC   *read_dof_real_vec(const char *, MESH *, FE_SPACE *);
DOF_REAL_D_VEC *read_dof_real_d_vec(const char *, MESH *, FE_SPACE *);
DOF_INT_VEC    *read_dof_int_vec(const char *, MESH *, FE_SPACE *);
DOF_SCHAR_VEC  *read_dof_schar_vec(const char *, MESH *, FE_SPACE *);
DOF_UCHAR_VEC  *read_dof_uchar_vec(const char *, MESH *, FE_SPACE *);

/***   file read_mesh_xdr.c  *********************************************/
MESH *read_mesh_xdr(const char *, REAL *, void (*)(LEAF_DATA_INFO *),
		    const BOUNDARY *(*)(MESH *, int));
DOF_REAL_VEC   *read_dof_real_vec_xdr(const char *, MESH *, FE_SPACE *);
DOF_REAL_D_VEC *read_dof_real_d_vec_xdr(const char *, MESH *, FE_SPACE *);
DOF_INT_VEC    *read_dof_int_vec_xdr(const char *, MESH *, FE_SPACE *);
DOF_SCHAR_VEC  *read_dof_schar_vec_xdr(const char *, MESH *, FE_SPACE *);
DOF_UCHAR_VEC  *read_dof_uchar_vec_xdr(const char *, MESH *, FE_SPACE *);

/***   file write_mesh.c  ********************************************/
int write_mesh(MESH *, const char *, REAL);
int write_dof_real_vec(const DOF_REAL_VEC *, const char *);
int write_dof_real_d_vec(const DOF_REAL_D_VEC *, const char *);
int write_dof_int_vec(const DOF_INT_VEC *, const char *);
int write_dof_schar_vec(const DOF_SCHAR_VEC *, const char *);
int write_dof_uchar_vec(const DOF_UCHAR_VEC *, const char *);

/***   file write_mesh_xdr.c  ********************************************/
int write_mesh_xdr(MESH *, const char *, REAL);
int write_dof_real_vec_xdr(const DOF_REAL_VEC *, const char *);
int write_dof_real_d_vec_xdr(const DOF_REAL_D_VEC *, const char *);
int write_dof_int_vec_xdr(const DOF_INT_VEC *, const char *);
int write_dof_schar_vec_xdr(const DOF_SCHAR_VEC *, const char *);
int write_dof_uchar_vec_xdr(const DOF_UCHAR_VEC *, const char *);

/*--  file write_mesh_ps.c  ------------------------------------------------*/
void write_mesh_ps(MESH *mesh, const char *filename, const char *title,
                   const REAL x[2], const REAL y[2], int keepaspect, 
                   int draw_bound);

/*--------------------------------------------------------------------------*/
/*  interface for Lagrange elements for the gltools                         */
/*  file  gltools.c                                                         */
/*--------------------------------------------------------------------------*/
typedef void*    GLTOOLS_WINDOW;

GLTOOLS_WINDOW open_gltools_window(const char *, const char *, const REAL *,
				   MESH *, int);
void close_gltools_window(GLTOOLS_WINDOW);

void gltools_est(GLTOOLS_WINDOW, MESH *, REAL (*)(EL *), REAL, REAL);
void gltools_disp_mesh(GLTOOLS_WINDOW, MESH *, int, const DOF_REAL_D_VEC *);
void gltools_mesh(GLTOOLS_WINDOW win, MESH *, int);
void gltools_disp_drv(GLTOOLS_WINDOW, const DOF_REAL_VEC *,
		      REAL, REAL, const DOF_REAL_D_VEC *);
void gltools_drv(GLTOOLS_WINDOW, const DOF_REAL_VEC *, REAL, REAL);
void gltools_disp_drv_d(GLTOOLS_WINDOW, const DOF_REAL_D_VEC *, 
			REAL, REAL, const DOF_REAL_D_VEC *);
void gltools_drv_d(GLTOOLS_WINDOW, const DOF_REAL_D_VEC *, REAL, REAL);
void gltools_disp_vec(GLTOOLS_WINDOW, const DOF_REAL_D_VEC *, 
		      REAL, REAL, const DOF_REAL_D_VEC *);
void gltools_vec(GLTOOLS_WINDOW, const DOF_REAL_D_VEC *, REAL, REAL);

#if 0
/*--------------------------------------------------------------------------*/
/*  fast quadrature on faces  (only preliminary)                            */
/*--------------------------------------------------------------------------*/

typedef struct face_quad_fast FACE_QUAD_FAST;
struct face_quad_fast
{
  const QUAD        *quad;
  const BAS_FCTS    *bas_fcts;
		  
  int               n_points;
  int               n_bas_fcts;
  const double      *w;
#if DIM == 2	  
  const double      (*lambda[N_EDGES])[DIM+1];
#elif DIM == 3	  
  const double      (*lambda[N_FACES])[DIM+1];
#endif		  
  U_CHAR            init_flag;
		  
#if DIM == 2	  
  REAL              **phi[N_EDGES];
#elif DIM == 3	  
  REAL              **phi[N_FACES];
#endif		  
		  
  FACE_QUAD_FAST    *next;
};
const FACE_QUAD_FAST *get_face_quad_fast(const BAS_FCTS *, const QUAD *, 
					 U_CHAR init_flag);
#endif

#ifdef __cplusplus
}
#endif

#endif  /* !_ALBERTA_H_ */
