/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     error.c                                                        */
/*                                                                          */
/* description:  routines for computing different types of errors           */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include "alberta.h"

static const QUAD_FAST  *quad_fast = nil;
static int              relative;

/*--------------------------------------------------------------------------*/
/*  discrete solution                                                       */
/*--------------------------------------------------------------------------*/
static const DOF_REAL_VEC    *err_uh;
static const DOF_REAL_D_VEC  *err_uh_d;
static const BAS_FCTS  *bas_fcts;

/*--------------------------------------------------------------------------*/
/*  data on a single element                                                */
/*--------------------------------------------------------------------------*/
static const EL_INFO *elinfo;
static const REAL    *(*get_real_vec_el)(const EL *, const DOF_REAL_VEC *,
					   REAL *);
static const REAL_D  *(*get_real_d_vec_el)(const EL *, const DOF_REAL_D_VEC *,
					   REAL_D *);
static REAL          *(*rw_error)(EL *el);

static PARAMETRIC    *el_parametric = nil;

/*--------------------------------------------------------------------------*/
/*  functions handed over to the error routines evaluated in local          */
/*  coordinates                                                             */
/*--------------------------------------------------------------------------*/

static REAL   (*p_u)(const REAL_D);

static REAL err_u(const REAL lambda[DIM+1])
{
  REAL_D  x[1];

  if (el_parametric) {
    el_parametric->coord_to_world(elinfo, nil, 1,
				  (const REAL (*)[DIM+1])lambda, x);
  }
  else {
    coord_to_world(elinfo, lambda, *x);
  }
  return((*p_u)(*x));
}

static const REAL   *(*p_u_d)(const REAL_D, REAL_D);

static const REAL *err_u_d(const REAL lambda[DIM+1])
{
  REAL_D  x[1];

  if (el_parametric) {
    el_parametric->coord_to_world(elinfo, nil, 1,
				  (const REAL (*)[DIM+1])lambda, x);
  }
  else {
    coord_to_world(elinfo, lambda, x[0]);
  }
  return((*p_u_d)(x[0], nil));
}

static const REAL  *(*p_grd_u)(const REAL_D);

static const REAL *err_grd_u(const REAL lambda[DIM+1])
{
  REAL_D  x[1];

  if (el_parametric) {
    el_parametric->coord_to_world(elinfo, nil, 1,
				  (const REAL (*)[DIM+1])lambda, x);
  }
  else {
    coord_to_world(elinfo, lambda, *x);
  }
  
  return((*p_grd_u)(*x));
}

static const REAL_D  *(*p_grd_u_d)(const REAL_D, REAL_D [DIM_OF_WORLD]);

static const REAL_D *err_grd_u_d(const REAL lambda[DIM+1])
{
  REAL_D  x[1];

  if (el_parametric) {
    el_parametric->coord_to_world(elinfo, nil, 1,
				  (const REAL (*)[DIM+1])lambda, x);
  }
  else {
    coord_to_world(elinfo, lambda, *x);
  }
  
  return((*p_grd_u_d)(*x, nil));
}

/*--------------------------------------------------------------------------*/
/*  and now, the error functions :-)))                                      */
/*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/*  max error at the quadrature points :-)))                                */
/*--------------------------------------------------------------------------*/

static REAL  max_err;

static void max_err_at_qp_fct(const EL_INFO *el_info)
{
  int         i;
  REAL        err;
  const REAL  *u_vec, *uh_vec, *uh_el;
  PARAMETRIC  *parametric = el_info->mesh->parametric;

  elinfo = el_info;
  if (parametric && parametric->init_element(el_info, parametric)) {
    el_parametric = parametric;
  }
  else {
    el_parametric = nil;
  }

  u_vec = f_at_qp(quad_fast->quad, err_u, nil);

  uh_el = (*get_real_vec_el)(el_info->el, err_uh, nil);
  uh_vec = uh_at_qp(quad_fast, uh_el, nil);

  for (i = 0; i < quad_fast->n_points; i++)
  {
    err = u_vec[i] > uh_vec[i] ? u_vec[i] - uh_vec[i] : uh_vec[i] - u_vec[i];
    max_err = MAX(max_err, err);
  }

  return;
}

/*--------------------------------------------------------------------------*/
/*  max error at the quadrutare points of quadrature formula of degree      */
/*  of the continous function u and discrete function uh                    */
/*--------------------------------------------------------------------------*/

REAL max_err_at_qp(REAL (*u)(const REAL_D), const DOF_REAL_VEC *uh,
		   const QUAD *quad)
{
  FUNCNAME("max_err_at_qp");
  const FE_SPACE *fe_space;

  if (!(p_u = u))
  {
    ERROR("no function u specified; doing nothing\n");
    return(-1.0);
  }
  if (!(err_uh = uh)  ||  !(fe_space = uh->fe_space))
  {
    ERROR("no discrete function or no fe_space for it; doing nothing\n");
    return(-1.0);
  }
  if (!uh->vec)
  {
    ERROR("no coefficient vector at discrete solution ; doing nothing\n");
    return(-1.0);
  }
  if (!(bas_fcts = fe_space->bas_fcts))
  {
    ERROR("no basis functions at discrete solution ; doing nothing\n");
    return(-1.0);
  }
  if (!quad)
    quad = get_quadrature(DIM, 2*fe_space->bas_fcts->degree-2);
  quad_fast = get_quad_fast(bas_fcts, quad, INIT_PHI);

  get_real_vec_el = fe_space->bas_fcts->get_real_vec;

  max_err = 0.0;
  mesh_traverse(fe_space->mesh,-1,FILL_COORDS|CALL_LEAF_EL, max_err_at_qp_fct);

  return(max_err);
}

/*--------------------------------------------------------------------------*/
/*  max error at the vertices of the grid :-)))                             */
/*--------------------------------------------------------------------------*/
#if DIM == 1
static REAL vl[DIM+1][DIM+1] = {{1,0},{0,1}};
#endif
#if DIM == 2
static REAL vl[DIM+1][DIM+1] = {{1,0,0},{0,1,0},{0,0,1}};
#endif
#if DIM == 3
static REAL vl[DIM+1][DIM+1] = {{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}};
#endif

static void max_err_at_v_fct(const EL_INFO *el_info)
{
  int         i;
  REAL        err;
  const REAL  *uh_el;
  PARAMETRIC  *parametric = el_info->mesh->parametric;

  elinfo = el_info;
  if (parametric && parametric->init_element(el_info, parametric)) {
    el_parametric = parametric;
  }
  else {
    el_parametric = nil;
  }

  uh_el = get_real_vec_el(el_info->el, err_uh, nil);
  for (i = 0; i < N_VERTICES; i++)
  {
    err = err_u(vl[i]) - eval_uh(vl[i], uh_el, bas_fcts);
    err = ABS(err);
    max_err = MAX(max_err, err);
  }

  return;
}

/*--------------------------------------------------------------------------*/
/*  max error at the quadrutare points of quadrature formula of degree      */
/*  of the continous function u and discrete function uh                    */
/*--------------------------------------------------------------------------*/

REAL max_err_at_vert(REAL (*u)(const REAL_D), const DOF_REAL_VEC *uh)
{
  FUNCNAME("max_err_at_vert");
  const FE_SPACE *fe_space;

  if (!(p_u = u))
  {
    ERROR("no function u specified; doing nothing\n");
    return(-1.0);
  }
  if (!(err_uh = uh)  ||  !(fe_space = uh->fe_space))
  {
    ERROR("no discrete function or no fe_space for it; doing nothing\n");
    return(-1.0);
  }
  if (!uh->vec)
  {
    ERROR("no coefficient vector at discrete solution ; doing nothing\n");
    return(-1.0);
  }
  if (!(bas_fcts = fe_space->bas_fcts))
  {
    ERROR("no basis functions at discrete solution ; doing nothing\n");
    return(-1.0);
  }

  get_real_vec_el = fe_space->bas_fcts->get_real_vec;

  max_err = 0.0;
  mesh_traverse(fe_space->mesh,-1,FILL_COORDS|CALL_LEAF_EL, max_err_at_v_fct);

  return(max_err);
}

/*--------------------------------------------------------------------------*/
/*  function for setting relative error on the elements: used by both       */
/*  h1_err and l2_err                                                       */
/*--------------------------------------------------------------------------*/
static REAL  rel_norm2;

static void rel_fct(const EL_INFO *el_info)
{
  REAL *exact = rw_error(el_info->el);
  *exact /= rel_norm2;
  return;
}

/*--------------------------------------------------------------------------*/
/*  L2 error on the mesh                                                    */
/*--------------------------------------------------------------------------*/
static REAL l2_err_2, l2_norm2;

static void l2_err_fct(const EL_INFO *el_info)
{
  int          i;
  REAL         err, det, l2_err_el, norm_el, exact;
  const REAL   *uh_el, *u_vec, *uh_vec;
  PARAMETRIC   *parametric = el_info->mesh->parametric;

  elinfo = el_info;
  if (parametric && parametric->init_element(el_info, parametric)) {
    el_parametric = parametric;
  }
  else {
    el_parametric = nil;
  }

  u_vec = f_at_qp(quad_fast->quad, err_u, nil);

  uh_el = (*get_real_vec_el)(el_info->el, err_uh, nil);
  uh_vec = uh_at_qp(quad_fast, uh_el, nil);

  if (el_parametric) {
    el_parametric->det(el_info, nil, 1, nil, &det);
  }
  else {
    det = el_det(el_info);
  }

  for (l2_err_el = i = 0; i < quad_fast->n_points; i++)
  {
    err = u_vec[i] - uh_vec[i];
    l2_err_el += quad_fast->w[i]*SQR(err);
  }

  exact = det*l2_err_el;
  l2_err_2 += exact;
  max_err = MAX(max_err, exact);

  if (rw_error) *(*rw_error)(el_info->el) = exact;

  if (relative)
  {
    for (norm_el = i = 0; i < quad_fast->n_points; i++)
      norm_el += quad_fast->w[i]*SQR(u_vec[i]);
    l2_norm2 += det*norm_el;
  }

  return;
}

REAL L2_err(REAL (*u)(const REAL_D), const DOF_REAL_VEC *uh, const QUAD *quad,
	    int rel_err, REAL *(*rw_err_el)(EL *), REAL *max_l2_err2)
{
  FUNCNAME("L2_err");
  const FE_SPACE *fe_space;
  
  if (!(p_u = u))
  {
    ERROR("no function u specified; doing nothing\n");
    return(0.0);
  }
  if (!(err_uh = uh)  ||  !(fe_space = uh->fe_space))
  {
    ERROR("no discrete function or no fe_space for it; doing nothing\n");
    return(0.0);
  }
  if (!uh->vec)
  {
    ERROR("no coefficient vector at discrete solution ; doing nothing\n");
    return(0.0);
  }
  if (!(bas_fcts = fe_space->bas_fcts))
  {
    ERROR("no basis functions at discrete solution ; doing nothing\n");
    return(0.0);
  }
  if (!quad)
    quad = get_quadrature(DIM, 2*fe_space->bas_fcts->degree -2);
  quad_fast = get_quad_fast(bas_fcts, quad, INIT_PHI);

  get_real_vec_el = fe_space->bas_fcts->get_real_vec;

  relative = rel_err;

  rw_error = rw_err_el;

  max_err = l2_err_2 = l2_norm2 = 0.0;
  mesh_traverse(fe_space->mesh, -1, FILL_COORDS|CALL_LEAF_EL, l2_err_fct);

  if (relative)
  {
    rel_norm2 = l2_norm2+1.e-15;
    if (rw_error)  mesh_traverse(fe_space->mesh, -1, CALL_LEAF_EL, rel_fct);
    l2_err_2 /= rel_norm2;
  }

  if (max_l2_err2)  *max_l2_err2 = max_err;

  return(sqrt(l2_err_2));
}

/*--------------------------------------------------------------------------*/
/*  H1 error on the mesh                                                    */
/*--------------------------------------------------------------------------*/

static REAL h1_err_2, h1_norm2;

static void h1_err_fct(const EL_INFO *el_info)
{
  int          i, j;
  REAL         err, err_2, h1_err_el, norm_el, norm2, det, exact;
  REAL_D       Lambda[1][DIM+1];
  const REAL   *uh_el;
  const REAL_D *grdu_vec, *grduh_vec;
  PARAMETRIC  *parametric = el_info->mesh->parametric;

  elinfo = el_info;
  if (parametric && parametric->init_element(el_info, parametric)) {
    el_parametric = parametric;
  }
  else {
    el_parametric = nil;
  }

  grdu_vec = grd_f_at_qp(quad_fast->quad, err_grd_u, nil);

  uh_el = (*get_real_vec_el)(el_info->el, err_uh, nil);

  if (el_parametric) {
    el_parametric->grd_lambda(el_info, nil, 1, nil, Lambda, &det);
  }
  else {
    det = el_grd_lambda(el_info, *Lambda);
  }

  grduh_vec = grd_uh_at_qp(quad_fast, (const REAL_D *) *Lambda, uh_el, nil);

  for (h1_err_el = i = 0; i < quad_fast->n_points; i++)
  {
    for (err_2 = j = 0; j < DIM_OF_WORLD; j++) 
    {
      err = grdu_vec[i][j] - grduh_vec[i][j];
      err_2 += SQR(err);
    }
    h1_err_el += quad_fast->w[i]*err_2;
  }

  exact = det*h1_err_el;
  h1_err_2 += exact;
  max_err = MAX(max_err, exact);

  if (rw_error) *(*rw_error)(el_info->el) = exact;

  if (relative)
  {
    for (norm_el = i = 0; i < quad_fast->n_points; i++)
    {
      for (norm2 = j = 0; j < DIM_OF_WORLD; j++) 
	norm2 += SQR(grdu_vec[i][j]);
      norm_el += quad_fast->w[i]*norm2;
    }
    h1_norm2 += det*norm_el;
  }

  return;
}

REAL H1_err(const REAL *(*grd_u)(const REAL_D), const DOF_REAL_VEC *uh, 
	    const QUAD *quad, int rel_err, REAL *(*rw_err_el)(EL *),
	    REAL *max_h1_err2)
{
  FUNCNAME("H1_err");
  const FE_SPACE *fe_space;
  
  if (!(p_grd_u = grd_u))
  {
    ERROR("no gradient function grd_u specified; doing nothing\n");
    return(0.0);
  }
  if (!(err_uh = uh)  ||  !(fe_space = uh->fe_space))
  {
    ERROR("no discrete function or no fe_space for it; doing nothing\n");
    return(0.0);
  }
  if (!uh->vec)
  {
    ERROR("no coefficient vector at discrete solution ; doing nothing\n");
    return(0.0);
  }
  if (!(bas_fcts = fe_space->bas_fcts))
  {
    ERROR("no basis functions at discrete solution ; doing nothing\n");
    return(0.0);
  }
  if (!quad)
    quad = get_quadrature(DIM, 2*fe_space->bas_fcts->degree-2);
  quad_fast = get_quad_fast(bas_fcts, quad, INIT_GRD_PHI);

  get_real_vec_el = fe_space->bas_fcts->get_real_vec;

  relative = rel_err;

  rw_error = rw_err_el;

/*
  if (!rw_error) 
    MSG("rw_err_el nil pointer; can not write errors to elements\n");
*/

  max_err = h1_err_2 = h1_norm2 = 0.0;
  mesh_traverse(fe_space->mesh, -1, FILL_COORDS|CALL_LEAF_EL, h1_err_fct);

  if (relative)
  {
    rel_norm2 = h1_norm2+1.e-15;
    if (rw_error)  mesh_traverse(fe_space->mesh, -1, CALL_LEAF_EL, rel_fct);
    h1_err_2 /= rel_norm2;
    max_err /= rel_norm2;
  }

  if (max_h1_err2)  *max_h1_err2 = max_err;

  return(sqrt(h1_err_2));
}

/*--------------------------------------------------------------------------*/
/*  and now, the error functions for _d :-)))                               */
/*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/*  max error at the quadrature points :-)))                                */
/*--------------------------------------------------------------------------*/

static REAL  max_err;

static void max_err_d_at_qp_fct(const EL_INFO *el_info)
{
  int           i, k;
  REAL          err;
  const REAL_D  *u_vec, *uh_vec, *uh_el;
  PARAMETRIC  *parametric = el_info->mesh->parametric;

  elinfo = el_info;
  if (parametric && parametric->init_element(el_info, parametric)) {
    el_parametric = parametric;
  }
  else {
    el_parametric = nil;
  }

  u_vec = f_d_at_qp(quad_fast->quad, err_u_d, nil);

  uh_el = (*get_real_d_vec_el)(el_info->el, err_uh_d, nil);
  uh_vec = uh_d_at_qp(quad_fast, uh_el, nil);

  for (i = 0; i < quad_fast->n_points; i++)
  {
    for (err = k = 0; k < DIM_OF_WORLD; k++)
      err += SQR(u_vec[i][k] - uh_vec[i][k]);

    max_err = MAX(max_err, err);
  }

  return;
}

/*--------------------------------------------------------------------------*/
/*  max error at the quadrutare points of quadrature formula of degree      */
/*  of the continous function u and discrete function uh                    */
/*--------------------------------------------------------------------------*/

REAL max_err_d_at_qp(const REAL *(*u)(const REAL_D, REAL_D),
		     const DOF_REAL_D_VEC *uh, const QUAD *quad)
{
  FUNCNAME("max_err_d_at_qp");
  const FE_SPACE *fe;
  
  if (!(p_u_d = u))
  {
    ERROR("no function u specified; doing nothing\n");
    return(0.0);
  }
  if (!(err_uh_d = uh)  ||  !(fe = uh->fe_space))
  {
    ERROR("no discrete function or no fe_space for it; doing nothing\n");
    return(0.0);
  }
  if (!uh->vec)
  {
    ERROR("no coefficient vector at discrete solution ; doing nothing\n");
    return(0.0);
  }
  if (!(bas_fcts = fe->bas_fcts))
  {
    ERROR("no basis functions at discrete solution ; doing nothing\n");
    return(0.0);
  }
  if (!quad)
    quad = get_quadrature(DIM, 2*bas_fcts->degree-2);
  quad_fast = get_quad_fast(bas_fcts, quad, INIT_PHI);

  get_real_d_vec_el = bas_fcts->get_real_d_vec;

  max_err = 0.0;
  mesh_traverse(fe->mesh, -1, FILL_COORDS|CALL_LEAF_EL,
		max_err_d_at_qp_fct);

  return(max_err);
}

/*--------------------------------------------------------------------------*/
/*  L2_d error on the mesh                                                  */
/*--------------------------------------------------------------------------*/

static void l2_err_fct_d(const EL_INFO *el_info)
{
  int            i, k;
  REAL           err = 0, det, l2_err_el, norm_el, exact;
  const REAL_D   *uh_el, *u_vec, *uh_vec;
  PARAMETRIC  *parametric = el_info->mesh->parametric;

  elinfo = el_info;
  if (parametric && parametric->init_element(el_info, parametric)) {
    el_parametric = parametric;
  }
  else {
    el_parametric = nil;
  }

  u_vec = f_d_at_qp(quad_fast->quad, err_u_d, nil);

  uh_el = (*get_real_d_vec_el)(el_info->el, err_uh_d, nil);
  uh_vec = uh_d_at_qp(quad_fast, uh_el, nil);

  if (el_parametric) {
    el_parametric->det(el_info, nil, 1, nil, &det);
  }
  else {
    det = el_det(el_info);
  }

  for (l2_err_el = i = 0; i < quad_fast->n_points; i++)
  {
    for (err = k = 0; k < DIM_OF_WORLD; k++)
      err += SQR(u_vec[i][k] - uh_vec[i][k]);

    l2_err_el += quad_fast->w[i]*err;
  }

  exact = det*l2_err_el;
  l2_err_2 += exact;
  max_err = MAX(max_err, exact);

  if (rw_error) *(*rw_error)(el_info->el) = exact;

  if (relative)
  {
    for (norm_el = i = 0; i < quad_fast->n_points; i++)
    {
      for (err = k = 0; k < DIM_OF_WORLD; k++)
	err += SQR(u_vec[i][k]);
      norm_el += quad_fast->w[i]*err;
    }
    l2_norm2 += det*norm_el;
  }

  return;
}

REAL L2_err_d(const REAL *(*u)(const REAL_D, REAL_D), 
	       const DOF_REAL_D_VEC *uh,
	       const QUAD *quad, int rel_err, REAL *(*rw_err_el)(EL *),
	       REAL *max_l2_err2)
{
  FUNCNAME("L2_err_d");
  const FE_SPACE *fe;

  if (!(p_u_d = u))
  {
    ERROR("no function u specified; doing nothing\n");
    return(0.0);
  }
  if (!(err_uh_d = uh)  ||  !(fe = uh->fe_space))
  {
    ERROR("no discrete function or no fe_space for it; doing nothing\n");
    return(0.0);
  }
  if (!uh->vec)
  {
    ERROR("no coefficient vector at discrete solution ; doing nothing\n");
    return(0.0);
  }
  if (!(bas_fcts = fe->bas_fcts))
  {
    ERROR("no basis functions at discrete solution ; doing nothing\n");
    return(0.0);
  }
  if (!quad)
    quad = get_quadrature(DIM, 2*bas_fcts->degree -2);
  quad_fast = get_quad_fast(bas_fcts, quad, INIT_PHI);

  get_real_d_vec_el = bas_fcts->get_real_d_vec;

  relative = rel_err;

  rw_error = rw_err_el;

  max_err = l2_err_2 = l2_norm2 = 0.0;
  mesh_traverse(fe->mesh, -1, FILL_COORDS|CALL_LEAF_EL, l2_err_fct_d);

  if (relative)
  {
    rel_norm2 = l2_norm2+1.e-15;
    if (rw_error)  mesh_traverse(fe->mesh, -1, CALL_LEAF_EL, rel_fct);
    l2_err_2 /= rel_norm2;
  }

  if (max_l2_err2) *max_l2_err2 = max_err;

  return(sqrt(l2_err_2));
}

/*--------------------------------------------------------------------------*/
/*  H1_d error on the mesh                                                  */
/*--------------------------------------------------------------------------*/

static void h1_err_fct_d(const EL_INFO *el_info)
{
  int            i, j, k;
  REAL           err, err_2, h1_err_el, norm_el, norm2, det, exact;
  REAL_D         Lambda[1][DIM+1];
  const REAL_D   *uh_el;
  const REAL_DD  *grdu_vec, *grduh_vec;
  PARAMETRIC  *parametric = el_info->mesh->parametric;

  elinfo = el_info;
  if (parametric && parametric->init_element(el_info, parametric)) {
    el_parametric = parametric;
  }
  else {
    el_parametric = nil;
  }

  grdu_vec = grd_f_d_at_qp(quad_fast->quad, err_grd_u_d, nil);

  uh_el = (*get_real_d_vec_el)(el_info->el, err_uh_d, nil);

  if (el_parametric) {
    el_parametric->grd_lambda(el_info, nil, 1, nil, Lambda, &det);
  }
  else {
    det = el_grd_lambda(el_info, *Lambda);
  }

  grduh_vec = grd_uh_d_at_qp(quad_fast, (void *) *Lambda, uh_el, nil);

  for (h1_err_el = i = 0; i < quad_fast->n_points; i++)
  {
    for (err_2 = k = 0; k < DIM_OF_WORLD; k++)
      for (j = 0; j < DIM_OF_WORLD; j++) 
      {
	err = grdu_vec[i][k][j] - grduh_vec[i][k][j];
	err_2 += SQR(err);
      }

    h1_err_el += quad_fast->w[i]*err_2;
  }

  exact = det*h1_err_el;
  h1_err_2 += exact;
  max_err = MAX(max_err, exact);

  if (rw_error) *(*rw_error)(el_info->el) = exact;

  if (relative)
  {
    for (norm_el = i = 0; i < quad_fast->n_points; i++)
    {
      for (norm2 = j = 0; j < DIM_OF_WORLD; j++) 
	for (k = 0; k < DIM_OF_WORLD; k++)
	  norm2 += SQR(grdu_vec[i][k][j]);

      norm_el += quad_fast->w[i]*norm2;
    }
    h1_norm2 += det*norm_el;
  }

  return;
}

REAL H1_err_d(const REAL_D *(*grd_u)(const REAL_D, REAL_DD),
	       const DOF_REAL_D_VEC *uh, const QUAD *quad, int rel_err, 
	       REAL *(*rw_err_el)(EL *), REAL *max_h1_err2)
{
  FUNCNAME("H1_err_d");
  const FE_SPACE *fe;
  
  if (!(p_grd_u_d = grd_u))
  {
    ERROR("no gradient function grd_u specified; doing nothing\n");
    return(0.0);
  }
  if (!(err_uh_d = uh)  ||  !(fe = uh->fe_space))
  {
    ERROR("no discrete function or no admin for it; doing nothing\n");
    return(0.0);
  }
  if (!uh->vec)
  {
    ERROR("no coefficient vector at discrete solution ; doing nothing\n");
    return(0.0);
  }
  if (!(bas_fcts = fe->bas_fcts))
  {
    ERROR("no basis functions at discrete solution ; doing nothing\n");
    return(0.0);
  }
  if (!quad)
    quad = get_quadrature(DIM, 2*bas_fcts->degree-2);
  quad_fast = get_quad_fast(bas_fcts, quad, INIT_GRD_PHI);

  get_real_d_vec_el = bas_fcts->get_real_d_vec;

  relative = rel_err;
  rw_error = rw_err_el;
  
  max_err = h1_err_2 = h1_norm2 = 0.0;
  mesh_traverse(fe->mesh, -1, FILL_COORDS|CALL_LEAF_EL, h1_err_fct_d);

  if (relative)
  {
    rel_norm2 = h1_norm2+1.e-15;
    if (rw_error)  mesh_traverse(fe->mesh, -1, CALL_LEAF_EL, rel_fct);
    h1_err_2 /= rel_norm2;
    max_err /= rel_norm2;
  }

  if (max_h1_err2) *max_h1_err2 = max_err;

  return(sqrt(h1_err_2));
}
