/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     face_quad_fast.c                                               */
/*                                                                          */
/* description:  Fast (i.e. caching) quadrature over faces                  */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*             Claus-Justus Heine                                           */
/*             Abteilung fuer Angewandte Mathematik                         */
/*             Albert-Ludwigs-Universitaet Freiburg                         */
/*             Hermann-Herder-Str. 10                                       */
/*             D-79104 Freiburg im Breisgau, Germany                        */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*         C.-J. Heine (1998-2004)                                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#include <alberta.h>
#include "face_quad_fast.h"

const FACE_QUAD_FAST *get_face_quad_fast(const BAS_FCTS *bas_fcts, 
					 const QUAD *quad, U_CHAR init_flag)
{
  FUNCNAME("get_face_quad_fast");
  static FACE_QUAD_FAST  *first = nil;
  FACE_QUAD_FAST   *quad_fast;
  int              d, i, j, k, l, m, n_points, n_bas_fcts;
  size_t           size, size_p;
  REAL             **g_phi;
  REAL             (**g_grd_phi)[DIM+1];
  REAL             (**g_D2_phi)[DIM+1][DIM+1];
  REAL             *lambda;
  void             *mem = nil, *mem_p = nil, *mem_last = nil;
  /*  ALBERTA_LOCAL_RWLOCK_DEF(fqfast_lock); */

  if (quad->dim != DIM-1) {
    ERROR("FACE_QUAD_FAST not for %d != DIM-1\n");
    return(nil);
  }

  /* ALBERTA_RDLOCK(&fqfast_lock); */
  for (quad_fast = first; quad_fast; quad_fast = quad_fast->next) {
    if (quad_fast->quad == quad  &&  quad_fast->bas_fcts == bas_fcts) {
      break;
    }
  }

  if (quad_fast  &&  ((quad_fast->init_flag & init_flag) == init_flag)) {
    /* ALBERTA_RWUNLOCK(&fqfast_lock); */
    return((const FACE_QUAD_FAST *) quad_fast);
  }

  /* ALBERTA_RWUNLOCK(&fqfast_lock); */

  /* ok, remainder either will modify the qfast structure, or even add
   * a new one. Get the write lock. Of course, as we loose the lock in
   * between, it might (and will) happen that there are duplicated
   * quad_fast structures. But these cases will be rare.
   */
  /* ALBERTA_WRLOCK(&fqfast_lock); */

  if (!quad_fast) {
    quad_fast = MEM_ALLOC(1, FACE_QUAD_FAST);
    quad_fast->quad = quad;
    quad_fast->n_points = quad->n_points;
    quad_fast->n_bas_fcts = bas_fcts->n_bas_fcts;
    quad_fast->w = quad->w;
    quad_fast->init_flag = 0;
    for (k = 0; k < N_FACES; k++) {
      quad_fast->lambda[k]    = nil;
      quad_fast->phi[k]       = nil;
      quad_fast->grd_phi[k]   = NULL;
      quad_fast->D2_phi[k]    = NULL;
      quad_fast->sorted[0][0] = NULL;
    }
    quad_fast->next = first;
    quad_fast->bas_fcts = (BAS_FCTS *) bas_fcts;
    first = quad_fast;
  }

  n_points = quad->n_points;
  n_bas_fcts = bas_fcts->n_bas_fcts;

  size = size_p = 0;

  /*  for barycentric coordinates  */
  if (!quad_fast->lambda[0]) {
    size += N_FACES*quad_fast->n_points*sizeof(REAL [DIM+1]);
  }

  if (!quad_fast->sorted[0][0]) {
    size += N_FACES*N_FACES*quad_fast->n_points*sizeof(int);
    /* there may be alignment problems. At least on Super-Sparc CPUs
     * doubles must be aligned at 8-byte boundaries.
     */
    if (size % sizeof(REAL) != 0) {
      size += sizeof(REAL) - size % sizeof(REAL);
    }
  }

  if (!quad_fast->phi[0]  &&  (init_flag & INIT_PHI)) {
    size +=  N_FACES*n_points*n_bas_fcts*sizeof(REAL);
    size_p += N_FACES*n_points*sizeof(REAL *);
  }
  if (!quad_fast->grd_phi[0]  &&  (init_flag & INIT_GRD_PHI)) {
    size += N_FACES*n_points*n_bas_fcts*sizeof(REAL [DIM+1]); 
    size_p += N_FACES*n_points*sizeof(REAL (*)[DIM+1]);
  }
  if (!quad_fast->D2_phi[0]  &&  (init_flag & INIT_D2_PHI)) {
    size += N_FACES*n_points*n_bas_fcts*sizeof(REAL [DIM+1][DIM+1]);
    size_p += N_FACES*n_points*sizeof(REAL (*)[DIM+1][DIM+1]);
  }

  if (size) {
    mem = (void *)MEM_ALLOC(size, char);
    mem_last = (void *) (((char *) mem)+size);
    mem_p = (void *)MEM_ALLOC(size_p, char);

    /* get space for barycentric coordinates */
    if (!quad_fast->lambda[0]) {
      for (k = 0; k < N_FACES; k++) {
	quad_fast->lambda[k] = mem;
	mem = (void *) (quad_fast->lambda[k]+n_points);
      }
    }

    /* initialize barycentric coordinates relative to the reference
     * element. Use the orientation induced by the orientation of the
     * reference element.
     */
    for (k = 0; k < N_FACES; k++) {
      for (i = 0; i < n_points; i++) {
	lambda = (REAL *)quad_fast->lambda[k][i];
	lambda[k] = 0.0;
	for (d = 0; d < DIM; d++) {
	  lambda[(k + 1 + d) % (DIM+1)] = quad->lambda[i][d];
	}
      }
    }

    if (!quad_fast->sorted[0][0]) {
      for (i = 0; i < N_FACES; i++) {
	for (k = 0; k < N_FACES; k++) {
	  quad_fast->sorted[i][k] = mem;
	  mem = (void *) (quad_fast->sorted[i][k] + n_points);
	}
      }

      if ((unsigned long)mem % sizeof(REAL) != 0) {
	mem  = (void *)((char *)mem
			+
			sizeof(REAL) - (unsigned long)mem % sizeof(REAL));
      }

      /* now find out about the face-pairing for neighbouring
       * elements. We assume that the elements have the same
       * orientation. In that case a point (0, l1, l2) on face 0 of an
       * element has the barycentric coordinates (l1, 0, l2) if this
       * face is face 1 of the neighbour of the element.
       *
       * This is terribly ugly, and could be done much faster as the
       * ordering of the vertices is known in advance.
       *
       * Mmmh. Can this be done faster?
       * a) for this to work at all the quadrature points must be symmetric
       *    w.r.t. to pairwise-interchange of barycentric coordinates
       * b) if the quadrature-nodes are in arbitrary order, then this
       *    can hardly be done faster, IMHO.
       */
      for (i = 0; i < N_FACES; i++) { 
	for (j = 0; j < N_FACES; j++) {
	  for (k = 0; k < n_points; k++) {
	    REAL l_n[DIM+1];

	    /* construct the barycentric coordinates relative to a
	     * neighbour with the same orientation. This means that
	     * the orientation of the face induced by the neighbour is
	     * different from the orientation induced by this element, as
	     * inside/outside directions are interchanged.
	     */
	    l_n[j] = 0.0; /* j-th face of neighbour */
#if DIM == 2
	    /* then interchange remaining coordinates */
	    l_n[(j+1) % (DIM+1)] = quad_fast->lambda[i][k][(i+2) % (DIM+1)];
	    l_n[(j+2) % (DIM+1)] = quad_fast->lambda[i][k][(i+1) % (DIM+1)];
#elif DIM == 3
	    /* not too difficult, but in 3 dimensions a rotation of
	     * the face doesn't change its orientation.
	     *
	     * Argh. This probably means that face_qfast->sorted has
	     * to be modified, too. There would be 3 possibilities how
	     * the faces could be renumbered relative to each other
	     * without changing their relative orientation
	     * (w.r.t. each other)
	     */
# error not yet implemented
#endif

	    /* find the corresponding quadrature point for this face,
	     * if this face is face "k" of the neighbour.
	     */
	    for (l = 0; l < n_points; l++) {
	      for (m = 0;
		   m < DIM+1 && l_n[m] == quad_fast->lambda[j][l][m];
		   m++);
	      if (m == DIM+1) {
		break; /* found it */
	      }
	    }
	    TEST_EXIT(l < n_points)
	      ("qp not found\n"); /* there MUST be a corresponding point */

	    /* ok, now we know that 
	     *
	     * quad_fast->lambda[j][l] designates the same point as
	     * quad_fast->lambda[i][k] if the face in question is face
	     * "i" on this element, face "k" on the neighbour and if
	     * both elements are oriented the same way.
	     */
	    quad_fast->sorted[i][j][k] = l;
	  }
	}
      }
    }

    /* initialize values of basis functions at quadrature points */
    if (!quad_fast->phi[0]  &&  (init_flag & INIT_PHI)) {
      BAS_FCT    **phi = bas_fcts->phi;
      
      for (k = 0; k < N_FACES; k++) {
	g_phi = quad_fast->phi[k] = (REAL **) mem_p;
	mem_p = (void *) (g_phi+n_points);

	for (i = 0; i < n_points; i++) {
	  lambda = (REAL *)quad_fast->lambda[k][i];
	  g_phi[i] = (REAL *) mem;
	  mem = (void *) (g_phi[i]+n_bas_fcts);

	  for (j = 0; j < n_bas_fcts; j++) {
	    g_phi[i][j] = phi[j](lambda);
	  }
	}
      }
      quad_fast->init_flag |= INIT_PHI;
    }

    if (!quad_fast->grd_phi[0]  && (init_flag & INIT_GRD_PHI)) {
      GRD_BAS_FCT  **grd_phi = bas_fcts->grd_phi;
      const REAL   *grd;
      
      for (k = 0; k < N_FACES; k++) {
	g_grd_phi =
	  quad_fast->grd_phi[k]  =
	  (REAL (**)[DIM+1]) mem_p;
	mem_p = (void *) (g_grd_phi+n_points);

	for (i = 0; i < n_points; i++) { 
	  lambda = (REAL *)quad_fast->lambda[k][i];
	  g_grd_phi[i] = (REAL (*)[DIM+1]) mem;
	  mem = (void *) (g_grd_phi[i]+n_bas_fcts);
	  
	  for (j = 0; j < n_bas_fcts; j++) {
	    grd = grd_phi[j](lambda);
	    for (l = 0; l < DIM+1; l++) {
	      g_grd_phi[i][j][l] = grd[l];
	    }	    
	  }
	}
      }
      quad_fast->init_flag |= INIT_GRD_PHI;
    }

    if (!quad_fast->D2_phi[0]  && (init_flag & INIT_D2_PHI)) {
      D2_BAS_FCT   **D2_phi = bas_fcts->D2_phi;
      const REAL   (*D2)[DIM+1];

      for (k = 0; k < N_FACES; k++) {
	g_D2_phi = quad_fast->D2_phi[k] = (REAL (**)[DIM+1][DIM+1]) mem_p;
	mem_p = (void *) (quad_fast->D2_phi+n_points);

	for (i = 0; i < n_points; i++) {
	  lambda = (REAL *)quad_fast->lambda[k][i];
	  g_D2_phi[i] = (REAL (*)[DIM+1][DIM+1]) mem;
	  mem = (void *) (g_D2_phi[i]+n_bas_fcts);

	  for (j = 0; j < n_bas_fcts; j++) {
	    D2 = D2_phi[j](lambda);
	    for (m = 0; m < DIM+1; m++) {
	      for (l = 0; l < DIM+1; l++) {
		g_D2_phi[i][j][m][l] = D2[m][l];
	      }
	    }
	  }
	}
      }
      quad_fast->init_flag |= INIT_D2_PHI;
    }
  }

  /* ALBERTA_RWUNLOCK(&fqfast_lock); */

  if (mem > mem_last)
    ERROR_EXIT("allocated size too small\n");
  if (mem < mem_last)
    ERROR_EXIT("allocated size too big\n");

  if ((quad_fast->init_flag & init_flag) != init_flag)
  {
    ERROR("could not initialize quad_fast, returning pointer to nil\n");
    return(nil);
  }

  return((const FACE_QUAD_FAST *) quad_fast);
}

#if 0
/* below is some demo-code how to use a FACE_QUAD_FAST for implementing
 * an error estimator (Neumann + singular residual).
 *
 * This illustrates also the use of the "sorted" component.
 *
 * This code WILL NOT COMPILE 'cause it is taken from a modified
 * version of ALBERTAused for parametric surfaces.
 *
 * BTW, you can fold it away using folding-mode in (X)emacs with 
 * "<<<" ">>>" as folding marks.
 */

/* <<< singluar_residual, "DOF_REAL_VEC *" version */

/* <<< sing_res_data_t */

typedef struct sing_res_data {
  const REAL           *uh_neigh;
  const REAL           *uh_el;
  const BAS_FCTS       *bas_fcts;
  const QUAD           *quad;
  const FACE_QUAD_FAST *qfast;
} SING_RES_DATA;

/* >>> */

static REAL neumann_res2(EL_QUAD_INFO *qinfo, int qp, int face, void *data);
static REAL face_res2(EL_QUAD_INFO *qinfo, int qp, int face, void *data);

/* <<< singular_residual, function definition */

REAL singular_residual(const EL_INFO *el_info,
		       REAL exponent,
		       REAL neumann_weight,
		       const DOF_REAL_VEC   *uh,
		       EL_QUAD_INFO         *qinfo,
		       const QUAD           *quad,
		       const FACE_QUAD_FAST *qfast,
		       const REAL           *uh_el,
		       const REAL           *uh_neigh)
{
  const FE_SPACE  *fe_space = uh->fe_space;
  EL              *neigh;
  REAL             value;
  REAL             sing_res = 0.0;
  SING_RES_DATA    data;
  EL_QUAD_INFO    *qi;
  int              face;

  if (qinfo == NULL) {
    qi = GET_QINFO(sing_res_qinfo, el_info->mesh);
  } else {
    qi = qinfo;
  }
  if (quad == NULL) {
    data.quad = get_quadrature(DIM-1, 2*fe_space->bas_fcts->degree - 2);
  } else {
    data.quad = quad;
  }
  if (qfast == NULL) {
    data.qfast = get_face_quad_fast(fe_space->bas_fcts, data.quad, 
				    INIT_PHI|INIT_GRD_PHI);
  } else {
    data.qfast = qfast;
  }
  if (uh_el) {
    data.uh_el = uh_el;
  } else {
    data.uh_el = (const REAL *)GET_TEMP_BUFFER(sing_res_uh_el,
					       fe_space->bas_fcts->n_bas_fcts
					       *
					       sizeof(REAL));
    fe_space->bas_fcts->get_real_vec(el_info->el, uh, (REAL *)data.uh_el);
  }
  if (uh_neigh) {
    data.uh_neigh = uh_neigh;
  } else {
    data.uh_neigh = GET_TEMP_BUFFER(sing_res_uh_neigh,
				    fe_space->bas_fcts->n_bas_fcts
				    *
				    sizeof(REAL));
  }
  data.bas_fcts = fe_space->bas_fcts;

  for (face = 0; face < N_NEIGH; face++) {
    
    if (IS_NEUMANN(el_info->boundary[face])) {

      fill_el_quad_info(qi, el_info, face, data.quad,
			EL_FILL_FACE_NORMAL |
			EL_FILL_GRD_LAM     |
			EL_FILL_H2);

      value = integrate_face_no_det(data.quad, qi, neumann_res2, face, &data);
      value *= pow(qi->h2, exponent/2.0);
      value *= SQR(neumann_weight);

    } else if ((neigh = NEIGH(el_info->el,el_info)[face])) {

      fill_el_quad_info(qi, el_info, face, data.quad,
			EL_FILL_FACE_NORMAL   |
			EL_FILL_FACE_NORMAL_N |
			EL_FILL_FACE_DET      |
			EL_FILL_GRD_LAM       |
			EL_FILL_GRD_LAM_N     |
			EL_FILL_H2            |
			EL_FILL_H2_N);


      data.bas_fcts->get_real_vec(neigh, uh, (REAL *)data.uh_neigh);
      value = integrate_face(data.quad, qi, face_res2, face, (void *)&data);
      value *= pow(0.5*(qinfo->h2 + qinfo->h2_n), exponent/2.0);

    } else {
/****************************************************************************/
/*  either neigh is pointer to NULL: => DIRICHLET boundary: do not compute  */
/*    jump of the normal derivatives, no contribution to the estimator      */
/****************************************************************************/
      value = 0.0;
    }
    sing_res += value;
  }

  return sing_res;
}

/* >>> */

/* <<< face_res2, function definition */

static REAL face_res2(EL_QUAD_INFO *qinfo, int qp, int face, void *_data)
{
  FUNCNAME("face_res2");
  SING_RES_DATA *data = (SING_RES_DATA *)_data;
  REAL_D         grd_uh_el, grd_uh_n, mean_normal, diff_grd;
  REAL           val;
  int            face_n;
  int            qp_n;

  eval_grd_uh_fast(*qinfo->grd_lam[qp], data->uh_el,
		   (const REAL (*)[DIM+1])data->qfast->grd_phi[face][qp],
		   data->bas_fcts->n_bas_fcts, grd_uh_el);

  face_n = qinfo->opp_vertex;
  qp_n   = data->qfast->sorted[face][face_n][qp];
  eval_grd_uh_fast(*qinfo->grd_lam_n[qp], data->uh_neigh,
		   (const REAL (*)[DIM+1])data->qfast->grd_phi[face_n][qp_n],
		   data->bas_fcts->n_bas_fcts, grd_uh_n);

  if (qinfo->el_info->mesh->parametric) {
    /* the gradients themselves cannot be compared directly, as they
     * point into different directions, because grd_uh_el and grd_uh_n
     * hold the covariant gradients of the FEM function u_h
     */
    
    /* we know that the face normals have the same length */
    REAL_D_AXPBY(1.0, *qinfo->face_normal[qp], -1.0, *qinfo->face_normal_n[qp],
		 mean_normal);
    REAL_D_NORMALIZE(mean_normal);      
    val = REAL_D_SKP(REAL_D_AXPBY(1.0, grd_uh_el, -1.0, grd_uh_n, diff_grd),
		     mean_normal);      
    val *= val;
  } else {
    val = REAL_D_DIST2(grd_uh_el, grd_uh_n);
  }

  return val;
}

/* >>> */

/* <<< neumann_res2, function definition */

static REAL neumann_res2(EL_QUAD_INFO *qinfo, int qp, int face, void *_data)
{
  FUNCNAME("neumann_res2");
  SING_RES_DATA *data = (SING_RES_DATA *)_data;
  REAL_D         grd_uh;
  REAL           val;

  eval_grd_uh_fast(*qinfo->grd_lam[qp], data->uh_el,
		   (const REAL (*)[DIM+1])data->qfast->grd_phi[face][qp],
		   data->bas_fcts->n_bas_fcts, grd_uh);
  
  val = REAL_D_SKP(grd_uh,  *qinfo->face_normal[qp]);
  
  return SQR(val);
}

/* >>> */

/* >>> */

#endif /* demo-code */

