/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     traverse_r_common.c                                            */
/*                                                                          */
/*                                                                          */
/* description:                                                             */
/*           recursive mesh traversal routines - common 2d/3d part          */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Univesitaet Bremen                                           */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/* common (static) traversal routines for 2d and 3d                         */
/* to be #included in traverse_r.c                                          */
/*--------------------------------------------------------------------------*/

/* traverse hierarchical mesh,  call el_fct() for each/some element */

typedef struct traverse_info TRAVERSE_INFO;

struct traverse_info 
{
  MESH     *mesh;
  FLAGS    flag; 
  U_CHAR   level;
  void     (*el_fct)(const EL_INFO *);
};


static void recursive_traverse(EL_INFO *, TRAVERSE_INFO *);

/*--------------------------------------------------------------------------*/
/*   mesh_traverse:                                            		    */
/*   --------------                                     		    */
/*   run through all macro elements and start mesh traversal      	    */
/*   call recursive_traverse() for each macro element's element		    */
/*--------------------------------------------------------------------------*/

void mesh_traverse(MESH *mesh, int level, FLAGS flag, 
		   void (*el_fct)(const EL_INFO *))
{
  FUNCNAME("mesh_traverse");
  MACRO_EL *mel;
  EL_INFO elinfo;
  TRAVERSE_INFO traverse_info;

  if (mesh==nil) return;

  elinfo.fill_flag = (flag & FILL_ANY);

  traverse_info.mesh   = mesh;
  traverse_info.level  = level;
  traverse_info.el_fct = el_fct;
  traverse_info.flag = flag;
  elinfo.mesh = mesh;

  if (flag & (CALL_LEAF_EL_LEVEL | CALL_EL_LEVEL | CALL_MG_LEVEL)) {
    TEST(level >= 0)("invalid level: %d\n",level);
  }

  for (mel = mesh->first_macro_el; mel; mel = mel->next) {
    fill_macro_info(mesh, mel, &elinfo);
    recursive_traverse(&elinfo, &traverse_info);
  }

}

/*--------------------------------------------------------------------------*/
/*   recursive_traverse:                                		    */
/*   -------------------                                		    */
/*   recursively traverse the mesh hierarchy tree       		    */
/*   call the routine traverse_info->el_fct(el_info) with    		    */
/*    - all tree leaves                                 		    */
/*    - all tree leaves at a special level              		    */
/*    - all tree elements in pre-/in-/post-order        		    */
/*   depending on the traverse_info->level variable     		    */
/*--------------------------------------------------------------------------*/

static void recursive_traverse(EL_INFO *elinfo, TRAVERSE_INFO *trinfo)
{
  EL      *el = elinfo->el;
  EL_INFO elinfo_new;
  int     mg_level;

  if (trinfo->flag & CALL_LEAF_EL) 
  {
    if (el->child[0]) 
    {
      fill_elinfo(0, elinfo, &elinfo_new);
      recursive_traverse(&elinfo_new, trinfo);
      fill_elinfo(1, elinfo, &elinfo_new);
      recursive_traverse(&elinfo_new, trinfo);
    }
    else 
    {
      (*trinfo->el_fct)(elinfo);
    }
    return;
  }

  if (trinfo->flag & CALL_LEAF_EL_LEVEL) 
  {
    if (el->child[0]) 
    {
      if ((elinfo->level >= trinfo->level))  return;
      fill_elinfo(0, elinfo, &elinfo_new);
      recursive_traverse(&elinfo_new, trinfo);
      fill_elinfo(1, elinfo, &elinfo_new);
      recursive_traverse(&elinfo_new, trinfo);
    }
    else 
    {
      if (elinfo->level == trinfo->level) (trinfo->el_fct)(elinfo);
    }
    return;
  }

  if (trinfo->flag & CALL_EL_LEVEL) 
  {
    if (elinfo->level == trinfo->level)
    {
      (*trinfo->el_fct)(elinfo);
    }
    else 
    {
      if (elinfo->level > trinfo->level)
      {
	return;
      }
      else
      {
	if (el->child[0]) 
	{
	  fill_elinfo(0, elinfo, &elinfo_new);
	  recursive_traverse(&elinfo_new, trinfo);
	  fill_elinfo(1, elinfo, &elinfo_new);
	  recursive_traverse(&elinfo_new, trinfo);
	}
      }
    }
    return;
  }

  if (trinfo->flag & CALL_MG_LEVEL) 
  {
    mg_level = ((int)elinfo->level + DIM-1) / DIM;

    if (mg_level > (int)trinfo->level) return;

    if (!(el->child[0])) 
    {
      (*trinfo->el_fct)(elinfo);
      return;
    }
    
    if ((mg_level == trinfo->level) && ((elinfo->level % DIM) == 0))
    {
      (*trinfo->el_fct)(elinfo);
      return;
    }

    fill_elinfo(0, elinfo, &elinfo_new);
    recursive_traverse(&elinfo_new, trinfo);

    fill_elinfo(1, elinfo, &elinfo_new);
    recursive_traverse(&elinfo_new, trinfo);

    return;
  }

  if (trinfo->flag & CALL_EVERY_EL_PREORDER)  (*trinfo->el_fct)(elinfo);

  if (el->child[0]) 
  {
    fill_elinfo(0, elinfo, &elinfo_new);
    recursive_traverse(&elinfo_new, trinfo);

    if (trinfo->flag & CALL_EVERY_EL_INORDER) (*trinfo->el_fct)(elinfo);

    fill_elinfo(1, elinfo, &elinfo_new);
    recursive_traverse(&elinfo_new, trinfo);
  }
  else 
  {
    if (trinfo->flag & CALL_EVERY_EL_INORDER) (*trinfo->el_fct)(elinfo);
  }

  if (trinfo->flag & CALL_EVERY_EL_POSTORDER)  (*trinfo->el_fct)(elinfo);

  return;

}

/*--------------------------------------------------------------------------*/
/*   test_traverse_fct:                              		            */
/*   ------------------                                 		    */
/*   display information about an element from EL and EL_INFO structures    */
/*--------------------------------------------------------------------------*/

static void test_traverse_fct(const EL_INFO *elinfo)
{
  FUNCNAME("test_traverse_fct");
  EL *el = elinfo->el;
  int i,j;

  MSG("\n");
  MSG("traversing element %d: at %8X -----------------------\n",
      INDEX(el), el);

  print_msg("level:        %3d\n",elinfo->level);

  print_msg("macro_el:     %8X\n", elinfo->macro_el);
  
  if (el->child[0]) {
    print_msg("childs:        ");
    for (i=0;i<2;i++) 
      if (el->child[i])
	         print_msg(" %3d at %8X",INDEX(el->child[i]), el->child[i]);
      else       print_msg(" ---");
    print_msg("\n");
  }

  if (elinfo->fill_flag & FILL_COORDS) {
    print_msg("coords:      ");
    for (i=0; i<N_VERTICES; i++) 
    {
      print_msg("%1d: (", i);
      for (j=0; j<DIM_OF_WORLD; j++)
	print_msg("%10.6lf%s", elinfo->coord[i][j],
		 j < (DIM_OF_WORLD-1) ? ", " : ")\n");
      if (i < (N_VERTICES-1)) print_msg("             ");
    }
  }

  if (elinfo->fill_flag & FILL_NEIGH) {
#if EL_INDEX
    print_msg("neigh index :");
    for (i=0; i<N_NEIGH; i++) 
      if (NEIGH(el,elinfo)[i]) print_msg(" %3d",INDEX(NEIGH(el,elinfo)[i]));
      else                  print_msg(" ---");
    print_msg("\n");
#endif
    print_msg("opp_vertex:  ");
    for (i=0; i<N_NEIGH; i++) 
      if (NEIGH(el,elinfo)[i]) print_msg(" %3d",OPP_VERTEX(el,elinfo)[i]);
      else                  print_msg(" ---");
    print_msg("\n");

    print_msg("neigh:      ");
    for (i=0; i<N_NEIGH; i++) print_msg(" %8X",NEIGH(el,elinfo)[i]);
    print_msg("\n");
  }
   
  if (elinfo->fill_flag & FILL_OPP_COORDS) {
    print_msg("opp_coords:  ");
    for (i=0; i<N_NEIGH; i++) {
      if (NEIGH(el,elinfo)[i]) {
	print_msg("%1d (ov=%1d): (", i, OPP_VERTEX(el,elinfo)[i]);
	for (j=0; j<DIM_OF_WORLD; j++)
	  print_msg("%10.6lf%s", elinfo->opp_coord[i][j],
		   j < (DIM_OF_WORLD-1) ? ", " : ")\n");
      }
      else {
	print_msg("%1d             : ---\n",i);
      }
      if (i < (N_VERTICES-1)) print_msg("             ");
    }
  }
}

/*--------------------------------------------------------------------------*/
/*   test_traverse:                                  		            */
/*   --------------                                     		    */
/*   display information about all elements from EL and EL_INFO structures  */
/*   by calling traverse with test_traverse_fct().      		    */
/*--------------------------------------------------------------------------*/

void test_traverse(MESH *mesh, int level, FLAGS fill_flag)
{
  FUNCNAME("test_traverse");

  MSG("with level    : %3d\n", level);
  MSG("with fill_flag:");
  if (fill_flag & FILL_ANY) {
    if (fill_flag & FILL_COORDS)   print_msg(" FILL_COORDS");
    if (fill_flag & FILL_BOUND)    print_msg(" FILL_BOUND");
    if (fill_flag & FILL_NEIGH)    print_msg(" FILL_NEIGH");
    if (fill_flag & FILL_OPP_COORDS)  print_msg(" FILL_OPP_COORDS");
  }
  else print_msg(" none");
  print_msg("\n");

  mesh_traverse(mesh, level, fill_flag, test_traverse_fct);
  MSG("done.\n");
 }
