/*--------------------------------------------------------------------------*/
/* ALBERTA:  an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques for scientific Applications                         */
/*                                                                          */
/* file:     lagrange_4.c                                                   */
/*                                                                          */
/* description:  piecewise quartic Lagrange elements in 3d                  */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*                                                                          */
/*  authors:   Alfred Schmidt                                               */
/*             Zentrum fuer Technomathematik                                */
/*             Fachbereich 3 Mathematik/Informatik                          */
/*             Universitaet Bremen                                          */
/*             Bibliothekstr. 2                                             */
/*             D-28359 Bremen, Germany                                      */
/*                                                                          */
/*             Kunibert G. Siebert                                          */
/*             Institut fuer Mathematik                                     */
/*             Universitaet Augsburg                                        */
/*             Universitaetsstr. 14                                         */
/*             D-86159 Augsburg, Germany                                    */
/*                                                                          */
/*  http://www.alberta-fem.de/                                              */
/*                                                                          */
/*--------------------------------------------------------------------------*/
/*  (c) by A. Schmidt and K.G. Siebert (1996-2004)                          */
/*                                                                          */
/*     This program is free software; you can redistribute it and/or modify */
/*     it under the terms of the GNU General Public License as published by */
/*     the Free Software Foundation; either version 2 of the License, or    */
/*     any later version.                                                   */
/*                                                                          */
/*     This program is distributed in the hope that it will be useful,      */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/*     GNU General Public License for more details.                         */
/*--------------------------------------------------------------------------*/

#define N_BAS4  35

static const REAL   bary4[N_BAS4][DIM+1] = {
                  {    1.0,     0.0,     0.0,     0.0},
		  {    0.0,     1.0,     0.0,     0.0},
		  {    0.0,     0.0,     1.0,     0.0},
		  {    0.0,     0.0,     0.0,     1.0},
		  {3.0/4.0, 1.0/4.0,     0.0,     0.0},
		  {    0.5,     0.5,     0.0,     0.0},
		  {1.0/4.0, 3.0/4.0,     0.0,     0.0},
		  {3.0/4.0,     0.0, 1.0/4.0,     0.0},
		  {    0.5,     0.0,     0.5,     0.0},
		  {1.0/4.0,     0.0, 3.0/4.0,     0.0},
		  {3.0/4.0,     0.0,     0.0, 1.0/4.0},
		  {    0.5,     0.0,     0.0,     0.5},
		  {1.0/4.0,     0.0,     0.0, 3.0/4.0},
		  {    0.0, 3.0/4.0, 1.0/4.0,     0.0},
		  {    0.0,     0.5,     0.5,     0.0},
		  {    0.0, 1.0/4.0, 3.0/4.0,     0.0},
		  {    0.0, 3.0/4.0,     0.0, 1.0/4.0},
		  {    0.0,     0.5,     0.0,     0.5},
		  {    0.0, 1.0/4.0,     0.0, 3.0/4.0},
		  {    0.0,     0.0, 3.0/4.0, 1.0/4.0},
		  {    0.0,     0.0,     0.5,     0.5},
		  {    0.0,     0.0, 1.0/4.0, 3.0/4.0},
		  {    0.0, 1.0/2.0, 1.0/4.0, 1.0/4.0},
		  {    0.0, 1.0/4.0, 1.0/2.0, 1.0/4.0},
		  {    0.0, 1.0/4.0, 1.0/4.0, 1.0/2.0},
		  {1.0/2.0,    0.0,  1.0/4.0, 1.0/4.0},
		  {1.0/4.0,    0.0,  1.0/2.0, 1.0/4.0},
		  {1.0/4.0,    0.0,  1.0/4.0, 1.0/2.0},
		  {1.0/2.0, 1.0/4.0,    0.0,  1.0/4.0},
		  {1.0/4.0, 1.0/2.0,    0.0,  1.0/4.0},
		  {1.0/4.0, 1.0/4.0,    0.0,  1.0/2.0},
		  {1.0/2.0, 1.0/4.0, 1.0/4.0,     0.0},
		  {1.0/4.0, 1.0/2.0, 1.0/4.0,     0.0},
		  {1.0/4.0, 1.0/4.0, 1.0/2.0,     0.0},
		  {1.0/4.0, 1.0/4.0, 1.0/4.0, 1.0/4.0}};

#if 0
static const int    fct_at_node4[N_BAS4] = {0,1,2,3,
					      4,4,4,
					      5,5,5,
					      6,6,6,
					      7,7,7,
					      8,8,8,
					      9,9,9,
					      10,10,10,
					      11,11,11,
					      12,12,12,
					      13,13,13,
					      14};

static const int    n_dof4[DIM+1] = {1, 3, 3, 1};
#endif

/*--------------------------------------------------------------------------*/
/*  basisfunction at vertex 0                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4v0(const REAL l[DIM+1])
{
  return((((32.0*l[0] - 48.0)*l[0] + 22.0)*l[0] - 3.0)*l[0]/3.0);
}

static const REAL *grd_phi4v0(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[0] = ((128.0*l[0] - 144.0)*l[0] + 44.0)*l[0]/3.0 - 1.0;
  return((const REAL *) grd);
}

static const REAL (*D2_phi4v0(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0.0}};
  D2[0][0] = (128.0*l[0] - 96.0)*l[0] + 44.0/3.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at vertex 1                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4v1(const REAL l[DIM+1])
{
  return((((32.0*l[1] - 48.0)*l[1] + 22.0)*l[1] - 3.0)*l[1]/3.0);
}

static const REAL *grd_phi4v1(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[1] = ((128.0*l[1] - 144.0)*l[1] + 44.0)*l[1]/3.0 - 1.0;
  return((const REAL *) grd);
}

static const REAL (*D2_phi4v1(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[1][1] = (128.0*l[1] - 96.0)*l[1] + 44.0/3.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at vertex 2                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4v2(const REAL l[DIM+1])
{
  return((((32.0*l[2] - 48.0)*l[2] + 22.0)*l[2] - 3.0)*l[2]/3.0);
}

static const REAL *grd_phi4v2(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[2] = ((128.0*l[2] - 144.0)*l[2] + 44.0)*l[2]/3.0 - 1.0;
  return((const REAL *) grd);
}

static const REAL (*D2_phi4v2(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[2][2] = (128.0*l[2] - 96.0)*l[2] + 44.0/3.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction at vertex 3                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4v3(const REAL l[DIM+1])
{
  return((((32.0*l[3] - 48.0)*l[3] + 22.0)*l[3] - 3.0)*l[3]/3.0);
}

static const REAL *grd_phi4v3(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[3] = ((128.0*l[3] - 144.0)*l[3] + 44.0)*l[3]/3.0 - 1.0;
  return((const REAL *) grd);
}

static const REAL (*D2_phi4v3(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[3][3] = (128.0*l[3] - 96.0)*l[3] + 44.0/3.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 0 at edge 0                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4e00(const REAL l[DIM+1])
{
  return(((128.0*l[0] - 96.0)*l[0] + 16.0)*l[0]*l[1]/3.0);
}

static const REAL *grd_phi4e00(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[0] = ((128*l[0] - 64.0)*l[0] + 16.0/3.0)*l[1];
  grd[1] = ((128*l[0] - 96.0)*l[0] + 16.0)*l[0]/3.0;
  return((const REAL *) grd);
}

static const REAL (*D2_phi4e00(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[0][0] = (256.0*l[0] - 64.0)*l[1];
  D2[0][1] = D2[1][0] = (128.0*l[0] - 64.0)*l[0] + 16.0/3.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 1 at edge 0                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4e01(const REAL l[DIM+1])
{
  return((4.0*l[0] - 1.0)*l[0]*(4.0*l[1] - 1.0)*l[1]*4.0);
}

static const REAL *grd_phi4e01(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[0] = 4.0*(8.0*l[0] - 1.0)*l[1]*(4.0*l[1] - 1.0);
  grd[1] = 4.0*l[0]*(4.0*l[0] - 1.0)*(8.0*l[1] - 1.0);
  return((const REAL *) grd);
}

static const REAL (*D2_phi4e01(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[0][0] = 32.0*l[1]*(4.0*l[1] - 1.0);
  D2[0][1] = D2[1][0] = 4.0*(8.0*l[0] - 1.0)*(8.0*l[1] - 1.0);
  D2[1][1] = 32.0*l[0]*(4.0*l[0] - 1.0);
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 2 at edge 2                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4e02(const REAL l[DIM+1])
{
  return(((128.0*l[1] - 96.0)*l[1] + 16.0)*l[1]*l[0]/3.0);
}

static const REAL *grd_phi4e02(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[0] = ((128*l[1] - 96.0)*l[1] + 16.0)*l[1]/3.0;
  grd[1] = ((128*l[1] - 64.0)*l[1] + 16.0/3.0)*l[0];
  return((const REAL *) grd);
}

static const REAL (*D2_phi4e02(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0, 0, 0},{0, 0, 0}, {0, 0, 0}};
  D2[1][1] = (256.0*l[1] - 64.0)*l[0];
  D2[0][1] = D2[1][0] = (128.0*l[1] - 64.0)*l[1] + 16.0/3.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 0 at edge 1                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4e10(const REAL l[DIM+1])
{
  return(((128.0*l[0] - 96.0)*l[0] + 16.0)*l[0]*l[2]/3.0);
}

static const REAL *grd_phi4e10(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[0] = ((128*l[0] - 64.0)*l[0] + 16.0/3.0)*l[2];
  grd[2] = ((128*l[0] - 96.0)*l[0] + 16.0)*l[0]/3.0;
  return((const REAL *) grd);
}

static const REAL (*D2_phi4e10(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[0][0] = (256.0*l[0] - 64.0)*l[2];
  D2[0][2] = D2[2][0] = (128.0*l[0] - 64.0)*l[0] + 16.0/3.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 1 at edge 1                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4e11(const REAL l[DIM+1])
{
  return((4.0*l[0] - 1.0)*l[0]*(4.0*l[2] - 1.0)*l[2]*4.0);
}

static const REAL *grd_phi4e11(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[0] = 4.0*(8.0*l[0] - 1.0)*l[2]*(4.0*l[2] - 1.0);
  grd[2] = 4.0*l[0]*(4.0*l[0] - 1.0)*(8.0*l[2] - 1.0);

  return((const REAL *) grd);
}

static const REAL (*D2_phi4e11(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[0][0] = 32.0*l[2]*(4.0*l[2] - 1.0);
  D2[0][2] = D2[2][0] = 4.0*(8.0*l[0] - 1.0)*(8.0*l[2] - 1.0);
  D2[2][2] = 32.0*l[0]*(4.0*l[0] - 1.0);
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 2 at edge 2                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4e12(const REAL l[DIM+1])
{
  return(((128.0*l[2] - 96.0)*l[2] + 16.0)*l[2]*l[0]/3.0);
}

static const REAL *grd_phi4e12(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[0] = ((128*l[2] - 96.0)*l[2] + 16.0)*l[2]/3.0;
  grd[2] = ((128*l[2] - 64.0)*l[2] + 16.0/3.0)*l[0];
  return((const REAL *) grd);
}

static const REAL (*D2_phi4e12(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0, 0, 0},{0, 0, 0}, {0, 0, 0}};
  D2[2][2] = (256.0*l[2] - 64.0)*l[0];
  D2[0][2] = D2[2][0] = (128.0*l[2] - 64.0)*l[2] + 16.0/3.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 0 at edge 2                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4e20(const REAL l[DIM+1])
{
  return(((128.0*l[0] - 96.0)*l[0] + 16.0)*l[0]*l[3]/3.0);
}

static const REAL *grd_phi4e20(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[0] = ((128*l[0] - 64.0)*l[0] + 16.0/3.0)*l[3];
  grd[3] = ((128*l[0] - 96.0)*l[0] + 16.0)*l[0]/3.0;
  return((const REAL *) grd);
}

static const REAL (*D2_phi4e20(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[0][0] = (256.0*l[0] - 64.0)*l[3];
  D2[0][3] = D2[3][0] = (128.0*l[0] - 64.0)*l[0] + 16.0/3.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 1 at edge 2                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4e21(const REAL l[DIM+1])
{
  return((4.0*l[0] - 1.0)*l[0]*(4.0*l[3] - 1.0)*l[3]*4.0);
}

static const REAL *grd_phi4e21(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[0] = 4.0*(8.0*l[0] - 1.0)*l[3]*(4.0*l[3] - 1.0);
  grd[3] = 4.0*l[0]*(4.0*l[0] - 1.0)*(8.0*l[3] - 1.0);
  return((const REAL *) grd);
}

static const REAL (*D2_phi4e21(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[0][0] = 32.0*l[3]*(4.0*l[3] - 1.0);
  D2[0][3] = D2[3][0] = 4.0*(8.0*l[0] - 1.0)*(8.0*l[3] - 1.0);
  D2[3][3] = 32.0*l[0]*(4.0*l[0] - 1.0);
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 2 at edge 2                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4e22(const REAL l[DIM+1])
{
  return(((128.0*l[3] - 96.0)*l[3] + 16.0)*l[3]*l[0]/3.0);
}

static const REAL *grd_phi4e22(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[0] = ((128*l[3] - 96.0)*l[3] + 16.0)*l[3]/3.0;
  grd[3] = ((128*l[3] - 64.0)*l[3] + 16.0/3.0)*l[0];
  return((const REAL *) grd);
}

static const REAL (*D2_phi4e22(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0, 0, 0},{0, 0, 0}, {0, 0, 0}};
  D2[3][3] = (256.0*l[3] - 64.0)*l[0];
  D2[0][3] = D2[3][0] = (128.0*l[3] - 64.0)*l[3] + 16.0/3.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 0 at edge 3                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4e30(const REAL l[DIM+1])
{
  return(((128.0*l[1] - 96.0)*l[1] + 16.0)*l[1]*l[2]/3.0);
}

static const REAL *grd_phi4e30(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[1] = ((128*l[1] - 64.0)*l[1] + 16.0/3.0)*l[2];
  grd[2] = ((128*l[1] - 96.0)*l[1] + 16.0)*l[1]/3.0;
  return((const REAL *) grd);
}

static const REAL (*D2_phi4e30(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[1][1] = (256.0*l[1] - 64.0)*l[2];
  D2[1][2] = D2[2][1] = (128.0*l[1] - 64.0)*l[1] + 16.0/3.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 1 at edge 3                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4e31(const REAL l[DIM+1])
{
  return((4.0*l[1] - 1.0)*l[1]*(4.0*l[2] - 1.0)*l[2]*4.0);
}

static const REAL *grd_phi4e31(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[1] = 4.0*(8.0*l[1] - 1.0)*l[2]*(4.0*l[2] - 1.0);
  grd[2] = 4.0*l[1]*(4.0*l[1] - 1.0)*(8.0*l[2] - 1.0);
  return((const REAL *) grd);
}

static const REAL (*D2_phi4e31(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[1][1] = 32.0*l[2]*(4.0*l[2] - 1.0);
  D2[1][2] = D2[2][1] = 4.0*(8.0*l[1] - 1.0)*(8.0*l[2] - 1.0);
  D2[2][2] = 32.0*l[1]*(4.0*l[1] - 1.0);
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 2 at edge 3                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4e32(const REAL l[DIM+1])
{
  return(((128.0*l[2] - 96.0)*l[2] + 16.0)*l[2]*l[1]/3.0);
}

static const REAL *grd_phi4e32(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[1] = ((128*l[2] - 96.0)*l[2] + 16.0)*l[2]/3.0;
  grd[2] = ((128*l[2] - 64.0)*l[2] + 16.0/3.0)*l[1];
  return((const REAL *) grd);
}

static const REAL (*D2_phi4e32(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0, 0, 0},{0, 0, 0}, {0, 0, 0}};
  D2[2][2] = (256.0*l[2] - 64.0)*l[1];
  D2[1][2] = D2[2][1] = (128.0*l[2] - 64.0)*l[2] + 16.0/3.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 0 at edge 4                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4e40(const REAL l[DIM+1])
{
  return(((128.0*l[1] - 96.0)*l[1] + 16.0)*l[1]*l[3]/3.0);
}

static const REAL *grd_phi4e40(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[1] = ((128*l[1] - 64.0)*l[1] + 16.0/3.0)*l[3];
  grd[3] = ((128*l[1] - 96.0)*l[1] + 16.0)*l[1]/3.0;
  return((const REAL *) grd);
}

static const REAL (*D2_phi4e40(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[1][1] = (256.0*l[1] - 64.0)*l[3];
  D2[1][3] = D2[3][1] = (128.0*l[1] - 64.0)*l[1] + 16.0/3.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 1 at edge 4                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4e41(const REAL l[DIM+1])
{
  return((4.0*l[1] - 1.0)*l[1]*(4.0*l[3] - 1.0)*l[3]*4.0);
}

static const REAL *grd_phi4e41(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[1] = 4.0*(8.0*l[1] - 1.0)*l[3]*(4.0*l[3] - 1.0);
  grd[3] = 4.0*l[1]*(4.0*l[1] - 1.0)*(8.0*l[3] - 1.0);
  return((const REAL *) grd);
}

static const REAL (*D2_phi4e41(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[1][1] = 32.0*l[3]*(4.0*l[3] - 1.0);
  D2[1][3] = D2[3][1] = 4.0*(8.0*l[1] - 1.0)*(8.0*l[3] - 1.0);
  D2[3][3] = 32.0*l[1]*(4.0*l[1] - 1.0);
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 2 at edge 4                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4e42(const REAL l[DIM+1])
{
  return(((128.0*l[3] - 96.0)*l[3] + 16.0)*l[3]*l[1]/3.0);
}

static const REAL *grd_phi4e42(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[1] = ((128*l[3] - 96.0)*l[3] + 16.0)*l[3]/3.0;
  grd[3] = ((128*l[3] - 64.0)*l[3] + 16.0/3.0)*l[1];
  return((const REAL *) grd);
}

static const REAL (*D2_phi4e42(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0, 0, 0},{0, 0, 0}, {0, 0, 0}};
  D2[3][3] = (256.0*l[3] - 64.0)*l[1];
  D2[1][3] = D2[3][1] = (128.0*l[3] - 64.0)*l[3] + 16.0/3.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 0 at edge 5                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4e50(const REAL l[DIM+1])
{
  return(((128.0*l[2] - 96.0)*l[2] + 16.0)*l[2]*l[3]/3.0);
}

static const REAL *grd_phi4e50(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[2] = ((128*l[2] - 64.0)*l[2] + 16.0/3.0)*l[3];
  grd[3] = ((128*l[2] - 96.0)*l[2] + 16.0)*l[2]/3.0;
  return((const REAL *) grd);
}

static const REAL (*D2_phi4e50(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[2][2] = (256.0*l[2] - 64.0)*l[3];
  D2[2][3] = D2[3][2] = (128.0*l[2] - 64.0)*l[2] + 16.0/3.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 1 at edge 5                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4e51(const REAL l[DIM+1])
{
  return((4.0*l[2] - 1.0)*l[2]*(4.0*l[3] - 1.0)*l[3]*4.0);
}

static const REAL *grd_phi4e51(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[2] = 4.0*(8.0*l[2] - 1.0)*l[3]*(4.0*l[3] - 1.0);
  grd[3] = 4.0*l[2]*(4.0*l[2] - 1.0)*(8.0*l[3] - 1.0);
  return((const REAL *) grd);
}

static const REAL (*D2_phi4e51(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[2][2] = 32.0*l[3]*(4.0*l[3] - 1.0);
  D2[2][3] = D2[3][2] = 4.0*(8.0*l[2] - 1.0)*(8.0*l[3] - 1.0);
  D2[3][3] = 32.0*l[2]*(4.0*l[2] - 1.0);
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 2 at edge 5                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4e52(const REAL l[DIM+1])
{
  return(((128.0*l[3] - 96.0)*l[3] + 16.0)*l[3]*l[2]/3.0);
}

static const REAL *grd_phi4e52(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[2] = ((128*l[3] - 96.0)*l[3] + 16.0)*l[3]/3.0;
  grd[3] = ((128*l[3] - 64.0)*l[3] + 16.0/3.0)*l[2];
  return((const REAL *) grd);
}

static const REAL (*D2_phi4e52(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0, 0, 0},{0, 0, 0}, {0, 0, 0}};
  D2[3][3] = (256.0*l[3] - 64.0)*l[2];
  D2[2][3] = D2[3][2] = (128.0*l[3] - 64.0)*l[3] + 16.0/3.0;
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 0 at face 3                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4f30(const REAL l[DIM+1])
{
  return((4.0*l[0] - 1.0)*l[0]*l[1]*l[2]*32.0);
}

static const REAL *grd_phi4f30(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[0] = 32.0*(8.0*l[0] - 1.0)*l[1]*l[2];
  grd[1] = 32.0*(4.0*l[0] - 1.0)*l[0]*l[2];
  grd[2] = 32.0*(4.0*l[0] - 1.0)*l[0]*l[1];
  return((const REAL *) grd);
}

static const REAL (*D2_phi4f30(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[0][0] = 256.0*l[1]*l[2];
  D2[0][1] = D2[1][0] = 32.0*(8.0*l[0] - 1.0)*l[2];
  D2[0][2] = D2[2][0] = 32.0*(8.0*l[0] - 1.0)*l[1];
  D2[1][2] = D2[2][1] = 32.0*(4.0*l[0] - 1.0)*l[0];
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 1 at face 3                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4f31(const REAL l[DIM+1])
{
  return(l[0]*(4.0*l[1] - 1.0)*l[1]*l[2]*32.0);
}

static const REAL *grd_phi4f31(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[0] = 32.0*(4.0*l[1] - 1.0)*l[1]*l[2];
  grd[1] = 32.0*(8.0*l[1] - 1.0)*l[0]*l[2];
  grd[2] = 32.0*(4.0*l[1] - 1.0)*l[0]*l[1];
  return((const REAL *) grd);
}

static const REAL (*D2_phi4f31(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[0][1] = D2[1][0] = 32.0*(8.0*l[1] - 1.0)*l[2];
  D2[0][2] = D2[2][0] = 32.0*(4.0*l[1] - 1.0)*l[1];
  D2[1][1] = 256.0*l[0]*l[2];
  D2[1][2] = D2[2][1] = 32.0*(8.0*l[1] - 1.0)*l[0];
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 2 at face 3                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4f32(const REAL l[DIM+1])
{
  return(l[0]*l[1]*(4.0*l[2] - 1.0)*l[2]*32.0);
}

static const REAL *grd_phi4f32(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[0] = 32.0*(4.0*l[2] - 1.0)*l[1]*l[2];
  grd[1] = 32.0*(4.0*l[2] - 1.0)*l[0]*l[2];
  grd[2] = 32.0*(8.0*l[2] - 1.0)*l[0]*l[1];
  return((const REAL *) grd);
}

static const REAL (*D2_phi4f32(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[0][1] = D2[1][0] = 32.0*(4.0*l[2] - 1.0)*l[2];
  D2[0][2] = D2[2][0] = 32.0*(8.0*l[2] - 1.0)*l[1];
  D2[1][2] = D2[2][1] = 32.0*(8.0*l[2] - 1.0)*l[0];
  D2[2][2] = 256.0*l[0]*l[1];
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 0 at face 2                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4f20(const REAL l[DIM+1])
{
  return((4.0*l[0] - 1.0)*l[0]*l[1]*l[3]*32.0);
}

static const REAL *grd_phi4f20(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[0] = 32.0*(8.0*l[0] - 1.0)*l[1]*l[3];
  grd[1] = 32.0*(4.0*l[0] - 1.0)*l[0]*l[3];
  grd[3] = 32.0*(4.0*l[0] - 1.0)*l[0]*l[1];
  return((const REAL *) grd);
}

static const REAL (*D2_phi4f20(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[0][0] = 256.0*l[1]*l[3];
  D2[0][1] = D2[1][0] = 32.0*(8.0*l[0] - 1.0)*l[3];
  D2[0][3] = D2[3][0] = 32.0*(8.0*l[0] - 1.0)*l[1];
  D2[1][3] = D2[3][1] = 32.0*(4.0*l[0] - 1.0)*l[0];
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 1 at face 2                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4f21(const REAL l[DIM+1])
{
  return(l[0]*(4.0*l[1] - 1.0)*l[1]*l[3]*32.0);
}

static const REAL *grd_phi4f21(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[0] = 32.0*(4.0*l[1] - 1.0)*l[1]*l[3];
  grd[1] = 32.0*(8.0*l[1] - 1.0)*l[0]*l[3];
  grd[3] = 32.0*(4.0*l[1] - 1.0)*l[0]*l[1];
  return((const REAL *) grd);
}

static const REAL (*D2_phi4f21(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[0][1] = D2[1][0] = 32.0*(8.0*l[1] - 1.0)*l[3];
  D2[0][3] = D2[3][0] = 32.0*(4.0*l[1] - 1.0)*l[1];
  D2[1][1] = 256.0*l[0]*l[3];
  D2[1][3] = D2[3][1] = 32.0*(8.0*l[1] - 1.0)*l[0];
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 2 at face 2                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4f22(const REAL l[DIM+1])
{
  return(l[0]*l[1]*(4.0*l[3] - 1.0)*l[3]*32.0);
}

static const REAL *grd_phi4f22(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[0] = 32.0*(4.0*l[3] - 1.0)*l[1]*l[3];
  grd[1] = 32.0*(4.0*l[3] - 1.0)*l[0]*l[3];
  grd[3] = 32.0*(8.0*l[3] - 1.0)*l[0]*l[1];
  return((const REAL *) grd);
}

static const REAL (*D2_phi4f22(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[0][1] = D2[1][0] = 32.0*(4.0*l[3] - 1.0)*l[3];
  D2[0][3] = D2[3][0] = 32.0*(8.0*l[3] - 1.0)*l[1];
  D2[1][3] = D2[3][1] = 32.0*(8.0*l[3] - 1.0)*l[0];
  D2[3][3] = 256.0*l[0]*l[1];
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 0 at face 1                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4f10(const REAL l[DIM+1])
{
  return((4.0*l[0] - 1.0)*l[0]*l[2]*l[3]*32.0);
}

static const REAL *grd_phi4f10(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[0] = 32.0*(8.0*l[0] - 1.0)*l[2]*l[3];
  grd[2] = 32.0*(4.0*l[0] - 1.0)*l[0]*l[3];
  grd[3] = 32.0*(4.0*l[0] - 1.0)*l[0]*l[2];
  return((const REAL *) grd);
}

static const REAL (*D2_phi4f10(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[0][0] = 256.0*l[2]*l[3];
  D2[0][2] = D2[2][0] = 32.0*(8.0*l[0] - 1.0)*l[3];
  D2[0][3] = D2[3][0] = 32.0*(8.0*l[0] - 1.0)*l[2];
  D2[2][3] = D2[3][2] = 32.0*(4.0*l[0] - 1.0)*l[0];
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 1 at face 1                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4f11(const REAL l[DIM+1])
{
  return(l[0]*(4.0*l[2] - 1.0)*l[2]*l[3]*32.0);
}

static const REAL *grd_phi4f11(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[0] = 32.0*(4.0*l[2] - 1.0)*l[2]*l[3];
  grd[2] = 32.0*(8.0*l[2] - 1.0)*l[0]*l[3];
  grd[3] = 32.0*(4.0*l[2] - 1.0)*l[0]*l[2];
  return((const REAL *) grd);
}

static const REAL (*D2_phi4f11(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[0][2] = D2[2][0] = 32.0*(8.0*l[2] - 1.0)*l[3];
  D2[0][3] = D2[3][0] = 32.0*(4.0*l[2] - 1.0)*l[2];
  D2[2][2] = 256.0*l[0]*l[3];
  D2[2][3] = D2[3][2] = 32.0*(8.0*l[2] - 1.0)*l[0];
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 2 at face 1                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4f12(const REAL l[DIM+1])
{
  return(l[0]*l[2]*(4.0*l[3] - 1.0)*l[3]*32.0);
}

static const REAL *grd_phi4f12(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[0] = 32.0*(4.0*l[3] - 1.0)*l[2]*l[3];
  grd[2] = 32.0*(4.0*l[3] - 1.0)*l[0]*l[3];
  grd[3] = 32.0*(8.0*l[3] - 1.0)*l[0]*l[2];
  return((const REAL *) grd);
}

static const REAL (*D2_phi4f12(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[0][2] = D2[2][0] = 32.0*(4.0*l[3] - 1.0)*l[3];
  D2[0][3] = D2[3][0] = 32.0*(8.0*l[3] - 1.0)*l[2];
  D2[2][3] = D2[3][2] = 32.0*(8.0*l[3] - 1.0)*l[0];
  D2[3][3] = 256.0*l[0]*l[2];
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 0 at face 0                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4f00(const REAL l[DIM+1])
{
  return((4.0*l[1] - 1.0)*l[1]*l[2]*l[3]*32.0);
}

static const REAL *grd_phi4f00(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[1] = 32.0*(8.0*l[1] - 1.0)*l[2]*l[3];
  grd[2] = 32.0*(4.0*l[1] - 1.0)*l[1]*l[3];
  grd[3] = 32.0*(4.0*l[1] - 1.0)*l[1]*l[2];
  return((const REAL *) grd);
}

static const REAL (*D2_phi4f00(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[1][1] = 256.0*l[2]*l[3];
  D2[1][2] = D2[2][1] = 32.0*(8.0*l[1] - 1.0)*l[3];
  D2[1][3] = D2[3][1] = 32.0*(8.0*l[1] - 1.0)*l[2];
  D2[2][3] = D2[3][2] = 32.0*(4.0*l[1] - 1.0)*l[1];
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 1 at face 0                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4f01(const REAL l[DIM+1])
{
  return(l[1]*(4.0*l[2] - 1.0)*l[2]*l[3]*32.0);
}

static const REAL *grd_phi4f01(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[1] = 32.0*(4.0*l[2] - 1.0)*l[2]*l[3];
  grd[2] = 32.0*(8.0*l[2] - 1.0)*l[1]*l[3];
  grd[3] = 32.0*(4.0*l[2] - 1.0)*l[1]*l[2];
  return((const REAL *) grd);
}

static const REAL (*D2_phi4f01(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[1][2] = D2[2][1] = 32.0*(8.0*l[2] - 1.0)*l[3];
  D2[1][3] = D2[3][1] = 32.0*(4.0*l[2] - 1.0)*l[2];
  D2[2][2] = 256.0*l[1]*l[3];
  D2[2][3] = D2[3][2] = 32.0*(8.0*l[2] - 1.0)*l[1];
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 2 at face 0                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4f02(const REAL l[DIM+1])
{
  return(l[1]*l[2]*(4.0*l[3] - 1.0)*l[3]*32.0);
}

static const REAL *grd_phi4f02(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[1] = 32.0*(4.0*l[3] - 1.0)*l[2]*l[3];
  grd[2] = 32.0*(4.0*l[3] - 1.0)*l[1]*l[3];
  grd[3] = 32.0*(8.0*l[3] - 1.0)*l[1]*l[2];
  return((const REAL *) grd);
}

static const REAL (*D2_phi4f02(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[1][2] = D2[2][1] = 32.0*(4.0*l[3] - 1.0)*l[3];
  D2[1][3] = D2[3][1] = 32.0*(8.0*l[3] - 1.0)*l[2];
  D2[2][3] = D2[3][2] = 32.0*(8.0*l[3] - 1.0)*l[1];
  D2[3][3] = 256.0*l[1]*l[2];
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  basisfunction 2 at center                                               */
/*--------------------------------------------------------------------------*/

static REAL phi4c(const REAL l[DIM+1])
{
  return(256.0*l[0]*l[1]*l[2]*l[3]);
}

static const REAL *grd_phi4c(const REAL l[DIM+1])
{
  static REAL grd[DIM+1] = {0,0,0,0};
  grd[0] = 256.0*l[1]*l[2]*l[3];
  grd[1] = 256.0*l[0]*l[2]*l[3];
  grd[2] = 256.0*l[0]*l[1]*l[3];
  grd[3] = 256.0*l[0]*l[1]*l[2];
  return((const REAL *) grd);
}

static const REAL (*D2_phi4c(const REAL l[DIM+1]))[DIM+1]
{
  static  REAL D2[DIM+1][DIM+1] = {{0,0,0,0},{0,0,0,0},{0,0,0,0},{0,0,0,0}};
  D2[0][1] = D2[1][0] = 256.0*l[2]*l[3];
  D2[0][2] = D2[2][0] = 256.0*l[1]*l[3];
  D2[0][3] = D2[3][0] = 256.0*l[1]*l[2];
  D2[1][2] = D2[2][1] = 256.0*l[0]*l[3];
  D2[1][3] = D2[3][1] = 256.0*l[0]*l[2];
  D2[2][3] = D2[3][2] = 256.0*l[0]*l[1];
  return((const REAL (*)[DIM+1]) D2);
}

/*--------------------------------------------------------------------------*/
/*  functions for combining basisfunctions with coefficients                */
/*--------------------------------------------------------------------------*/

static const int *order_e_indices(const EL *el, int edge)
{
  static int  sorted[2][3] = {{0,1,2}, {2,1,0}};
  DOF    **dof = el->dof;

  if (dof[vertex_of_edge[edge][0]][0] < dof[vertex_of_edge[edge][1]][0])
    return((const int *) sorted[0]);
  else
    return((const int *) sorted[1]);
}

static const int *order_f_indices(const EL *el, int face)
{
  FUNCNAME("order_f_indices");
  static int  sorted[7][3] = {{0,0,0}, {0,2,1}, {1,0,2}, {0,1,2},
			      {2,1,0}, {1,2,0}, {2,0,1}};
  static int  vertex_of_face[4][3] = {{1,2,3}, {0,2,3}, {0,1,3}, {0,1,2}};
  int    no, *vof;
  DOF    **dof = el->dof;


  vof = vertex_of_face[face];
  no = 0;
  if (dof[vof[0]][0] < dof[vof[1]][0])
    no++;
  if (dof[vof[1]][0] < dof[vof[2]][0])
    no += 2;
  if (dof[vof[2]][0] < dof[vof[0]][0])
    no += 4;
  
  if (no >= 1  &&  no <= 6)
    return((const int *) sorted[no]);
  else
  {
    MSG("can not sort face indices of element %d at face %d\n", 
	INDEX(el), face);
    return(nil);
  }
}

static const DOF *get_dof_indices4(const EL *el, const DOF_ADMIN *admin,
				   DOF *idof)
{
  FUNCNAME("get_dof_indices4");
  static DOF  index_vec[N_BAS4];
  DOF         *rvec = idof ? idof : index_vec;
  int         i, j, k, n0, node0;
  const int   *indi;
  DOF         **dof = el->dof;

  n0 = admin->n0_dof[VERTEX];
  for (j = i = 0; i < N_VERTICES; i++)
    rvec[j++] = dof[i][n0];

  n0 = admin->n0_dof[EDGE];
  node0 = admin->mesh->node[EDGE];
  for (i = 0; i < N_EDGES; node0++, i++)
  {
    indi = order_e_indices(el, i);
    for (k = 0; k < 3; k++)
      rvec[j++] = dof[node0][n0+indi[k]];
  }

  n0 = admin->n0_dof[FACE];
  node0 = admin->mesh->node[FACE];
  for (i = 0; i < N_FACES; node0++, i++)
  {
    indi = order_f_indices(el, i);

    for (k = 0; k < 3; k++)
      rvec[j++] = dof[node0][n0+indi[k]];
  }

  TEST_EXIT(j == N_BAS4-1)("found not enough dofs\n");

  n0 = admin->n0_dof[CENTER];
  node0 = admin->mesh->node[CENTER];
  rvec[j] = dof[node0][n0];

  return((const DOF *) rvec);
}


static const S_CHAR *get_bound4(const EL_INFO *el_info, S_CHAR *bound)
{
  FUNCNAME("get_bound4");
  static S_CHAR  bound_vec[N_BAS4];
  S_CHAR         *rvec = bound ? bound : bound_vec;
  int            i, j;

  TEST_FLAG(FILL_BOUND, el_info);

  for (j = i = 0; i < N_VERTICES; j++, i++)
    rvec[j] = el_info->bound[i];
  for (i = N_FACES; i < N_FACES+N_EDGES; j += 3, i++)
    rvec[j] = rvec[j+1] = rvec[j+2] =  GET_BOUND(el_info->boundary[i]);

  for (i = 0; i < N_FACES; j += 3, i++)
    rvec[j] = rvec[j+1] = rvec[j+2] = GET_BOUND(el_info->boundary[i]);
  
  TEST_EXIT(j == N_BAS4-1)("found not enough boundarys\n");
  rvec[j] = INTERIOR;

  return((const S_CHAR *) rvec);
}

static const int *get_int_vec4(const EL *el, const DOF_INT_VEC *vec,
				int *ivec)
{
  FUNCNAME("get_int_vec4");
  int             i, j, k, n0, node0;
  const int       *indi;
  static int      local_vec[N_BAS4];
  int             *v = nil, *rvec = ivec ? ivec : local_vec;
  DOF             **dof = el->dof;
  const DOF_ADMIN *admin = nil;

  GET_DOF_VEC(v, vec);
  GET_STRUCT(admin,vec->fe_space);

  n0 = admin->n0_dof[VERTEX];
  for (j = i = 0; i < N_VERTICES; i++)
    rvec[j++] = v[dof[i][n0]];

  n0 = admin->n0_dof[EDGE];
  node0 = admin->mesh->node[EDGE];
  for (i = 0; i < N_EDGES; node0++, i++)
  {
    indi = order_e_indices(el, i);
    for (k = 0; k < 3; k++)
      rvec[j++] = v[dof[node0][n0+indi[k]]];
  }

  n0 = admin->n0_dof[FACE];
  node0 = admin->mesh->node[FACE];
  for (i = 0; i < N_FACES; node0++, i++)
  {
    indi = order_f_indices(el, i);

    for (k = 0; k < 3; k++)
      rvec[j++] = v[dof[node0][n0+indi[k]]];

  }

  TEST_EXIT(j == N_BAS4-1)("found not enough dofs\n");

  n0 = admin->n0_dof[CENTER];
  node0 = admin->mesh->node[CENTER];
  rvec[j] = v[dof[node0][n0]];

  return((const int *) rvec);
}

static const REAL *get_real_vec4(const EL *el, const DOF_REAL_VEC *vec,
				  REAL *Rvec)
{
  FUNCNAME("get_real_vec4");
  int             i, j, k, n0, node0;
  const int       *indi;
  static REAL     local_vec[N_BAS4];
  REAL            *v = nil, *rvec = Rvec ? Rvec : local_vec;
  DOF             **dof = el->dof;
  const DOF_ADMIN *admin = nil;

  GET_DOF_VEC(v, vec);
  GET_STRUCT(admin,vec->fe_space);

  n0 = admin->n0_dof[VERTEX];
  for (j = i = 0; i < N_VERTICES; i++)
    rvec[j++] = v[dof[i][n0]];

  n0 = admin->n0_dof[EDGE];
  node0 = admin->mesh->node[EDGE];
  for (i = 0; i < N_EDGES; node0++, i++)
  {
    indi = order_e_indices(el, i);
    for (k = 0; k < 3; k++)
      rvec[j++] = v[dof[node0][n0+indi[k]]];
  }

  n0 = admin->n0_dof[FACE];
  node0 = admin->mesh->node[FACE];
  for (i = 0; i < N_FACES; node0++, i++)
  {
    indi = order_f_indices(el, i);

    for (k = 0; k < 3; k++)
      rvec[j++] = v[dof[node0][n0+indi[k]]];

  }

  TEST_EXIT(j == N_BAS4-1)("found not enough dofs\n");

  n0 = admin->n0_dof[CENTER];
  node0 = admin->mesh->node[CENTER];
  rvec[j] = v[dof[node0][n0]];

  return((const REAL *) rvec);
}

static const REAL_D *get_real_d_vec4(const EL *el, const DOF_REAL_D_VEC *vec,
				     REAL_D *Rvec)
{
  FUNCNAME("get_real_d_vec4");
  int             i, j, k, l, n0, node0;
  const int       *indi;
  static REAL_D   local_vec[N_BAS4];
  REAL_D          *v = nil, *rvec = Rvec ? Rvec : local_vec;
  DOF             **dof = el->dof;
  const DOF_ADMIN *admin = nil;

  GET_DOF_VEC(v, vec);
  GET_STRUCT(admin,vec->fe_space);

  n0 = admin->n0_dof[VERTEX];
  for (j = i = 0; i < N_VERTICES; i++, j++)
    for (l = 0; l < DIM_OF_WORLD; l++)
      rvec[j][l] = v[dof[i][n0]][l];

  n0 = admin->n0_dof[EDGE];
  node0 = admin->mesh->node[EDGE];
  for (i = 0; i < N_EDGES; node0++, i++)
  {
    indi = order_e_indices(el, i);
    for (k = 0; k < 3; k++, j++)
      for (l = 0; l < DIM_OF_WORLD; l++)
	rvec[j][l] = v[dof[node0][n0+indi[k]]][l];
  }

  n0 = admin->n0_dof[FACE];
  node0 = admin->mesh->node[FACE];
  for (i = 0; i < N_FACES; node0++, i++)
  {
    indi = order_f_indices(el, i);

    for (k = 0; k < 3; k++, j++)
      for (l = 0; l < DIM_OF_WORLD; l++)
	rvec[j][l] = v[dof[node0][n0+indi[k]]][l];

  }

  TEST_EXIT(j == N_BAS4-1)("found not enough dofs\n");

  n0 = admin->n0_dof[CENTER];
  node0 = admin->mesh->node[CENTER];
  for (l = 0; l < DIM_OF_WORLD; l++)
    rvec[j] [l]= v[dof[node0][n0]][l];

  return((const REAL_D *) rvec);
}

static const U_CHAR *get_uchar_vec4(const EL *el, const DOF_UCHAR_VEC *vec,
				     U_CHAR *uvec)
{
  FUNCNAME("get_uchar_vec4");
  int             i, j, k, n0, node0;
  const int       *indi;
  static U_CHAR   local_vec[N_BAS4];
  U_CHAR          *v = nil, *rvec = uvec ? uvec : local_vec;
  DOF             **dof = el->dof;
  const DOF_ADMIN *admin = nil;

  GET_DOF_VEC(v, vec);
  GET_STRUCT(admin,vec->fe_space);

  n0 = admin->n0_dof[VERTEX];
  for (j = i = 0; i < N_VERTICES; i++)
    rvec[j++] = v[dof[i][n0]];

  n0 = admin->n0_dof[EDGE];
  node0 = admin->mesh->node[EDGE];
  for (i = 0; i < N_EDGES; node0++, i++)
  {
    indi = order_e_indices(el, i);
    for (k = 0; k < 3; k++)
      rvec[j++] = v[dof[node0][n0+indi[k]]];
  }

  n0 = admin->n0_dof[FACE];
  node0 = admin->mesh->node[FACE];
  for (i = 0; i < N_FACES; node0++, i++)
  {
    indi = order_f_indices(el, i);

    for (k = 0; k < 3; k++)
      rvec[j++] = v[dof[node0][n0+indi[k]]];

  }

  TEST_EXIT(j == N_BAS4-1)("found not enough dofs\n");

  n0 = admin->n0_dof[CENTER];
  node0 = admin->mesh->node[CENTER];
  rvec[j] = v[dof[node0][n0]];

  return((const U_CHAR *) rvec);
}

static const S_CHAR *get_schar_vec4(const EL *el, const DOF_SCHAR_VEC *vec,
				     S_CHAR *svec)
{
  FUNCNAME("get_schar_vec4");
  int             i, j, k, n0, node0;
  const int       *indi;
  static S_CHAR   local_vec[N_BAS4];
  S_CHAR          *v = nil, *rvec = svec ? svec : local_vec;
  DOF             **dof = el->dof;
  const DOF_ADMIN *admin = nil;

  GET_DOF_VEC(v, vec);
  GET_STRUCT(admin,vec->fe_space);

  n0 = admin->n0_dof[VERTEX];
  for (j = i = 0; i < N_VERTICES; i++)
    rvec[j++] = v[dof[i][n0]];

  n0 = admin->n0_dof[EDGE];
  node0 = admin->mesh->node[EDGE];
  for (i = 0; i < N_EDGES; node0++, i++)
  {
    indi = order_e_indices(el, i);
    for (k = 0; k < 3; k++)
      rvec[j++] = v[dof[node0][n0+indi[k]]];
  }

  n0 = admin->n0_dof[FACE];
  node0 = admin->mesh->node[FACE];
  for (i = 0; i < N_FACES; node0++, i++)
  {
    indi = order_f_indices(el, i);

    for (k = 0; k < 3; k++)
      rvec[j++] = v[dof[node0][n0+indi[k]]];

  }

  TEST_EXIT(j == N_BAS4-1)("found not enough dofs\n");

  n0 = admin->n0_dof[CENTER];
  node0 = admin->mesh->node[CENTER];
  rvec[j] = v[dof[node0][n0]];

  return((const S_CHAR *) rvec);
}


static const REAL *interpol4(const EL_INFO *el_info, int no, 
			      const int *b_no, REAL (*f)(const REAL_D),
			      REAL *vec)
{
  FUNCNAME("interpol4");
  static REAL  inter[N_BAS4];
  REAL         *rvec = vec ? vec : inter;
  REAL_D       x;
  const REAL_D *coord;
  int          i;
  const PARAMETRIC *parametric = el_info->mesh->parametric;

  if (parametric) 
    ERROR_EXIT("Sorry, parametric support for lagrange4 not implemented yet!\n");

  TEST_FLAG(FILL_COORDS, el_info);
  coord = el_info->coord;

  if (b_no)
  {
    if (no <= 0  ||  no > N_BAS4)
    {
      MSG("something is wrong, doing nothing\n");
      rvec[0] = 0.0;
      return((const REAL *) rvec);
    }
    for (i = 0; i < no; i++)
    {
      if (b_no[i] < N_VERTICES)
      {
	rvec[i] = f(coord[b_no[i]]);
      }
      else 
      {
	coord_to_world(el_info, bary4[b_no[i]], x);
	rvec[i] = f(x);
      }
    }
  }
  else
  {
    for (i = 0; i < N_VERTICES; i++)
      rvec[i] = f(coord[i]);
    for (i = N_VERTICES; i < N_BAS4; i++)
    {
      coord_to_world(el_info, bary4[i], x);
      rvec[i] = f(x);
    }
  }  
    
  return((const REAL *) rvec);
}

static const REAL_D *interpol_d4(const EL_INFO *el_info, int no, 
				 const int *b_no,
				 const REAL *(*f)(const REAL_D, REAL_D),
				 REAL_D *vec)
{
  FUNCNAME("interpol4");
  static REAL_D inter[N_BAS4];
  REAL_D        *rvec = vec ? vec : inter;
  REAL_D        x;
  const REAL_D  *coord;
  int           i;
  const PARAMETRIC *parametric = el_info->mesh->parametric;

  if (parametric) 
    ERROR_EXIT("Sorry, parametric support for lagrange4 not implemented yet!\n");

  TEST_FLAG(FILL_COORDS, el_info);
  coord = el_info->coord;

  if (b_no)
  {
    if (no <= 0  ||  no > N_BAS4)
    {
      ERROR("something is wrong, doing nothing\n");

      for (i = 0; i < DIM_OF_WORLD; i++)
	rvec[0][i] = 0.0;
      return((const REAL_D *) rvec);
    }
    for (i = 0; i < no; i++)
    {
      if (b_no[i] < N_VERTICES)
      {
	(*f)(coord[b_no[i]], rvec[i]);
      }
      else 
      {
	coord_to_world(el_info, bary4[b_no[i]], x);
	(*f)(x, rvec[i]);
      }
    }
  }
  else
  {
    for (i = 0; i < N_VERTICES; i++)
      (*f)(coord[i], rvec[i]);
    for (i = N_VERTICES; i < N_BAS4; i++)
    {
      coord_to_world(el_info, bary4[i], x);
      (*f)(x, rvec[i]);
    }
  }  
    
  return((const REAL_D *) rvec);
}


/*--------------------------------------------------------------------------*/
/*  functions for interpolation/ restriction during refinement/coarsening   */
/*--------------------------------------------------------------------------*/

static void real_refine_inter4(DOF_REAL_VEC *drv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_refine_inter4");
  DOF       pd[N_BAS4];
  const DOF *cd;
  EL        *el;
  REAL      *v = nil;
  int       i, typ, lr_set;
  const DOF       *(*get_dof_indices)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF_ADMIN *admin;

  if (n < 1) return;
  el = list->el;
  typ = list->el_type;

  GET_DOF_VEC(v, drv);
  if (!drv->fe_space)
  {
    ERROR("no fe_space in dof_real_vec %s\n", NAME(drv));
    return;
  }
  else if (!drv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drv->fe_space));
    return;
  }
  get_dof_indices = drv->fe_space->bas_fcts->get_dof_indices;
  GET_STRUCT(admin,drv->fe_space);

  get_dof_indices(el, admin, pd);

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

  cd = get_dof_indices(el->child[0], admin, nil);

  v[cd[3]] = (v[pd[5]]);
  v[cd[10]] = (0.2734375*v[pd[0]] - 0.0390625*v[pd[1]]
	       + 1.09375*v[pd[4]] - 0.546875*v[pd[5]]
	       + 0.21875*v[pd[6]]);
  v[cd[11]] = (v[pd[4]]);
  v[cd[12]] = (-0.0390625*v[pd[0]] + 0.0234375*v[pd[1]]
	       + 0.46875*v[pd[4]] + 0.703125*v[pd[5]]
	       - 0.15625*v[pd[6]]);
  v[cd[16]] = (0.0390625*(-v[pd[0]] - v[pd[1]])
	       + 0.03125*(v[pd[4]] + v[pd[6]])
	       + 0.015625*v[pd[5]] +
	       0.1875*(v[pd[7]]+v[pd[13]]-v[pd[31]]-v[pd[32]])
	       + 0.375*(-v[pd[8]] - v[pd[14]])
	       + 0.5*(v[pd[9]] + v[pd[15]]) + 0.75*v[pd[33]]);
  v[cd[17]] = (v[pd[33]]);
  v[cd[18]] = (0.0234375*(v[pd[0]] + v[pd[1]]) 
	       + 0.09375*(-v[pd[4]] - v[pd[6]]) + 0.140625*v[pd[5]]
	       + 0.0625*(-v[pd[7]] - v[pd[13]])
	       + 0.5625*(v[pd[31]] + v[pd[32]]));
  v[cd[19]] = (0.0390625*(-v[pd[0]] - v[pd[1]])
	       + 0.03125*(v[pd[4]] + v[pd[6]]) + 0.015625*v[pd[5]]
	       + 0.1875*(v[pd[10]]+v[pd[16]]-v[pd[28]]-v[pd[29]])
	       + 0.375*(-v[pd[11]] - v[pd[17]])
	       + 0.5*(v[pd[12]] + v[pd[18]]) + 0.75*v[pd[30]]);
  v[cd[20]] = (v[pd[30]]);
  v[cd[21]] = (0.0234375*(v[pd[0]] + v[pd[1]])
	       + 0.09375*(-v[pd[4]] - v[pd[6]]) + 0.140625*v[pd[5]]
	       + 0.0625*(-v[pd[10]] - v[pd[16]])
	       + 0.5625*(v[pd[28]] + v[pd[29]]));
  v[cd[22]] = (0.0390625*(-v[pd[0]] - v[pd[1]])
	       + 0.03125*(v[pd[4]] + v[pd[6]]) + 0.015625*v[pd[5]]
	       + 0.125*(v[pd[7]]-v[pd[8]]+v[pd[13]]-v[pd[14]]
			-v[pd[31]]-v[pd[32]])
	       + 0.0625*(v[pd[10]]+v[pd[16]]-v[pd[28]]-v[pd[29]])
	       + 0.25*(-v[pd[22]] - v[pd[25]] + v[pd[33]])
	       + 0.5*(v[pd[23]] + v[pd[26]] + v[pd[34]]));
  v[cd[23]] = (0.0390625*(-v[pd[0]] - v[pd[1]])
	       + 0.03125*(v[pd[4]] + v[pd[6]]) + 0.015625*v[pd[5]]
	       + 0.0625*(v[pd[7]]+v[pd[13]]-v[pd[31]]-v[pd[32]])
	       + 0.125*(v[pd[10]]-v[pd[11]]+v[pd[16]]-v[pd[17]]
			-v[pd[28]]-v[pd[29]])
	       + 0.25*(-v[pd[22]] - v[pd[25]] + v[pd[30]])
	       + 0.5*(v[pd[24]] + v[pd[27]] + v[pd[34]]));
  v[cd[24]] = (v[pd[34]]);
  v[cd[25]] = (0.0390625*(-v[pd[0]] - v[pd[1]])
	       + 0.15625*(v[pd[4]] + v[pd[6]]) - 0.234375*v[pd[5]]
	       + 0.3125*(v[pd[10]] - v[pd[29]]) + 0.0625*v[pd[16]]
	       + 0.9375*v[pd[28]]);
  v[cd[26]] = (0.0234375*v[pd[0]] - 0.0390625*v[pd[1]]
	       - 0.03125*v[pd[4]] - 0.046875*v[pd[5]] 
	       + 0.09375*v[pd[6]]
	       + 0.125*(-v[pd[10]] + v[pd[16]] - v[pd[17]])
	       + 0.375*(v[pd[11]] + v[pd[28]] - v[pd[29]])
	       + 0.75*v[pd[30]]);
  v[cd[27]] = (v[pd[28]]);
  v[cd[28]] = (0.0390625*(-v[pd[0]] - v[pd[1]])
	       + 0.15625*(v[pd[4]] + v[pd[6]]) - 0.234375*v[pd[5]]
	       + 0.3125*(v[pd[7]] - v[pd[32]]) + 0.0625*v[pd[13]]
	       + 0.9375*v[pd[31]]);
  v[cd[29]] = (0.0234375*v[pd[0]] - 0.0390625*v[pd[1]]
	       - 0.03125*v[pd[4]] - 0.046875*v[pd[5]]
	       + 0.09375*v[pd[6]]
	       + 0.125*(-v[pd[7]] + v[pd[13]] - v[pd[14]])
	       + 0.375*(v[pd[8]] + v[pd[31]] - v[pd[32]])
	       + 0.75*v[pd[33]]);
  v[cd[30]] = (v[pd[31]]);
  v[cd[34]] = (0.0234375*v[pd[0]] - 0.0390625*v[pd[1]]
	       - 0.03125*v[pd[4]] - 0.046875*v[pd[5]]
	       + 0.09375*v[pd[6]]
	       + 0.0625*(-v[pd[7]]-v[pd[10]]+v[pd[13]]+v[pd[16]])
	       - 0.125*v[pd[22]] + 0.375*v[pd[25]]
	       + 0.1875*(v[pd[28]]-v[pd[29]]+v[pd[31]]-v[pd[32]])
	       + 0.75*v[pd[34]]);

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

  cd = get_dof_indices(el->child[1], admin, nil);
  if (typ == 0)
  {
/*--------------------------------------------------------------------------*/
/*  parent of el_type 0                                                     */
/*--------------------------------------------------------------------------*/

    v[cd[10]] = (-0.0390625*v[pd[0]] + 0.2734375*v[pd[1]]
		 + 0.21875*v[pd[4]] - 0.546875*v[pd[5]]
		 + 1.09375*v[pd[6]]);
    v[cd[11]] = v[pd[6]];
    v[cd[12]] = (0.0234375*v[pd[0]] - 0.0390625*v[pd[1]]
		 - 0.15625*v[pd[4]] + 0.703125*v[pd[5]]
		 + 0.46875*v[pd[6]]);
    v[cd[25]] = (0.0390625*(-v[pd[0]] - v[pd[1]]) 
		 + 0.15625*(v[pd[4]] + v[pd[6]])
		 - 0.234375*v[pd[5]] + 0.0625*v[pd[7]]
		 + 0.3125*(v[pd[13]] - v[pd[31]])
		 + 0.9375*v[pd[32]]);
    v[cd[26]] = (-0.0390625*v[pd[0]] + 0.0234375*v[pd[1]]
		 + 0.09375*v[pd[4]] - 0.046875*v[pd[5]]
		 - 0.03125*v[pd[6]]
		 + 0.125*(v[pd[7]] - v[pd[8]] - v[pd[13]])
		 + 0.375*(v[pd[14]] - v[pd[31]] + v[pd[32]])
		 + 0.75*v[pd[33]]);
    v[cd[27]] = v[pd[32]];
    v[cd[28]] = (0.0390625*(-v[pd[0]] - v[pd[1]])
		 + 0.15625*(v[pd[4]] + v[pd[6]])
		 - 0.234375*v[pd[5]] + 0.0625*v[pd[10]]
		 + 0.3125*(v[pd[16]] - v[pd[28]])
		 + 0.9375*v[pd[29]]);
    v[cd[29]] = (-0.0390625*v[pd[0]] + 0.0234375*v[pd[1]]
		 + 0.09375*v[pd[4]] - 0.046875*v[pd[5]]
		 - 0.03125*v[pd[6]]
		 + 0.125*(v[pd[10]] - v[pd[11]] - v[pd[16]])
		 + 0.375*(v[pd[17]] - v[pd[28]] + v[pd[29]])
		 + 0.75*v[pd[30]]);
    v[cd[30]] = v[pd[29]];
    v[cd[34]] = (-0.0390625*v[pd[0]] + 0.0234375*v[pd[1]]
		 + 0.09375*v[pd[4]] - 0.046875*v[pd[5]]
		 - 0.03125*v[pd[6]]
		 + 0.0625*(v[pd[7]] + v[pd[10]]
			   -v[pd[13]] - v[pd[16]])
		 + 0.375*v[pd[22]] - 0.125*v[pd[25]]
		 + 0.1875*(-v[pd[28]] + v[pd[29]]
			   -v[pd[31]] + v[pd[32]])
		 + 0.75*v[pd[34]]);
  }
  else
  {
/*--------------------------------------------------------------------------*/
/*  parent of el_type 1|2                                                   */
/*--------------------------------------------------------------------------*/

    v[cd[10]] = (-0.0390625*v[pd[0]] + 0.2734375*v[pd[1]]
		 + 0.21875*v[pd[4]] - 0.546875*v[pd[5]]
		 + 1.09375*v[pd[6]]);
    v[cd[11]] = v[pd[6]];
    v[cd[12]] = (0.0234375*v[pd[0]] - 0.0390625*v[pd[1]]
		 - 0.15625*v[pd[4]] + 0.703125*v[pd[5]]
		 + 0.46875*v[pd[6]]);
    v[cd[25]] = (0.0390625*(-v[pd[0]] - v[pd[1]])
		 + 0.15625*(v[pd[4]] + v[pd[6]])
		 - 0.234375*v[pd[5]] + 0.0625*v[pd[10]]
		 + 0.3125*(v[pd[16]] - v[pd[28]])
		 + 0.9375*v[pd[29]]);
    v[cd[26]] = (-0.0390625*v[pd[0]] + 0.0234375*v[pd[1]]
		 + 0.09375*v[pd[4]] - 0.046875*v[pd[5]]
		 - 0.03125*v[pd[6]] 
		 + 0.125*(v[pd[10]] - v[pd[11]] - v[pd[16]])
		 + 0.375*(v[pd[17]] - v[pd[28]] + v[pd[29]])
		 + 0.75*v[pd[30]]);
    v[cd[27]] = v[pd[29]];
    v[cd[28]] = (0.0390625*(-v[pd[0]] - v[pd[1]])
		 + 0.15625*(v[pd[4]] + v[pd[6]])
		 - 0.234375*v[pd[5]] + 0.0625*v[pd[7]]
		 + 0.3125*(v[pd[13]] - v[pd[31]])
		 + 0.9375*v[pd[32]]);
    v[cd[29]] = (-0.0390625*v[pd[0]] + 0.0234375*v[pd[1]]
		 + 0.09375*v[pd[4]] - 0.046875*v[pd[5]]
		 - 0.03125*v[pd[6]]
		 + 0.125*(v[pd[7]] - v[pd[8]] - v[pd[13]])
		 + 0.375*(v[pd[14]] - v[pd[31]] + v[pd[32]])
		 + 0.75*v[pd[33]]);
    v[cd[30]] = v[pd[32]];
    v[cd[34]] = (-0.0390625*v[pd[0]] + 0.0234375*v[pd[1]]
		 + 0.09375*v[pd[4]] - 0.046875*v[pd[5]]
		 - 0.03125*v[pd[6]]
		 + 0.0625*(v[pd[7]] + v[pd[10]]
			   - v[pd[13]] - v[pd[16]])
		 + 0.375*v[pd[22]] - 0.125*v[pd[25]]
		 + 0.1875*(-v[pd[28]] + v[pd[29]]
			   - v[pd[31]] + v[pd[32]])
		 + 0.75*v[pd[34]]);
  }

/*--------------------------------------------------------------------------*/
/*   adjust neighbour values                                                */
/*--------------------------------------------------------------------------*/
  
  for (i = 1; i < n; i++)
  {
    el = list[i].el;
    typ = list[i].el_type;
    get_dof_indices(el, admin, pd);

    lr_set = 0;
    if (list[i].neigh[0]  &&  list[i].neigh[0]->no < i)
      lr_set = 1;

    if (list[i].neigh[1]  &&  list[i].neigh[1]->no < i)
      lr_set += 2;

    TEST_EXIT(lr_set)("no values set on both neighbours\n");

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

    cd = get_dof_indices(el->child[0], admin, nil);

    switch(lr_set)
    {
    case 1:
      v[cd[16]] = (0.0390625*(-v[pd[0]] - v[pd[1]])
		   + 0.03125*(v[pd[4]] + v[pd[6]])
		   + 0.015625*v[pd[5]]
		   + 0.1875*(v[pd[7]] + v[pd[13]]
			     - v[pd[31]] - v[pd[32]])
		   + 0.375*(-v[pd[8]] - v[pd[14]])
		   + 0.5*(v[pd[9]] + v[pd[15]])
		   + 0.75*v[pd[33]]);
      v[cd[17]] = v[pd[33]];
      v[cd[18]] = (0.0234375*(v[pd[0]] + v[pd[1]])
		   + 0.09375*(-v[pd[4]] - v[pd[6]])
		   + 0.140625*v[pd[5]]
		   + 0.0625*(-v[pd[7]] - v[pd[13]])
		   + 0.5625*(v[pd[31]] + v[pd[32]]));
      v[cd[22]] = (0.0390625*(-v[pd[0]] - v[pd[1]])
		   + 0.03125*(v[pd[4]] + v[pd[6]])
		   + 0.015625*v[pd[5]]
		   + 0.125*(v[pd[7]] - v[pd[8]] + v[pd[13]]
			    - v[pd[14]] - v[pd[31]] - v[pd[32]])
		   + 0.0625*(v[pd[10]] + v[pd[16]]
			     - v[pd[28]] - v[pd[29]])
		   + 0.25*(-v[pd[22]] - v[pd[25]] + v[pd[33]])
		   + 0.5*(v[pd[23]] + v[pd[26]] + v[pd[34]]));
      v[cd[23]] = (0.0390625*(-v[pd[0]] - v[pd[1]])
		   + 0.03125*(v[pd[4]] + v[pd[6]])
		   + 0.015625*v[pd[5]]
		   + 0.0625*(v[pd[7]] + v[pd[13]]
			     - v[pd[31]] - v[pd[32]])
		   + 0.125*(v[pd[10]] - v[pd[11]] + v[pd[16]]
			    - v[pd[17]] - v[pd[28]] - v[pd[29]])
		   + 0.25*(-v[pd[22]] - v[pd[25]] + v[pd[30]])
		   + 0.5*(v[pd[24]] + v[pd[27]] + v[pd[34]]));
      v[cd[24]] = v[pd[34]];
      v[cd[28]] = (0.0390625*(-v[pd[0]] - v[pd[1]])
		   + 0.15625*(v[pd[4]] + v[pd[6]])
		   - 0.234375*v[pd[5]]
		   + 0.3125*(v[pd[7]] - v[pd[32]])
		   + 0.0625*v[pd[13]] + 0.9375*v[pd[31]]);
      v[cd[29]] = (0.0234375*v[pd[0]] - 0.0390625*v[pd[1]]
		   - 0.03125*v[pd[4]] - 0.046875*v[pd[5]]
		   + 0.09375*v[pd[6]]
		   + 0.125*(-v[pd[7]] + v[pd[13]] - v[pd[14]])
		   + 0.375*(v[pd[8]] + v[pd[31]] - v[pd[32]])
		   + 0.75*v[pd[33]]);
      v[cd[30]] = v[pd[31]];
      v[cd[34]] = (0.0234375*v[pd[0]] - 0.0390625*v[pd[1]]
		   - 0.03125*v[pd[4]] - 0.046875*v[pd[5]]
		   + 0.09375*v[pd[6]]
		   + 0.0625*(-v[pd[7]] - v[pd[10]]
			     + v[pd[13]] + v[pd[16]])
		   - 0.125*v[pd[22]] + 0.375*v[pd[25]]
		   + 0.1875*(v[pd[28]] - v[pd[29]]
			     + v[pd[31]] - v[pd[32]])
		   + 0.75*v[pd[34]]);
      break;
    case 2:
      v[cd[19]] = (0.0390625*(-v[pd[0]] - v[pd[1]])
		   + 0.03125*(v[pd[4]] + v[pd[6]])
		   + 0.015625*v[pd[5]]
		   + 0.1875*(v[pd[10]] + v[pd[16]]
			     - v[pd[28]] - v[pd[29]])
		   + 0.375*(-v[pd[11]] - v[pd[17]])
		   + 0.5*(v[pd[12]] + v[pd[18]])
		   + 0.75*v[pd[30]]);
      v[cd[20]] = v[pd[30]];
      v[cd[21]] = (0.0234375*(v[pd[0]] + v[pd[1]])
		   + 0.09375*(-v[pd[4]] - v[pd[6]])
		   + 0.140625*v[pd[5]]
		   + 0.0625*(-v[pd[10]] - v[pd[16]])
		   + 0.5625*(v[pd[28]] + v[pd[29]]));
      v[cd[22]] = (0.0390625*(-v[pd[0]] - v[pd[1]])
		   + 0.03125*(v[pd[4]] + v[pd[6]])
		   + 0.015625*v[pd[5]]
		   + 0.125*(v[pd[7]] - v[pd[8]] + v[pd[13]]
			    - v[pd[14]] - v[pd[31]] - v[pd[32]])
		   + 0.0625*(v[pd[10]] + v[pd[16]]
			     - v[pd[28]] - v[pd[29]])
		   + 0.25*(-v[pd[22]] - v[pd[25]] + v[pd[33]])
		   + 0.5*(v[pd[23]] + v[pd[26]] + v[pd[34]]));
      v[cd[23]] = (0.0390625*(-v[pd[0]] - v[pd[1]])
		   + 0.03125*(v[pd[4]] + v[pd[6]])
		   + 0.015625*v[pd[5]]
		   + 0.0625*(v[pd[7]] + v[pd[13]]
			     - v[pd[31]] - v[pd[32]])
		   + 0.125*(v[pd[10]] - v[pd[11]] + v[pd[16]]
			    - v[pd[17]] - v[pd[28]] - v[pd[29]])
		   + 0.25*(-v[pd[22]] - v[pd[25]] + v[pd[30]])
		   + 0.5*(v[pd[24]] + v[pd[27]] + v[pd[34]]));
      v[cd[24]] = v[pd[34]];
      v[cd[25]] = (0.0390625*(-v[pd[0]] - v[pd[1]])
		   + 0.15625*(v[pd[4]] + v[pd[6]])
		   - 0.234375*v[pd[5]]
		   + 0.3125*(v[pd[10]] - v[pd[29]])
		   + 0.0625*v[pd[16]] + 0.9375*v[pd[28]]);
      v[cd[26]] = (0.0234375*v[pd[0]] - 0.0390625*v[pd[1]]
		   - 0.03125*v[pd[4]] - 0.046875*v[pd[5]]
		   + 0.09375*v[pd[6]]
		   + 0.125*(-v[pd[10]] + v[pd[16]] - v[pd[17]])
		   + 0.375*(v[pd[11]] + v[pd[28]] - v[pd[29]])
		   + 0.75*v[pd[30]]);
      v[cd[27]] = v[pd[28]];
      v[cd[34]] = (0.0234375*v[pd[0]] - 0.0390625*v[pd[1]]
		   - 0.03125*v[pd[4]] - 0.046875*v[pd[5]]
		   + 0.09375*v[pd[6]]
		   + 0.0625*(-v[pd[7]] - v[pd[10]]
			     + v[pd[13]] + v[pd[16]])
		   - 0.125*v[pd[22]] + 0.375*v[pd[25]]
		   + 0.1875*(v[pd[28]] - v[pd[29]]
			     + v[pd[31]] - v[pd[32]])
		   + 0.75*v[pd[34]]);
      break;
    case 3:
      v[cd[22]] = (0.0390625*(-v[pd[0]] - v[pd[1]])
		   + 0.03125*(v[pd[4]] + v[pd[6]])
		   + 0.015625*v[pd[5]]
		   + 0.125*(v[pd[7]] - v[pd[8]] + v[pd[13]]
			    - v[pd[14]] - v[pd[31]] - v[pd[32]])
		   + 0.0625*(v[pd[10]] + v[pd[16]]
			     - v[pd[28]] - v[pd[29]])
		   + 0.25*(-v[pd[22]] - v[pd[25]] + v[pd[33]])
		   + 0.5*(v[pd[23]] + v[pd[26]] + v[pd[34]]));
      v[cd[23]] = (0.0390625*(-v[pd[0]] - v[pd[1]])
		   + 0.03125*(v[pd[4]] + v[pd[6]])
		   + 0.015625*v[pd[5]]
		   + 0.0625*(v[pd[7]] + v[pd[13]]
			     - v[pd[31]] - v[pd[32]])
		   + 0.125*(v[pd[10]] - v[pd[11]] + v[pd[16]]
			    - v[pd[17]] - v[pd[28]] - v[pd[29]])
		   + 0.25*(-v[pd[22]] - v[pd[25]] + v[pd[30]])
		   + 0.5*(v[pd[24]] + v[pd[27]] + v[pd[34]]));
      v[cd[24]] = v[pd[34]];
      v[cd[34]] = (0.0234375*v[pd[0]] - 0.0390625*v[pd[1]]
		   - 0.03125*v[pd[4]] - 0.046875*v[pd[5]]
		   + 0.09375*v[pd[6]]
		   + 0.0625*(-v[pd[7]] - v[pd[10]]
			     + v[pd[13]] + v[pd[16]])
		   - 0.125*v[pd[22]] + 0.375*v[pd[25]]
		   + 0.1875*(v[pd[28]] - v[pd[29]]
			     + v[pd[31]] - v[pd[32]])
		   + 0.75*v[pd[34]]);
      break;
    }

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

    cd = get_dof_indices(el->child[1], admin, nil);

    if (typ == 0)
    {
      switch(lr_set)
      {
      case 1:
	v[cd[25]] = (0.0390625*(-v[pd[0]] - v[pd[1]])
		     + 0.15625*(v[pd[4]] + v[pd[6]])
		     - 0.234375*v[pd[5]] + 0.0625*v[pd[7]]
		     + 0.3125*(v[pd[13]] - v[pd[31]])
		     + 0.9375*v[pd[32]]);
	v[cd[26]] = (-0.0390625*v[pd[0]] + 0.0234375*v[pd[1]]
		     + 0.09375*v[pd[4]] - 0.046875*v[pd[5]]
		     - 0.03125*v[pd[6]]
		     + 0.125*(v[pd[7]] - v[pd[8]] - v[pd[13]])
		     + 0.375*(v[pd[14]] - v[pd[31]] + v[pd[32]])
		     + 0.75*v[pd[33]]);
	v[cd[27]] = v[pd[32]];
	v[cd[34]] = (-0.0390625*v[pd[0]] + 0.0234375*v[pd[1]]
		     + 0.09375*v[pd[4]] - 0.046875*v[pd[5]]
		     - 0.03125*v[pd[6]]
		     + 0.0625*(v[pd[7]] + v[pd[10]]
			       - v[pd[13]] - v[pd[16]])
		     + 0.375*v[pd[22]] - 0.125*v[pd[25]]
		     + 0.1875*(-v[pd[28]] + v[pd[29]]
			       - v[pd[31]] + v[pd[32]])
		     + 0.75*v[pd[34]]);
	break;
      case 2:
	v[cd[28]] = (0.0390625*(-v[pd[0]] - v[pd[1]])
		     + 0.15625*(v[pd[4]] + v[pd[6]])
		     - 0.234375*v[pd[5]] + 0.0625*v[pd[10]]
		     + 0.3125*(v[pd[16]] - v[pd[28]])
		     + 0.9375*v[pd[29]]);
	v[cd[29]] = (-0.0390625*v[pd[0]] + 0.0234375*v[pd[1]]
		     + 0.09375*v[pd[4]] - 0.046875*v[pd[5]]
		     - 0.03125*v[pd[6]]
		     + 0.125*(v[pd[10]] - v[pd[11]] - v[pd[16]])
		     + 0.375*(v[pd[17]] - v[pd[28]] + v[pd[29]])
		     + 0.75*v[pd[30]]);
	v[cd[30]] = v[pd[29]];
	v[cd[34]] = (-0.0390625*v[pd[0]] + 0.0234375*v[pd[1]]
		     + 0.09375*v[pd[4]] - 0.046875*v[pd[5]]
		     - 0.03125*v[pd[6]]
		     + 0.0625*(v[pd[7]] + v[pd[10]]
			       - v[pd[13]] - v[pd[16]])
		     + 0.375*v[pd[22]] - 0.125*v[pd[25]]
		     + 0.1875*(-v[pd[28]] + v[pd[29]]
			       - v[pd[31]] + v[pd[32]])
		     + 0.75*v[pd[34]]);
	break;
      case 3:
	v[cd[34]] = (-0.0390625*v[pd[0]] + 0.0234375*v[pd[1]]
		     + 0.09375*v[pd[4]] - 0.046875*v[pd[5]]
		     - 0.03125*v[pd[6]]
		     + 0.0625*(v[pd[7]] + v[pd[10]]
			       - v[pd[13]] - v[pd[16]])
		     + 0.375*v[pd[22]] - 0.125*v[pd[25]]
		     + 0.1875*(-v[pd[28]] + v[pd[29]]
			       - v[pd[31]] + v[pd[32]])
		     + 0.75*v[pd[34]]);
	break;
      }
    }
    else
    {
      switch(lr_set)
      {
      case 1:
	v[cd[28]] = (0.0390625*(-v[pd[0]] - v[pd[1]])
		     + 0.15625*(v[pd[4]] + v[pd[6]])
		     - 0.234375*v[pd[5]] + 0.0625*v[pd[7]]
		     + 0.3125*(v[pd[13]] - v[pd[31]])
		     + 0.9375*v[pd[32]]);
	v[cd[29]] = (-0.0390625*v[pd[0]] + 0.0234375*v[pd[1]]
		     + 0.09375*v[pd[4]] - 0.046875*v[pd[5]]
		     - 0.03125*v[pd[6]]
		     + 0.125*(v[pd[7]] - v[pd[8]] - v[pd[13]])
		     + 0.375*(v[pd[14]] - v[pd[31]] + v[pd[32]])
		     + 0.75*v[pd[33]]);
	v[cd[30]] = v[pd[32]];
	v[cd[34]] = (-0.0390625*v[pd[0]] + 0.0234375*v[pd[1]]
		     + 0.09375*v[pd[4]] - 0.046875*v[pd[5]]
		     - 0.03125*v[pd[6]]
		     + 0.0625*(v[pd[7]] + v[pd[10]]
			       - v[pd[13]] - v[pd[16]])
		     + 0.375*v[pd[22]] - 0.125*v[pd[25]]
		     + 0.1875*(-v[pd[28]] + v[pd[29]]
			       - v[pd[31]] + v[pd[32]])
		     + 0.75*v[pd[34]]);
	break;
      case 2:
	v[cd[25]] = (0.0390625*(-v[pd[0]] - v[pd[1]])
		     + 0.15625*(v[pd[4]] + v[pd[6]])
		     - 0.234375*v[pd[5]] + 0.0625*v[pd[10]]
		     + 0.3125*(v[pd[16]] - v[pd[28]])
		     + 0.9375*v[pd[29]]);
	v[cd[26]] = (-0.0390625*v[pd[0]] + 0.0234375*v[pd[1]]
		     + 0.09375*v[pd[4]] - 0.046875*v[pd[5]]
		     - 0.03125*v[pd[6]]
		     + 0.125*(v[pd[10]] - v[pd[11]] - v[pd[16]])
		     + 0.375*(v[pd[17]] - v[pd[28]] + v[pd[29]])
		     + 0.75*v[pd[30]]);
	v[cd[27]] = v[pd[29]];
	v[cd[34]] = (-0.0390625*v[pd[0]] + 0.0234375*v[pd[1]]
		     + 0.09375*v[pd[4]] - 0.046875*v[pd[5]]
		     - 0.03125*v[pd[6]]
		     + 0.0625*(v[pd[7]] + v[pd[10]]
			       - v[pd[13]] - v[pd[16]])
		     + 0.375*v[pd[22]] - 0.125*v[pd[25]]
		     + 0.1875*(-v[pd[28]] + v[pd[29]]
			       - v[pd[31]] + v[pd[32]])
		     + 0.75*v[pd[34]]);
	break;
      case 3:
	v[cd[34]] = (-0.0390625*v[pd[0]] + 0.0234375*v[pd[1]]
		     + 0.09375*v[pd[4]] - 0.046875*v[pd[5]]
		     - 0.03125*v[pd[6]]
		     + 0.0625*(v[pd[7]] + v[pd[10]]
			       - v[pd[13]] - v[pd[16]])
		     + 0.375*v[pd[22]] - 0.125*v[pd[25]]
		     + 0.1875*(-v[pd[28]] + v[pd[29]]
			       - v[pd[31]] + v[pd[32]])
		     + 0.75*v[pd[34]]);
	break;
      }
    }
  }

  return;
}

static void real_coarse_inter4(DOF_REAL_VEC *drv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_coarse_inter4");
  DOF       pd[N_BAS4];
  const DOF *cd;
  EL        *el;
  REAL      *v = nil;
  int       i, typ, lr_set;
  const DOF       *(*get_dof_indices)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF_ADMIN *admin;

  if (n < 1) return;
  el = list->el;
  typ = list->el_type;

  GET_DOF_VEC(v, drv);
  if (!drv->fe_space)
  {
    ERROR("no fe_space in dof_real_vec %s\n", NAME(drv));
    return;
  }
  else if (!drv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drv->fe_space));
    return;
  }
  get_dof_indices = drv->fe_space->bas_fcts->get_dof_indices;
  GET_STRUCT(admin,drv->fe_space);

  get_dof_indices(el, admin, pd);

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

  cd = get_dof_indices(el->child[0], admin, nil);

  v[pd[4]] = v[cd[11]];
  v[pd[5]] = v[cd[3]];
  v[pd[28]] = v[cd[27]];
  v[pd[30]] = v[cd[20]];
  v[pd[31]] = v[cd[30]];
  v[pd[33]] = v[cd[17]];
  v[pd[34]] = v[cd[24]];

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

  cd = get_dof_indices(el->child[1], admin, nil);

  if (typ == 0)
  {
/*--------------------------------------------------------------------------*/
/*  parent of el_type 0                                                     */
/*--------------------------------------------------------------------------*/

    v[pd[6]] = v[cd[11]];
    v[pd[29]] = v[cd[30]];
    v[pd[32]] = v[cd[27]];
  }
  else
  {
/*--------------------------------------------------------------------------*/
/*  parent of el_type 1|2                                                   */
/*--------------------------------------------------------------------------*/

    v[pd[6]] = v[cd[11]];
    v[pd[29]] = v[cd[27]];
    v[pd[32]] = v[cd[30]];
  }

/*--------------------------------------------------------------------------*/
/*   adjust neighbour values                                                */
/*--------------------------------------------------------------------------*/
  
  for (i = 1; i < n; i++)
  {
    el = list[i].el;
    typ = list[i].el_type;
    get_dof_indices(el, admin, pd);

    lr_set = 0;
    if (list[i].neigh[0]  &&  list[i].neigh[0]->no < i)
      lr_set = 1;

    if (list[i].neigh[1]  &&  list[i].neigh[1]->no < i)
      lr_set += 2;

    TEST_EXIT(lr_set)("no values set on both neighbours\n");

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

    cd = get_dof_indices(el->child[0], admin, nil);

    switch(lr_set)
    {
    case 1:
      v[pd[31]] = v[cd[30]];
      v[pd[33]] = v[cd[17]];
      v[pd[34]] = v[cd[24]];
      break;
    case 2:
      v[pd[28]] = v[cd[27]];
      v[pd[30]] = v[cd[20]];
      v[pd[34]] = v[cd[24]];
      break;
    case 3:
      v[pd[34]] = v[cd[24]];
      break;
    }

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

    cd = get_dof_indices(el->child[1], admin, nil);

    if (typ == 0)
    {
      switch(lr_set)
      {
      case 1:
	v[pd[32]] = v[cd[27]];
	break;
      case 2:
	v[pd[29]] = v[cd[30]];
	break;
      }
    }
    else
    {
      switch(lr_set)
      {
      case 1:
	v[pd[32]] = v[cd[30]];
	break;
      case 2:
	v[pd[29]] = v[cd[27]];
	break;
      }
    }
  }

  return;
}



static void real_coarse_restr4(DOF_REAL_VEC *drv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_coarse_restr4");
  DOF       pd[N_BAS4];
  const DOF *cd;
  EL        *el;
  REAL      *v = nil;
  int       i, typ, lr_set;
  const DOF       *(*get_dof_indices)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF_ADMIN *admin;

  if (n < 1) return;
  el = list->el;
  typ = list->el_type;

  GET_DOF_VEC(v, drv);
  GET_DOF_VEC(v, drv);
  if (!drv->fe_space)
  {
    ERROR("no fe_space in dof_real_vec %s\n", NAME(drv));
    return;
  }
  else if (!drv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drv->fe_space));
    return;
  }
  get_dof_indices = drv->fe_space->bas_fcts->get_dof_indices;
  GET_STRUCT(admin,drv->fe_space);

  get_dof_indices(el, admin, pd);

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

  cd = get_dof_indices(el->child[0], admin, nil);

  v[pd[0]] += (0.2734375*v[cd[10]]
	       + 0.0390625*(-v[cd[12]] - v[cd[16]] - v[cd[19]]
			    - v[cd[22]] - v[cd[23]] - v[cd[25]]
			    - v[cd[28]])
	       + 0.0234375*(v[cd[18]] + v[cd[21]] + v[cd[26]]
			    + v[cd[29]] + v[cd[34]]));
  v[pd[1]] += (0.0390625*(-v[cd[10]] - v[cd[16]] - v[cd[19]]
			  - v[cd[22]] - v[cd[23]] - v[cd[25]]
			  - v[cd[26]] - v[cd[28]] - v[cd[29]]
			  - v[cd[34]])
	       + 0.0234375*(v[cd[12]] + v[cd[18]] + v[cd[21]]));
  v[pd[4]] = (v[cd[11]] + 1.09375*v[cd[10]] + 0.46875*v[cd[12]]
	      + 0.03125*(v[cd[16]] + v[cd[19]] + v[cd[22]]
			 + v[cd[23]] - v[cd[26]] - v[cd[29]]
			 - v[cd[34]])
	      + 0.09375*(-v[cd[18]] - v[cd[21]])
	      + 0.15625*(v[cd[25]] + v[cd[28]]));
  v[pd[5]] = (v[cd[3]] - 0.546875*v[cd[10]]
	      + 0.703125*v[cd[12]]
	      + 0.015625*(v[cd[16]] + v[cd[19]]
			  + v[cd[22]] + v[cd[23]])
	      + 0.140625*(v[cd[18]] + v[cd[21]])
	      + 0.234375*(-v[cd[25]] - v[cd[28]])
	      + 0.046875*(-v[cd[26]] - v[cd[29]] - v[cd[34]]));
  v[pd[6]] = (0.21875*v[cd[10]]
	      + 0.15625*(-v[cd[12]] + v[cd[25]] + v[cd[28]])
	      + 0.03125*(v[cd[16]] + v[cd[19]]
			 + v[cd[22]] + v[cd[23]])
	      + 0.09375*(-v[cd[18]] - v[cd[21]] + v[cd[26]]
			 + v[cd[29]] + v[cd[34]]));
  v[pd[7]] += (0.1875*v[cd[16]]
	       + 0.0625*(-v[cd[18]] + v[cd[23]] - v[cd[34]])
	       + 0.125*(v[cd[22]] - v[cd[29]]) + 0.3125*v[cd[28]]);
  v[pd[8]] += (0.375*(-v[cd[16]] + v[cd[29]]) - 0.125*v[cd[22]]);
  v[pd[9]] += 0.5*v[cd[16]];
  v[pd[10]] += (0.1875*v[cd[19]]
		+ 0.0625*(-v[cd[21]] + v[cd[22]] - v[cd[34]])
		+ 0.125*(v[cd[23]] - v[cd[26]])
		+ 0.3125*v[cd[25]]);
  v[pd[11]] += (0.375*(-v[cd[19]] + v[cd[26]])
		- 0.125*v[cd[23]]);
  v[pd[12]] += 0.5*v[cd[19]];
  v[pd[13]] += (0.1875*v[cd[16]]
		+ 0.0625*(-v[cd[18]] + v[cd[23]]
			  + v[cd[28]] + v[cd[34]])
		+ 0.125*(v[cd[22]] + v[cd[29]]));
  v[pd[14]] += (-0.375*v[cd[16]] + 0.125*(-v[cd[22]] - v[cd[29]]));
  v[pd[15]] += 0.5*v[cd[16]];
  v[pd[16]] += (0.1875*v[cd[19]]
		+ 0.0625*(-v[cd[21]] + v[cd[22]]
			  + v[cd[25]] + v[cd[34]])
		+ 0.125*(v[cd[23]] + v[cd[26]]));
  v[pd[17]] += (-0.375*v[cd[19]] + 0.125*(-v[cd[23]] - v[cd[26]]));
  v[pd[18]] += 0.5*v[cd[19]];
  v[pd[22]] += (0.25*(-v[cd[22]] - v[cd[23]]) - 0.125*v[cd[34]]);
  v[pd[23]] += 0.5*v[cd[22]];
  v[pd[24]] += 0.5*v[cd[23]];
  v[pd[25]] += (0.25*(-v[cd[22]] - v[cd[23]]) + 0.375*v[cd[34]]);
  v[pd[26]] += 0.5*v[cd[22]];
  v[pd[27]] += 0.5*v[cd[23]];
  v[pd[28]] = (v[cd[27]] + 0.1875*(-v[cd[19]] + v[cd[34]]) 
	       + 0.5625*v[cd[21]] - 0.0625*v[cd[22]]
	       - 0.125*v[cd[23]] + 0.9375*v[cd[25]]
	       + 0.375*v[cd[26]]);
  v[pd[29]] = (0.1875*(-v[cd[19]] - v[cd[34]])
	       + 0.5625*v[cd[21]] - 0.0625*v[cd[22]]
	       - 0.125*v[cd[23]] - 0.3125*v[cd[25]]
	       - 0.375*v[cd[26]]);
  v[pd[30]] = (v[cd[20]] + 0.75*(v[cd[19]] + v[cd[26]])
	       + 0.25*v[cd[23]]);
  v[pd[31]] = (v[cd[30]] + 0.1875*(-v[cd[16]] + v[cd[34]])
	       + 0.5625*v[cd[18]] - 0.125*v[cd[22]]
	       - 0.0625*v[cd[23]] + 0.9375*v[cd[28]]
	       + 0.375*v[cd[29]]);
  v[pd[32]] = (0.1875*(-v[cd[16]] - v[cd[34]])
	       + 0.5625*v[cd[18]] - 0.125*v[cd[22]]
	       - 0.0625*v[cd[23]] - 0.3125*v[cd[28]]
	       - 0.375*v[cd[29]]);
  v[pd[33]] = (v[cd[17]] + 0.75*(v[cd[16]] + v[cd[29]])
	       + 0.25*v[cd[22]]);
  v[pd[34]] = (v[cd[24]] + 0.5*(v[cd[22]] + v[cd[23]])
	       + 0.75*v[cd[34]]);

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

  cd = get_dof_indices(el->child[1], admin, nil);

  if (typ == 0)
  {
/*--------------------------------------------------------------------------*/
/*  parent of el_type 0                                                     */
/*--------------------------------------------------------------------------*/

    v[pd[0]] += (0.0390625*(-v[cd[10]] - v[cd[25]] - v[cd[26]]
			    - v[cd[28]] - v[cd[29]] - v[cd[34]])
		 + 0.0234375*v[cd[12]]);
    v[pd[1]] += (0.2734375*v[cd[10]]
		 + 0.0390625*(-v[cd[12]] - v[cd[25]] - v[cd[28]])
		 + 0.0234375*(v[cd[26]] + v[cd[29]] + v[cd[34]]));
    v[pd[4]] += (0.21875*v[cd[10]]
		 + 0.15625*(-v[cd[12]] + v[cd[25]] + v[cd[28]])
		 + 0.09375*(v[cd[26]] + v[cd[29]] + v[cd[34]]));
    v[pd[5]] += (-0.546875*v[cd[10]] + 0.703125*v[cd[12]]
		 + 0.234375*(-v[cd[25]] - v[cd[28]])
		 + 0.046875*(-v[cd[26]] - v[cd[29]] - v[cd[34]]));
    v[pd[6]] += (v[cd[11]] + 1.09375*v[cd[10]] + 0.46875*v[cd[12]]
		 + 0.15625*(v[cd[25]] + v[cd[28]])
		 + 0.03125*(-v[cd[26]] - v[cd[29]] - v[cd[34]]));
    v[pd[7]] += (0.0625*(v[cd[25]] + v[cd[34]]) + 0.125*v[cd[26]]);
    v[pd[8]] += -0.125*v[cd[26]];
    v[pd[10]] += (0.0625*(v[cd[28]] + v[cd[34]]) + 0.125*v[cd[29]]);
    v[pd[11]] += -0.125*v[cd[29]];
    v[pd[13]] += (0.3125*v[cd[25]] - 0.125*v[cd[26]]
		  - 0.0625*v[cd[34]]);
    v[pd[14]] += 0.375*v[cd[26]];
    v[pd[16]] += (0.3125*v[cd[28]] - 0.125*v[cd[29]]
		  - 0.0625*v[cd[34]]);
    v[pd[17]] += 0.375*v[cd[29]];
    v[pd[22]] += 0.375*v[cd[34]];
    v[pd[25]] += -0.125*v[cd[34]];
    v[pd[28]] += (-0.3125*v[cd[28]] - 0.375*v[cd[29]]
		  - 0.1875*v[cd[34]]);
    v[pd[29]] += (v[cd[30]] + 0.9375*v[cd[28]] + 0.375*v[cd[29]]
		  + 0.1875*v[cd[34]]);
    v[pd[30]] += 0.75*v[cd[29]];
    v[pd[31]] += (-0.3125*v[cd[25]] - 0.375*v[cd[26]]
		  - 0.1875*v[cd[34]]);
    v[pd[32]] += (v[cd[27]] + 0.9375*v[cd[25]]
		  + 0.375*v[cd[26]] + 0.1875*v[cd[34]]);
    v[pd[33]] += 0.75*v[cd[26]];
    v[pd[34]] += 0.75*v[cd[34]];
  }
  else
  {
/*--------------------------------------------------------------------------*/
/*  parent of el_type 1|2                                                   */
/*--------------------------------------------------------------------------*/

    v[pd[0]] += (0.0390625*(-v[cd[10]] - v[cd[25]] - v[cd[26]]
			    - v[cd[28]] - v[cd[29]] - v[cd[34]])
		 + 0.0234375*v[cd[12]]);
    v[pd[1]] += (0.2734375*v[cd[10]]
		 + 0.0390625*(-v[cd[12]] - v[cd[25]] - v[cd[28]])
		 + 0.0234375*(v[cd[26]] + v[cd[29]] + v[cd[34]]));
    v[pd[4]] += (0.21875*v[cd[10]]
		 + 0.15625*(-v[cd[12]] + v[cd[25]] + v[cd[28]])
		 + 0.09375*(v[cd[26]] + v[cd[29]] + v[cd[34]]));
    v[pd[5]] += (-0.546875*v[cd[10]] + 0.703125*v[cd[12]]
		 + 0.234375*(-v[cd[25]] - v[cd[28]])
		 + 0.046875*(-v[cd[26]] - v[cd[29]] - v[cd[34]]));
    v[pd[6]] += (v[cd[11]] + 1.09375*v[cd[10]] + 0.46875*v[cd[12]]
		 + 0.15625*(v[cd[25]] + v[cd[28]])
		 + 0.03125*(-v[cd[26]] - v[cd[29]] - v[cd[34]]));
    v[pd[7]] += (0.0625*(v[cd[28]] + v[cd[34]]) + 0.125*v[cd[29]]);
    v[pd[8]] += -0.125*v[cd[29]];
    v[pd[10]] += (0.0625*(v[cd[25]] + v[cd[34]]) + 0.125*v[cd[26]]);
    v[pd[11]] += -0.125*v[cd[26]];
    v[pd[13]] += (0.3125*v[cd[28]] - 0.125*v[cd[29]]
		  - 0.0625*v[cd[34]]);
    v[pd[14]] += 0.375*v[cd[29]];
    v[pd[16]] += (0.3125*v[cd[25]] - 0.125*v[cd[26]]
		  - 0.0625*v[cd[34]]);
    v[pd[17]] += 0.375*v[cd[26]];
    v[pd[22]] += 0.375*v[cd[34]];
    v[pd[25]] += -0.125*v[cd[34]];
    v[pd[28]] += (-0.3125*v[cd[25]] - 0.375*v[cd[26]]
		  - 0.1875*v[cd[34]]);
    v[pd[29]] += (v[cd[27]] + 0.9375*v[cd[25]] + 0.375*v[cd[26]]
		  + 0.1875*v[cd[34]]);
    v[pd[30]] += 0.75*v[cd[26]];
    v[pd[31]] += (-0.3125*v[cd[28]] - 0.375*v[cd[29]]
		  - 0.1875*v[cd[34]]);
    v[pd[32]] += (v[cd[30]] + 0.9375*v[cd[28]]
		  + 0.375*v[cd[29]] + 0.1875*v[cd[34]]);
    v[pd[33]] += 0.75*v[cd[29]];
    v[pd[34]] += 0.75*v[cd[34]];
  }

/*--------------------------------------------------------------------------*/
/*   adjust neighbour values                                                */
/*--------------------------------------------------------------------------*/
  
  for (i = 1; i < n; i++)
  {
    el = list[i].el;
    typ = list[i].el_type;
    get_dof_indices(el, admin, pd);

    lr_set = 0;
    if (list[i].neigh[0]  &&  list[i].neigh[0]->no < i)
      lr_set = 1;

    if (list[i].neigh[1]  &&  list[i].neigh[1]->no < i)
      lr_set += 2;

    TEST_EXIT(lr_set)("no values set on both neighbours\n");

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

    cd = get_dof_indices(el->child[0], admin, nil);

    switch(lr_set)
    {
    case 1:
      v[pd[0]] += (0.0390625*(-v[cd[16]] - v[cd[22]]
			      - v[cd[23]] - v[cd[28]])
		   + 0.0234375*(v[cd[18]] + v[cd[29]] + v[cd[34]]));
      v[pd[1]] += (0.0390625*(-v[cd[16]] - v[cd[22]] - v[cd[23]]
			      - v[cd[28]] - v[cd[29]] - v[cd[34]])
		   + 0.0234375*v[cd[18]]);
      v[pd[4]] += (0.03125*(v[cd[16]] + v[cd[22]] + v[cd[23]]
			    - v[cd[29]] - v[cd[34]])
		   - 0.09375*v[cd[18]] + 0.15625*v[cd[28]]);
      v[pd[5]] += (0.015625*(v[cd[16]] + v[cd[22]] + v[cd[23]])
		   + 0.140625*v[cd[18]] - 0.234375*v[cd[28]]
		   + 0.046875*(-v[cd[29]] - v[cd[34]]));
      v[pd[6]] += (0.03125*(v[cd[16]] + v[cd[22]] + v[cd[23]])
		   + 0.09375*(-v[cd[18]] + v[cd[29]] + v[cd[34]])
		   + 0.15625*v[cd[28]]);
      v[pd[7]] += (0.1875*v[cd[16]]
		   + 0.0625*(-v[cd[18]] + v[cd[23]] - v[cd[34]])
		   + 0.125*(v[cd[22]] - v[cd[29]])
		   + 0.3125*v[cd[28]]);
      v[pd[8]] += (0.375*(-v[cd[16]] + v[cd[29]])
		   - 0.125*v[cd[22]]);
      v[pd[9]] += 0.5*v[cd[16]];
      v[pd[10]] += (0.0625*(v[cd[22]] - v[cd[34]]) + 0.125*v[cd[23]]);
      v[pd[11]] += -0.125*v[cd[23]];
      v[pd[13]] += (0.1875*v[cd[16]]
		    + 0.0625*(-v[cd[18]] + v[cd[23]]
			      + v[cd[28]] + v[cd[34]])
		    + 0.125*(v[cd[22]] + v[cd[29]]));
      v[pd[14]] += (-0.375*v[cd[16]]
		    + 0.125*(-v[cd[22]] - v[cd[29]]));
      v[pd[15]] += 0.5*v[cd[16]];
      v[pd[16]] += (0.0625*(v[cd[22]] + v[cd[34]]) + 0.125*v[cd[23]]);
      v[pd[17]] += -0.125*v[cd[23]];
      v[pd[22]] += (0.25*(-v[cd[22]] - v[cd[23]]) - 0.125*v[cd[34]]);
      v[pd[23]] += 0.5*v[cd[22]];
      v[pd[24]] += 0.5*v[cd[23]];
      v[pd[25]] += (0.25*(-v[cd[22]] - v[cd[23]]) + 0.375*v[cd[34]]);
      v[pd[26]] += 0.5*v[cd[22]];
      v[pd[27]] += 0.5*v[cd[23]];
      v[pd[28]] += (-0.0625*v[cd[22]] - 0.125*v[cd[23]]
		    + 0.1875*v[cd[34]]);
      v[pd[29]] += (-0.0625*v[cd[22]] - 0.125*v[cd[23]]
		    - 0.1875*v[cd[34]]);
      v[pd[30]] += 0.25*v[cd[23]];
      v[pd[31]] = (v[cd[30]] + 0.1875*(-v[cd[16]] + v[cd[34]])
		   + 0.5625*v[cd[18]] - 0.125*v[cd[22]]
		   - 0.0625*v[cd[23]] + 0.9375*v[cd[28]]
		   + 0.375*v[cd[29]]);
      v[pd[32]] = (0.1875*(-v[cd[16]] - v[cd[34]])
		   + 0.5625*v[cd[18]] - 0.125*v[cd[22]]
		   - 0.0625*v[cd[23]] - 0.3125*v[cd[28]]
		   - 0.375*v[cd[29]]);
      v[pd[33]] = (v[cd[17]] + 0.75*(v[cd[16]] + v[cd[29]])
		   + 0.25*v[cd[22]]);
      v[pd[34]] = (v[cd[24]] + 0.5*(v[cd[22]] + v[cd[23]])
		   + 0.75*v[cd[34]]);
      break;
    case 2:
      v[pd[0]] += (0.0390625*(-v[cd[19]] - v[cd[22]]
			      - v[cd[23]] - v[cd[25]])
		   + 0.0234375*(v[cd[21]] + v[cd[26]] + v[cd[34]]));
      v[pd[1]] += (0.0390625*(-v[cd[19]] - v[cd[22]] - v[cd[23]]
			      - v[cd[25]] - v[cd[26]] - v[cd[34]])
		   + 0.0234375*v[cd[21]]);
      v[pd[4]] += (0.03125*(v[cd[19]] + v[cd[22]] + v[cd[23]]
			    - v[cd[26]] - v[cd[34]])
		   - 0.09375*v[cd[21]] + 0.15625*v[cd[25]]);
      v[pd[5]] += (0.015625*(v[cd[19]] + v[cd[22]] + v[cd[23]])
		   + 0.140625*v[cd[21]] - 0.234375*v[cd[25]]
		   + 0.046875*(-v[cd[26]] - v[cd[34]]));
      v[pd[6]] += (0.03125*(v[cd[19]] + v[cd[22]] + v[cd[23]])
		   + 0.09375*(-v[cd[21]] + v[cd[26]] + v[cd[34]])
		   + 0.15625*v[cd[25]]);
      v[pd[7]] += (0.125*v[cd[22]] + 0.0625*(v[cd[23]] - v[cd[34]]));
      v[pd[8]] += -0.125*v[cd[22]];
      v[pd[10]] += (0.1875*v[cd[19]]
		    + 0.0625*(-v[cd[21]] + v[cd[22]] - v[cd[34]])
		    + 0.125*(v[cd[23]] - v[cd[26]])
		    + 0.3125*v[cd[25]]);
      v[pd[11]] += (0.375*(-v[cd[19]] + v[cd[26]]) - 0.125*v[cd[23]]);
      v[pd[12]] += 0.5*v[cd[19]];
      v[pd[13]] += (0.125*v[cd[22]] + 0.0625*(v[cd[23]] + v[cd[34]]));
      v[pd[14]] += -0.125*v[cd[22]];
      v[pd[16]] += (0.1875*v[cd[19]]
		    + 0.0625*(-v[cd[21]] + v[cd[22]]
			      + v[cd[25]] + v[cd[34]])
		    + 0.125*(v[cd[23]] + v[cd[26]]));
      v[pd[17]] += (-0.375*v[cd[19]] 
		    + 0.125*(-v[cd[23]] - v[cd[26]]));
      v[pd[18]] += 0.5*v[cd[19]];
      v[pd[22]] += (0.25*(-v[cd[22]] - v[cd[23]]) - 0.125*v[cd[34]]);
      v[pd[23]] += (0.5*v[cd[22]]);
      v[pd[24]] += (0.5*v[cd[23]]);
      v[pd[25]] += (0.25*(-v[cd[22]] - v[cd[23]]) + 0.375*v[cd[34]]);
      v[pd[26]] += (0.5*v[cd[22]]);
      v[pd[27]] += (0.5*v[cd[23]]);
      v[pd[28]] = (v[cd[27]] + 0.1875*(-v[cd[19]] + v[cd[34]])
		   + 0.5625*v[cd[21]] - 0.0625*v[cd[22]]
		   - 0.125*v[cd[23]] + 0.9375*v[cd[25]]
		   + 0.375*v[cd[26]]);
      v[pd[29]] = (0.1875*(-v[cd[19]] - v[cd[34]])
		   + 0.5625*v[cd[21]] - 0.0625*v[cd[22]]
		   - 0.125*v[cd[23]] - 0.3125*v[cd[25]]
		   - 0.375*v[cd[26]]);
      v[pd[30]] = (v[cd[20]] + 0.75*(v[cd[19]] + v[cd[26]])
		   + 0.25*v[cd[23]]);
      v[pd[31]] += (-0.125*v[cd[22]] - 0.0625*v[cd[23]]
		    + 0.1875*v[cd[34]]);
      v[pd[32]] += (-0.125*v[cd[22]] - 0.0625*v[cd[23]]
		    - 0.1875*v[cd[34]]);
      v[pd[33]] += 0.25*v[cd[22]];
      v[pd[34]] = (v[cd[24]] + 0.5*(v[cd[22]] + v[cd[23]])
		   + 0.75*v[cd[34]]);
      break;
    case 3:
      v[pd[0]] += (0.0390625*(-v[cd[22]] - v[cd[23]])
		   + 0.0234375*v[cd[34]]);
      v[pd[1]] += (0.0390625*(-v[cd[22]] - v[cd[23]] - v[cd[34]]));
      v[pd[4]] += (0.03125*(v[cd[22]] + v[cd[23]] - v[cd[34]]));
      v[pd[5]] += (0.015625*(v[cd[22]] + v[cd[23]])
		   - 0.046875*v[cd[34]]);
      v[pd[6]] += (0.03125*(v[cd[22]] + v[cd[23]])
		   + 0.09375*v[cd[34]]);
      v[pd[7]] += (0.125*v[cd[22]] + 0.0625*(v[cd[23]] - v[cd[34]]));
      v[pd[8]] += -0.125*v[cd[22]];
      v[pd[10]] += (0.0625*(v[cd[22]] - v[cd[34]]) + 0.125*v[cd[23]]);
      v[pd[11]] += -0.125*v[cd[23]];
      v[pd[13]] += (0.125*v[cd[22]] + 0.0625*(v[cd[23]] + v[cd[34]]));
      v[pd[14]] += -0.125*v[cd[22]];
      v[pd[16]] += (0.0625*(v[cd[22]] + v[cd[34]]) + 0.125*v[cd[23]]);
      v[pd[17]] += -0.125*v[cd[23]];
      v[pd[22]] += (0.25*(-v[cd[22]] - v[cd[23]]) - 0.125*v[cd[34]]);
      v[pd[23]] += 0.5*v[cd[22]];
      v[pd[24]] += 0.5*v[cd[23]];
      v[pd[25]] += (0.25*(-v[cd[22]] - v[cd[23]]) + 0.375*v[cd[34]]);
      v[pd[26]] += 0.5*v[cd[22]];
      v[pd[27]] += 0.5*v[cd[23]];
      v[pd[28]] += (-0.0625*v[cd[22]] - 0.125*v[cd[23]]
		    + 0.1875*v[cd[34]]);
      v[pd[29]] += (-0.0625*v[cd[22]] - 0.125*v[cd[23]]
		    - 0.1875*v[cd[34]]);
      v[pd[30]] += 0.25*v[cd[23]];
      v[pd[31]] += (-0.125*v[cd[22]] - 0.0625*v[cd[23]]
		    + 0.1875*v[cd[34]]);
      v[pd[32]] += (-0.125*v[cd[22]] - 0.0625*v[cd[23]]
		    - 0.1875*v[cd[34]]);
      v[pd[33]] += 0.25*v[cd[22]];
      v[pd[34]] = (v[cd[24]] + 0.5*(v[cd[22]] + v[cd[23]])
		   + 0.75*v[cd[34]]);
      break;
    }

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

    cd = get_dof_indices(el->child[1], admin, nil);

    if (typ == 0)
    {
      switch(lr_set)
      {
      case 1:
	v[pd[0]] += (0.0390625*(-v[cd[25]] - v[cd[26]] - v[cd[34]]));
	v[pd[1]] += (-0.0390625*v[cd[25]]
		     + 0.0234375*(v[cd[26]] + v[cd[34]]));
	v[pd[4]] += (0.15625*v[cd[25]]
		     + 0.09375*(v[cd[26]] + v[cd[34]]));
	v[pd[5]] += (-0.234375*v[cd[25]]
		     + 0.046875*(-v[cd[26]] - v[cd[34]]));
	v[pd[6]] += (0.15625*v[cd[25]]
		     + 0.03125*(-v[cd[26]] - v[cd[34]]));
	v[pd[7]] += (0.0625*(v[cd[25]] + v[cd[34]])
		     + 0.125*v[cd[26]]);
	v[pd[8]] += -0.125*v[cd[26]];
	v[pd[10]] += 0.0625*v[cd[34]];
	v[pd[13]] += (0.3125*v[cd[25]] - 0.125*v[cd[26]]
		      - 0.0625*v[cd[34]]);
	v[pd[14]] += 0.375*v[cd[26]];
	v[pd[16]] += -0.0625*v[cd[34]];
	v[pd[22]] += 0.375*v[cd[34]];
	v[pd[25]] += -0.125*v[cd[34]];
	v[pd[28]] += -0.1875*v[cd[34]];
	v[pd[29]] += 0.1875*v[cd[34]];
	v[pd[31]] += (-0.3125*v[cd[25]] - 0.375*v[cd[26]]
		      - 0.1875*v[cd[34]]);
	v[pd[32]] += (v[cd[27]] + 0.9375*v[cd[25]]
		      + 0.375*v[cd[26]] + 0.1875*v[cd[34]]);
	v[pd[33]] += 0.75*v[cd[26]];
	v[pd[34]] += 0.75*v[cd[34]];
	break;
      case 2:
	v[pd[0]] += (0.0390625*(-v[cd[28]] - v[cd[29]] - v[cd[34]]));
	v[pd[1]] += (-0.0390625*v[cd[28]]
		     + 0.0234375*(v[cd[29]] + v[cd[34]]));
	v[pd[4]] += (0.15625*v[cd[28]]
		     + 0.09375*(v[cd[29]] + v[cd[34]]));
	v[pd[5]] += (-0.234375*v[cd[28]]
		     + 0.046875*(-v[cd[29]] - v[cd[34]]));
	v[pd[6]] += (0.15625*v[cd[28]]
		     + 0.03125*(-v[cd[29]] - v[cd[34]]));
	v[pd[7]] += 0.0625*v[cd[34]];
	v[pd[10]] += (0.0625*(v[cd[28]] + v[cd[34]])
		      + 0.125*v[cd[29]]);
	v[pd[11]] += -0.125*v[cd[29]];
	v[pd[13]] += -0.0625*v[cd[34]];
	v[pd[16]] += (0.3125*v[cd[28]] - 0.125*v[cd[29]]
		      - 0.0625*v[cd[34]]);
	v[pd[17]] += 0.375*v[cd[29]];
	v[pd[22]] += 0.375*v[cd[34]];
	v[pd[25]] += -0.125*v[cd[34]];
	v[pd[28]] += (-0.3125*v[cd[28]] - 0.375*v[cd[29]]
		      - 0.1875*v[cd[34]]);
	v[pd[29]] += (v[cd[30]] + 0.9375*v[cd[28]] + 0.375*v[cd[29]]
		      + 0.1875*v[cd[34]]);
	v[pd[30]] += 0.75*v[cd[29]];
	v[pd[31]] += -0.1875*v[cd[34]];
	v[pd[32]] += 0.1875*v[cd[34]];
	v[pd[34]] += 0.75*v[cd[34]];
	break;
      case 3:
	v[pd[0]] += -0.0390625*v[cd[34]];
	v[pd[1]] += 0.0234375*v[cd[34]];
	v[pd[4]] += 0.09375*v[cd[34]];
	v[pd[5]] += -0.046875*v[cd[34]];
	v[pd[6]] += -0.03125*v[cd[34]];
	v[pd[7]] += 0.0625*v[cd[34]];
	v[pd[10]] += 0.0625*v[cd[34]];
	v[pd[13]] += -0.0625*v[cd[34]];
	v[pd[16]] += -0.0625*v[cd[34]];
	v[pd[22]] += 0.375*v[cd[34]];
	v[pd[25]] += -0.125*v[cd[34]];
	v[pd[28]] += -0.1875*v[cd[34]];
	v[pd[29]] += 0.1875*v[cd[34]];
	v[pd[31]] += -0.1875*v[cd[34]];
	v[pd[32]] += 0.1875*v[cd[34]];
	v[pd[34]] += 0.75*v[cd[34]];
	break;
      }
    }
    else
    {
      switch(lr_set)
      {
      case 1:
	v[pd[0]] += (0.0390625*(-v[cd[28]] - v[cd[29]] - v[cd[34]]));
	v[pd[1]] += (-0.0390625*v[cd[28]]
		     + 0.0234375*(v[cd[29]] + v[cd[34]]));
	v[pd[4]] += (0.15625*v[cd[28]]
		     + 0.09375*(v[cd[29]] + v[cd[34]]));
	v[pd[5]] += (-0.234375*v[cd[28]]
		     + 0.046875*(-v[cd[29]] - v[cd[34]]));
	v[pd[6]] += (0.15625*v[cd[28]]
		     + 0.03125*(-v[cd[29]] - v[cd[34]]));
	v[pd[7]] += (0.0625*(v[cd[28]] + v[cd[34]])
		     + 0.125*v[cd[29]]);
	v[pd[8]] += -0.125*v[cd[29]];
	v[pd[10]] += 0.0625*v[cd[34]];
	v[pd[13]] += (0.3125*v[cd[28]] - 0.125*v[cd[29]]
		      - 0.0625*v[cd[34]]);
	v[pd[14]] += 0.375*v[cd[29]];
	v[pd[16]] += -0.0625*v[cd[34]];
	v[pd[22]] += 0.375*v[cd[34]];
	v[pd[25]] += -0.125*v[cd[34]];
	v[pd[28]] += -0.1875*v[cd[34]];
	v[pd[29]] += 0.1875*v[cd[34]];
	v[pd[31]] += (-0.3125*v[cd[28]] - 0.375*v[cd[29]]
		      - 0.1875*v[cd[34]]);
	v[pd[32]] += (v[cd[30]] + 0.9375*v[cd[28]] + 0.375*v[cd[29]]
		      + 0.1875*v[cd[34]]);
	v[pd[33]] += 0.75*v[cd[29]];
	v[pd[34]] += 0.75*v[cd[34]];
	break;
      case 2:
	v[pd[0]] += (0.0390625*(-v[cd[25]] - v[cd[26]] - v[cd[34]]));
	v[pd[1]] += (-0.0390625*v[cd[25]]
		     + 0.0234375*(v[cd[26]] + v[cd[34]]));
	v[pd[4]] += (0.15625*v[cd[25]]
		     + 0.09375*(v[cd[26]] + v[cd[34]]));
	v[pd[5]] += (-0.234375*v[cd[25]]
		     + 0.046875*(-v[cd[26]] - v[cd[34]]));
	v[pd[6]] += (0.15625*v[cd[25]]
		     + 0.03125*(-v[cd[26]] - v[cd[34]]));
	v[pd[7]] += 0.0625*v[cd[34]];
	v[pd[10]] += (0.0625*(v[cd[25]] + v[cd[34]])
		      + 0.125*v[cd[26]]);
	v[pd[11]] += -0.125*v[cd[26]];
	v[pd[13]] += -0.0625*v[cd[34]];
	v[pd[16]] += (0.3125*v[cd[25]] - 0.125*v[cd[26]]
		      - 0.0625*v[cd[34]]);
	v[pd[17]] += 0.375*v[cd[26]];
	v[pd[22]] += 0.375*v[cd[34]];
	v[pd[25]] += -0.125*v[cd[34]];
	v[pd[28]] += (-0.3125*v[cd[25]] - 0.375*v[cd[26]]
		      - 0.1875*v[cd[34]]);
	v[pd[29]] += (v[cd[27]] + 0.9375*v[cd[25]]
		      + 0.375*v[cd[26]] + 0.1875*v[cd[34]]);
	v[pd[30]] += 0.75*v[cd[26]];
	v[pd[31]] += -0.1875*v[cd[34]];
	v[pd[32]] += 0.1875*v[cd[34]];
	v[pd[34]] += 0.75*v[cd[34]];
	break;
      case 3:
	v[pd[0]] += -0.0390625*v[cd[34]];
	v[pd[1]] += 0.0234375*v[cd[34]];
	v[pd[4]] += 0.09375*v[cd[34]];
	v[pd[5]] += -0.046875*v[cd[34]];
	v[pd[6]] += -0.03125*v[cd[34]];
	v[pd[7]] += 0.0625*v[cd[34]];
	v[pd[10]] += 0.0625*v[cd[34]];
	v[pd[13]] += -0.0625*v[cd[34]];
	v[pd[16]] += -0.0625*v[cd[34]];
	v[pd[22]] += 0.375*v[cd[34]];
	v[pd[25]] += -0.125*v[cd[34]];
	v[pd[28]] += -0.1875*v[cd[34]];
	v[pd[29]] += 0.1875*v[cd[34]];
	v[pd[31]] += -0.1875*v[cd[34]];
	v[pd[32]] += 0.1875*v[cd[34]];
	v[pd[34]] += 0.75*v[cd[34]];
	break;
      }
    }
  }

  return;
}


static void real_d_refine_inter4(DOF_REAL_D_VEC *drdv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_d_refine_inter4");
  DOF       pd[N_BAS4];
  const DOF *cd;
  EL        *el;
  REAL_D    *v = nil;
  int       i, k, typ, lr_set;
  const DOF       *(*get_dof_indices)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF_ADMIN *admin;

  if (n < 1) return;
  el = list->el;
  typ = list->el_type;

  GET_DOF_VEC(v, drdv);
  if (!drdv->fe_space)
  {
    ERROR("no fe_space in dof_real_d_vec %s\n", NAME(drdv));
    return;
  }
  else if (!drdv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drdv->fe_space));
    return;
  }
  get_dof_indices = drdv->fe_space->bas_fcts->get_dof_indices;
  GET_STRUCT(admin,drdv->fe_space);

  get_dof_indices(el, admin, pd);

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

  cd = get_dof_indices(el->child[0], admin, nil);

  for (k = 0; k < DIM_OF_WORLD; k++)
  {
    v[cd[3]][k] = (v[pd[5]][k]);
    v[cd[10]][k] = (0.2734375*v[pd[0]][k] - 0.0390625*v[pd[1]][k]
		    + 1.09375*v[pd[4]][k] - 0.546875*v[pd[5]][k]
		    + 0.21875*v[pd[6]][k]);
    v[cd[11]][k] = (v[pd[4]][k]);
    v[cd[12]][k] = (-0.0390625*v[pd[0]][k] + 0.0234375*v[pd[1]][k]
		    + 0.46875*v[pd[4]][k] + 0.703125*v[pd[5]][k]
		    - 0.15625*v[pd[6]][k]);
    v[cd[16]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k])
		    + 0.03125*(v[pd[4]][k] + v[pd[6]][k])
		    + 0.015625*v[pd[5]][k] +
		    0.1875*(v[pd[7]][k]+v[pd[13]][k]
			    -v[pd[31]][k]-v[pd[32]][k])
		    + 0.375*(-v[pd[8]][k] - v[pd[14]][k])
		    + 0.5*(v[pd[9]][k] + v[pd[15]][k]) + 0.75*v[pd[33]][k]);
    v[cd[17]][k] = (v[pd[33]][k]);
    v[cd[18]][k] = (0.0234375*(v[pd[0]][k] + v[pd[1]][k]) 
		    + 0.09375*(-v[pd[4]][k] - v[pd[6]][k])
		    + 0.140625*v[pd[5]][k]
		    + 0.0625*(-v[pd[7]][k] - v[pd[13]][k])
		    + 0.5625*(v[pd[31]][k] + v[pd[32]][k]));
    v[cd[19]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k])
		    + 0.03125*(v[pd[4]][k] + v[pd[6]][k]) 
		    + 0.015625*v[pd[5]][k]
		    + 0.1875*(v[pd[10]][k]+v[pd[16]][k]
			      -v[pd[28]][k]-v[pd[29]][k])
		    + 0.375*(-v[pd[11]][k] - v[pd[17]][k])
		    + 0.5*(v[pd[12]][k] + v[pd[18]][k]) + 0.75*v[pd[30]][k]);
    v[cd[20]][k] = (v[pd[30]][k]);
    v[cd[21]][k] = (0.0234375*(v[pd[0]][k] + v[pd[1]][k])
		    + 0.09375*(-v[pd[4]][k] - v[pd[6]][k])
		    + 0.140625*v[pd[5]][k] 
		    + 0.0625*(-v[pd[10]][k] - v[pd[16]][k])
		    + 0.5625*(v[pd[28]][k] + v[pd[29]][k]));
    v[cd[22]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k])
		    + 0.03125*(v[pd[4]][k] + v[pd[6]][k])
		    + 0.015625*v[pd[5]][k]
		    + 0.125*(v[pd[7]][k]-v[pd[8]][k]+v[pd[13]][k]-v[pd[14]][k]
			     -v[pd[31]][k]-v[pd[32]][k])
		    + 0.0625*(v[pd[10]][k]+v[pd[16]][k]
			      -v[pd[28]][k]-v[pd[29]][k])
		    + 0.25*(-v[pd[22]][k] - v[pd[25]][k] + v[pd[33]][k])
		    + 0.5*(v[pd[23]][k] + v[pd[26]][k] + v[pd[34]][k]));
    v[cd[23]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k])
		    + 0.03125*(v[pd[4]][k] + v[pd[6]][k])
		    + 0.015625*v[pd[5]][k]
		    + 0.0625*(v[pd[7]][k]+v[pd[13]][k]
			      -v[pd[31]][k]-v[pd[32]][k])
		    + 0.125*(v[pd[10]][k] - v[pd[11]][k] + v[pd[16]][k]
			     - v[pd[17]][k] - v[pd[28]][k] - v[pd[29]][k])
		    + 0.25*(-v[pd[22]][k] - v[pd[25]][k] + v[pd[30]][k])
		    + 0.5*(v[pd[24]][k] + v[pd[27]][k] + v[pd[34]][k]));
    v[cd[24]][k] = (v[pd[34]][k]);
    v[cd[25]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k])
		    + 0.15625*(v[pd[4]][k] + v[pd[6]][k])
		    - 0.234375*v[pd[5]][k]
		    + 0.3125*(v[pd[10]][k] - v[pd[29]][k])
		    + 0.0625*v[pd[16]][k] + 0.9375*v[pd[28]][k]);
    v[cd[26]][k] = (0.0234375*v[pd[0]][k] - 0.0390625*v[pd[1]][k]
		    - 0.03125*v[pd[4]][k] - 0.046875*v[pd[5]][k] 
		    + 0.09375*v[pd[6]][k]
		    + 0.125*(-v[pd[10]][k] + v[pd[16]][k] - v[pd[17]][k])
		    + 0.375*(v[pd[11]][k] + v[pd[28]][k] - v[pd[29]][k])
		    + 0.75*v[pd[30]][k]);
    v[cd[27]][k] = (v[pd[28]][k]);
    v[cd[28]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k])
		    + 0.15625*(v[pd[4]][k] + v[pd[6]][k])
		    - 0.234375*v[pd[5]][k]
		    + 0.3125*(v[pd[7]][k] - v[pd[32]][k])
		    + 0.0625*v[pd[13]][k] + 0.9375*v[pd[31]][k]);
    v[cd[29]][k] = (0.0234375*v[pd[0]][k] - 0.0390625*v[pd[1]][k]
		    - 0.03125*v[pd[4]][k] - 0.046875*v[pd[5]][k]
		    + 0.09375*v[pd[6]][k]
		    + 0.125*(-v[pd[7]][k] + v[pd[13]][k] - v[pd[14]][k])
		    + 0.375*(v[pd[8]][k] + v[pd[31]][k] - v[pd[32]][k])
		    + 0.75*v[pd[33]][k]);
    v[cd[30]][k] = (v[pd[31]][k]);
    v[cd[34]][k] = (0.0234375*v[pd[0]][k] - 0.0390625*v[pd[1]][k]
		    - 0.03125*v[pd[4]][k] - 0.046875*v[pd[5]][k]
		    + 0.09375*v[pd[6]][k]
		    + 0.0625*(-v[pd[7]][k] - v[pd[10]][k]
			      + v[pd[13]][k] + v[pd[16]][k])
		    - 0.125*v[pd[22]][k] + 0.375*v[pd[25]][k]
		    + 0.1875*(v[pd[28]][k] - v[pd[29]][k]
			      + v[pd[31]][k] - v[pd[32]][k])
		    + 0.75*v[pd[34]][k]);
  }
/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

  cd = get_dof_indices(el->child[1], admin, nil);
  if (typ == 0)
  {
/*--------------------------------------------------------------------------*/
/*  parent of el_type 0                                                     */
/*--------------------------------------------------------------------------*/

    for (k = 0; k < DIM_OF_WORLD; k++)
    {
      v[cd[10]][k] = (-0.0390625*v[pd[0]][k] + 0.2734375*v[pd[1]][k]
		      + 0.21875*v[pd[4]][k] - 0.546875*v[pd[5]][k]
		      + 1.09375*v[pd[6]][k]);
      v[cd[11]][k] = v[pd[6]][k];
      v[cd[12]][k] = (0.0234375*v[pd[0]][k] - 0.0390625*v[pd[1]][k]
		      - 0.15625*v[pd[4]][k] + 0.703125*v[pd[5]][k]
		      + 0.46875*v[pd[6]][k]);
      v[cd[25]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k]) 
		      + 0.15625*(v[pd[4]][k] + v[pd[6]][k])
		      - 0.234375*v[pd[5]][k] + 0.0625*v[pd[7]][k]
		      + 0.3125*(v[pd[13]][k] - v[pd[31]][k])
		      + 0.9375*v[pd[32]][k]);
      v[cd[26]][k] = (-0.0390625*v[pd[0]][k] + 0.0234375*v[pd[1]][k]
		      + 0.09375*v[pd[4]][k] - 0.046875*v[pd[5]][k]
		      - 0.03125*v[pd[6]][k]
		      + 0.125*(v[pd[7]][k] - v[pd[8]][k] - v[pd[13]][k])
		      + 0.375*(v[pd[14]][k] - v[pd[31]][k] + v[pd[32]][k])
		      + 0.75*v[pd[33]][k]);
      v[cd[27]][k] = v[pd[32]][k];
      v[cd[28]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k])
		      + 0.15625*(v[pd[4]][k] + v[pd[6]][k])
		      - 0.234375*v[pd[5]][k] + 0.0625*v[pd[10]][k]
		      + 0.3125*(v[pd[16]][k] - v[pd[28]][k])
		      + 0.9375*v[pd[29]][k]);
      v[cd[29]][k] = (-0.0390625*v[pd[0]][k] + 0.0234375*v[pd[1]][k]
		      + 0.09375*v[pd[4]][k] - 0.046875*v[pd[5]][k]
		      - 0.03125*v[pd[6]][k]
		      + 0.125*(v[pd[10]][k] - v[pd[11]][k] - v[pd[16]][k])
		      + 0.375*(v[pd[17]][k] - v[pd[28]][k] + v[pd[29]][k])
		      + 0.75*v[pd[30]][k]);
      v[cd[30]][k] = v[pd[29]][k];
      v[cd[34]][k] = (-0.0390625*v[pd[0]][k] + 0.0234375*v[pd[1]][k]
		      + 0.09375*v[pd[4]][k] - 0.046875*v[pd[5]][k]
		      - 0.03125*v[pd[6]][k]
		      + 0.0625*(v[pd[7]][k] + v[pd[10]][k]
				-v[pd[13]][k] - v[pd[16]][k])
		      + 0.375*v[pd[22]][k] - 0.125*v[pd[25]][k]
		      + 0.1875*(-v[pd[28]][k] + v[pd[29]][k]
				-v[pd[31]][k] + v[pd[32]][k])
		      + 0.75*v[pd[34]][k]);
    }
  }
  else
  {
/*--------------------------------------------------------------------------*/
/*  parent of el_type 1|2                                                   */
/*--------------------------------------------------------------------------*/

    for (k = 0; k < DIM_OF_WORLD; k++)
    {
      v[cd[10]][k] = (-0.0390625*v[pd[0]][k] + 0.2734375*v[pd[1]][k]
		      + 0.21875*v[pd[4]][k] - 0.546875*v[pd[5]][k]
		      + 1.09375*v[pd[6]][k]);
      v[cd[11]][k] = v[pd[6]][k];
      v[cd[12]][k] = (0.0234375*v[pd[0]][k] - 0.0390625*v[pd[1]][k]
		      - 0.15625*v[pd[4]][k] + 0.703125*v[pd[5]][k]
		      + 0.46875*v[pd[6]][k]);
      v[cd[25]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k])
		      + 0.15625*(v[pd[4]][k] + v[pd[6]][k])
		      - 0.234375*v[pd[5]][k] + 0.0625*v[pd[10]][k]
		      + 0.3125*(v[pd[16]][k] - v[pd[28]][k])
		      + 0.9375*v[pd[29]][k]);
      v[cd[26]][k] = (-0.0390625*v[pd[0]][k] + 0.0234375*v[pd[1]][k]
		      + 0.09375*v[pd[4]][k] - 0.046875*v[pd[5]][k]
		      - 0.03125*v[pd[6]][k] 
		      + 0.125*(v[pd[10]][k] - v[pd[11]][k] - v[pd[16]][k])
		      + 0.375*(v[pd[17]][k] - v[pd[28]][k] + v[pd[29]][k])
		      + 0.75*v[pd[30]][k]);
      v[cd[27]][k] = v[pd[29]][k];
      v[cd[28]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k])
		      + 0.15625*(v[pd[4]][k] + v[pd[6]][k])
		      - 0.234375*v[pd[5]][k] + 0.0625*v[pd[7]][k]
		      + 0.3125*(v[pd[13]][k] - v[pd[31]][k])
		      + 0.9375*v[pd[32]][k]);
      v[cd[29]][k] = (-0.0390625*v[pd[0]][k] + 0.0234375*v[pd[1]][k]
		      + 0.09375*v[pd[4]][k] - 0.046875*v[pd[5]][k]
		      - 0.03125*v[pd[6]][k]
		      + 0.125*(v[pd[7]][k] - v[pd[8]][k] - v[pd[13]][k])
		      + 0.375*(v[pd[14]][k] - v[pd[31]][k] + v[pd[32]][k])
		      + 0.75*v[pd[33]][k]);
      v[cd[30]][k] = v[pd[32]][k];
      v[cd[34]][k] = (-0.0390625*v[pd[0]][k] + 0.0234375*v[pd[1]][k]
		      + 0.09375*v[pd[4]][k] - 0.046875*v[pd[5]][k]
		      - 0.03125*v[pd[6]][k]
		      + 0.0625*(v[pd[7]][k] + v[pd[10]][k]
				- v[pd[13]][k] - v[pd[16]][k])
		      + 0.375*v[pd[22]][k] - 0.125*v[pd[25]][k]
		      + 0.1875*(-v[pd[28]][k] + v[pd[29]][k]
				- v[pd[31]][k] + v[pd[32]][k])
		      + 0.75*v[pd[34]][k]);
    }
  }

/*--------------------------------------------------------------------------*/
/*   adjust neighbour values                                                */
/*--------------------------------------------------------------------------*/
  
  for (i = 1; i < n; i++)
  {
    el = list[i].el;
    typ = list[i].el_type;
    get_dof_indices(el, admin, pd);

    lr_set = 0;
    if (list[i].neigh[0]  &&  list[i].neigh[0]->no < i)
      lr_set = 1;

    if (list[i].neigh[1]  &&  list[i].neigh[1]->no < i)
      lr_set += 2;

    TEST_EXIT(lr_set)("no values set on both neighbours\n");

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

    cd = get_dof_indices(el->child[0], admin, nil);

    switch(lr_set)
    {
    case 1:
      for (k = 0; k < DIM_OF_WORLD; k++)
      {
	v[cd[16]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k])
			+ 0.03125*(v[pd[4]][k] + v[pd[6]][k])
			+ 0.015625*v[pd[5]][k]
			+ 0.1875*(v[pd[7]][k] + v[pd[13]][k]
				  - v[pd[31]][k] - v[pd[32]][k])
			+ 0.375*(-v[pd[8]][k] - v[pd[14]][k])
			+ 0.5*(v[pd[9]][k] + v[pd[15]][k])
			+ 0.75*v[pd[33]][k]);
	v[cd[17]][k] = v[pd[33]][k];
	v[cd[18]][k] = (0.0234375*(v[pd[0]][k] + v[pd[1]][k])
			+ 0.09375*(-v[pd[4]][k] - v[pd[6]][k])
			+ 0.140625*v[pd[5]][k]
			+ 0.0625*(-v[pd[7]][k] - v[pd[13]][k])
			+ 0.5625*(v[pd[31]][k] + v[pd[32]][k]));
	v[cd[22]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k])
			+ 0.03125*(v[pd[4]][k] + v[pd[6]][k])
			+ 0.015625*v[pd[5]][k]
			+ 0.125*(v[pd[7]][k] - v[pd[8]][k] + v[pd[13]][k]
				 - v[pd[14]][k] - v[pd[31]][k] - v[pd[32]][k])
			+ 0.0625*(v[pd[10]][k] + v[pd[16]][k]
				  - v[pd[28]][k] - v[pd[29]][k])
			+ 0.25*(-v[pd[22]][k] - v[pd[25]][k] + v[pd[33]][k])
			+ 0.5*(v[pd[23]][k] + v[pd[26]][k] + v[pd[34]][k]));
	v[cd[23]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k])
			+ 0.03125*(v[pd[4]][k] + v[pd[6]][k])
			+ 0.015625*v[pd[5]][k]
			+ 0.0625*(v[pd[7]][k] + v[pd[13]][k]
				  - v[pd[31]][k] - v[pd[32]][k])
			+ 0.125*(v[pd[10]][k] - v[pd[11]][k] + v[pd[16]][k]
				 - v[pd[17]][k] - v[pd[28]][k] - v[pd[29]][k])
			+ 0.25*(-v[pd[22]][k] - v[pd[25]][k] + v[pd[30]][k])
			+ 0.5*(v[pd[24]][k] + v[pd[27]][k] + v[pd[34]][k]));
	v[cd[24]][k] = v[pd[34]][k];
	v[cd[28]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k])
			+ 0.15625*(v[pd[4]][k] + v[pd[6]][k])
			- 0.234375*v[pd[5]][k]
			+ 0.3125*(v[pd[7]][k] - v[pd[32]][k])
			+ 0.0625*v[pd[13]][k] + 0.9375*v[pd[31]][k]);
	v[cd[29]][k] = (0.0234375*v[pd[0]][k] - 0.0390625*v[pd[1]][k]
			- 0.03125*v[pd[4]][k] - 0.046875*v[pd[5]][k]
			+ 0.09375*v[pd[6]][k]
			+ 0.125*(-v[pd[7]][k] + v[pd[13]][k] - v[pd[14]][k])
			+ 0.375*(v[pd[8]][k] + v[pd[31]][k] - v[pd[32]][k])
			+ 0.75*v[pd[33]][k]);
	v[cd[30]][k] = v[pd[31]][k];
	v[cd[34]][k] = (0.0234375*v[pd[0]][k] - 0.0390625*v[pd[1]][k]
			- 0.03125*v[pd[4]][k] - 0.046875*v[pd[5]][k]
			+ 0.09375*v[pd[6]][k]
			+ 0.0625*(-v[pd[7]][k] - v[pd[10]][k]
				  + v[pd[13]][k] + v[pd[16]][k])
			- 0.125*v[pd[22]][k] + 0.375*v[pd[25]][k]
			+ 0.1875*(v[pd[28]][k] - v[pd[29]][k]
				  + v[pd[31]][k] - v[pd[32]][k])
			+ 0.75*v[pd[34]][k]);
      }
      break;
    case 2:
      for (k = 0; k < DIM_OF_WORLD; k++)
      {
      v[cd[19]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k])
		     + 0.03125*(v[pd[4]][k] + v[pd[6]][k])
		     + 0.015625*v[pd[5]][k]
		     + 0.1875*(v[pd[10]][k] + v[pd[16]][k]
			       - v[pd[28]][k] - v[pd[29]][k])
		     + 0.375*(-v[pd[11]][k] - v[pd[17]][k])
		     + 0.5*(v[pd[12]][k] + v[pd[18]][k])
		     + 0.75*v[pd[30]][k]);
      v[cd[20]][k] = v[pd[30]][k];
      v[cd[21]][k] = (0.0234375*(v[pd[0]][k] + v[pd[1]][k])
		     + 0.09375*(-v[pd[4]][k] - v[pd[6]][k])
		     + 0.140625*v[pd[5]][k]
		     + 0.0625*(-v[pd[10]][k] - v[pd[16]][k])
		     + 0.5625*(v[pd[28]][k] + v[pd[29]][k]));
      v[cd[22]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k])
		     + 0.03125*(v[pd[4]][k] + v[pd[6]][k])
		     + 0.015625*v[pd[5]][k]
		     + 0.125*(v[pd[7]][k] - v[pd[8]][k] + v[pd[13]][k]
			      - v[pd[14]][k] - v[pd[31]][k] - v[pd[32]][k])
		     + 0.0625*(v[pd[10]][k] + v[pd[16]][k]
			       - v[pd[28]][k] - v[pd[29]][k])
		     + 0.25*(-v[pd[22]][k] - v[pd[25]][k] + v[pd[33]][k])
		     + 0.5*(v[pd[23]][k] + v[pd[26]][k] + v[pd[34]][k]));
      v[cd[23]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k])
		     + 0.03125*(v[pd[4]][k] + v[pd[6]][k])
		     + 0.015625*v[pd[5]][k]
		     + 0.0625*(v[pd[7]][k] + v[pd[13]][k]
			       - v[pd[31]][k] - v[pd[32]][k])
		     + 0.125*(v[pd[10]][k] - v[pd[11]][k] + v[pd[16]][k]
			      - v[pd[17]][k] - v[pd[28]][k] - v[pd[29]][k])
		     + 0.25*(-v[pd[22]][k] - v[pd[25]][k] + v[pd[30]][k])
		     + 0.5*(v[pd[24]][k] + v[pd[27]][k] + v[pd[34]][k]));
      v[cd[24]][k] = v[pd[34]][k];
      v[cd[25]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k])
		     + 0.15625*(v[pd[4]][k] + v[pd[6]][k])
		     - 0.234375*v[pd[5]][k]
		     + 0.3125*(v[pd[10]][k] - v[pd[29]][k])
		     + 0.0625*v[pd[16]][k] + 0.9375*v[pd[28]][k]);
      v[cd[26]][k] = (0.0234375*v[pd[0]][k] - 0.0390625*v[pd[1]][k]
		     - 0.03125*v[pd[4]][k] - 0.046875*v[pd[5]][k]
		     + 0.09375*v[pd[6]][k]
		     + 0.125*(-v[pd[10]][k] + v[pd[16]][k] - v[pd[17]][k])
		     + 0.375*(v[pd[11]][k] + v[pd[28]][k] - v[pd[29]][k])
		      + 0.75*v[pd[30]][k]);
      v[cd[27]][k] = v[pd[28]][k];
      v[cd[34]][k] = (0.0234375*v[pd[0]][k] - 0.0390625*v[pd[1]][k]
		      - 0.03125*v[pd[4]][k] - 0.046875*v[pd[5]][k]
		      + 0.09375*v[pd[6]][k]
		      + 0.0625*(-v[pd[7]][k] - v[pd[10]][k]
				+ v[pd[13]][k] + v[pd[16]][k])
		      - 0.125*v[pd[22]][k] + 0.375*v[pd[25]][k]
		      + 0.1875*(v[pd[28]][k] - v[pd[29]][k]
				+ v[pd[31]][k] - v[pd[32]][k])
		      + 0.75*v[pd[34]][k]);
      }
      break;
    case 3:
      for (k = 0; k < DIM_OF_WORLD; k++)
      {
	v[cd[22]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k])
			+ 0.03125*(v[pd[4]][k] + v[pd[6]][k])
			+ 0.015625*v[pd[5]][k]
			+ 0.125*(v[pd[7]][k] - v[pd[8]][k] + v[pd[13]][k]
				 - v[pd[14]][k] - v[pd[31]][k] - v[pd[32]][k])
			+ 0.0625*(v[pd[10]][k] + v[pd[16]][k]
				  - v[pd[28]][k] - v[pd[29]][k])
			+ 0.25*(-v[pd[22]][k] - v[pd[25]][k] + v[pd[33]][k])
			+ 0.5*(v[pd[23]][k] + v[pd[26]][k] + v[pd[34]][k]));
	v[cd[23]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k])
			+ 0.03125*(v[pd[4]][k] + v[pd[6]][k])
			+ 0.015625*v[pd[5]][k]
			+ 0.0625*(v[pd[7]][k] + v[pd[13]][k]
				  - v[pd[31]][k] - v[pd[32]][k])
			+ 0.125*(v[pd[10]][k] - v[pd[11]][k] + v[pd[16]][k]
				 - v[pd[17]][k] - v[pd[28]][k] - v[pd[29]][k])
			+ 0.25*(-v[pd[22]][k] - v[pd[25]][k] + v[pd[30]][k])
			+ 0.5*(v[pd[24]][k] + v[pd[27]][k] + v[pd[34]][k]));
	v[cd[24]][k] = v[pd[34]][k];
	v[cd[34]][k] = (0.0234375*v[pd[0]][k] - 0.0390625*v[pd[1]][k]
			- 0.03125*v[pd[4]][k] - 0.046875*v[pd[5]][k]
			+ 0.09375*v[pd[6]][k]
			+ 0.0625*(-v[pd[7]][k] - v[pd[10]][k]
				  + v[pd[13]][k] + v[pd[16]][k])
			- 0.125*v[pd[22]][k] + 0.375*v[pd[25]][k]
			+ 0.1875*(v[pd[28]][k] - v[pd[29]][k]
				  + v[pd[31]][k] - v[pd[32]][k])
			+ 0.75*v[pd[34]][k]);
      }
      break;
    }

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

    cd = get_dof_indices(el->child[1], admin, nil);

    if (typ == 0)
    {
      switch(lr_set)
      {
      case 1:
	for (k = 0; k < DIM_OF_WORLD; k++)
	{
	  v[cd[25]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k])
			  + 0.15625*(v[pd[4]][k] + v[pd[6]][k])
			  - 0.234375*v[pd[5]][k] + 0.0625*v[pd[7]][k]
			  + 0.3125*(v[pd[13]][k] - v[pd[31]][k])
			  + 0.9375*v[pd[32]][k]);
	  v[cd[26]][k] = (-0.0390625*v[pd[0]][k] + 0.0234375*v[pd[1]][k]
			  + 0.09375*v[pd[4]][k] - 0.046875*v[pd[5]][k]
			  - 0.03125*v[pd[6]][k]
			  + 0.125*(v[pd[7]][k] - v[pd[8]][k] - v[pd[13]][k])
			  + 0.375*(v[pd[14]][k] - v[pd[31]][k] + v[pd[32]][k])
			  + 0.75*v[pd[33]][k]);
	  v[cd[27]][k] = v[pd[32]][k];
	  v[cd[34]][k] = (-0.0390625*v[pd[0]][k] + 0.0234375*v[pd[1]][k]
			  + 0.09375*v[pd[4]][k] - 0.046875*v[pd[5]][k]
			  - 0.03125*v[pd[6]][k]
			  + 0.0625*(v[pd[7]][k] + v[pd[10]][k]
				    - v[pd[13]][k] - v[pd[16]][k])
			  + 0.375*v[pd[22]][k] - 0.125*v[pd[25]][k]
			  + 0.1875*(-v[pd[28]][k] + v[pd[29]][k]
				    - v[pd[31]][k] + v[pd[32]][k])
			  + 0.75*v[pd[34]][k]);
	}
	break;
      case 2:
	for (k = 0; k < DIM_OF_WORLD; k++)
	{
	  v[cd[28]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k])
			  + 0.15625*(v[pd[4]][k] + v[pd[6]][k])
			  - 0.234375*v[pd[5]][k] + 0.0625*v[pd[10]][k]
			  + 0.3125*(v[pd[16]][k] - v[pd[28]][k])
			  + 0.9375*v[pd[29]][k]);
	  v[cd[29]][k] = (-0.0390625*v[pd[0]][k] + 0.0234375*v[pd[1]][k]
			  + 0.09375*v[pd[4]][k] - 0.046875*v[pd[5]][k]
			  - 0.03125*v[pd[6]][k]
			  + 0.125*(v[pd[10]][k] - v[pd[11]][k] - v[pd[16]][k])
			  + 0.375*(v[pd[17]][k] - v[pd[28]][k] + v[pd[29]][k])
			  + 0.75*v[pd[30]][k]);
	  v[cd[30]][k] = v[pd[29]][k];
	  v[cd[34]][k] = (-0.0390625*v[pd[0]][k] + 0.0234375*v[pd[1]][k]
			  + 0.09375*v[pd[4]][k] - 0.046875*v[pd[5]][k]
			  - 0.03125*v[pd[6]][k]
			  + 0.0625*(v[pd[7]][k] + v[pd[10]][k]
				    - v[pd[13]][k] - v[pd[16]][k])
			  + 0.375*v[pd[22]][k] - 0.125*v[pd[25]][k]
			  + 0.1875*(-v[pd[28]][k] + v[pd[29]][k]
				    - v[pd[31]][k] + v[pd[32]][k])
			  + 0.75*v[pd[34]][k]);
	}
	break;
      case 3:
	for (k = 0; k < DIM_OF_WORLD; k++)
	{
	  v[cd[34]][k] = (-0.0390625*v[pd[0]][k] + 0.0234375*v[pd[1]][k]
			  + 0.09375*v[pd[4]][k] - 0.046875*v[pd[5]][k]
			  - 0.03125*v[pd[6]][k]
			  + 0.0625*(v[pd[7]][k] + v[pd[10]][k]
				    - v[pd[13]][k] - v[pd[16]][k])
			  + 0.375*v[pd[22]][k] - 0.125*v[pd[25]][k]
			  + 0.1875*(-v[pd[28]][k] + v[pd[29]][k]
				    - v[pd[31]][k] + v[pd[32]][k])
			  + 0.75*v[pd[34]][k]);
	}
	break;
      }
    }
    else
    {
      switch(lr_set)
      {
      case 1:
	for (k = 0; k < DIM_OF_WORLD; k++)
	{
	  v[cd[28]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k])
			  + 0.15625*(v[pd[4]][k] + v[pd[6]][k])
			  - 0.234375*v[pd[5]][k] + 0.0625*v[pd[7]][k]
			  + 0.3125*(v[pd[13]][k] - v[pd[31]][k])
			  + 0.9375*v[pd[32]][k]);
	  v[cd[29]][k] = (-0.0390625*v[pd[0]][k] + 0.0234375*v[pd[1]][k]
			  + 0.09375*v[pd[4]][k] - 0.046875*v[pd[5]][k]
			  - 0.03125*v[pd[6]][k]
			  + 0.125*(v[pd[7]][k] - v[pd[8]][k] - v[pd[13]][k])
			  + 0.375*(v[pd[14]][k] - v[pd[31]][k] + v[pd[32]][k])
			  + 0.75*v[pd[33]][k]);
	  v[cd[30]][k] = v[pd[32]][k];
	  v[cd[34]][k] = (-0.0390625*v[pd[0]][k] + 0.0234375*v[pd[1]][k]
			  + 0.09375*v[pd[4]][k] - 0.046875*v[pd[5]][k]
			  - 0.03125*v[pd[6]][k]
			  + 0.0625*(v[pd[7]][k] + v[pd[10]][k]
				    - v[pd[13]][k] - v[pd[16]][k])
			  + 0.375*v[pd[22]][k] - 0.125*v[pd[25]][k]
			  + 0.1875*(-v[pd[28]][k] + v[pd[29]][k]
				    - v[pd[31]][k] + v[pd[32]][k])
			  + 0.75*v[pd[34]][k]);
	}
	break;
      case 2:
	for (k = 0; k < DIM_OF_WORLD; k++)
	{
	  v[cd[25]][k] = (0.0390625*(-v[pd[0]][k] - v[pd[1]][k])
			  + 0.15625*(v[pd[4]][k] + v[pd[6]][k])
			  - 0.234375*v[pd[5]][k] + 0.0625*v[pd[10]][k]
			  + 0.3125*(v[pd[16]][k] - v[pd[28]][k])
			  + 0.9375*v[pd[29]][k]);
	  v[cd[26]][k] = (-0.0390625*v[pd[0]][k] + 0.0234375*v[pd[1]][k]
			  + 0.09375*v[pd[4]][k] - 0.046875*v[pd[5]][k]
			  - 0.03125*v[pd[6]][k]
			  + 0.125*(v[pd[10]][k] - v[pd[11]][k] - v[pd[16]][k])
			  + 0.375*(v[pd[17]][k] - v[pd[28]][k] + v[pd[29]][k])
			  + 0.75*v[pd[30]][k]);
	  v[cd[27]][k] = v[pd[29]][k];
	  v[cd[34]][k] = (-0.0390625*v[pd[0]][k] + 0.0234375*v[pd[1]][k]
			  + 0.09375*v[pd[4]][k] - 0.046875*v[pd[5]][k]
			  - 0.03125*v[pd[6]][k]
			  + 0.0625*(v[pd[7]][k] + v[pd[10]][k]
				    - v[pd[13]][k] - v[pd[16]][k])
			  + 0.375*v[pd[22]][k] - 0.125*v[pd[25]][k]
			  + 0.1875*(-v[pd[28]][k] + v[pd[29]][k]
				    - v[pd[31]][k] + v[pd[32]][k])
			  + 0.75*v[pd[34]][k]);
	}
	break;
      case 3:
	for (k = 0; k < DIM_OF_WORLD; k++)
	{
	  v[cd[34]][k] = (-0.0390625*v[pd[0]][k] + 0.0234375*v[pd[1]][k]
			  + 0.09375*v[pd[4]][k] - 0.046875*v[pd[5]][k]
			  - 0.03125*v[pd[6]][k]
			  + 0.0625*(v[pd[7]][k] + v[pd[10]][k]
				    - v[pd[13]][k] - v[pd[16]][k])
			  + 0.375*v[pd[22]][k] - 0.125*v[pd[25]][k]
			  + 0.1875*(-v[pd[28]][k] + v[pd[29]][k]
				    - v[pd[31]][k] + v[pd[32]][k])
			  + 0.75*v[pd[34]][k]);
	}
	break;
      }
    }
  }

  return;
}

static void real_d_coarse_inter4(DOF_REAL_D_VEC *drdv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_d_coarse_inter4");
  DOF       pd[N_BAS4];
  const DOF *cd;
  EL        *el;
  REAL_D    *v = nil;
  int       i, k, typ, lr_set;
  const DOF       *(*get_dof_indices)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF_ADMIN *admin;

  if (n < 1) return;
  el = list->el;
  typ = list->el_type;

  GET_DOF_VEC(v, drdv);
  if (!drdv->fe_space)
  {
    ERROR("no fe_space in dof_real_vec %s\n", NAME(drdv));
    return;
  }
  else if (!drdv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drdv->fe_space));
    return;
  }
  get_dof_indices = drdv->fe_space->bas_fcts->get_dof_indices;
  GET_STRUCT(admin,drdv->fe_space);

  get_dof_indices(el, admin, pd);

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

  cd = get_dof_indices(el->child[0], admin, nil);

  for (k = 0; k < DIM_OF_WORLD; k++)
  {
    v[pd[4]][k] = v[cd[11]][k];
    v[pd[5]][k] = v[cd[3]][k];
    v[pd[28]][k] = v[cd[27]][k];
    v[pd[30]][k] = v[cd[20]][k];
    v[pd[31]][k] = v[cd[30]][k];
    v[pd[33]][k] = v[cd[17]][k];
    v[pd[34]][k] = v[cd[24]][k];
  }

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

  cd = get_dof_indices(el->child[1], admin, nil);

  if (typ == 0)
  {
/*--------------------------------------------------------------------------*/
/*  parent of el_type 0                                                     */
/*--------------------------------------------------------------------------*/

    for (k = 0; k < DIM_OF_WORLD; k++)
    {
      v[pd[6]][k] = v[cd[11]][k];
      v[pd[29]][k] = v[cd[30]][k];
      v[pd[32]][k] = v[cd[27]][k];
    }
  }
  else
  {
/*--------------------------------------------------------------------------*/
/*  parent of el_type 1|2                                                   */
/*--------------------------------------------------------------------------*/

    for (k = 0; k < DIM_OF_WORLD; k++)
    {
      v[pd[6]][k] = v[cd[11]][k];
      v[pd[29]][k] = v[cd[27]][k];
      v[pd[32]][k] = v[cd[30]][k];
    }
  }

/*--------------------------------------------------------------------------*/
/*   adjust neighbour values                                                */
/*--------------------------------------------------------------------------*/
  
  for (i = 1; i < n; i++)
  {
    el = list[i].el;
    typ = list[i].el_type;
    get_dof_indices(el, admin, pd);

    lr_set = 0;
    if (list[i].neigh[0]  &&  list[i].neigh[0]->no < i)
      lr_set = 1;

    if (list[i].neigh[1]  &&  list[i].neigh[1]->no < i)
      lr_set += 2;

    TEST_EXIT(lr_set)("no values set on both neighbours\n");

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

    cd = get_dof_indices(el->child[0], admin, nil);

    switch(lr_set)
    {
    case 1:
      for (k = 0; k < DIM_OF_WORLD; k++)
      {
	v[pd[31]][k] = v[cd[30]][k];
	v[pd[33]][k] = v[cd[17]][k];
	v[pd[34]][k] = v[cd[24]][k];
      }
      break;
    case 2:
      for (k = 0; k < DIM_OF_WORLD; k++)
      {
	v[pd[28]][k] = v[cd[27]][k];
	v[pd[30]][k] = v[cd[20]][k];
	v[pd[34]][k] = v[cd[24]][k];
      }
      break;
    case 3:
      for (k = 0; k < DIM_OF_WORLD; k++)
	v[pd[34]][k] = v[cd[24]][k];
      break;
    }

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

    cd = get_dof_indices(el->child[1], admin, nil);

    if (typ == 0)
    {
      switch(lr_set)
      {
      case 1:
	for (k = 0; k < DIM_OF_WORLD; k++)
	  v[pd[32]][k] = v[cd[27]][k];
	break;
      case 2:
	for (k = 0; k < DIM_OF_WORLD; k++)
	  v[pd[29]][k] = v[cd[30]][k];
	break;
      }
    }
    else
    {
      switch(lr_set)
      {
      case 1:
	for (k = 0; k < DIM_OF_WORLD; k++)
	  v[pd[32]][k] = v[cd[30]][k];
	break;
      case 2:
	for (k = 0; k < DIM_OF_WORLD; k++)
	  v[pd[29]][k] = v[cd[27]][k];
	break;
      }
    }
  }

  return;
}



static void real_d_coarse_restr4(DOF_REAL_D_VEC *drdv, RC_LIST_EL *list, int n)
{
  FUNCNAME("real_d_coarse_restr4");
  DOF       pd[N_BAS4];
  const DOF *cd;
  EL        *el;
  REAL_D    *v = nil;
  int       i, k, typ, lr_set;
  const DOF       *(*get_dof_indices)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF_ADMIN *admin;

  if (n < 1) return;
  el = list->el;
  typ = list->el_type;

  GET_DOF_VEC(v, drdv);
  if (!drdv->fe_space)
  {
    ERROR("no fe_space in dof_real_vec %s\n", NAME(drdv));
    return;
  }
  else if (!drdv->fe_space->bas_fcts)
  {
    ERROR("no basis functions in fe_space %s\n", NAME(drdv->fe_space));
    return;
  }
  get_dof_indices = drdv->fe_space->bas_fcts->get_dof_indices;
  GET_STRUCT(admin,drdv->fe_space);

  get_dof_indices(el, admin, pd);

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

  cd = get_dof_indices(el->child[0], admin, nil);

  for (k = 0; k < DIM_OF_WORLD; k++)
  {
    v[pd[0]][k] += (0.2734375*v[cd[10]][k]
		    + 0.0390625*(-v[cd[12]][k] - v[cd[16]][k] - v[cd[19]][k]
				 - v[cd[22]][k] - v[cd[23]][k] - v[cd[25]][k]
				 - v[cd[28]][k])
		    + 0.0234375*(v[cd[18]][k] + v[cd[21]][k] + v[cd[26]][k]
				 + v[cd[29]][k] + v[cd[34]][k]));
    v[pd[1]][k] += (0.0390625*(-v[cd[10]][k] - v[cd[16]][k] - v[cd[19]][k]
			       - v[cd[22]][k] - v[cd[23]][k] - v[cd[25]][k]
			       - v[cd[26]][k] - v[cd[28]][k] - v[cd[29]][k]
			       - v[cd[34]][k])
		    + 0.0234375*(v[cd[12]][k] + v[cd[18]][k] + v[cd[21]][k]));
    v[pd[4]][k] = (v[cd[11]][k] + 1.09375*v[cd[10]][k] + 0.46875*v[cd[12]][k]
		   + 0.03125*(v[cd[16]][k] + v[cd[19]][k] + v[cd[22]][k]
			      + v[cd[23]][k] - v[cd[26]][k] - v[cd[29]][k]
			      - v[cd[34]][k])
		   + 0.09375*(-v[cd[18]][k] - v[cd[21]][k])
		   + 0.15625*(v[cd[25]][k] + v[cd[28]][k]));
    v[pd[5]][k] = (v[cd[3]][k] - 0.546875*v[cd[10]][k]
		   + 0.703125*v[cd[12]][k]
		   + 0.015625*(v[cd[16]][k] + v[cd[19]][k]
			       + v[cd[22]][k] + v[cd[23]][k])
		   + 0.140625*(v[cd[18]][k] + v[cd[21]][k])
		   + 0.234375*(-v[cd[25]][k] - v[cd[28]][k])
		   + 0.046875*(-v[cd[26]][k] - v[cd[29]][k] - v[cd[34]][k]));
    v[pd[6]][k] = (0.21875*v[cd[10]][k]
		   + 0.15625*(-v[cd[12]][k] + v[cd[25]][k] + v[cd[28]][k])
		   + 0.03125*(v[cd[16]][k] + v[cd[19]][k]
			      + v[cd[22]][k] + v[cd[23]][k])
		   + 0.09375*(-v[cd[18]][k] - v[cd[21]][k] + v[cd[26]][k]
			      + v[cd[29]][k] + v[cd[34]][k]));
    v[pd[7]][k] += (0.1875*v[cd[16]][k]
		    + 0.0625*(-v[cd[18]][k] + v[cd[23]][k] - v[cd[34]][k])
		    + 0.125*(v[cd[22]][k] - v[cd[29]][k])
		    + 0.3125*v[cd[28]][k]);
    v[pd[8]][k] += (0.375*(-v[cd[16]][k] + v[cd[29]][k])
		    - 0.125*v[cd[22]][k]);
    v[pd[9]][k] += 0.5*v[cd[16]][k];
    v[pd[10]][k] += (0.1875*v[cd[19]][k]
		     + 0.0625*(-v[cd[21]][k] + v[cd[22]][k] - v[cd[34]][k])
		     + 0.125*(v[cd[23]][k] - v[cd[26]][k])
		     + 0.3125*v[cd[25]][k]);
    v[pd[11]][k] += (0.375*(-v[cd[19]][k] + v[cd[26]][k])
		     - 0.125*v[cd[23]][k]);
    v[pd[12]][k] += 0.5*v[cd[19]][k];
    v[pd[13]][k] += (0.1875*v[cd[16]][k]
		     + 0.0625*(-v[cd[18]][k] + v[cd[23]][k]
			       + v[cd[28]][k] + v[cd[34]][k])
		     + 0.125*(v[cd[22]][k] + v[cd[29]][k]));
    v[pd[14]][k] += (-0.375*v[cd[16]][k] 
		     + 0.125*(-v[cd[22]][k] - v[cd[29]][k]));
    v[pd[15]][k] += 0.5*v[cd[16]][k];
    v[pd[16]][k] += (0.1875*v[cd[19]][k]
		     + 0.0625*(-v[cd[21]][k] + v[cd[22]][k]
			       + v[cd[25]][k] + v[cd[34]][k])
		     + 0.125*(v[cd[23]][k] + v[cd[26]][k]));
    v[pd[17]][k] += (-0.375*v[cd[19]][k] 
		     + 0.125*(-v[cd[23]][k] - v[cd[26]][k]));
    v[pd[18]][k] += 0.5*v[cd[19]][k];
    v[pd[22]][k] += (0.25*(-v[cd[22]][k] - v[cd[23]][k])
		     - 0.125*v[cd[34]][k]);
    v[pd[23]][k] += 0.5*v[cd[22]][k];
    v[pd[24]][k] += 0.5*v[cd[23]][k];
    v[pd[25]][k] += (0.25*(-v[cd[22]][k] - v[cd[23]][k])
		     + 0.375*v[cd[34]][k]);
    v[pd[26]][k] += 0.5*v[cd[22]][k];
    v[pd[27]][k] += 0.5*v[cd[23]][k];
    v[pd[28]][k] = (v[cd[27]][k] + 0.1875*(-v[cd[19]][k] + v[cd[34]][k]) 
		    + 0.5625*v[cd[21]][k] - 0.0625*v[cd[22]][k]
		    - 0.125*v[cd[23]][k] + 0.9375*v[cd[25]][k]
		    + 0.375*v[cd[26]][k]);
    v[pd[29]][k] = (0.1875*(-v[cd[19]][k] - v[cd[34]][k])
		    + 0.5625*v[cd[21]][k] - 0.0625*v[cd[22]][k]
		    - 0.125*v[cd[23]][k] - 0.3125*v[cd[25]][k]
		    - 0.375*v[cd[26]][k]);
    v[pd[30]][k] = (v[cd[20]][k] + 0.75*(v[cd[19]][k] + v[cd[26]][k])
		    + 0.25*v[cd[23]][k]);
    v[pd[31]][k] = (v[cd[30]][k] + 0.1875*(-v[cd[16]][k] + v[cd[34]][k])
		    + 0.5625*v[cd[18]][k] - 0.125*v[cd[22]][k]
		    - 0.0625*v[cd[23]][k] + 0.9375*v[cd[28]][k]
		    + 0.375*v[cd[29]][k]);
    v[pd[32]][k] = (0.1875*(-v[cd[16]][k] - v[cd[34]][k])
		    + 0.5625*v[cd[18]][k] - 0.125*v[cd[22]][k]
		    - 0.0625*v[cd[23]][k] - 0.3125*v[cd[28]][k]
		    - 0.375*v[cd[29]][k]);
    v[pd[33]][k] = (v[cd[17]][k] + 0.75*(v[cd[16]][k] + v[cd[29]][k])
		    + 0.25*v[cd[22]][k]);
    v[pd[34]][k] = (v[cd[24]][k] + 0.5*(v[cd[22]][k] + v[cd[23]][k])
		    + 0.75*v[cd[34]][k]);
  }
/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

  cd = get_dof_indices(el->child[1], admin, nil);

  if (typ == 0)
  {
/*--------------------------------------------------------------------------*/
/*  parent of el_type 0                                                     */
/*--------------------------------------------------------------------------*/
    for (k = 0; k < DIM_OF_WORLD; k++)
    {
      v[pd[0]][k] += (0.0390625*(-v[cd[10]][k] - v[cd[25]][k] - v[cd[26]][k]
				 - v[cd[28]][k] - v[cd[29]][k] - v[cd[34]][k])
		      + 0.0234375*v[cd[12]][k]);
      v[pd[1]][k] += (0.2734375*v[cd[10]][k]
		      + 0.0390625*(-v[cd[12]][k] - v[cd[25]][k]
				   - v[cd[28]][k])
		      + 0.0234375*(v[cd[26]][k] + v[cd[29]][k]
				   + v[cd[34]][k]));
      v[pd[4]][k] += (0.21875*v[cd[10]][k]
		      + 0.15625*(-v[cd[12]][k] + v[cd[25]][k] + v[cd[28]][k])
		      + 0.09375*(v[cd[26]][k] + v[cd[29]][k] + v[cd[34]][k]));
      v[pd[5]][k] += (-0.546875*v[cd[10]][k] + 0.703125*v[cd[12]][k]
		      + 0.234375*(-v[cd[25]][k] - v[cd[28]][k])
		      + 0.046875*(-v[cd[26]][k] - v[cd[29]][k]
				  - v[cd[34]][k]));
      v[pd[6]][k] += (v[cd[11]][k] + 1.09375*v[cd[10]][k]
		      + 0.46875*v[cd[12]][k] 
		      + 0.15625*(v[cd[25]][k] + v[cd[28]][k])
		      + 0.03125*(-v[cd[26]][k] - v[cd[29]][k]
				 - v[cd[34]][k]));
      v[pd[7]][k] += (0.0625*(v[cd[25]][k] + v[cd[34]][k])
		      + 0.125*v[cd[26]][k]);
      v[pd[8]][k] += -0.125*v[cd[26]][k];
      v[pd[10]][k] += (0.0625*(v[cd[28]][k] + v[cd[34]][k])
		       + 0.125*v[cd[29]][k]);
      v[pd[11]][k] += -0.125*v[cd[29]][k];
      v[pd[13]][k] += (0.3125*v[cd[25]][k] - 0.125*v[cd[26]][k]
		       - 0.0625*v[cd[34]][k]);
      v[pd[14]][k] += 0.375*v[cd[26]][k];
      v[pd[16]][k] += (0.3125*v[cd[28]][k] - 0.125*v[cd[29]][k]
		       - 0.0625*v[cd[34]][k]);
      v[pd[17]][k] += 0.375*v[cd[29]][k];
      v[pd[22]][k] += 0.375*v[cd[34]][k];
      v[pd[25]][k] += -0.125*v[cd[34]][k];
      v[pd[28]][k] += (-0.3125*v[cd[28]][k] - 0.375*v[cd[29]][k]
		       - 0.1875*v[cd[34]][k]);
      v[pd[29]][k] += (v[cd[30]][k] + 0.9375*v[cd[28]][k] + 0.375*v[cd[29]][k]
		       + 0.1875*v[cd[34]][k]);
      v[pd[30]][k] += 0.75*v[cd[29]][k];
      v[pd[31]][k] += (-0.3125*v[cd[25]][k] - 0.375*v[cd[26]][k]
		       - 0.1875*v[cd[34]][k]);
      v[pd[32]][k] += (v[cd[27]][k] + 0.9375*v[cd[25]][k]
		       + 0.375*v[cd[26]][k] + 0.1875*v[cd[34]][k]);
      v[pd[33]][k] += 0.75*v[cd[26]][k];
      v[pd[34]][k] += 0.75*v[cd[34]][k];
    }
  }
  else
  {
/*--------------------------------------------------------------------------*/
/*  parent of el_type 1|2                                                   */
/*--------------------------------------------------------------------------*/
    for (k = 0; k < DIM_OF_WORLD; k++)
    {
      v[pd[0]][k] += (0.0390625*(-v[cd[10]][k] - v[cd[25]][k] - v[cd[26]][k]
				 - v[cd[28]][k] - v[cd[29]][k] - v[cd[34]][k])
		      + 0.0234375*v[cd[12]][k]);
      v[pd[1]][k] += (0.2734375*v[cd[10]][k]
		      + 0.0390625*(-v[cd[12]][k] - v[cd[25]][k]
				   - v[cd[28]][k])
		      + 0.0234375*(v[cd[26]][k] + v[cd[29]][k]
				   + v[cd[34]][k]));
      v[pd[4]][k] += (0.21875*v[cd[10]][k]
		      + 0.15625*(-v[cd[12]][k] + v[cd[25]][k] + v[cd[28]][k])
		      + 0.09375*(v[cd[26]][k] + v[cd[29]][k] + v[cd[34]][k]));
      v[pd[5]][k] += (-0.546875*v[cd[10]][k] + 0.703125*v[cd[12]][k]
		      + 0.234375*(-v[cd[25]][k] - v[cd[28]][k])
		      + 0.046875*(-v[cd[26]][k] - v[cd[29]][k]
				  - v[cd[34]][k]));
      v[pd[6]][k] += (v[cd[11]][k] + 1.09375*v[cd[10]][k]
		      + 0.46875*v[cd[12]][k]
		      + 0.15625*(v[cd[25]][k] + v[cd[28]][k])
		      + 0.03125*(-v[cd[26]][k] - v[cd[29]][k]
				 - v[cd[34]][k]));
      v[pd[7]][k] += (0.0625*(v[cd[28]][k] + v[cd[34]][k])
		      + 0.125*v[cd[29]][k]);
      v[pd[8]][k] += -0.125*v[cd[29]][k];
      v[pd[10]][k] += (0.0625*(v[cd[25]][k] + v[cd[34]][k])
		       + 0.125*v[cd[26]][k]);
      v[pd[11]][k] += -0.125*v[cd[26]][k];
      v[pd[13]][k] += (0.3125*v[cd[28]][k] - 0.125*v[cd[29]][k]
		       - 0.0625*v[cd[34]][k]);
      v[pd[14]][k] += 0.375*v[cd[29]][k];
      v[pd[16]][k] += (0.3125*v[cd[25]][k] - 0.125*v[cd[26]][k]
		       - 0.0625*v[cd[34]][k]);
      v[pd[17]][k] += 0.375*v[cd[26]][k];
      v[pd[22]][k] += 0.375*v[cd[34]][k];
      v[pd[25]][k] += -0.125*v[cd[34]][k];
      v[pd[28]][k] += (-0.3125*v[cd[25]][k] - 0.375*v[cd[26]][k]
		       - 0.1875*v[cd[34]][k]);
      v[pd[29]][k] += (v[cd[27]][k] + 0.9375*v[cd[25]][k] + 0.375*v[cd[26]][k]
		       + 0.1875*v[cd[34]][k]);
      v[pd[30]][k] += 0.75*v[cd[26]][k];
      v[pd[31]][k] += (-0.3125*v[cd[28]][k] - 0.375*v[cd[29]][k]
		       - 0.1875*v[cd[34]][k]);
      v[pd[32]][k] += (v[cd[30]][k] + 0.9375*v[cd[28]][k]
		       + 0.375*v[cd[29]][k] + 0.1875*v[cd[34]][k]);
      v[pd[33]][k] += 0.75*v[cd[29]][k];
      v[pd[34]][k] += 0.75*v[cd[34]][k];
    }
  }

/*--------------------------------------------------------------------------*/
/*   adjust neighbour values                                                */
/*--------------------------------------------------------------------------*/
  
  for (i = 1; i < n; i++)
  {
    el = list[i].el;
    typ = list[i].el_type;
    get_dof_indices(el, admin, pd);

    lr_set = 0;
    if (list[i].neigh[0]  &&  list[i].neigh[0]->no < i)
      lr_set = 1;

    if (list[i].neigh[1]  &&  list[i].neigh[1]->no < i)
      lr_set += 2;

    TEST_EXIT(lr_set)("no values set on both neighbours\n");

/*--------------------------------------------------------------------------*/
/*  values on child[0]                                                      */
/*--------------------------------------------------------------------------*/

    cd = get_dof_indices(el->child[0], admin, nil);

    switch(lr_set)
    {
    case 1:
      for (k = 0; k < DIM_OF_WORLD; k++)
      {
	v[pd[0]][k] += (0.0390625*(-v[cd[16]][k] - v[cd[22]][k]
				   - v[cd[23]][k] - v[cd[28]][k])
			+ 0.0234375*(v[cd[18]][k] + v[cd[29]][k]
				     + v[cd[34]][k]));
	v[pd[1]][k] += (0.0390625*(-v[cd[16]][k] - v[cd[22]][k] - v[cd[23]][k]
				   - v[cd[28]][k]-v[cd[29]][k]-v[cd[34]][k])
			+ 0.0234375*v[cd[18]][k]);
	v[pd[4]][k] += (0.03125*(v[cd[16]][k] + v[cd[22]][k] + v[cd[23]][k]
				 - v[cd[29]][k] - v[cd[34]][k])
			- 0.09375*v[cd[18]][k] + 0.15625*v[cd[28]][k]);
	v[pd[5]][k] += (0.015625*(v[cd[16]][k] + v[cd[22]][k] + v[cd[23]][k])
			+ 0.140625*v[cd[18]][k] - 0.234375*v[cd[28]][k]
			+ 0.046875*(-v[cd[29]][k] - v[cd[34]][k]));
	v[pd[6]][k] += (0.03125*(v[cd[16]][k] + v[cd[22]][k] + v[cd[23]][k])
			+ 0.09375*(-v[cd[18]][k]+v[cd[29]][k]+v[cd[34]][k])
			+ 0.15625*v[cd[28]][k]);
	v[pd[7]][k] += (0.1875*v[cd[16]][k]
			+ 0.0625*(-v[cd[18]][k] + v[cd[23]][k] - v[cd[34]][k])
			+ 0.125*(v[cd[22]][k] - v[cd[29]][k])
			+ 0.3125*v[cd[28]][k]);
	v[pd[8]][k] += (0.375*(-v[cd[16]][k] + v[cd[29]][k])
			- 0.125*v[cd[22]][k]);
	v[pd[9]][k] += 0.5*v[cd[16]][k];
	v[pd[10]][k] += (0.0625*(v[cd[22]][k] - v[cd[34]][k])
			 + 0.125*v[cd[23]][k]);
	v[pd[11]][k] += -0.125*v[cd[23]][k];
	v[pd[13]][k] += (0.1875*v[cd[16]][k]
			 + 0.0625*(-v[cd[18]][k] + v[cd[23]][k]
				   + v[cd[28]][k] + v[cd[34]][k])
			 + 0.125*(v[cd[22]][k] + v[cd[29]][k]));
	v[pd[14]][k] += (-0.375*v[cd[16]][k]
			 + 0.125*(-v[cd[22]][k] - v[cd[29]][k]));
	v[pd[15]][k] += 0.5*v[cd[16]][k];
	v[pd[16]][k] += (0.0625*(v[cd[22]][k] + v[cd[34]][k])
			 + 0.125*v[cd[23]][k]);
	v[pd[17]][k] += -0.125*v[cd[23]][k];
	v[pd[22]][k] += (0.25*(-v[cd[22]][k] - v[cd[23]][k])
			 - 0.125*v[cd[34]][k]);
	v[pd[23]][k] += 0.5*v[cd[22]][k];
	v[pd[24]][k] += 0.5*v[cd[23]][k];
	v[pd[25]][k] += (0.25*(-v[cd[22]][k] - v[cd[23]][k])
			 + 0.375*v[cd[34]][k]);
	v[pd[26]][k] += 0.5*v[cd[22]][k];
	v[pd[27]][k] += 0.5*v[cd[23]][k];
	v[pd[28]][k] += (-0.0625*v[cd[22]][k] - 0.125*v[cd[23]][k]
			 + 0.1875*v[cd[34]][k]);
	v[pd[29]][k] += (-0.0625*v[cd[22]][k] - 0.125*v[cd[23]][k]
			 - 0.1875*v[cd[34]][k]);
	v[pd[30]][k] += 0.25*v[cd[23]][k];
	v[pd[31]][k] = (v[cd[30]][k] + 0.1875*(-v[cd[16]][k] + v[cd[34]][k])
			+ 0.5625*v[cd[18]][k] - 0.125*v[cd[22]][k]
			- 0.0625*v[cd[23]][k] + 0.9375*v[cd[28]][k]
			+ 0.375*v[cd[29]][k]);
	v[pd[32]][k] = (0.1875*(-v[cd[16]][k] - v[cd[34]][k])
			+ 0.5625*v[cd[18]][k] - 0.125*v[cd[22]][k]
			- 0.0625*v[cd[23]][k] - 0.3125*v[cd[28]][k]
			- 0.375*v[cd[29]][k]);
	v[pd[33]][k] = (v[cd[17]][k] + 0.75*(v[cd[16]][k] + v[cd[29]][k])
			+ 0.25*v[cd[22]][k]);
	v[pd[34]][k] = (v[cd[24]][k] + 0.5*(v[cd[22]][k] + v[cd[23]][k])
			+ 0.75*v[cd[34]][k]);
      }
      break;
    case 2:
      for (k = 0; k < DIM_OF_WORLD; k++)
      {
	v[pd[0]][k] += (0.0390625*(-v[cd[19]][k] - v[cd[22]][k]
				   - v[cd[23]][k] - v[cd[25]][k])
			+ 0.0234375*(v[cd[21]][k] + v[cd[26]][k]
				     + v[cd[34]][k]));
	v[pd[1]][k] += (0.0390625*(-v[cd[19]][k]-v[cd[22]][k]-v[cd[23]][k]
				   - v[cd[25]][k]-v[cd[26]][k]-v[cd[34]][k])
			+ 0.0234375*v[cd[21]][k]);
	v[pd[4]][k] += (0.03125*(v[cd[19]][k] + v[cd[22]][k] + v[cd[23]][k]
				 - v[cd[26]][k] - v[cd[34]][k])
			- 0.09375*v[cd[21]][k] + 0.15625*v[cd[25]][k]);
	v[pd[5]][k] += (0.015625*(v[cd[19]][k] + v[cd[22]][k] + v[cd[23]][k])
			+ 0.140625*v[cd[21]][k] - 0.234375*v[cd[25]][k]
			+ 0.046875*(-v[cd[26]][k] - v[cd[34]][k]));
	v[pd[6]][k] += (0.03125*(v[cd[19]][k] + v[cd[22]][k] + v[cd[23]][k])
			+ 0.09375*(-v[cd[21]][k]+v[cd[26]][k]+v[cd[34]][k])
			+ 0.15625*v[cd[25]][k]);
	v[pd[7]][k] += (0.125*v[cd[22]][k]
			+ 0.0625*(v[cd[23]][k]-v[cd[34]][k]));
	v[pd[8]][k] += -0.125*v[cd[22]][k];
	v[pd[10]][k] += (0.1875*v[cd[19]][k]
			 + 0.0625*(-v[cd[21]][k]+v[cd[22]][k]-v[cd[34]][k])
			 + 0.125*(v[cd[23]][k] - v[cd[26]][k])
		      + 0.3125*v[cd[25]][k]);
	v[pd[11]][k] += (0.375*(-v[cd[19]][k] + v[cd[26]][k])
			 - 0.125*v[cd[23]][k]);
	v[pd[12]][k] += 0.5*v[cd[19]][k];
	v[pd[13]][k] += (0.125*v[cd[22]][k]
			 + 0.0625*(v[cd[23]][k] + v[cd[34]][k]));
	v[pd[14]][k] += -0.125*v[cd[22]][k];
	v[pd[16]][k] += (0.1875*v[cd[19]][k]
			 + 0.0625*(-v[cd[21]][k] + v[cd[22]][k]
				   + v[cd[25]][k] + v[cd[34]][k])
			 + 0.125*(v[cd[23]][k] + v[cd[26]][k]));
	v[pd[17]][k] += (-0.375*v[cd[19]][k] 
			 + 0.125*(-v[cd[23]][k] - v[cd[26]][k]));
	v[pd[18]][k] += 0.5*v[cd[19]][k];
	v[pd[22]][k] += (0.25*(-v[cd[22]][k] - v[cd[23]][k])
			 - 0.125*v[cd[34]][k]);
	v[pd[23]][k] += (0.5*v[cd[22]][k]);
	v[pd[24]][k] += (0.5*v[cd[23]][k]);
	v[pd[25]][k] += (0.25*(-v[cd[22]][k] - v[cd[23]][k])
			 + 0.375*v[cd[34]][k]);
	v[pd[26]][k] += (0.5*v[cd[22]][k]);
	v[pd[27]][k] += (0.5*v[cd[23]][k]);
	v[pd[28]][k] = (v[cd[27]][k] + 0.1875*(-v[cd[19]][k] + v[cd[34]][k])
			+ 0.5625*v[cd[21]][k] - 0.0625*v[cd[22]][k]
			- 0.125*v[cd[23]][k] + 0.9375*v[cd[25]][k]
			+ 0.375*v[cd[26]][k]);
	v[pd[29]][k] = (0.1875*(-v[cd[19]][k] - v[cd[34]][k])
			+ 0.5625*v[cd[21]][k] - 0.0625*v[cd[22]][k]
			- 0.125*v[cd[23]][k] - 0.3125*v[cd[25]][k]
			- 0.375*v[cd[26]][k]);
	v[pd[30]][k] = (v[cd[20]][k] + 0.75*(v[cd[19]][k] + v[cd[26]][k])
			+ 0.25*v[cd[23]][k]);
	v[pd[31]][k] += (-0.125*v[cd[22]][k] - 0.0625*v[cd[23]][k]
			 + 0.1875*v[cd[34]][k]);
	v[pd[32]][k] += (-0.125*v[cd[22]][k] - 0.0625*v[cd[23]][k]
			 - 0.1875*v[cd[34]][k]);
	v[pd[33]][k] += 0.25*v[cd[22]][k];
	v[pd[34]][k] = (v[cd[24]][k] + 0.5*(v[cd[22]][k] + v[cd[23]][k])
			+ 0.75*v[cd[34]][k]);
      }
      break;
    case 3:
      for (k = 0; k < DIM_OF_WORLD; k++)
      {
	v[pd[0]][k] += (0.0390625*(-v[cd[22]][k] - v[cd[23]][k])
			+ 0.0234375*v[cd[34]][k]);
	v[pd[1]][k] += (0.0390625*(-v[cd[22]][k] - v[cd[23]][k]
				   - v[cd[34]][k]));
	v[pd[4]][k] += (0.03125*(v[cd[22]][k] + v[cd[23]][k] - v[cd[34]][k]));
	v[pd[5]][k] += (0.015625*(v[cd[22]][k] + v[cd[23]][k])
			- 0.046875*v[cd[34]][k]);
	v[pd[6]][k] += (0.03125*(v[cd[22]][k] + v[cd[23]][k])
			+ 0.09375*v[cd[34]][k]);
	v[pd[7]][k] += (0.125*v[cd[22]][k]
			+ 0.0625*(v[cd[23]][k] - v[cd[34]][k]));
	v[pd[8]][k] += -0.125*v[cd[22]][k];
	v[pd[10]][k] += (0.0625*(v[cd[22]][k] - v[cd[34]][k])
			 + 0.125*v[cd[23]][k]);
	v[pd[11]][k] += -0.125*v[cd[23]][k];
	v[pd[13]][k] += (0.125*v[cd[22]][k]
			 + 0.0625*(v[cd[23]][k] + v[cd[34]][k]));
	v[pd[14]][k] += -0.125*v[cd[22]][k];
	v[pd[16]][k] += (0.0625*(v[cd[22]][k] + v[cd[34]][k])
			 + 0.125*v[cd[23]][k]);
	v[pd[17]][k] += -0.125*v[cd[23]][k];
	v[pd[22]][k] += (0.25*(-v[cd[22]][k] - v[cd[23]][k])
			 - 0.125*v[cd[34]][k]);
	v[pd[23]][k] += 0.5*v[cd[22]][k];
	v[pd[24]][k] += 0.5*v[cd[23]][k];
	v[pd[25]][k] += (0.25*(-v[cd[22]][k] - v[cd[23]][k])
			 + 0.375*v[cd[34]][k]);
	v[pd[26]][k] += 0.5*v[cd[22]][k];
	v[pd[27]][k] += 0.5*v[cd[23]][k];
	v[pd[28]][k] += (-0.0625*v[cd[22]][k] - 0.125*v[cd[23]][k]
			 + 0.1875*v[cd[34]][k]);
	v[pd[29]][k] += (-0.0625*v[cd[22]][k] - 0.125*v[cd[23]][k]
			 - 0.1875*v[cd[34]][k]);
	v[pd[30]][k] += 0.25*v[cd[23]][k];
	v[pd[31]][k] += (-0.125*v[cd[22]][k] - 0.0625*v[cd[23]][k]
			 + 0.1875*v[cd[34]][k]);
	v[pd[32]][k] += (-0.125*v[cd[22]][k] - 0.0625*v[cd[23]][k]
			 - 0.1875*v[cd[34]][k]);
	v[pd[33]][k] += 0.25*v[cd[22]][k];
	v[pd[34]][k] = (v[cd[24]][k] + 0.5*(v[cd[22]][k] + v[cd[23]][k])
			+ 0.75*v[cd[34]][k]);
      }
      break;
    }

/*--------------------------------------------------------------------------*/
/*  values on child[1]                                                      */
/*--------------------------------------------------------------------------*/

    cd = get_dof_indices(el->child[1], admin, nil);

    if (typ == 0)
    {
      switch(lr_set)
      {
      case 1:
	for (k = 0; k < DIM_OF_WORLD; k++)
	{
	  v[pd[0]][k] += (0.0390625*(-v[cd[25]][k] - v[cd[26]][k]
				     - v[cd[34]][k]));
	  v[pd[1]][k] += (-0.0390625*v[cd[25]][k]
			  + 0.0234375*(v[cd[26]][k] + v[cd[34]][k]));
	  v[pd[4]][k] += (0.15625*v[cd[25]][k]
			  + 0.09375*(v[cd[26]][k] + v[cd[34]][k]));
	  v[pd[5]][k] += (-0.234375*v[cd[25]][k]
			  + 0.046875*(-v[cd[26]][k] - v[cd[34]][k]));
	  v[pd[6]][k] += (0.15625*v[cd[25]][k]
			  + 0.03125*(-v[cd[26]][k] - v[cd[34]][k]));
	  v[pd[7]][k] += (0.0625*(v[cd[25]][k] + v[cd[34]][k])
			  + 0.125*v[cd[26]][k]);
	  v[pd[8]][k] += -0.125*v[cd[26]][k];
	  v[pd[10]][k] += 0.0625*v[cd[34]][k];
	  v[pd[13]][k] += (0.3125*v[cd[25]][k] - 0.125*v[cd[26]][k]
			   - 0.0625*v[cd[34]][k]);
	  v[pd[14]][k] += 0.375*v[cd[26]][k];
	  v[pd[16]][k] += -0.0625*v[cd[34]][k];
	  v[pd[22]][k] += 0.375*v[cd[34]][k];
	  v[pd[25]][k] += -0.125*v[cd[34]][k];
	  v[pd[28]][k] += -0.1875*v[cd[34]][k];
	  v[pd[29]][k] += 0.1875*v[cd[34]][k];
	  v[pd[31]][k] += (-0.3125*v[cd[25]][k] - 0.375*v[cd[26]][k]
			   - 0.1875*v[cd[34]][k]);
	  v[pd[32]][k] += (v[cd[27]][k] + 0.9375*v[cd[25]][k]
			   + 0.375*v[cd[26]][k] + 0.1875*v[cd[34]][k]);
	  v[pd[33]][k] += 0.75*v[cd[26]][k];
	  v[pd[34]][k] += 0.75*v[cd[34]][k];
	}
	break;
      case 2:
	for (k = 0; k < DIM_OF_WORLD; k++)
	{
	  v[pd[0]][k] += (0.0390625*(-v[cd[28]][k] - v[cd[29]][k]
				     - v[cd[34]][k]));
	  v[pd[1]][k] += (-0.0390625*v[cd[28]][k]
			  + 0.0234375*(v[cd[29]][k] + v[cd[34]][k]));
	  v[pd[4]][k] += (0.15625*v[cd[28]][k]
			  + 0.09375*(v[cd[29]][k] + v[cd[34]][k]));
	  v[pd[5]][k] += (-0.234375*v[cd[28]][k]
			  + 0.046875*(-v[cd[29]][k] - v[cd[34]][k]));
	  v[pd[6]][k] += (0.15625*v[cd[28]][k]
			  + 0.03125*(-v[cd[29]][k] - v[cd[34]][k]));
	  v[pd[7]][k] += 0.0625*v[cd[34]][k];
	  v[pd[10]][k] += (0.0625*(v[cd[28]][k] + v[cd[34]][k])
			   + 0.125*v[cd[29]][k]);
	  v[pd[11]][k] += -0.125*v[cd[29]][k];
	  v[pd[13]][k] += -0.0625*v[cd[34]][k];
	  v[pd[16]][k] += (0.3125*v[cd[28]][k] - 0.125*v[cd[29]][k]
			   - 0.0625*v[cd[34]][k]);
	  v[pd[17]][k] += 0.375*v[cd[29]][k];
	  v[pd[22]][k] += 0.375*v[cd[34]][k];
	  v[pd[25]][k] += -0.125*v[cd[34]][k];
	  v[pd[28]][k] += (-0.3125*v[cd[28]][k] - 0.375*v[cd[29]][k]
			   - 0.1875*v[cd[34]][k]);
	  v[pd[29]][k] += (v[cd[30]][k] + 0.9375*v[cd[28]][k]
			   + 0.375*v[cd[29]][k] + 0.1875*v[cd[34]][k]);
	  v[pd[30]][k] += 0.75*v[cd[29]][k];
	  v[pd[31]][k] += -0.1875*v[cd[34]][k];
	  v[pd[32]][k] += 0.1875*v[cd[34]][k];
	  v[pd[34]][k] += 0.75*v[cd[34]][k];
	}
	break;
      case 3:
	for (k = 0; k < DIM_OF_WORLD; k++)
	{
	  v[pd[0]][k] += -0.0390625*v[cd[34]][k];
	  v[pd[1]][k] += 0.0234375*v[cd[34]][k];
	  v[pd[4]][k] += 0.09375*v[cd[34]][k];
	  v[pd[5]][k] += -0.046875*v[cd[34]][k];
	  v[pd[6]][k] += -0.03125*v[cd[34]][k];
	  v[pd[7]][k] += 0.0625*v[cd[34]][k];
	  v[pd[10]][k] += 0.0625*v[cd[34]][k];
	  v[pd[13]][k] += -0.0625*v[cd[34]][k];
	  v[pd[16]][k] += -0.0625*v[cd[34]][k];
	  v[pd[22]][k] += 0.375*v[cd[34]][k];
	  v[pd[25]][k] += -0.125*v[cd[34]][k];
	  v[pd[28]][k] += -0.1875*v[cd[34]][k];
	  v[pd[29]][k] += 0.1875*v[cd[34]][k];
	  v[pd[31]][k] += -0.1875*v[cd[34]][k];
	  v[pd[32]][k] += 0.1875*v[cd[34]][k];
	  v[pd[34]][k] += 0.75*v[cd[34]][k];
	}
	break;
      }
    }
    else
    {
      switch(lr_set)
      {
      case 1:
	for (k = 0; k < DIM_OF_WORLD; k++)
	{
	  v[pd[0]][k] += (0.0390625*(-v[cd[28]][k] - v[cd[29]][k]
				     - v[cd[34]][k]));
	  v[pd[1]][k] += (-0.0390625*v[cd[28]][k]
			  + 0.0234375*(v[cd[29]][k] + v[cd[34]][k]));
	  v[pd[4]][k] += (0.15625*v[cd[28]][k]
			  + 0.09375*(v[cd[29]][k] + v[cd[34]][k]));
	  v[pd[5]][k] += (-0.234375*v[cd[28]][k]
			  + 0.046875*(-v[cd[29]][k] - v[cd[34]][k]));
	  v[pd[6]][k] += (0.15625*v[cd[28]][k]
			  + 0.03125*(-v[cd[29]][k] - v[cd[34]][k]));
	  v[pd[7]][k] += (0.0625*(v[cd[28]][k] + v[cd[34]][k])
			  + 0.125*v[cd[29]][k]);
	  v[pd[8]][k] += -0.125*v[cd[29]][k];
	  v[pd[10]][k] += 0.0625*v[cd[34]][k];
	  v[pd[13]][k] += (0.3125*v[cd[28]][k] - 0.125*v[cd[29]][k]
			   - 0.0625*v[cd[34]][k]);
	  v[pd[14]][k] += 0.375*v[cd[29]][k];
	  v[pd[16]][k] += -0.0625*v[cd[34]][k];
	  v[pd[22]][k] += 0.375*v[cd[34]][k];
	  v[pd[25]][k] += -0.125*v[cd[34]][k];
	  v[pd[28]][k] += -0.1875*v[cd[34]][k];
	  v[pd[29]][k] += 0.1875*v[cd[34]][k];
	  v[pd[31]][k] += (-0.3125*v[cd[28]][k] - 0.375*v[cd[29]][k]
			   - 0.1875*v[cd[34]][k]);
	  v[pd[32]][k] += (v[cd[30]][k] + 0.9375*v[cd[28]][k]
			   + 0.375*v[cd[29]][k] + 0.1875*v[cd[34]][k]);
	  v[pd[33]][k] += 0.75*v[cd[29]][k];
	  v[pd[34]][k] += 0.75*v[cd[34]][k];
	}
	break;
      case 2:
	for (k = 0; k < DIM_OF_WORLD; k++)
	{
	  v[pd[0]][k] += (0.0390625*(-v[cd[25]][k] - v[cd[26]][k]
				     - v[cd[34]][k]));
	  v[pd[1]][k] += (-0.0390625*v[cd[25]][k]
			  + 0.0234375*(v[cd[26]][k] + v[cd[34]][k]));
	  v[pd[4]][k] += (0.15625*v[cd[25]][k]
			  + 0.09375*(v[cd[26]][k] + v[cd[34]][k]));
	  v[pd[5]][k] += (-0.234375*v[cd[25]][k]
			  + 0.046875*(-v[cd[26]][k] - v[cd[34]][k]));
	  v[pd[6]][k] += (0.15625*v[cd[25]][k]
			  + 0.03125*(-v[cd[26]][k] - v[cd[34]][k]));
	  v[pd[7]][k] += 0.0625*v[cd[34]][k];
	  v[pd[10]][k] += (0.0625*(v[cd[25]][k] + v[cd[34]][k])
			   + 0.125*v[cd[26]][k]);
	  v[pd[11]][k] += -0.125*v[cd[26]][k];
	  v[pd[13]][k] += -0.0625*v[cd[34]][k];
	  v[pd[16]][k] += (0.3125*v[cd[25]][k] - 0.125*v[cd[26]][k]
			   - 0.0625*v[cd[34]][k]);
	  v[pd[17]][k] += 0.375*v[cd[26]][k];
	  v[pd[22]][k] += 0.375*v[cd[34]][k];
	  v[pd[25]][k] += -0.125*v[cd[34]][k];
	  v[pd[28]][k] += (-0.3125*v[cd[25]][k] - 0.375*v[cd[26]][k]
			   - 0.1875*v[cd[34]][k]);
	  v[pd[29]][k] += (v[cd[27]][k] + 0.9375*v[cd[25]][k]
			   + 0.375*v[cd[26]][k] + 0.1875*v[cd[34]][k]);
	  v[pd[30]][k] += 0.75*v[cd[26]][k];
	  v[pd[31]][k] += -0.1875*v[cd[34]][k];
	  v[pd[32]][k] += 0.1875*v[cd[34]][k];
	  v[pd[34]][k] += 0.75*v[cd[34]][k];
	}
	break;
      case 3:
	for (k = 0; k < DIM_OF_WORLD; k++)
	{
	  v[pd[0]][k] += -0.0390625*v[cd[34]][k];
	  v[pd[1]][k] += 0.0234375*v[cd[34]][k];
	  v[pd[4]][k] += 0.09375*v[cd[34]][k];
	  v[pd[5]][k] += -0.046875*v[cd[34]][k];
	  v[pd[6]][k] += -0.03125*v[cd[34]][k];
	  v[pd[7]][k] += 0.0625*v[cd[34]][k];
	  v[pd[10]][k] += 0.0625*v[cd[34]][k];
	  v[pd[13]][k] += -0.0625*v[cd[34]][k];
	  v[pd[16]][k] += -0.0625*v[cd[34]][k];
	  v[pd[22]][k] += 0.375*v[cd[34]][k];
	  v[pd[25]][k] += -0.125*v[cd[34]][k];
	  v[pd[28]][k] += -0.1875*v[cd[34]][k];
	  v[pd[29]][k] += 0.1875*v[cd[34]][k];
	  v[pd[31]][k] += -0.1875*v[cd[34]][k];
	  v[pd[32]][k] += 0.1875*v[cd[34]][k];
	  v[pd[34]][k] += 0.75*v[cd[34]][k];
	}
	break;
      }
    }
  }

  return;
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

static BAS_FCT      *phi4[N_BAS4]     = {phi4v0,  phi4v1,
					 phi4v2,  phi4v3,
					 phi4e00, phi4e01, phi4e02,
					 phi4e10, phi4e11, phi4e12,
					 phi4e20, phi4e21, phi4e22,
					 phi4e30, phi4e31, phi4e32,
					 phi4e40, phi4e41, phi4e42,
					 phi4e50, phi4e51, phi4e52,
					 phi4f00, phi4f01, phi4f02,
					 phi4f10, phi4f11, phi4f12,
					 phi4f20, phi4f21, phi4f22,
					 phi4f30, phi4f31, phi4f32,
					 phi4c};
static GRD_BAS_FCT  *grd_phi4[N_BAS4] = {
                                grd_phi4v0,  grd_phi4v1, 
				grd_phi4v2,  grd_phi4v3,
				grd_phi4e00, grd_phi4e01, grd_phi4e02,
				grd_phi4e10, grd_phi4e11, grd_phi4e12,
				grd_phi4e20, grd_phi4e21, grd_phi4e22,
				grd_phi4e30, grd_phi4e31, grd_phi4e32,
				grd_phi4e40, grd_phi4e41, grd_phi4e42,
				grd_phi4e50, grd_phi4e51, grd_phi4e52,
				grd_phi4f00, grd_phi4f01, grd_phi4f02,
				grd_phi4f10, grd_phi4f11, grd_phi4f12,
				grd_phi4f20, grd_phi4f21, grd_phi4f22,
				grd_phi4f30, grd_phi4f31, grd_phi4f32,
				grd_phi4c};

static D2_BAS_FCT  *D2_phi4[N_BAS4] = {D2_phi4v0,  D2_phi4v1,
				       D2_phi4v2,  D2_phi4v3,
				       D2_phi4e00, D2_phi4e01, D2_phi4e02,
				       D2_phi4e10, D2_phi4e11, D2_phi4e12,
				       D2_phi4e20, D2_phi4e21, D2_phi4e22,
				       D2_phi4e30, D2_phi4e31, D2_phi4e32,
				       D2_phi4e40, D2_phi4e41, D2_phi4e42,
				       D2_phi4e50, D2_phi4e51, D2_phi4e52,
				       D2_phi4f00, D2_phi4f01, D2_phi4f02,
				       D2_phi4f10, D2_phi4f11, D2_phi4f12,
				       D2_phi4f20, D2_phi4f21, D2_phi4f22,
				       D2_phi4f30, D2_phi4f31, D2_phi4f32,
				       D2_phi4c};

static const BAS_FCTS lagrange4 = {"lagrange4", N_BAS4, 4, {1, 3, 3, 1},
				   nil, phi4, grd_phi4, D2_phi4,
				   get_dof_indices4, 
				   get_bound4,
				   interpol4,
				   interpol_d4,
				   get_int_vec4,
				   get_real_vec4,
				   get_real_d_vec4,
				   get_uchar_vec4,
				   get_schar_vec4,
				   real_refine_inter4, 
				   real_coarse_inter4,
				   real_coarse_restr4,
				   real_d_refine_inter4, 
				   real_d_coarse_inter4,
				   real_d_coarse_restr4,
				   (void *) bary4};

